/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CApplication.cpp
  Info   : $Id: CApplication.cpp,v 1.91 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// cxxx headers
#include <cxxx/CApplication.h>
#include <cxxx/CMessageBox.h>
#include <cxxx/CTimer.h>
#include <cxxx/CToolButton.h>
#include <cxxx/CTopLevelWindow.h>

// C headers
#include <sys/stat.h> // stat
#include <unistd.h> // access

#include <pwd.h> // getpwuid
#include <langinfo.h>

// C++ headers
#include <iostream>
#include <fstream>
#include <cerrno>
#include <cstdio>
#include <string>
#include <stdexcept>

// namespace declarations
using cxxx::cxAlphaFormat;
using cxxx::cxApp;
using cxxx::cxAppIntl;
using cxxx::cxARGBFormat;
using cxxx::cxCxxxIntl;
using cxxx::cxDefaultPic;
using cxxx::cxDnDManager;
using cxxx::cxDpy;
using cxxx::cxError;
using cxxx::cxFileManager;
using cxxx::cxGrayPic;
using cxxx::cxHomeDir;
using cxxx::cxSysIconDir;
using cxxx::cxMainWin;
using cxxx::cxLocale;
using cxxx::cxCxxxLocalePath;
using cxxx::cxAppLocalePath;
using cxxx::cxSelManager;
using cxxx::cxServerTime;
using cxxx::cxStipplePic;
using cxxx::cxTextFont;
using cxxx::cxTitleFont;
using cxxx::cxServerTime;
using cxxx::cxUpdateServerTime;
using cxxx::cxUtf8Atom;
using cxxx::cxWindowFormat;
using cxxx::cxStartupID;
using cxxx::lib::cxgettext;
using std::string;
using std::vector;
using std::cerr;
using std::endl;
using std::ios;
using std::ofstream;
using std::locale;
using std::runtime_error;

// constructor
CApplication::CApplication(const string& name)
: CMessageTransceiver(this, IDAPP),
  mModalWindow(NULL),
  mPopup(NULL),
  mShutDown(false),
  mAppname(name)
{
	cxAssert(cxApp == NULL);
	cxApp = this;
	cxServerTime = 0;
	try {
		locale loc("");
		cxLocale = loc;
	}
	catch (const runtime_error& e) {
		cerr << "C++X: Error:\nFailed to set locale:\n" << e.what() << "\nDefaulting to \'C\' locale." << endl;
	}
	locale::global(cxLocale);
	if (!XSupportsLocale()) {
		cerr << "C++X: Locale not supported by X Server." << std::endl;
		exit(EXIT_FAILURE);
	}
	if (cxCxxxLocalePath.empty()) cxCxxxLocalePath = LOCALEDIR "/%L/LC_MESSAGES/cxxx.msg";
	if (cxAppLocalePath.empty()) cxAppLocalePath = LOCALEDIR "/%L/LC_MESSAGES/" + mAppname + ".msg";
 	int err = cxCxxxIntl.open(cxLocale.name().c_str(), cxCxxxLocalePath.c_str(), nl_langinfo(CODESET));
	if (err) {
		cerr << "C++X: Failed to open C++X localization file. Trying C locale." << endl;
		err = cxCxxxIntl.open("C", cxCxxxLocalePath.c_str(), nl_langinfo(CODESET));
		if (err) {
			cerr << cxCxxxIntl.error_text(err) << endl;
			exit (EXIT_FAILURE);
		}
	}
	string id = "C++X ";
	id += CXXX_VERSION;
	if (id != cxCxxxIntl.get_id()) cerr << "C++X: Version mismatch in localization file." << endl;
	if (!mAppname.empty()) {
		err = cxAppIntl.open(cxLocale.name().c_str(), cxAppLocalePath.c_str(), nl_langinfo(CODESET));
		if (err) {
			cerr << "C++X: Failed to open " << mAppname << " localization file. Trying C locale." << endl;
			err = cxAppIntl.open("C", cxAppLocalePath.c_str(), nl_langinfo(CODESET));
			if (err) {
				cerr << cxAppIntl.error_text(err) << endl;
				exit(EXIT_FAILURE);
			}
		}
	}
	// Get preferred writing direction
#ifdef CXXX_HAS_BIDI
	string writedir = _("LTR", 474);
	if (writedir != "LTR") {
		cxxx::cxRTL = true;
		cxxx::cxBidi = new CBidi;
	}
#endif // CXXX_HAS_BIDI

	const char* homedir = getenv("HOME");
	if (homedir == NULL || *homedir == 0) {
		cxError << _("Environment variable HOME not set. Set it to your home directory and run again.", 6) << CError::Fatal;
	}
	cxHomeDir = homedir;
	string cxxxdir = cxHomeDir + "/.cxxx";
	string cxxxconfigfile = cxxxdir + "/cxxx.conf";
	string appconfigfile = cxxxdir + "/" + mAppname + ".conf";
 	// open X
	cxDpy = XOpenDisplay("");
	if (cxDpy == NULL) cxError << _("Cannot access X server.", 7) << CError::Fatal;
	// retrieve visual information
	int count;
	XVisualInfo tmpl;
	tmpl.visualid = XVisualIDFromVisual(DefaultVisual(cxDpy, DefaultScreen(cxDpy)));
	XVisualInfo* vis = XGetVisualInfo(cxDpy, VisualIDMask, &tmpl, &count);
	if (count != 1) cxError << _("Cannot retrieve info for default visual", 8) << CError::Fatal;
	if (vis->c_class != TrueColor && (vis->c_class != PseudoColor || (vis->depth > 8))) {
		cxError << _("Visual not suitable for libcxxx.\nSupported visuals are TrueColor and PseudoColor", 9) << CError::Fatal;
	}
	// initialize colors
	CColor::Init(*vis);
	XFree(vis);
	// retrieve UTF8_STRING Atom
	if (XA_UTF8_STRING == None) {
		cxUtf8Atom = XInternAtom(cxDpy, "UTF8_STRING", False);
		if (XA_UTF8_STRING == None) cxError << _("Failed to register XA_UTF8_STRING atom.", 398) << CError::Warn;
	}
	// query XRender
	int vmajor, vminor, res;
	res = XRenderQueryVersion(cxDpy, &vmajor, &vminor);
	cxAssert(res != 0 && (vmajor | vminor) != 0);
	// find the formats needed by XRender
	cxAlphaFormat = XRenderFindStandardFormat(cxDpy, PictStandardA8);
	cxWindowFormat = XRenderFindVisualFormat(cxDpy, XDefaultVisual(cxDpy, XDefaultScreen(cxDpy)));
	cxARGBFormat = XRenderFindStandardFormat(cxDpy, PictStandardARGB32);
	// Create a 1 x 1 picture with 32 bit format and repeat attribute
	XRenderPictureAttributes pa;
	pa.repeat = True;
	Pixmap pix = XCreatePixmap(cxDpy, XDefaultRootWindow(cxDpy), 1, 1, 32);
	cxDefaultPic = XRenderCreatePicture (cxDpy, pix, cxARGBFormat, CPRepeat, &pa);
	// Create a GC for 32 bit format
	cxxx::cxGC32 = XCreateGC(cxDpy, pix, 0, NULL);
   	cxAssert(cxxx::cxGC32 != NULL);
	XFreePixmap (cxDpy, pix);
	// Create the grayed out picture
	pix = XCreatePixmap(cxDpy, XDefaultRootWindow(cxDpy), 1, 1, 8);
	cxGrayPic = XRenderCreatePicture(cxDpy, pix, cxAlphaFormat, CPRepeat, &pa);
	XRenderColor rcol = CColor(0, 0, 0, 0x60);
	XRenderFillRectangle(cxDpy, PictOpSrc, cxGrayPic, &rcol, 0, 0, 1, 1);
	XFreePixmap(cxDpy, pix);
	// Create the stippled picture
	pix = XCreatePixmap(cxDpy, XDefaultRootWindow(cxDpy), 2, 2, 8);
	rcol = CColor(0, 0, 0, 0xff);
	cxStipplePic = XRenderCreatePicture(cxDpy, pix, cxAlphaFormat, CPRepeat, &pa);
	XRenderFillRectangle(cxDpy, PictOpSrc, cxStipplePic, &rcol, 0, 0, 2, 2);
	rcol = CColor(0, 0, 0, 0);
	XRenderFillRectangle(cxDpy, PictOpSrc, cxStipplePic, &rcol, 0, 1, 1, 1);
	XRenderFillRectangle(cxDpy, PictOpSrc, cxStipplePic, &rcol, 1, 0, 1, 1);
	XFreePixmap(cxDpy, pix);
	// set profile paths
	struct stat st;
	if (stat(cxxxdir.c_str(), &st) != 0) {
		if (errno == ENOENT) {
			if (mkdir(cxxxdir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
				cxError << _("Could not create C++X configuration directory\n{1}.\n{2}", 10) << cxxxdir << strerror(errno) << CError::Fatal;
			}
			if (stat(cxxxdir.c_str(), &st)) {
				cxError << _("Could not read status of newly created\nC++X configuration directory\n{1}.\n{2}", 11) << cxxxdir << strerror(errno) << CError::Fatal;
			}
		} else cxError << _("Could not access {1}.\n{2}", 12) << cxxxdir << strerror(errno) << CError::Fatal;
	}
	if (!S_ISDIR(st.st_mode)) {
		string backup = cxxxdir;
		backup += ".bak";
		cxError << _("File {1} exists. It will be renamed to {2}", 13) << cxxxdir << backup << CError::Warn;
		if (rename(cxxxdir.c_str(), backup.c_str())) cxError << _("Could not rename {1} to {2}", 14) << cxxxdir << backup << CError::Fatal;
		if (mkdir(cxxxdir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) cxError << _("Could not create C++X configuration directory.\n{1}", 15) << strerror(errno) << CError::Fatal;
	}
	// read global profile
	if (access(cxxxconfigfile.c_str(), R_OK | W_OK) != 0) {
		if (errno != ENOENT) cxError << _("Could not access C++X configuration file {1}\n{2}", 16) << cxxxconfigfile << strerror(errno) << CError::Fatal;
		else {
			ofstream ost(cxxxconfigfile.c_str());
			if (!ost) cxError << _("Could not create C++X configuration file {1}\n{2}", 17) << cxxxconfigfile << strerror(errno) << CError::Fatal;
		}
	}
	mGlobalProfile.ReadFile(cxxxconfigfile);
	cxxx::lib::cxProfile = &mGlobalProfile;
	// read application profile
	if (!mAppname.empty()) {
		if (access(appconfigfile.c_str(), R_OK | W_OK) != 0) {
			if (errno != ENOENT) {
				cxError << _("Could not access application configuration file {1}\n{2}", 457) << appconfigfile << strerror(errno) << CError::Error;
			} else {
				ofstream ost(appconfigfile.c_str());
				if (!ost) {
					cxError << _("Could not create application configuration file {1}\n{2}", 455) << appconfigfile << strerror(errno) << CError::Error;
				}
			}
		}
		mAppProfile.ReadFile(appconfigfile);
	}
	cxxx::cxProfile = &mAppProfile;
	// load icon theme
	if (cxSysIconDir.empty()) cxSysIconDir = DATADIR "/cxxx/icons";
	string icontheme = mGlobalProfile.Read("Icon Theme", "C++X");
	CIcon::Init(icontheme);
	// initialize the application
	mDblClickTime = mGlobalProfile.Read("Doubleclick Time", 400);
	unsigned tipdelay = mGlobalProfile.Read("Tooltip Delay", 1000);
	CToolButton::SetDelay(tipdelay);
	// fonts
	string stdfont = mGlobalProfile.Read("Standard Font", "-adobe-helvetica-medium-r-*-*-12-*-*-*-*-*-iso8859-1");
 	string titlefont = mGlobalProfile.Read("Title Font", stdfont);
	if (!mTextFont.LoadFont(stdfont)) {
		cxError << _("Failed to load text font\n{1}", 18) << stdfont << CError::Fatal;
	}
	cxTextFont = &mTextFont;
	if(!mTitleFont.LoadFont(titlefont)) {
		cxError << _("Failed to load title font\n{1}", 19) << titlefont << CError::Error;
		cxTitleFont = cxTextFont;
	} else cxTitleFont = &mTitleFont;
	// colors
	unsigned col = mGlobalProfile.Read("Frame Color", 0xffc0c0c0);
	CColor ccol(0, 0, 0, 0xff);
	ccol.SetColor(TC_BLACK);
	ccol = CColor(0xff, 0xff, 0xff, 0xff);
	ccol.SetColor(TC_WHITE);
	ccol = CColor((col >> 16) & 0xff, (col >> 8) & 0xff, col & 0xff, (col >> 24) & 0xff);
	ccol.SetColor(TC_FRAME);
	col = mGlobalProfile.Read("Title Color", 0xff000000);
	ccol = CColor((col >> 16) & 0xff, (col >> 8) & 0xff, col & 0xff, (col >> 24) & 0xff);
	ccol.SetColor(TC_TITLE);
	col = mGlobalProfile.Read("Tooltip Color", 0xffffffff);
	ccol = CColor((col >> 16) & 0xff, (col >> 8) & 0xff, col & 0xff, (col >> 24) & 0xff);
	ccol.SetColor(TC_TOOLTIP);
	col = mGlobalProfile.Read("Background Color", 0xffffffff);
	ccol = CColor((col >> 16) & 0xff, (col >> 8) & 0xff, col & 0xff, (col >> 24) & 0xff);
	ccol.SetColor(TC_BACKGND);
	// initialize managers
	cxxxdir = mGlobalProfile.Read("C++X Directory", cxxxdir);
	if (!mFileManager.Init(cxxxdir)) cxError << _("Failed to initialize file handler.", 451) << CError::Fatal;
	cxFileManager = &mFileManager;
	if (!mSelectionManager.Init(cxxxdir)) cxError << _("Failed to initialize selection handler.", 20) << CError::Fatal;
	cxSelManager = &mSelectionManager;
	cxDnDManager = &mDnDManager;
#ifdef MAXHOSTNAMELEN
	char host [MAXHOSTNAMELEN];
#else
	char host [1024];
#endif // MAXHOSTNAMELEN
	if (gethostname(host, sizeof(host)) == 0) {
		mHostname = host;
	}
	mUsername = getpwuid(getuid())->pw_name;
	XSetErrorHandler(&cxxx::cxXError);
	// Get Startup ID
	const char* startup = getenv("DESKTOP_STARTUP_ID");
	if (startup != NULL) {
		cxStartupID = startup;
		unsetenv("DESKTOP_STARTUP_ID");
	}
}

// assignment operator
void CApplication::operator=(const CApplication& app)
{
	if (&app != this) {
		mDblClickTime = app.mDblClickTime;
		mModalWindow = app.mModalWindow;
		mPopup = app.mPopup;
		mShutDown = app.mShutDown;
		mTimeOut = app.mTimeOut;
		mErrorText = app.mErrorText;
		mIdleTarget = app.mIdleTarget;
		mUsername = app.mUsername;
		mHostname = app.mHostname;
		mAppname = app.mAppname;
		mGlobalProfile = app.mGlobalProfile;
		mAppProfile = app.mAppProfile;
		mTextFont = app.mTextFont;
		mTitleFont = app.mTitleFont;
		mFileManager = app.mFileManager;
		mSelectionManager = app.mSelectionManager;
		mDnDManager = app.mDnDManager;
	}
	cxError << "CApplication operator=() must not be used!" << CError::Fatal;
}


// destructor
CApplication::~CApplication()
{
	cxAssert(mPopup == NULL);
	// close connection to X-server
	cxAssert(cxMainWin == NULL);
	XRenderFreePicture(cxDpy, cxDefaultPic);
	XRenderFreePicture(cxDpy, cxGrayPic);
	XRenderFreePicture(cxDpy, cxStipplePic);
	XFreeGC(cxDpy, cxxx::cxGC32);
	XCloseDisplay(cxDpy);
	cxDpy = NULL;
	mGlobalProfile.Flush();
	if (cxxx::cxBidi) delete cxxx::cxBidi;
}

// functions
int CApplication::Run()
{
	if (cxMainWin == NULL) {
		cxError << _("Main Window does not exist on application startup.", 446) << CError::Fatal;
	}
	// a timeout is needed for the timers
	gettimeofday(&mTimeOut, NULL);
	// main loop
	while (!mShutDown) MainLoop();
	// end of the program
	if (mPopup != NULL) delete mPopup;
	mPopup = NULL;
	return 0;
}

void CApplication::MainLoop()
{
	// the application sleeps most of the time, only X events awake it
	// create a file descriptor and set it to zero
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(ConnectionNumber(cxDpy), &fds);
	// sleep while there is nothing to do
	if (!EventPending()) select(FD_SETSIZE, &fds, NULL, NULL, &mTimeOut);
	// handle events + messages
	if (EventPending()) DispatchEvent();
	// handle timers
	HandleTimers();
	// send an idle message
	HandleIdle();
}

/* Handle Idle mainly serves to update the state of the application. It is called once when
the application gets idle, that is, there are no pending X events, no enqueued messages and
no expired timers. However, this function is triggered by a X or timer event (Messages can be
considered responses of these events). A timer event is generated at least every minute,
assuring regular calls of this function. */

/* This function returns immediately if an X event is pending. An index to the next idle task
is stored in case of such an interruption. That way circular calling of all idle tasks is
achieved even if the system is loaded heavily. No priority is given to a particular task. */

void CApplication::HandleIdle()
{
	static unsigned Current;
	if (Current >= mIdleTarget.size()) Current = 0;
	if (mIdleTarget.empty()) {
		return;
	}
	unsigned end = Current;
	while (!EventPending()) {
		++Current;
		if (Current >= mIdleTarget.size()) Current = 0;
		mIdleTarget[Current]->OnMessage(IDAPP, MT_IDLE, 0);
		if (Current == end) return;
	}
}

void CApplication::HandleEvent(XEvent& e)
{
	if (e.xany.type == PropertyNotify) {
		cxUpdateServerTime(e.xproperty.time);
		cxSelManager->HandleProperty(e.xproperty);
	}
	if (e.xany.window == None) {
		if(e.xclient.type == CxxxMessage) OnMessage(e.xclient.data.l[0], (TMessage)e.xclient.message_type, e.xclient.data.l[1]);
		return;
	}
	// the window that is to receive the event
	CWindow* w = CWindow::GetWindowByID(e.xany.window);
	// not found: we discard this one
	if (w == NULL) return ;
	// now we analyse the event
	bool discard = true;
	switch (e.xany.type) {
	case ButtonPress:
		cxServerTime = e.xbutton.time;
		// mouse click: find out whether its a double click
		HandleButtonPress(e.xbutton);
		// events trapped in modal mode
		if (mPopup) {
			if (w->IsPopup() || (w->GetParent() && w->GetParent()->IsPopup())) {
				w->HandleEvent(e);
				return ;
			} else {
				mPopup->GetTopLevelWindow()->SetFocusWindow(NULL);
				SetPopupWindow(NULL);
			}
		}
		break;
	case KeyPress:
		cxUpdateServerTime(e.xkey.time);
		break;
	case ButtonRelease:
		cxUpdateServerTime(e.xbutton.time);
		break;
	case KeyRelease:
		cxUpdateServerTime(e.xkey.time);
		break;
	case MotionNotify:
		cxUpdateServerTime(e.xmotion.time);
		break;
	case SelectionRequest:
		cxUpdateServerTime(e.xselectionrequest.time);
		cxSelManager->HandleRequest(e.xselectionrequest);
		return;
	case SelectionClear:
		cxUpdateServerTime(e.xselectionclear.time);
		cxSelManager->HandleClear(e.xselectionclear);
		return;
	case SelectionNotify:
		cxUpdateServerTime(e.xselection.time);
		cxSelManager->HandleNotify(e.xselection);
		return;
	case ClientMessage:
		break;
	case CxxxMessage:
		break;
	default:
		discard = false;
		break;
	}
	if (discard && mModalWindow != NULL) {
		// if there is a modal window all events to other windows
		// than the modal window and its childs are discarded
		if (w->GetTopLevelWindow() != mModalWindow) return;
	}
	// the dnd manager is interested in motion events
	if ((e.xany.type == MotionNotify) && cxDnDManager->OnMotion(*w, e.xmotion)) return;
	// let CWindow take control
	w->HandleEvent(e);
}

void CApplication::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
	// avoid recursion
	static bool errorbox;
	switch (msgid) {
	case IDAPP:
		switch (type) {
		case MT_ERROR:
			if (errorbox) {
				cerr << "C++X:" << mErrorText << endl;
			} else {
				errorbox = true;
				CMessageBox mb(mErrorText, "C++X", CMessageBox::OK, CIcon::CRITICAL);
				mb.FocusButton(CMessageBox::OK);
				mb.DoModal();
				errorbox = false;
			}
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

void CApplication::HandleTimers()
{
	CTimer::Refresh(mTimeOut);
}

void CApplication::HandleButtonPress(XButtonEvent& e)
{
	// these static variables hold the information about the last mouse click
	static int x;
	static int y;
	static unsigned button;
	static Time t;

	// do not handle send events
	if (e.send_event) {
		return;
	}
	// same place as before, time within double click time, same button?
	if ((abs(x - e.x_root) < 4) && (abs(y - e.y_root) < 4) && ((e.time - t) < mDblClickTime) && (e.button == button)) {
		// double click, we change the event type
		e.type = ButtonDblClick;
		t = 0;
		return ;
	}
	// single click, store data
	x = e.x_root;
	y = e.y_root;
	button = e.button;
	t = e.time;
}

void CApplication::SetPopupWindow(CWindow* w)
{
	// if there is already a popup window and w is not a child of it, it must be deleted
	// prior to setting a new one. CWindow::GetTopLevelWindow() refuses to work here because
	// popup windows are reparented.
	if (mPopup != NULL) {
		if ((w != NULL) && w->IsChildOf(mPopup)) return;
		else delete mPopup;
	}
	mPopup = w;
}

CTopLevelWindow* CApplication::SetModalWindow(CTopLevelWindow* w)
{
	CTopLevelWindow* tmp = mModalWindow;
	mModalWindow = w;
	// the return value is needed for proper nesting of multiple modal windows.
	// a modal window must store the return value and call SetModalWindow with it on exiting
	// modal mode
	return tmp;
}

void CApplication::Error(const string& text, bool fatal)
{
	// building the error string
	if (fatal) mErrorText = _("Fatal", 21);
	else mErrorText = _("Error", 22);
	mErrorText += ":\n";
	mErrorText += text;
	// if the main window already exists it is possible to use a message box
	if ((cxMainWin != NULL) && cxMainWin->IsVisible()) {
		// fatal error: must not return
		if (fatal) {
			CMessageBox mb(mErrorText, "C++X", CMessageBox::OK, CIcon::CRITICAL);
			mb.DoModal();
			exit(EXIT_FAILURE);
		} else {
			// non fatal error: return immediately while processing the error later in
			// the main loop
			PostMessage(MT_ERROR, 0);
        	}
	} else  {
		// the application is not sufficiently initialized, we have to use stderr
		cerr << "C++X: " << mErrorText << endl;
		if (fatal) exit(EXIT_FAILURE);
	}
}

int cxxx::cxXError(Display* d, XErrorEvent* e)
{
	char buf[1024];
	XGetErrorText(d, e->error_code, buf, 1024);
	cxError << _("X Server Error:\n{1}", 23) << buf << CError::Error;
	cxDnDManager->CancelDrag();
	return 0;
}
