/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CApplication.h
  Info   : $Id: CApplication.h,v 1.26 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CApplication_h
#define CApplication_h

// C++X headers
#include <cxxx/CDnDManager.h>
#include <cxxx/CFileHandler.h>
#include <cxxx/CFont.h>
#include <cxxx/CProfile.h>
#include <cxxx/CSelection.h>

// Forward declarations
class CTopLevelWindow;
class CTimer;
class CWindow;

// Linkage compatibility
namespace cxxx {
	extern "C" {
		int cxXError(Display*, XErrorEvent*);
	}
}

class CApplication: public CMessageTransceiver
{
	
public:
	// Message IDs
	enum
	{
		IDAPP = cxxx::msgid::MSGID_CApplication,
		IDDND,
		IDSEL
	};

typedef std::vector<CMessageTransceiver*> TIdleTargetCont;

private:
	// assignment and copy constructor are private to avoid
	// accidential use of them
	void operator=(const CApplication& app);

	CApplication(const CApplication& app)
	: CMessageTransceiver(NULL, 0)
	{
		*this = app;
	}

public:
	// Constructor
	CApplication(const std::string& name = "");

	// Destructor
	virtual ~CApplication();

	// currently used for error handling
	virtual void OnMessage(unsigned msgid, TMessage type, unsigned value);
	
	// starts main loop
	int Run();

	// quits main loop
	void Quit()
	{
		mShutDown = true;
	}

	// registers a modal window. SetModalWindow(NULL) unregisters the current modal window.
	// returns the current modal window or NULL if there is none.
	CTopLevelWindow* SetModalWindow(CTopLevelWindow* w);

	// dispatches e to the window that needs it.
	void HandleEvent(XEvent& e);
	
	// handles events, timers and messages, suspends application when there's nothing to do.
	void MainLoop();

	// registers a popup window. SetPopupWindow _DELETES_ the current popup in any case.
	// therefore a popup window must _NEVER_ be created on the stack.
	void SetPopupWindow(CWindow* w);

	// returns a pointer to the popup window, or NULL if no such window exists
	CWindow* GetPopupWindow() const
	{
		return mPopup;
	}

	// error handler. If fatal is false, an error message is posted. Otherwise the application
	// exits immediatly with returncode ret. If possible msg appears in a messagebox, otherwise
	// it gets printed out to stderr.
	void Error(const std::string& text, bool fatal);

	// Idle targets receive a idle message whenever the application gets idle.
	// An idle message has a msgid of IDAPP and a type of IDLE
	// Enable / Disable receiving idle messages
	void AddIdleTarget(CMessageTransceiver& target)
	{
		mIdleTarget.push_back(&target);
	}

	void RemoveIdleTarget(CMessageTransceiver& target)
	{
		TIdleTargetCont::iterator i = std::find(mIdleTarget.begin(), mIdleTarget.end(), &target);
		if (i == mIdleTarget.end()) return;
		mIdleTarget.erase(i);
	}

	// The user name 
	const std::string& GetUsername() const
	{
		return mUsername;
	}

	// The host name
	const std::string& GetHostname() const
	{
		return mHostname;
	}

	// The name of the application - the argument given to CApplication's constructor
	const std::string& GetAppname() const
	{
		return mAppname;
	}

	// The maximum time for two mouse button clicks to be interpreted as a double click
	Time GetDoubleClickTime() const
	{
		return mDblClickTime;
	}

	void SetDoubleClickTime(Time dct)
	{
		cxAssert (dct > 200);
		mDblClickTime = dct;
	}

	// Write entry / value pair to profile file
	void WriteProfile(const std::string& entry, const std::string& value)
	{
		mGlobalProfile.Write(entry, value);
	}

	void WriteProfile(const std::string& entry, unsigned value)
	{
		mGlobalProfile.Write(entry, value);
	}

protected:
	// polls the timers.
	void HandleTimers();

	// determines whether a mouse button was double clicked. if it was, e->xany->type is set
	// to ButtonDblClick.
	void HandleButtonPress(XButtonEvent& e);

	void HandleIdle();
    
	void DispatchEvent()
	{
		XEvent e;
		if (EventsQueued()) {
			e = PopEvent();
			HandleEvent(e);
		} else {
			while (XPending(cxxx::cxDpy)) {
				XNextEvent(cxxx::cxDpy, &e);
				PushEvent(e);
			}
			if (EventsQueued()) {
				e = PopEvent();
				HandleEvent(e);
			}
		}
	}

protected:
	// double click time
	Time mDblClickTime;
	// important windows
	CTopLevelWindow* mModalWindow;
	CWindow* mPopup;
	// main loop variable
	bool mShutDown;
	// the next timeout of the timers
	timeval mTimeOut;
	// needed by error handling routines
	std::string mErrorText;
	// needed by CToolButton
	TIdleTargetCont mIdleTarget;
	std::string mUsername;
	std::string mHostname;
	std::string mAppname;
	CProfile mGlobalProfile;
	CProfile mAppProfile;
	CFont mTextFont;
	CFont mTitleFont;
	CFileHandler mFileManager;
	CSelection mSelectionManager;
	CDnDManager mDnDManager;
};

#endif // CApplication_h
