/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CBZStreambuf.cpp
  Info   : $Id: CBZStreambuf.cpp,v 1.3 2012/06/26 08:05:52 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CBZStreambuf.h>

#ifdef CXXX_HAS_BZ2

// C++ headers
#include <cstdio>
#include <cstring>
#include <algorithm>

// Namespace declarations
using std::streambuf;
using std::streamoff;
using std::streampos;
using std::streamsize;
using std::ios_base;
using std::min;
using std::copy;
using std::memcmp;

CBZStreambuf::CBZStreambuf(std::streambuf& in, unsigned size)
: mIn(in),
  mBufferSize(size)
{
	mState = UNINITIALIZED;
	mInBuffer = new char[mBufferSize];
	mOutBuffer = new char[mBufferSize];
	if (!DecompressInit()) DecompressEnd();
}

/* virtual */ CBZStreambuf::~CBZStreambuf()
{
	DecompressEnd();
	delete[] mInBuffer;
	delete[] mOutBuffer;
}

/* virtual */ int CBZStreambuf::overflow(int c __attribute__ ((unused)))
{
	// write function disabled
	return EOF;
}

/* virtual */ streampos CBZStreambuf::seekoff(streamoff off, ios_base::seekdir way, ios_base::openmode which)
{
	streampos pos;
	if (way == ios_base::beg) pos = 0;
	else if (way == ios_base::cur) pos = mBZ.total_out_lo32;
	else if (way == ios_base::end) return -1; // way too expensive
	pos += off;
	return seekpos(pos, which);
}

/* virtual */ streampos CBZStreambuf::seekpos(streampos pos, ios_base::openmode which)
{
	unsigned upos = static_cast<unsigned>(pos);
	if ((which & ios_base::out) != 0) return -1;
	if ((which & ios_base::in) == 0) return -1;
	if (mBZ.total_out_lo32 >= upos) {
		// pos has already been read
		if (mBZ.total_out_lo32 - upos <= egptr() - eback()) {
			setg(eback(), egptr() - (mBZ.total_out_lo32 - upos), egptr());
			return pos;
		}
		if (!DecompressInit()) {
			DecompressEnd();
			return -1;
		}
	}
	// decompress until position
	while (mBZ.total_out_lo32 < upos) {
		if (!Decompress(upos - mBZ.total_out_lo32)) break;
	}
	if (mBZ.total_out_lo32 == upos) {
		setg(eback(), egptr(), egptr());
		return pos;
	} else return -1;
}

/* virtual */ streamsize CBZStreambuf::showmanyc()
{
	return egptr() - gptr();
}

/* virtual */ int CBZStreambuf::underflow()
{
	if (mState != GOOD) return EOF;
	if (gptr() < egptr()) return *gptr();
	while (egptr() == gptr()) {
		if (!Decompress(-1u)) break;
	}
	if (gptr() == egptr()) return EOF;
	else return *gptr() & 0xff;
}

/* virtual */ streamsize CBZStreambuf::xsgetn(char* s, streamsize n)
{
	if (mState != GOOD || n == 0) return 0;
	unsigned ret = min(showmanyc(), n);
	if (ret != 0) {
		copy(gptr(), gptr() + ret, s);
		setg(eback(), gptr() + ret, egptr());
		if (ret != n) ret += xsgetn(s + ret, n - ret);	return ret;
	}
	while (egptr() == gptr()) {
		if (!Decompress(n)) break;
	}
	ret = showmanyc();
	cxAssert(ret <= n);
	if (ret == 0) return 0;
	copy(gptr(), egptr(), s);
	setg(eback(), egptr(), egptr());
	if (ret != n) ret += xsgetn(s + ret, n - ret);
	return ret;
}

// Initializes the compression library, returns true if successfull
bool CBZStreambuf::DecompressInit()
{
	if (mState != UNINITIALIZED) DecompressEnd();
	if (mIn.pubseekpos(0, ios_base::in) == -1) {
		// failed to rewind mIn
		return false;
	}
	// set buffer pointers
	setg(mOutBuffer + 1, mOutBuffer + 1, mOutBuffer + 1);
	// initialize ZStream pointers
	mBZ.bzalloc = NULL;
	mBZ.bzfree = NULL;
	mBZ.opaque = NULL;
	mBZ.next_in = mInBuffer;
	mBZ.avail_in = 0;
	mBZ.total_out_lo32 = 0;
	mBZ.total_out_hi32 = 0;	
	mBZ.next_out = egptr();
	mBZ.avail_out = mBufferSize - 1;
	int small;
	if (mBufferSize <= CZStreambuf::SMALL) small = 1;
	else small = 0;	
	if (BZ2_bzDecompressInit(&mBZ, 0, small) == BZ_OK) {
		mState = GOOD;
		return true;
	} else return false;
}

// Decompresses a maximum of size characters, returns true if successfull
bool CBZStreambuf::Decompress(unsigned size)
{
	if (mState != GOOD) return false;
	// prepare input buffer
	if (!mBZ.avail_in) {
		mBZ.next_in = mInBuffer;
		mBZ.avail_in = mIn.sgetn(mInBuffer, mBufferSize);
	}
	// prepare output buffer
	if (!mBZ.avail_out) {
		mOutBuffer[0] = *(egptr() - 1);
		mBZ.next_out = eback();
		mBZ.avail_out = mBufferSize - 1;
		setg(eback(), eback(), eback());
	}
	// honor size parameter
	if (mBZ.avail_out > size) mBZ.avail_out = size;
	// do the actual decompression
	int ret;
	ret = BZ2_bzDecompress(&mBZ);
	if (ret != BZ_OK) {
		if (ret == BZ_STREAM_END) mState = END;
		else mState = FAULT;
	}
	// set streambuf pointers
	setg(eback(), gptr(), mBZ.next_out);
	return mState != FAULT;
}

#endif // CXXX_HAS_BZ2
