/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CBidi.cpp
  Info   : $Id: CBidi.cpp,v 1.5 2009/12/28 13:27:55 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2009 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CBidi.h>
#include <cxxx/Cxxx.h>

// C++ headers
#include <algorithm>
#include <cstring>

// Internationalization
using cxxx::lib::cxgettext;

// Static variables
unsigned CBidi::mRefCount = 0;
unsigned char** CBidi::mLookUpTable = NULL;
bool CBidi::mDisable = true;

CBidi::CBidi()
{
	if (!mRefCount) {
		cxAssert(mLookUpTable == NULL);
	}
	++mRefCount;
}

CBidi::~CBidi()
{
	if (--mRefCount == 0 ) {
		Clear();
	}
}

std::wstring CBidi::Convert(const wchar_t* s, TBidiType rundir)
{
	if (rundir == NONE) {
		rundir = GetLineDirection(s);
		if (rundir == NONE) rundir = RTL;
		else if (rundir == LTR_NUMBER) rundir = LTR;
		else if (rundir == RTL_NUMBER) rundir = RTL;
	}
	
	std::wstring res;
	TBidiType bidi;
	
	if (IsLTR(rundir)) {
		// LTR or LTR_RUN
		while (*s && *s != L'\n') {
			bidi = GetBidiType(*s);
			if (bidi == LTR) res += *s++;
			else if (IsNeutral(bidi)) {
				bidi = GetLineDirection(s);
				if (bidi == LTR || rundir == LTR) res += *s++;
				else break;
			} else {
				if (rundir == LTR) {
					std::wstring sub = Convert(s, (TBidiType)(bidi | RUN));
					res += sub;
					s += sub.size();
				} else break;
			}
		}
	} else if (IsRTL(rundir)) {
		// RTL or RTL_RUN
		while (*s && *s != L'\n') {
			bidi = GetBidiType(*s);
			if (bidi == RTL) res += *s++;
			else if (IsNeutral(bidi)) {
				wchar_t c = *s;
				if (bidi == MIRROR) DoMirror(c);
				bidi = GetLineDirection(s++);
			    if (bidi == RTL || rundir == RTL) res += c;
				else break;
			} else {
				if (rundir == RTL) {
					std::wstring sub = Convert(s, (TBidiType)(bidi | RUN));
					cxAssert(!sub.empty());
					std::reverse_copy(sub.begin(), sub.end(), std::back_inserter(res));
					s += sub.size();
				} else break;
			}
		}
		std::reverse(res.begin(), res.end());
	} else if (rundir == LTR_NUMBER) {
		// LTR number
		while (*s && *s != L'\n') {
			bidi = GetBidiType(*s);
			if( bidi == LTR_NUMBER) res += *s++;
			else if (bidi == OTHER) {
				bidi = GetBidiType(*(s + 1));
				if (bidi == LTR_NUMBER) res += *s++;
				else break;
			} else break;
		}
	} else if (rundir == RTL_NUMBER) {
		// RTL number
		while (*s && *s != L'\n') {
			bidi = GetBidiType(*s);
			if( bidi == RTL_NUMBER) res += *s++;
			else if (bidi == OTHER) {
				bidi = GetBidiType(*(s + 1));
				if (bidi == RTL_NUMBER) res += *s++;
				else break;
			} else break;
		}
		std::reverse(res.begin(), res.end());
	} else cxxx::cxError << _("Unknown bidi type {1}.", 473) << (unsigned) rundir << CError::Error;
	return res;

}
			
void CBidi::DoMirror(wchar_t& c)
{
#define MIRROR_SIZE 70
	
	static const wchar_t mirror[2 * MIRROR_SIZE] = 
	{ 
		L'\x0028', L'\x0029', L'\x003C', L'\x003E', L'\x005B', L'\x005D', L'\x007B', L'\x007D',
		L'\x00AB', L'\x00BB', L'\x2039', L'\x203A', L'\x2045', L'\x2046', L'\x2208', L'\x2209',
		L'\x220B', L'\x220C', L'\x2215', L'\x223C', L'\x223D', L'\x2243', L'\x2254', L'\x2255',
		L'\x2264', L'\x2265', L'\x2266', L'\x2267', L'\x226A', L'\x226B', L'\x2276', L'\x2277',
		L'\x227A', L'\x227B', L'\x2282', L'\x2283', L'\x2298', L'\x22A2', L'\x22A3', L'\x22CD',
		L'\x22DC', L'\x22DD', L'\x22DE', L'\x22DF', L'\x2308', L'\x2309', L'\x230A', L'\x230B',
		L'\x2329', L'\x232A', L'\x27E8', L'\x27E9', L'\x27EA', L'\x27EB', L'\x27EE', L'\x27EF',
		L'\x29B8', L'\x29C0', L'\x29C1', L'\x29C4', L'\x29C5', L'\x29F5', L'\x3008', L'\x3009',
		L'\x300A', L'\x300B', L'\x300C', L'\x300D', L'\x300E', L'\x300F',
		L'\x0029', L'\x0028', L'\x003E', L'\x003C', L'\x005D', L'\x005B', L'\x007D', L'\x007B',
		L'\x00BB', L'\x00AB', L'\x203A', L'\x2039', L'\x2046', L'\x2045', L'\x220B', L'\x220C',
		L'\x2208', L'\x2209', L'\x29F5', L'\x223D', L'\x223C', L'\x22CD', L'\x2255', L'\x2254',
		L'\x2265', L'\x2264', L'\x2267', L'\x2266', L'\x226B', L'\x226A', L'\x2277', L'\x2276',
		L'\x227B', L'\x227B', L'\x2283', L'\x2282', L'\x29B8', L'\x22A3', L'\x22A2', L'\x2243',
		L'\x22DD', L'\x22DC', L'\x22DF', L'\x22DE', L'\x2309', L'\x2308', L'\x230B', L'\x230A',
		L'\x232A', L'\x2329', L'\x27E9', L'\x27E8', L'\x27EB', L'\x27EA', L'\x27EF', L'\x27EE',
		L'\x2298', L'\x29C1', L'\x29C0', L'\x29C5', L'\x29C4', L'\x2215', L'\x3009', L'\x3008',
		L'\x300B', L'\x300A', L'\x300D', L'\x300C', L'\x300F', L'\x300E'
	};
	
	const wchar_t* pos = std::lower_bound(mirror, mirror + MIRROR_SIZE, c);
	cxAssert(*pos == c);
	c = *(pos + MIRROR_SIZE);
	
#undef MIRROR_SIZE
}

#ifdef SMALL_WCHAR_T
  #define BIDI_SIZE 594
#else // SMALL_WCHAR_T
  #define BIDI_SIZE 642
#endif // SMALL_WCHAR_T

static const wchar_t BidiRange[BIDI_SIZE] =
{
	L'\x000C', L'\x000D', L'\x0020', L'\x0021', L'\x0028', L'\x002A', L'\x0030', L'\x003A', L'\x003C', L'\x003D',
	L'\x003E', L'\x003F', L'\x0041', L'\x005B', L'\x005C', L'\x005D', L'\x005E', L'\x0061', L'\x007B', L'\x007C',
	L'\x007D', L'\x007E', L'\x00AA', L'\x00AB', L'\x00AC', L'\x00B2', L'\x00B4', L'\x00B5', L'\x00B6', L'\x00B9',
	L'\x00BA', L'\x00BB', L'\x00BC', L'\x00C0', L'\x00D7', L'\x00D8', L'\x00F7', L'\x00F8', L'\x02B9', L'\x02BB',
	L'\x02C2', L'\x02D0', L'\x02D2', L'\x02E0', L'\x02E5', L'\x02EE', L'\x02EF', L'\x0370', L'\x0374', L'\x0376',
	L'\x037E', L'\x0386', L'\x0387', L'\x0388', L'\x03F6', L'\x03F7', L'\x0483', L'\x048A', L'\x058A', L'\x05BE',
	L'\x05BF', L'\x05C0', L'\x05C1', L'\x05C3', L'\x05C4', L'\x05C6', L'\x05C7', L'\x05D0', L'\x0600', L'\x0606',
	L'\x0608', L'\x0609', L'\x060B', L'\x060C', L'\x060D', L'\x060E', L'\x061B', L'\x064B', L'\x0660', L'\x066A',
	L'\x066B', L'\x066D', L'\x0670', L'\x0671', L'\x06D6', L'\x06DD', L'\x06DE', L'\x06E5', L'\x06E7', L'\x06EE',
	L'\x06F0', L'\x06FA', L'\x070F', L'\x0710', L'\x0711', L'\x0712', L'\x0730', L'\x074D', L'\x07A6', L'\x07B1',
	L'\x07EB', L'\x07F4', L'\x07F6', L'\x07FA', L'\x0816', L'\x081A', L'\x081B', L'\x0824', L'\x0825', L'\x0828',
	L'\x0829', L'\x0830', L'\x0900', L'\x0903', L'\x093C', L'\x093D', L'\x0941', L'\x0949', L'\x094D', L'\x094E',
	L'\x0951', L'\x0958', L'\x0962', L'\x0964', L'\x0981', L'\x0982', L'\x09BC', L'\x09BD', L'\x09C1', L'\x09C7',
	L'\x09CD', L'\x09CE', L'\x09E2', L'\x09E6', L'\x09F2', L'\x09F4', L'\x09FB', L'\x0A03', L'\x0A3C', L'\x0A3E',
	L'\x0A41', L'\x0A59', L'\x0A70', L'\x0A72', L'\x0A75', L'\x0A83', L'\x0ABC', L'\x0ABD', L'\x0AC1', L'\x0AC9',
	L'\x0ACD', L'\x0AD0', L'\x0AE2', L'\x0AE6', L'\x0AF1', L'\x0B02', L'\x0B3C', L'\x0B3D', L'\x0B3F', L'\x0B40',
	L'\x0B41', L'\x0B47', L'\x0B4D', L'\x0B57', L'\x0B62', L'\x0B66', L'\x0B82', L'\x0B83', L'\x0BC0', L'\x0BC1',
	L'\x0BCD', L'\x0BD0', L'\x0BF3', L'\x0C01', L'\x0C3E', L'\x0C41', L'\x0C46', L'\x0C58', L'\x0C62', L'\x0C66',
	L'\x0C78', L'\x0C7F', L'\x0CBC', L'\x0CBD', L'\x0CCC', L'\x0CD5', L'\x0CE2', L'\x0CE6', L'\x0CF1', L'\x0D02',
	L'\x0D41', L'\x0D46', L'\x0D4D', L'\x0D57', L'\x0D62', L'\x0D66', L'\x0DCA', L'\x0DCF', L'\x0DD2', L'\x0DD8',
	L'\x0E31', L'\x0E32', L'\x0E34', L'\x0E40', L'\x0E47', L'\x0E4F', L'\x0EB1', L'\x0EB2', L'\x0EB4', L'\x0EBD',
	L'\x0EC8', L'\x0ED0', L'\x0F18', L'\x0F1A', L'\x0F35', L'\x0F36', L'\x0F37', L'\x0F38', L'\x0F39', L'\x0F3E',
	L'\x0F71', L'\x0F7F', L'\x0F80', L'\x0F85', L'\x0F86', L'\x0F88', L'\x0F90', L'\x0FBE', L'\x0FC6', L'\x0FC7',
	L'\x102D', L'\x1031', L'\x1032', L'\x1038', L'\x1039', L'\x103B', L'\x103D', L'\x103F', L'\x1058', L'\x105A',
	L'\x105E', L'\x1061', L'\x1071', L'\x1075', L'\x1082', L'\x1083', L'\x1085', L'\x1087', L'\x108D', L'\x108E',
	L'\x109D', L'\x109E', L'\x135F', L'\x1360', L'\x1390', L'\x13A0', L'\x1400', L'\x1401', L'\x1680', L'\x1681',
	L'\x169B', L'\x16A0', L'\x1712', L'\x1720', L'\x1732', L'\x1735', L'\x1752', L'\x1760', L'\x1772', L'\x1780',
	L'\x17B7', L'\x17BE', L'\x17C6', L'\x17C7', L'\x17C9', L'\x17D4', L'\x17DB', L'\x17DC', L'\x17DD', L'\x17E0',
	L'\x17F0', L'\x180E', L'\x1810', L'\x18A9', L'\x18AA', L'\x1920', L'\x1923', L'\x1927', L'\x1929', L'\x1932',
	L'\x1933', L'\x1939', L'\x1946', L'\x19DE', L'\x1A00', L'\x1A17', L'\x1A19', L'\x1A56', L'\x1A57', L'\x1A58',
	L'\x1A61', L'\x1A62', L'\x1A63', L'\x1A65', L'\x1A6D', L'\x1A73', L'\x1A80', L'\x1B00', L'\x1B04', L'\x1B34',
	L'\x1B35', L'\x1B36', L'\x1B3B', L'\x1B3C', L'\x1B3D', L'\x1B42', L'\x1B43', L'\x1B6B', L'\x1B74', L'\x1B80',
	L'\x1B82', L'\x1BA2', L'\x1BA6', L'\x1BA8', L'\x1BAA', L'\x1C2C', L'\x1C34', L'\x1C36', L'\x1C3B', L'\x1CD0',
	L'\x1CD3', L'\x1CD4', L'\x1CE1', L'\x1CE2', L'\x1CE9', L'\x1CED', L'\x1CEE', L'\x1DC0', L'\x1E00', L'\x1FBD',
	L'\x1FBE', L'\x1FBF', L'\x1FC2', L'\x1FCD', L'\x1FD0', L'\x1FDD', L'\x1FE0', L'\x1FED', L'\x1FF2', L'\x1FFD',
	L'\x2000', L'\x200B', L'\x200E', L'\x200F', L'\x2010', L'\x2028', L'\x2029', L'\x202A', L'\x202B', L'\x202C',
	L'\x202D', L'\x202E', L'\x202F', L'\x2039', L'\x203B', L'\x2045', L'\x2047', L'\x205F', L'\x2060', L'\x2070',
	L'\x2071', L'\x2074', L'\x207A', L'\x207F', L'\x2080', L'\x208A', L'\x2090', L'\x20A0', L'\x2102', L'\x2103',
	L'\x2107', L'\x2108', L'\x210A', L'\x2114', L'\x2115', L'\x2116', L'\x2119', L'\x211E', L'\x2124', L'\x2125',
	L'\x2126', L'\x2127', L'\x2128', L'\x2129', L'\x212A', L'\x212E', L'\x212F', L'\x213A', L'\x213C', L'\x2140',
	L'\x2145', L'\x214A', L'\x214E', L'\x2150', L'\x2160', L'\x2189', L'\x2208', L'\x220A', L'\x220B', L'\x220D',
	L'\x2215', L'\x2216', L'\x223C', L'\x223E', L'\x2243', L'\x2244', L'\x2254', L'\x2256', L'\x2264', L'\x2268',
	L'\x226A', L'\x226C', L'\x2276', L'\x2278', L'\x227A', L'\x227C', L'\x2282', L'\x2284', L'\x2298', L'\x2299',
	L'\x22A2', L'\x22A4', L'\x22CD', L'\x22CE', L'\x22DC', L'\x22E0', L'\x2308', L'\x230C', L'\x2329', L'\x232B',
	L'\x2336', L'\x237B', L'\x2395', L'\x2396', L'\x2488', L'\x249C', L'\x24EA', L'\x26AC', L'\x26AD', L'\x27E8',
	L'\x27EC', L'\x27EE', L'\x27F0', L'\x2800', L'\x2900', L'\x29B8', L'\x29B9', L'\x29C0', L'\x29C2', L'\x29C4',
	L'\x29C6', L'\x29F5', L'\x29F6', L'\x2C00', L'\x2CE5', L'\x2CEB', L'\x2CEF', L'\x2D00', L'\x2DE0', L'\x3000',
	L'\x3001', L'\x3005', L'\x3008', L'\x3010', L'\x3021', L'\x302A', L'\x3031', L'\x3036', L'\x3038', L'\x303D',
	L'\x3041', L'\x3099', L'\x309D', L'\x30A0', L'\x30A1', L'\x30FB', L'\x30FC', L'\x31C0', L'\x31F0', L'\x321D',
	L'\x3220', L'\x3250', L'\x3260', L'\x327C', L'\x327F', L'\x32B1', L'\x32C0', L'\x32CC', L'\x32D0', L'\x3377',
	L'\x337B', L'\x33DE', L'\x33E0', L'\x33FF', L'\x3400', L'\x4DC0', L'\x4E00', L'\xA490', L'\xA4D0', L'\xA60D',
	L'\xA610', L'\xA66F', L'\xA680', L'\xA6F0', L'\xA6F2', L'\xA700', L'\xA722', L'\xA788', L'\xA789', L'\xA802',
	L'\xA803', L'\xA806', L'\xA807', L'\xA80B', L'\xA80C', L'\xA825', L'\xA827', L'\xA828', L'\xA830', L'\xA838',
	L'\xA840', L'\xA874', L'\xA880', L'\xA8C4', L'\xA8CE', L'\xA8E0', L'\xA8F2', L'\xA926', L'\xA92E', L'\xA947',
	L'\xA952', L'\xA980', L'\xA983', L'\xA9B3', L'\xA9B4', L'\xA9B6', L'\xA9BA', L'\xA9BC', L'\xA9BD', L'\xAA29',
	L'\xAA2F', L'\xAA31', L'\xAA33', L'\xAA35', L'\xAA40', L'\xAA43', L'\xAA44', L'\xAA4C', L'\xAA4D', L'\xAAB0',
	L'\xAAB1', L'\xAAB2', L'\xAAB5', L'\xAAB7', L'\xAAB9', L'\xAABE', L'\xAAC0', L'\xAAC1', L'\xAAC2', L'\xABE5',
	L'\xABE6', L'\xABE8', L'\xABE9', L'\xABED', L'\xABF0', L'\xFB1D', L'\xFB1E', L'\xFB1F', L'\xFB29', L'\xFB2A',
	L'\xFD3E', L'\xFD50', L'\xFDFD', L'\xFE70', L'\xFEFF', L'\xFF10', L'\xFF1A', L'\xFF21', L'\xFF3B', L'\xFF41',
	L'\xFF5B', L'\xFF66', L'\xFFE0',
#ifdef SMALL_WCHAR_T
	L'\xFFFF'
#else // SMALL_WCHAR_T
    L'\x10000', L'\x10101', L'\x10102', L'\x10140', L'\x101D0', L'\x101FD',
	L'\x10280', L'\x10800', L'\x1091F', L'\x10920', L'\x10A01', L'\x10A10', L'\x10A38', L'\x10A40', L'\x10B39',
	L'\x10B40', L'\x10E60', L'\x11080', L'\x11082', L'\x110B3', L'\x110B7', L'\x110B9', L'\x110BB', L'\x1D167',
	L'\x1D16A', L'\x1D173', L'\x1D183', L'\x1D185', L'\x1D18C', L'\x1D1AA', L'\x1D1AE', L'\x1D200', L'\x1D360',
	L'\x1D6DB', L'\x1D6DC', L'\x1D715', L'\x1D716', L'\x1D74F', L'\x1D750', L'\x1D789', L'\x1D78A', L'\x1D7C3',
	L'\x1D7C4', L'\x1D7CE', L'\x1F000', L'\x1F100',	L'\x1F110', L'\xE0001', wchar_t(-1u)
#endif // SMALL_WCHAR_T
};

static const CBidi::TBidiType BidiType[BIDI_SIZE + 1] =
{
	CBidi::OTHER, CBidi::SPACE, CBidi::OTHER, CBidi::SPACE, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::LTR_NUMBER,
	CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::MIRROR, CBidi::OTHER,
	CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::LTR,
	CBidi::MIRROR, CBidi::OTHER, CBidi::LTR_NUMBER, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR_NUMBER,
	CBidi::LTR, CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL,
	CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::RTL_NUMBER, CBidi::OTHER,
	CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER,
	CBidi::RTL_NUMBER, CBidi::OTHER, CBidi::RTL_NUMBER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER,
	CBidi::RTL_NUMBER, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::LTR_NUMBER, CBidi::RTL,
	CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL,
	CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL,
	CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::SPACE, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::SPACE, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::SPACE, CBidi::OTHER, CBidi::LTR, CBidi::RTL, CBidi::OTHER, CBidi::SPACE,
	CBidi::OTHER, CBidi::LTR, CBidi::RTL, CBidi::OTHER, CBidi::LTR, CBidi::RTL, CBidi::OTHER, CBidi::MIRROR,
	CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::SPACE, CBidi::OTHER, CBidi::LTR_NUMBER, CBidi::LTR,
	CBidi::LTR_NUMBER, CBidi::OTHER, CBidi::LTR, CBidi::LTR_NUMBER, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER,
	CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR,
	CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER,
	CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR,
	CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER,
	CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR_NUMBER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER, CBidi::MIRROR, CBidi::OTHER,
	CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::SPACE, CBidi::OTHER, CBidi::LTR, CBidi::MIRROR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR,
	CBidi::OTHER, CBidi::LTR, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER,
	CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::LTR_NUMBER, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR
#ifndef SMALL_WCHAR_T
	, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL,
	CBidi::OTHER, CBidi::RTL, CBidi::OTHER, CBidi::RTL, CBidi::RTL_NUMBER, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER,
	CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::OTHER, CBidi::LTR, CBidi::LTR_NUMBER,
	CBidi::OTHER, CBidi::LTR_NUMBER, CBidi::LTR, CBidi::OTHER, CBidi::OTHER
#endif // SMALL_WCHAR_T
};

void CBidi::CreateTable(wchar_t c)
{
	cxAssert(sizeof(BidiRange) / sizeof(wchar_t) == BIDI_SIZE);
	cxAssert(sizeof(BidiType) / sizeof(TBidiType) == BIDI_SIZE + 1);
	unsigned page = (unsigned)c >> 8;
	if (mLookUpTable == NULL) {
		mLookUpTable = new unsigned char*[0x80];
		for (unsigned i = 0; i < 0x80; ++i) mLookUpTable[i] = NULL;
	}
	mLookUpTable[page] = new unsigned char[0x100];
	const wchar_t* res = std::upper_bound(BidiRange, BidiRange + BIDI_SIZE, c & 0x7f00);
	TBidiType t = BidiType[res - BidiRange];
	unsigned bound = (unsigned)(*res) - page * 256;
	for (unsigned i = 0; i < 0x100; ++i) {
		if (i == bound) {
			t = BidiType[++res - BidiRange];
			bound = (unsigned)(*res) - page * 256;
		}
		(mLookUpTable[page])[i] = (unsigned char)t;
	}
}

CBidi::TBidiType CBidi::CalculateBidiType(wchar_t c) const
{
	const wchar_t* res = std::upper_bound(BidiRange, BidiRange + BIDI_SIZE, c);
	return BidiType[res - BidiRange];
}

#undef BIDI_SIZE

CBidi::TBidiType CBidi::GetLineDirection(const wchar_t* c)
{
	cxAssert(cxxx::cxRTL);
	TBidiType typ = NONE;
	TBidiType tmp;
	while (*c && *c != L'\n') {
		tmp = GetBidiType(*c);
		if (tmp == LTR) return LTR;
		if (tmp == RTL) return RTL;
		if (typ == NONE) {
			if (tmp == LTR_NUMBER) typ = LTR_NUMBER;
			if (tmp == RTL_NUMBER) typ = RTL_NUMBER;
		}			
		++c;
	}
	return typ;
}

void CBidi::Clear()
{
	if (mLookUpTable != NULL) {
		for (unsigned i = 0; i < 0x80; ++i) {
			if (mLookUpTable[i] != NULL) delete[] mLookUpTable[i];
		}
		delete[] mLookUpTable;
	}
}
