/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CBidi.h
  Info   : $Id: CBidi.h,v 1.4 2009/12/28 13:27:55 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2009 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CBidi_h
#define CBidi_h

// C++X headers
#include <cxxx/Cxxx.h>

// C++ headers
#include <string>

class CBidi
{
public:
	typedef enum
	{
		NONE		= 0x00,
		OTHER		= 0x01,
		SPACE		= 0x02,
		MIRROR		= 0x04,
		NEUTRAL		= 0x07,
		RUN			= 0x08,
		LTR			= 0x10,
		LTR_RUN		= 0x18,
		RTL			= 0x20,
		RTL_RUN		= 0x28,
		LTR_NUMBER	= 0x48,
		RTL_NUMBER  = 0x88
	} TBidiType;
		
public:
	// Constructor
	CBidi();
	
	// Destructor
	~CBidi();
	
	// Converts a wide string
	const std::wstring Convert(const std::wstring& s);
			
	// Looks for c in the tables and creates table entry, if necessary
	TBidiType GetBidiType(wchar_t c);
		
	// Guesses the direction of a line
	TBidiType GetLineDirection(const wchar_t* c);

private:
	// Calculate the bidi type of c
	TBidiType CalculateBidiType(wchar_t c) const;

	// String concatenation
	void LeftAppend(std::wstring& s, const wchar_t c);
	void LeftAppend(std::wstring& s, const std::wstring& c);
	
	// Helper functions
	bool IsLTR(TBidiType t);
	bool IsRTL(TBidiType t);
	bool IsNeutral(TBidiType t);
	
	// Mirror one character
	void DoMirror(wchar_t& c);
	
	// Converts a wide C-string
	std::wstring Convert(const wchar_t* s, TBidiType rundir);
	
	// Creates a table entry for c
	void CreateTable(wchar_t c);
		
	// Clears the lookup table
	void Clear();
private:
	// Reference counter
	static unsigned mRefCount;
	
	// Lookup table
	static unsigned char** mLookUpTable;
	
	// Switch
	static bool mDisable;
};

inline bool CBidi::IsLTR(TBidiType t)
{
	return (t & LTR);
}

inline bool CBidi::IsRTL(TBidiType t)
{
	return (t & RTL);
}

inline bool CBidi::IsNeutral(TBidiType t)
{
	return (t & NEUTRAL);
}

inline const std::wstring CBidi::Convert(const std::wstring& s)
{
	cxAssert(cxxx::cxRTL);
	std::wstring res;
	const wchar_t* tmp = s.c_str();
	while (*tmp) {
		res += Convert(tmp, NONE);
		tmp = &s[0] + res.size();
		if (*tmp) res += *tmp++;
	}
	return res;
}

inline CBidi::TBidiType CBidi::GetBidiType(wchar_t c)
{
	cxAssert(cxxx::cxRTL);
	if (c > L'\x7fff') return CalculateBidiType(c);
	unsigned page = (unsigned)c >> 8;
	unsigned idx = (unsigned)c & 0xff;
	if (mLookUpTable == NULL || mLookUpTable[page] == NULL) CreateTable(c);
	return (TBidiType) (mLookUpTable[page])[idx];
}

#endif // CBidi_h
