/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CControl.h
  Info   : $Id: CControl.h,v 1.32 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2004 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CControl_h
#define CControl_h

// cxxx headers
#include <cxxx/CDnDWindow.h>
#include <cxxx/CScrollBar.h>
#include <cxxx/CControlItem.h>

class CControl : public CDnDWindow
{
	enum
	{
		// Message IDs
		IDHSCROLL = cxxx::msgid::MSGID_CControl,
		IDVSCROLL,
		IDVIEW,
		IDTIMER
	};

protected:
	enum
	{
		// Geometry
		ICON_WIDTH = 20,
		ICON_HEIGHT = 20,
		ICON_BORDER = 2,
		SCROLL_STEP = 20,
		BORDER_WIDTH = 2
	};

public:
	// Constructor
	CControl(CWindow* parent, unsigned msgid);

	// X event processing
	virtual void OnConfigure(XConfigureEvent& e);
	virtual void OnFocus(bool in);
	virtual void OnDblClick(XButtonEvent& e);
	virtual void OnButtonUp(XButtonEvent& e);
	virtual void OnMouseWheelUp(XButtonEvent& e);
 	virtual void OnMouseWheelDown(XButtonEvent& e);
	virtual void OnKey(wchar_t key, unsigned state);
	virtual void OnContext();
 	virtual void Recolor();

	// C++X message processing
	virtual void OnMessage(unsigned msgid, TMessage type, unsigned value);

	// Reload content and display it
	void Update()
	{
		// force calculation of the document
		mDocHeight = 0;
		// do the work
		Layout();
		//Redraw();
	}

	// Delete content
	void ClearList()
	{
		mList.clear();
		mSelection.clear();
	}
	
	// Get item count
	unsigned GetItemCount() const
	{
		return mList.size();
	}
	
	// Get count of items in selection
	unsigned GetSelCount() const
	{
		return mSelection.size();
	}
	
	// Select all items
	void SelectAll()
	{
		cxAssert(mMultipleSelection);
		for (unsigned i = 0; i < GetItemCount(); ++i) {
			if (!IsSelected(i)) {
				mList[i].Select(true);
				mSelection.push_back(i);
				DrawItem(i);
			}
		}
		DrawFrame();
	}
	
	// Select nothing
	void DeselectAll()
	{
		if (mSelection.empty()) return ;
		for (std::vector<unsigned>::iterator i = mSelection.begin(); i != mSelection.end(); ++i) {
			mList[*i].Select(false);
			DrawItem(*i);
		}
		mSelection.clear();
		DrawFrame();
	}
	
	// Select anything that is not selected, and vice versa
	void InvertSelection()
	{
		cxAssert(mMultipleSelection);
		mSelection.clear();
		for (unsigned i = 0; i < GetItemCount(); i++) {
			if (!IsSelected(i)) {
				mList[i].Select(true);
				mSelection.push_back(i);
			} else mList[i].Select(false);
			DrawItem(i);
		}
		DrawFrame();
	}

	// Returns true if more than one item at a time can be selected
	bool CanMultipleSelect() const
	{
		return mMultipleSelection;
	}
	
	// Returns selection as an array of indices into item list
	const std::vector<unsigned>& GetSelection() const
	{
		return mSelection;
	}

	
	// Select item with index idx
	void Select(unsigned idx)
	{
		cxAssert(idx < mList.size());
		if (IsSelected(idx)) return ;
		mList[idx].Select(true);
		mSelection.push_back(idx);
		mScrollTimer.SetTime(500);
		DrawItem(idx);
		DrawFrame();
	}

	//
	void SetIconSize(unsigned w, unsigned h)
	{
		mIconWidth = w + 2 * ICON_BORDER;
		mIconHeight = h + 2 * ICON_BORDER;
		if (w < h) mIconSize = w;
		else mIconSize = h;
	}
	
	// Returns true if item is selected
	bool IsSelected(unsigned idx)
	{
		cxAssert(idx < mList.size());
		return mList[idx].IsSelected();
	}
	
protected:
	// Calculate width and height of the document
	virtual void CalculateDocSize() = 0;

	// Returns the index of the item at screen position x, y or -1u if no such item exists
	virtual unsigned ItemFromPos(int x, int y) = 0;

	// Draws item with index idx
	virtual void DrawItem(unsigned idx) = 0;

	// Scrolls item with index idx into view
	virtual void ScrollIn(unsigned idx) = 0;

	// Calculates document size and sets scrollbars accordingly
	void Layout();

	// Scroll window content by x, y pixels
	void Scroll(int x, int y);

	// Select all items with indices in the range from lo to hi, inclusive
	void Select(unsigned lo, unsigned hi);

	// Deselect item with index idx
	void Deselect(unsigned idx)
	{
		cxAssert(idx < mList.size());
		if (!IsSelected(idx)) return;
		mList[idx].Select(false);
		std::vector<unsigned>::iterator it = std::find(mSelection.begin(), mSelection.end(), idx);
		if (it !=  mSelection.end()) mSelection.erase(it);
		DrawItem(idx);
		DrawFrame();
	}
	
	// Select item with index idx if it is deselected, or deselect it if it is selected
	void ToggleSelection(unsigned idx)
	{
		cxAssert(idx < GetItemCount());
		if (IsSelected(idx)) Deselect(idx);
		else Select(idx);
	}
	
	// Transform window coordinates to document coordinates
	void ScreenToWindow(int& x, int& y) const
	{
		x += mXOffset * SCROLL_STEP - BORDER_WIDTH;
		y += mYOffset * SCROLL_STEP - BORDER_WIDTH;
	}

	// Transform document coordinates to window coordinates
	void WindowToScreen(int& x, int& y) const
	{
		x -= mXOffset * SCROLL_STEP - BORDER_WIDTH;
		y -= mYOffset * SCROLL_STEP - BORDER_WIDTH;
	}
	
	// Draw the border of the window
	void DrawFrame()
	{
		CWindow::DrawFrame(0, 0, GetWidth(), GetHeight(), FS_SUNKEN_THICK, false);
		if (mHScrollBar.IsEnabled() && mVScrollBar.IsEnabled()) FillRectangle(GetWidth() - mVScrollBar.GetWidth() - BORDER_WIDTH, GetHeight() - mHScrollBar.GetHeight() - BORDER_WIDTH, mVScrollBar.GetWidth(), mHScrollBar.GetHeight(), CColor(TC_FRAME));
	}
	
protected:
	std::vector<CControlItem> mList;
	std::vector<unsigned> mSelection;
	std::vector<unsigned> mColumn;
	std::vector<unsigned> mRow;
	CHScrollBar mHScrollBar;
	CVScrollBar mVScrollBar;
	CTimer mScrollTimer;
	bool mMultipleSelection;
	int mXOffset;
	int mYOffset;
	unsigned mLastSel;
	unsigned mDocWidth;
	unsigned mDocHeight;
	unsigned mViewWidth;
	unsigned mViewHeight;
	unsigned mIconWidth;
	unsigned mIconHeight;
	unsigned mIconSize;
};

#endif // CControl_h
