/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CCursor.cpp
  Info   : $Id: CCursor.cpp,v 1.2 2012/09/14 19:16:47 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CCursor.h>
#include <cxxx/CImage.h>

// C++ headers
#include <locale>

// C headers
#include <unistd.h>
#include <dirent.h>

// Namespace declarations
using std::string;
using std::wstring;
using std::vector;
using std::locale;
using std::ifstream;
using cxxx::cxDpy;
using cxxx::cxHomeDir;
using cxxx::cxSysIconDir;
using cxxx::cxWiden;
using cxxx::lib::cxgettext;

// static variables
bool CCursor::mInitialized = false;
vector<string> CCursor::mThemeDirs;

bool CCursor::Create(const std::string& name)
{
	if (!mInitialized) {
		Init();
		mInitialized = true;
	}
	if (DoCreate(name)) return true;
	// try alternate names
	if (name == "move") {
		if (DoCreate("alias")) return true;
		if (DoCreate("dnd-move")) return true;
	} else if (name == "copy") {
		if (DoCreate("dnd-copy")) return true;
	} else if (name == "question_arrow") {
		if (DoCreate("whats_this")) return true;
		if (DoCreate("dnd-ask")) return true;
	} else if (name == "link") {
		if (DoCreate("dnd-link")) return true;
	} else if (name == "no-drop") {
		if (DoCreate("dnd-no-drop")) return true;
		if (DoCreate("dnd-none")) return true;
		if (DoCreate("crossed_circle")) return true;
		if (DoCreate("not-allowed")) return true;
		if (DoCreate("forbidden")) return true;
	}
	// not found: use C++X Cursors
	string defname = cxSysIconDir + "/0x0/" + name;
	return DoCreate(defname);		   
}

bool CCursor::DoCreate(const string& name)
{
	cxAssert(!name.empty());
	// absolute file name, just load it!
	if (name[0] == '/') {
		CImage im;
		im.ReadFile(name);
		if (im.IsValid()) {
			mCursor = XRenderCreateCursor(cxDpy, im.GetPicture(), im.GetXHot(), im.GetYHot());
			if (mCursor != None) return true;
		}
		return false;
	}
	// this is a short name, find it
	string fn;
	for (vector<string>::iterator it = mThemeDirs.begin(); it != mThemeDirs.end(); ++it) {
		fn = *it + name;
		if (!access(fn.c_str(), R_OK) && DoCreate(fn)) return true;
	}
	return false;		
}

void CCursor::AddTheme(const string& theme)
{
	wstring ws = cxxx::cxWiden(theme);
	const std::ctype<wchar_t>& ct = std::use_facet<std::ctype<wchar_t> >(locale::classic());
	ct.tolower(&ws[0], &ws[0] + ws.size());
	string lotheme = cxxx::cxNarrow(ws);
	
	bool up = true;
	for (unsigned i = 0; i < ws.size(); ++i) {
		if (!ct.is(std::ctype_base::alnum, ws[i])) up = true;
		else if (up) {
			ws[i] = ct.toupper(ws[i]);
			up = false;
		}
	}
	string hitheme = cxxx::cxNarrow(ws);
	string dir = cxHomeDir + "/.icons";
	AddThemeDir(dir, lotheme);
	AddThemeDir(dir, hitheme);
	AddThemeDir("/usr/local/share/icons", lotheme);
	AddThemeDir("/usr/local/share/icons", hitheme);
	AddThemeDir("/usr/share/icons", lotheme);
	AddThemeDir("/usr/share/icons", hitheme);
}

void CCursor::AddThemeDir(const string& dir, const string& theme)
{
	string themepath = dir + "/" + theme;

	// themepath is not accessible: give up
	if (access(themepath.c_str(), X_OK)) return;
	
	string index = themepath + "/index.theme";
	themepath += "/cursors/";
	
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		// this directory has already been processed, nothing to do
		if (themepath == *i) return;
	}
	
	if (!access(themepath.c_str(), X_OK)) mThemeDirs.push_back(themepath);
	
	string parent;
	ifstream in;
	in.open(index.c_str());
	while (in) {
		getline(in, parent);
		if (parent.substr(0, 8) == "Inherits") {
			string::size_type pos = 8;
			while (pos < parent.size() && parent[pos] == ' ') ++pos;
			if (++pos > parent.size()) return;
			while (pos < parent.size() && parent[pos] == ' ') ++pos;
			parent = parent.substr(pos);
			while ((pos = parent.find(',')) != string::npos) {
				AddTheme(parent.substr(0, pos));
				if (++pos < parent.size()) parent = parent.substr(pos);
				else break;
			}
			AddTheme(parent);
			break;
		}
	}
}

string CCursor::ToLower(const string& s)
{
	wstring ws = cxxx::cxWiden(s);
	std::use_facet<std::ctype<wchar_t> >(locale::classic()).tolower(&ws[0], &ws[0] + ws.size());
	string ret = cxxx::cxNarrow(ws);
	return ret;
}
