/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CDnDManager.cpp
  Info   : $Id: CDnDManager.cpp,v 1.30 2012/09/14 18:31:01 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CApplication.h>
#include <cxxx/CDnDWindow.h>
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/CDnDAskDialog.h>

// X11 headers
#include <X11/Xatom.h>

// Namespace declarations
using cxxx::cxApp;
using cxxx::cxDpy;
using cxxx::cxError;
using cxxx::cxIconv;
using cxxx::cxMainWin;
using cxxx::cxSelManager;
using cxxx::lib::cxgettext;
using cxxx::cxUpdateServerTime;
using std::string;
using std::wstring;
using std::vector;

// Static members
Atom CDnDManager::DND_MOVE = None;
Atom CDnDManager::DND_COPY = None;
Atom CDnDManager::DND_LINK = None;
Atom CDnDManager::DND_ASK = None;
Atom CDnDManager::DND_PRIVATE = None;
Atom CDnDManager::DND_SELECTION = None;


CDnDManager::CDnDManager()
: CMessageTransceiver(cxApp, CApplication::IDDND),
  mInitialized(false),
  mWaitForStatus(true),
  mTargetResponds(false),
  mTypeListValid(false),
  mActionListValid(false),
  mFinalStatus(false),
  mDropRefused(false),
  mVersion(0),
  mEventMask(0),
  mTimeout(this, IDTIMEOUT),
  mDragSource(NULL),
  mDropTarget(NULL),
  mSourceXID(None),
  mTargetXID(None),
  mProxy(None),
  mInitialAction(None),
  mPerformedAction(None),
  DND_AWARE(None),
  DND_ENTER(None),
  DND_POSITION(None),
  DND_STATUS(None),
  DND_LEAVE(None),
  DND_DROP(None),
  DND_FINISHED(None),
  DND_PROXY(None),
  DND_TYPES(None),
  DND_ACTIONS(None),
  DND_DESCRIPTIONS(None)
{
}

/* virtual */ CDnDManager::~CDnDManager()
{
}

/* virtual */ void CDnDManager::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
    switch (msgid) {
	case IDTIMEOUT:
		switch (type) {
		case MT_TIMER:
			CancelDrag();
			break;
		default:
			break;
		}
	default:
		break;
	}
}

bool CDnDManager::SetupDnD(CDnDWindow& w)
{
	if (!mInitialized) {
		if (!RegisterAtoms()) {
			cxError << _("Failed to register Xdnd atoms.\nDrag \'n\' Drop not supported.", 31) << CError::Error;
			return false;
		}
		Atom type;
		int fmt;
		unsigned long count;
		unsigned long rem;
		//Atom* version = NULL;
		Window tlw = w.GetTopLevelWindow()->GetXID();
		union {unsigned char* in; Atom* version;} buf; 
		XGetWindowProperty(cxDpy, tlw, DND_AWARE, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
		if ((count == 0) || (fmt != 32) || (type != XA_ATOM) || (*buf.version !=  MAX_VERSION)) {
			type = MAX_VERSION;
			XChangeProperty(cxDpy, tlw, DND_AWARE, XA_ATOM, 32, PropModeReplace, (unsigned char*)(&type), 1);
		}
		XFree(buf.in);
		XGetWindowProperty(cxDpy, tlw, DND_AWARE, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
		if ((count == 0) || (fmt != 32) || (type != XA_ATOM) || (*buf.version !=  MAX_VERSION)) {
			// failed!
			cxError << _("Failed to set Window property.\nDrag \'n\' Drop not supported.", 32) << CError::Error;
			XFree(buf.version);
			return false;
		}
		XFree(buf.in);
		if (!CreateCursors()) cxError << _("Failed to create Drag & Drop Cursors.\nCursor may not be shown corectly", 33) << CError::Error;
		mInitialized = true;
	}
	return true;
}

bool CDnDManager::RegisterAtoms()
{
	const char* names[] = {"XdndAware", "XdndEnter", "XdndPosition", "XdndStatus", "XdndLeave", "XdndDrop",
		"XdndFinished", "XdndActionMove", "XdndActionCopy", "XdndActionLink", "XdndActionAsk",
		"XdndActionPrivate", "XdndProxy", "XdndSelection", "XdndTypeList", "XdndActionList",
		"XdndActionDescription"};
	Atom atoms[17];
	if (XInternAtoms(cxDpy, const_cast<char**>(names), 17, False, atoms) == 0) return false;
	DND_AWARE        = atoms[0];
	DND_ENTER        = atoms[1];
	DND_POSITION     = atoms[2];
	DND_STATUS       = atoms[3];
	DND_LEAVE        = atoms[4];
	DND_DROP         = atoms[5];
	DND_FINISHED     = atoms[6];
	DND_MOVE         = atoms[7];
	DND_COPY         = atoms[8];
	DND_LINK         = atoms[9];
	DND_ASK          = atoms[10];
	DND_PRIVATE      = atoms[11];
	DND_PROXY        = atoms[12];
	DND_SELECTION    = atoms[13];
	DND_TYPES        = atoms[14];
	DND_ACTIONS      = atoms[15];
	DND_DESCRIPTIONS = atoms[16];
	return true;
}

bool CDnDManager::HandleClientMessage(XClientMessageEvent& e)
{
	if (e.message_type == DND_ENTER) {
		ReceiveEnter(e);
		return true;
	}
	if (e.message_type == DND_POSITION) {
		ReceivePosition(e);
		return true;
	}
	if (e.message_type == DND_STATUS) {
		ReceiveStatus(e);
		return true;
	}
	if (e.message_type == DND_LEAVE) {
		ReceiveLeave(e);
		return true;
	}
	if (e.message_type == DND_DROP) {
		ReceiveDrop(e);
		return true;
	}
	if (e.message_type == DND_FINISHED) {
		ReceiveFinished(e);
		return true;
	}
	return false;
}

void CDnDManager::HandleMotionEvent(XMotionEvent& e)
{
	mActualPosition = e;
	if (!mTargetResponds || !mWaitForStatus) HandleNewPosition();
}

void CDnDManager::HandleNewPosition()
{
	mLastPosition = mActualPosition;
	Window proxy = mProxy;
	Window target = FindTarget(mLastPosition);
	if (target != mTargetXID) {
		mPerformedAction = None;
		UpdateCursor();
		if (mTargetXID != None) {
			Window tmp = mProxy;
			mProxy = proxy;
			SendLeave();
			mProxy = tmp;
		}
		mTargetXID = target;
		if (mTargetXID != None) {
			mSourceXID = mDragSource->GetTopLevelWindow()->GetXID();
			if (mSourceXID != mTargetXID && mInitialAction == DND_ASK && !mActionListValid) {
				if (!SetActionList()) return;
			}
			SendEnter();
			SendPosition();
		}
	} else {
		if (mTargetXID == None || !mTargetResponds) return;
		int offx = mLastPosition.x_root - mIgnoreRect.x;
		int offy = mLastPosition.y_root - mIgnoreRect.y;
		if ((offx < 0) || (offx >= mIgnoreRect.width) || (offy < 0) || (offy >= mIgnoreRect.height)) SendPosition();
	}
}


bool CDnDManager::StartDrag(CDnDWindow& w, XMotionEvent& e, unsigned button)
{
	mInitialAction = None;
	mPerformedAction = None;
	mTypeListValid = false;
	mActionListValid = false;
	mWaitForStatus = true;
	mTargetResponds = false;
	mFinalStatus = false;
	mDropTarget = NULL;
	mDragSource = NULL;
	mTargetXID = None;
	mSourceXID = None;
	mActualPosition = e;
	mMimeList.clear();
	mActionList.clear();
	mDescriptionList.clear();
	if (!w.OnDragInit(button, mInitialAction) || mInitialAction == None) return false;
	if (!cxSelManager->Grab(&w, DND_SELECTION, e.time)) return false;
	if (mInitialAction == DND_ASK) { 	 
		if (!w.GetActionList(mActionList, mDescriptionList)) return false;
	}
	vector<unsigned char> v;
	if (!w.ConvertSelection(DND_SELECTION, CSelection::TARGETS, v)) return false;
	copy((Atom*)&v[0], (Atom*)&v[0] + v.size() / sizeof(Atom), back_inserter(mMimeList));
	mEventMask = ButtonReleaseMask;
	switch (button) {
	case 1:
		mEventMask |= Button1MotionMask;
		break;
	case 2:
		mEventMask |= Button2MotionMask;
		break;
	case 3:
		mEventMask |= Button3MotionMask;
		break;
	default:
		cxAssert(0);
		break;
	}
	UpdateCursor();
	mDragSource = &w;
	mSourceXID = w.GetTopLevelWindow()->GetXID();
	UpdateCursor();
	HandleNewPosition();
	return true;
}

void CDnDManager::StopDrag()
{
	if (!mTargetResponds) {
		mWaitForStatus = false;
		mPerformedAction = None;
	}
	if (mWaitForStatus) {
		mTimeout.SetTime(2000);
		mFinalStatus = true;
	return;
	} else mTimeout.Stop();
	if (mPerformedAction == None) {
		if (mDragSource != NULL) SendLeave();
		mDragSource = NULL;
	} else SendDrop();
}

void CDnDManager::CancelDrag()
{
	mTargetResponds = false;
	StopDrag();
}

Window CDnDManager::FindTarget(XMotionEvent& e)
{
	Window target = e.root;
	Window child;
	int dummy;
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	union {unsigned char* in; Atom* version; Window* proxy;} buf;
	mVersion = 0;
	mProxy = None;
	// search down window tree
	while (target != None) {
		XTranslateCoordinates(cxDpy, e.root, target, e.x_root, e.y_root, &dummy, &dummy, &child);
		// is a proxy specified for the window?
		if ((target != e.root) || (child != None)) {
			XGetWindowProperty(cxDpy, e.window, DND_PROXY, 0, 1, False, XA_WINDOW, &type, &fmt, &count, &rem, &buf.in);
			if (type == XA_WINDOW && fmt == 32 && count != 0) {
				// yes!
				mProxy = *buf.proxy;
				XFree(buf.in);
				// does the proxy window specify itself as proxy?
				XGetWindowProperty(cxDpy, mProxy, DND_PROXY, 0, 1, False, XA_WINDOW, &type, &fmt, &count, &rem, &buf.in);
				if (type == XA_WINDOW && fmt == 32 && count != 0 && mProxy == *buf.proxy) {
					// yes!
					XFree(buf.in);
					// is the proxy dnd aware?
					XGetWindowProperty(cxDpy, mProxy, DND_AWARE, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
					if (type == XA_ATOM && fmt == 32 && count != 0) {
						mVersion = *buf.version;
						XFree(buf.in);
						// o. k.
						break;
					}
				}
			}
			// no proxy...
			XFree(buf.in);
			mProxy = None;
		}
		// search down subwindows
		target = child;
		if (target == None) return None;
		// is the subwindow dnd aware?
		XGetWindowProperty(cxDpy, target, DND_AWARE, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
		if (type == XA_ATOM && fmt == 32 && count != 0) {
			// yes!
			mVersion = *buf.version;
			XFree(buf.in);
			// o. k.
			break;
		}
		XFree(buf.in);
	}
	// due to incompatibilities with older versions we have to reject those
	if (mVersion >= MIN_VERSION) {
		if (mVersion > MAX_VERSION) mVersion = MAX_VERSION;
		if (mProxy == None) mProxy = target;
		return target;
	} else {
		mProxy = None;
		return None;
	}
}


CDnDWindow* CDnDManager::GetTarget(int x, int y)
{
	if (mTargetXID == None) return NULL;
	int dummy;
	Window target;
	Window child = mTargetXID;
	while (child != None) {
		target = child;
		XTranslateCoordinates(cxDpy, DefaultRootWindow(cxDpy), target, x, y, &dummy, &dummy, &child);
	}
	CWindow* ret = CWindow::GetWindowByID(target);
	if (ret->IsDnDAware()) return static_cast<CDnDWindow*>(ret);
	else return NULL;
}

void CDnDManager::SendEnter()
{
	mWaitForStatus = true;
	mTargetResponds = false;
	cxAssert(mDragSource != NULL);
	SetIgnoreRect(0, 0, 0, 0);
	cxAssert(mTargetXID != None);
	if (mTargetXID == mSourceXID) return;
	XClientMessageEvent e;
	e.type = ClientMessage;
	e.display = cxDpy;
	e.window = mTargetXID;
	e.send_event = False;
	e.message_type = DND_ENTER;
	e.format = 32;
	e.data.l[0] = mSourceXID;
	e.data.l[1] = mVersion << 24;
	e.data.l[2] = 0;
	e.data.l[3] = 0;
	e.data.l[4] = 0;
	switch (mMimeList.size()) {
	default:
		e.data.l[1] |= 1;
		// fall through
	case 3:
		e.data.l[4] = mMimeList[2];
		// fall through
	case 2:
		e.data.l[3] = mMimeList[1];
		// fall through
	case 1:
		e.data.l[2] = mMimeList[0];
		break;
	case 0:
		cxAssert(0);
		break;
	}
	if (((e.data.l[1] & 1) != 0) && !mTypeListValid) {
		if (!SetTypeList()) mDragSource = NULL;
		mTargetXID = None;
		return;
	}
	XSendEvent(e.display, mProxy, False, NoEventMask, (XEvent*)&e);
}

void CDnDManager::ReceiveEnter(XClientMessageEvent& e)
{
	mVersion = (e.data.l[1] >> 24) & 0xff;
	if ((mVersion < MIN_VERSION) || (mVersion > MAX_VERSION)) return;
	mSourceXID = e.data.l[0];
	mTargetXID = e.window;
	mMimeList.clear();
	if (e.data.l[1] & 1) GetTypeList();
	else {
		if (e.data.l[2] != None) mMimeList.push_back(e.data.l[2]);
		if (e.data.l[3] != None) mMimeList.push_back(e.data.l[3]);
		if (e.data.l[4] != None) mMimeList.push_back(e.data.l[4]);
	}
	if (mMimeList.empty()) {
		mSourceXID = None;
		mTargetXID = None;
	}
	return;
}

void CDnDManager::SendPosition()
{
	if (mTargetXID == None) return;
	mWaitForStatus = true;
	XClientMessageEvent e;
	e.type = ClientMessage;
	e.display = cxDpy;
	e.window = mTargetXID;
	e.send_event = False;
	e.message_type = DND_POSITION;
	e.format = 32;
	e.data.l[0] = mSourceXID;
	e.data.l[1] = 0;
	e.data.l[2] = (mLastPosition.x_root << 16) | mLastPosition.y_root;
	e.data.l[3] = mLastPosition.time;
	e.data.l[4] = mInitialAction;
	cxAssert(mTargetXID != None);
	if (mTargetXID == mSourceXID) ReceivePosition(e);
	else XSendEvent(e.display, mProxy, False, NoEventMask, (XEvent*)&e);
}

void CDnDManager::ReceivePosition(XClientMessageEvent& e)
{
	cxUpdateServerTime((Time&)e.data.l[3]);
	int xroot = (e.data.l[2] >> 16) & 0xffff;
	int yroot = e.data.l[2] & 0xffff;
	mPerformedAction = e.data.l[4];
	SetIgnoreRect(0, 0, 0, 0);
	// we want to respond to any position event in order not to block the source
	// so in case of an error we send an status report with action == None
	if ((Window)e.data.l[0] != mSourceXID) mPerformedAction = None;
	if (mPerformedAction != None) {
		CDnDWindow* target = GetTarget(xroot, yroot);
		if (target != mDropTarget) {
			if (mDropTarget != NULL) mDropTarget->OnDrop(None);
			mDropTarget = target;
			if (mDropTarget != NULL) mDropRefused = !mDropTarget->OnDropInit(mMimeList);
			else mDropRefused = true;
		}
		if (mDropTarget != NULL) {
			int xwin, ywin;
			Window dummy;
			XTranslateCoordinates(cxDpy, DefaultRootWindow(cxDpy), mDropTarget->GetXID(), xroot, yroot, &xwin, &ywin, &dummy);
			if (mDropRefused) {
				SetIgnoreRect(0, 0, target->GetWidth(), target->GetHeight());
				mPerformedAction = None;
			} else {
				if (cxxx::cxRTL) mPerformedAction = mDropTarget->GetDnDAction(mDropTarget->GetWidth() - xwin, ywin, e.data.l[4]);
				else mPerformedAction = mDropTarget->GetDnDAction(xwin, ywin, e.data.l[4]);
			}
			mIgnoreRect.x += xroot - xwin;
			mIgnoreRect.y += yroot - ywin;
		} else mPerformedAction = None;
	}
	SendStatus();
}

void CDnDManager::SendStatus()
{
	if (mSourceXID == None) return;
	XEvent e;
	e.xclient.type = ClientMessage;
	e.xclient.display = cxDpy;
	e.xclient.window = mSourceXID;
	e.xclient.send_event = False;
	e.xclient.message_type = DND_STATUS;
	e.xclient.format = 32;
	e.xclient.data.l[0] = mTargetXID;
	e.xclient.data.l[1] = 0;
	if (mPerformedAction != None) e.xclient.data.l[1] |= 1;
	e.xclient.data.l[2] = (mIgnoreRect.x << 16) | mIgnoreRect.y;
	e.xclient.data.l[3] = ((mIgnoreRect.width) << 16) | (mIgnoreRect.height);
	e.xclient.data.l[4] = mPerformedAction;
	if (mSourceXID == mTargetXID) ReceiveStatus(e.xclient);
	else XSendEvent(cxDpy, mSourceXID, False, NoEventMask, &e);
}

void CDnDManager::ReceiveStatus(XClientMessageEvent& e)
{
	mWaitForStatus = false;
	// kde3 may return a DnDStatus Message with data.l[0] set to None to indicate that a drop will not be
	// accepted. This is really nasty, because we do not know which window responded. Ok, so we loose some
	// safety and assume that everyting went right.
	if ((Window)e.data.l[0] == None) e.data.l[0] = mTargetXID;
	if ((Window)e.data.l[0] != mTargetXID || mDragSource == NULL) return;
	mTargetResponds = true;
	if ((e.data.l[1] & 2) == 0) SetIgnoreRect((e.data.l[2] >> 16) & 0xffff, e.data.l[2]
	    & 0xffff, (e.data.l[3] >> 16) & 0xffff, e.data.l[3] & 0xffff);
	else SetIgnoreRect(0, 0, 0, 0);
	mPerformedAction = None;
	if ((e.data.l[1] & 1) != 0) mPerformedAction = e.data.l[4];
	if (mPerformedAction != mInitialAction && !mDragSource->CanAcceptAction(mPerformedAction)) mPerformedAction = None;
	UpdateCursor();
	if (mFinalStatus) {
		// the window is waiting for this event to get the final status message
		StopDrag();
		mFinalStatus = false;
		return;
	}
	// if there are cached Motion Events immediately process them
	if (memcmp(&mActualPosition, &mLastPosition, sizeof(XMotionEvent)) == 0) return;
	HandleNewPosition();
}

void CDnDManager::SendLeave()
{
	if (mTargetXID == None) return;
	XEvent e;
	e.xclient.type = ClientMessage;
	e.xclient.display = cxDpy;
	e.xclient.window = mTargetXID;
	e.xclient.send_event = False;
	e.xclient.message_type = DND_LEAVE;
	e.xclient.format = 32;
	e.xclient.data.l[0] = mSourceXID;
	e.xclient.data.l[1] = 0;
	e.xclient.data.l[2] = 0;
	e.xclient.data.l[3] = 0;
	e.xclient.data.l[4] = 0;
	if (mSourceXID == mTargetXID) ReceiveLeave(e.xclient);
	else XSendEvent(cxDpy, mProxy, False, NoEventMask, &e);
	mTargetXID = None;
}

void CDnDManager::ReceiveLeave(XClientMessageEvent& e)
{
	if ((Window)e.data.l[0] != mSourceXID) return;
	if (mDropTarget != NULL) mDropTarget->OnDrop(None);
	mSourceXID = None;
	mDropTarget = NULL;
}

void CDnDManager::SendDrop()
{
	if (mTargetXID == None) return;
	XEvent e;
	e.xclient.type = ClientMessage;
	e.xclient.display = cxDpy;
	e.xclient.window = mTargetXID;
	e.xclient.send_event = False;
	e.xclient.message_type = DND_DROP;
	e.xclient.format = 32;
	e.xclient.data.l[0] = mSourceXID;
	e.xclient.data.l[1] = 0;
	e.xclient.data.l[2] = cxxx::cxServerTime;
	e.xclient.data.l[3] = 0;
	e.xclient.data.l[4] = 0;
	if (mSourceXID == mTargetXID) ReceiveDrop(e.xclient);
	else XSendEvent(cxDpy, mProxy, False, NoEventMask, &e);
}

void CDnDManager::ReceiveDrop(XClientMessageEvent& e)
{
	if ((Window)e.data.l[0] != mSourceXID) return;
	cxAssert(mDropTarget != NULL);
	if (mPerformedAction == DND_ASK) mPerformedAction = OnAskDialog();
	if (!mDropTarget->OnDrop(mPerformedAction)) mPerformedAction = None;
	SendFinished();
	mDropTarget = NULL;
}

void CDnDManager::SendFinished()
{
	if (mSourceXID == None) return;
	XEvent e;
	e.xclient.type = ClientMessage;
	e.xclient.display = cxDpy;
	e.xclient.window = mSourceXID;
	e.xclient.send_event = False;
	e.xclient.message_type = DND_FINISHED;
	e.xclient.format = 32;
	e.xclient.data.l[0] = mTargetXID;
	e.xclient.data.l[1] = 0;
	if ((mPerformedAction & 1) != 0) e.xclient.data.l[1] |= 1;
	e.xclient.data.l[2] = mPerformedAction;
	e.xclient.data.l[3] = 0;
	e.xclient.data.l[4] = 0;
	if (mSourceXID == mTargetXID) ReceiveFinished(e.xclient);
	else XSendEvent(cxDpy, mSourceXID, False, NoEventMask, &e);
}

void CDnDManager::ReceiveFinished(XClientMessageEvent& e)
{
	if ((Window)e.data.l[0] != mTargetXID) return;
	cxSelManager->Release(DND_SELECTION, CurrentTime);
	if (mDragSource != NULL) mDragSource->OnDropFinished();
	mDragSource = NULL;
	mTargetXID = None;
}

void CDnDManager::GetTypeList()
{
	cxAssert(mSourceXID != None);
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	union {unsigned char* in; Atom* data;} buf;
	XGetWindowProperty(cxDpy, mSourceXID, DND_TYPES, 0, 8, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
	if ((type != XA_ATOM) || (fmt != 32) || (count <= 3)) {
		XFree(buf.data);
		return;
	}
	for (unsigned i = 0; i < count; ++i) mMimeList.push_back(buf.data[i]);
	XFree(buf.in);
	if (rem != 0) {
		XGetWindowProperty(cxDpy, mSourceXID, DND_TYPES, 8, (rem / 4) + 1, False, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
		for (unsigned i = 0; i < count; ++i) mMimeList.push_back(buf.data[i]);
		XFree(buf.in);
	}
}

bool CDnDManager::SetTypeList()
{
	cxAssert(mSourceXID != None);
	XDeleteProperty(cxDpy, mSourceXID, DND_TYPES);
	XChangeProperty(cxDpy, mSourceXID, DND_TYPES, XA_ATOM, 32, PropModeReplace, (unsigned char*)&mMimeList[0], mMimeList.size());
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	unsigned char* data = NULL;
	XGetWindowProperty(cxDpy, mSourceXID, DND_TYPES, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &data);
	XFree(data);
	mTypeListValid = (type == XA_ATOM && fmt == 32 && count == 1 && rem == (4 * mMimeList.size() - 1));
	return mTypeListValid;
}

bool CDnDManager::SetActionList()
{
	cxAssert(mSourceXID != None);
	XDeleteProperty(cxDpy, mSourceXID, DND_ACTIONS);
	XDeleteProperty(cxDpy, mSourceXID, DND_DESCRIPTIONS);
	XChangeProperty(cxDpy, mSourceXID, DND_ACTIONS, XA_ATOM, 32, PropModeReplace, (unsigned char*)&mActionList[0], mActionList.size());
	string s;
	for (vector<string>::iterator i = mDescriptionList.begin(); i != mDescriptionList.end(); ++i) {
		s += *i;
		s += '\0';
	}
	XChangeProperty(cxDpy, mSourceXID, DND_DESCRIPTIONS, XA_STRING, 8, PropModeReplace, (unsigned char*)&s[0], s.size());
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	unsigned char* data;
	XGetWindowProperty(cxDpy, mSourceXID, DND_ACTIONS, 0, 1, False, XA_ATOM, &type, &fmt, &count, &rem, &data);
	XFree(data);
	mActionListValid = (type == XA_ATOM && fmt == 32 && count == 1 && rem == (4 * (mActionList.size() - 1)));
	if (!mActionListValid) return false;
	XGetWindowProperty(cxDpy, mSourceXID, DND_DESCRIPTIONS, 0, 1, False, XA_STRING, &type, &fmt, &count, &rem, &data);
	XFree(data);
	mActionListValid = (type == XA_STRING && fmt == 8 && (count + rem) == s.size());
	return mActionListValid;
}

Atom CDnDManager::OnAskDialog()
{
	if (mSourceXID != mTargetXID && mPerformedAction == DND_ASK) {
		// acquire from source
		Atom type;
		int fmt;
		unsigned long count;
		unsigned long rem;
		unsigned char* data;
		vector<Atom> avec;
		vector<string> svec;
		XGetWindowProperty(cxDpy, mSourceXID, DND_ACTIONS, 0, -1u, False, XA_ATOM, &type, &fmt, &count, &rem, &data);
		if (type != XA_ATOM || fmt != 32 || count == 0) return None;
		Atom* abeg = (Atom*)data;
		copy(abeg, abeg + count, back_inserter(avec));
		XFree(data);
		XGetWindowProperty(cxDpy, mSourceXID, DND_DESCRIPTIONS, 0, -1u, False, XA_STRING, &type, &fmt, &count, &rem, &data);
		if (type != XA_STRING || fmt != 8 || count == 0) return None;
		char* sbeg = (char*)data;
		char* send = sbeg + count;
		while (sbeg < send) {
			string s(sbeg);
			svec.push_back(s);
			sbeg += s.size() + 1;
		}
		if (avec.size() != svec.size()) return None;
		mActionList.clear();
		mDescriptionList.clear();
		for (unsigned i = 0; i < avec.size(); ++i) {
			if (mDropTarget->CanPerformAction(avec[i])) {
				mActionList.push_back(avec[i]);
				mDescriptionList.push_back(svec[i]);
			}
		}
	} else {
		// acquire from target
		mActionList.clear();
		mDescriptionList.clear();
		mDropTarget->GetActionList(mActionList, mDescriptionList);
	}
	if (mDescriptionList.empty()) return None;
	CDnDAskDialog ad(mDescriptionList);
	int x = (mDropTarget->GetWidth()  - ad.GetWidth()) / 2;
	int y = (mDropTarget->GetHeight() - ad.GetHeight()) / 2;
	mDropTarget->TranslateCoordinates(x, y, NULL);
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	unsigned dw = DisplayWidth(cxDpy, DefaultScreen(cxDpy));
	unsigned dh = DisplayHeight(cxDpy, DefaultScreen(cxDpy));
	if ((x + ad.GetWidth()) > dw) x = dw - ad.GetWidth();
	if ((y + ad.GetHeight()) > dh) y = dh - ad.GetHeight();
	ad.Move(x, y);
	ad.Show();
	unsigned ret = ad.DoModal();
	if (ret == CDnDAskDialog::IDCANCEL) return None;
	return (mActionList[ret]);
}

void CDnDManager::UpdateCursor()
{
	static Atom state;
	if (mDragSource == NULL) {
		// initialize function
		state = None + 1;
	} else {
		if (state == mPerformedAction) return;
		Cursor cur = mDragSource->GetDnDCursor(mPerformedAction);
		XChangeActivePointerGrab(cxDpy, mEventMask, cur, CurrentTime);
	}
}

bool CDnDManager::CreateCursors()
{
	bool ret = true;
    if (!mMoveCursor.Create("move")) ret = false;
    if (!mNotCursor.Create("no-drop")) ret = false;
    if (!mCopyCursor.Create("copy")) ret = false;
	if (!mAskCursor.Create("question_arrow")) ret = false;
    if (!mLinkCursor.Create("link")) ret = false;
	return ret;
}
