/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CDropDownPopup.cpp
  Info   : $Id: CDropDownPopup.cpp,v 1.30 2012/06/26 08:05:52 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// cxxx headers
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/CDropDownPopup.h>
#include <cxxx/CApplication.h>

// X11 headers
#include <X11/cursorfont.h>

// namespace declarations
using cxxx::cxApp;
using cxxx::cxDpy;
using cxxx::cxWiden;
using std::string;
using std::vector;
using std::wstring;

CDropDownPopup::CDropDownPopup(CWindow* w, unsigned msgid, const std::vector<string>& list)
: CWindow(NULL),
  mSelected(-1u),
  mOffset(0),
  mCursor(None),
  mScrollbar(this, IDSCROLL)
{
	mWindowState.set(ISPOPUP);
	Reparent(w);
	SetMessageTarget(w);
	SetMessageID(msgid);
	if (!IsChildOf(cxApp->GetPopupWindow())) cxApp->SetPopupWindow(NULL);
	CColor col(TC_BACKGND);
	mAttributes.background_pixel = col.GetPixel();
	mAttributes.event_mask |= PointerMotionMask | ButtonPressMask | ButtonReleaseMask | LeaveWindowMask | EnterWindowMask;
	mAttributes.override_redirect = true;
	mAttributes.save_under = true;
	ChangeAttributes(CWOverrideRedirect | CWSaveUnder | CWEventMask | CWBackPixel);
	mCursor = XCreateFontCursor(cxDpy, XC_left_ptr);
	XDefineCursor(cxDpy, GetXID(), mCursor);
	unsigned width = 0;
	for (vector<string>::const_iterator i = list.begin(); i < list.end(); ++i) {
		wstring wi = cxWiden(*i);
		mList.push_back(wi);
		unsigned w = GetTextWidth(wi);
		if (w > width) width = w;
	}
	width += 10;
	unsigned h = 20 * mList.size() + 4;
	unsigned dmax = DisplayHeight(cxDpy, DefaultScreen(cxDpy)) / 3;
	mScrolling = h > dmax;
	if (mScrolling) {
		width += mScrollbar.GetWidth();
		h = dmax / 20;
		mScrollbar.SetPage(h);
		h *= 20;
		h += 4;
		mScrollbar.SetRange(mList.size());
	}
	Resize(width, h);
	cxApp->SetPopupWindow(this);
}

/* virtual */ CDropDownPopup::~CDropDownPopup()
{
	PostMessage(MT_CONFIRM_SELECTION, mSelected);
	XFreeCursor(cxDpy, mCursor);
}

/* virtual */ void CDropDownPopup::Recolor()
{
	CColor col(TC_BACKGND);
	mAttributes.background_pixel = col.GetPixel();
	ChangeAttributes(CWBackPixel);
	Redraw();
}

/* virtual */ void CDropDownPopup::OnExpose(XExposeEvent& e)
{
	ClearArea(e);
	if (!mList.empty()) {
		bool draw = true;
		if ((e.x + e.width) < 2) draw = false;
		if (mScrolling) {
			if (e.x > (int)(mWidth - mScrollbar.GetWidth() - 2)) draw = false;
		} else {
			if (e.x > (int)(mWidth - 2)) draw = false; 
		}
		if (draw) {
			int lo = e.y - 2;
			if (lo < 0) lo = 0;
			int hi = e.y + e.height - 2;
			if (hi > (int)mHeight - 5) hi = mHeight - 5;
			lo += mOffset;
			hi += mOffset;
			lo /= 20;
			hi /= 20;
			for (unsigned i = lo; i <= (unsigned)hi; i++) DrawItem(i, i == mSelected);
		}
	}
	DrawRectangle(0, 0, GetWidth(), GetHeight(), CColor(TC_BACKGND_TEXT));
}

/* virtual */ void CDropDownPopup::OnLButtonDown(XButtonEvent& e __attribute__ ((unused)))
{
	if (mSelected == -1u) GetTopLevelWindow()->SetFocusWindow(NULL);
}

/* virtual */ void CDropDownPopup::OnButtonUp(XButtonEvent& e)
{
	CTopLevelWindow* w = GetTopLevelWindow();
	if (e.button != 1) mSelected = -1u;
	w->SetFocusWindow(NULL);
}

/* virtual */ void CDropDownPopup::OnEvent(XEvent& e)
{
	switch (e.xany.type) {
	case MotionNotify: {
			unsigned i = GetItemFromPos(e.xmotion.x, e.xmotion.y);
			if (i == mSelected) return;
			else Select(i);
		}
		break;
	case LeaveNotify:
		Select(-1u);
		break;
	default:
		break;
	}
}

/* virtual */ void CDropDownPopup::OnMessage(unsigned msgid, TMessage type __attribute__ ((unused)), unsigned value)
{
	if (msgid == IDSCROLL) {
		value *= 20;
		int delta = value - mOffset;
		mOffset = value;
		Scroll(0, delta);
	}
}

void CDropDownPopup::Position(int x, int y, int w, int h)
{
	if ((int)(x + mWidth) >= DisplayWidth(cxDpy, DefaultScreen(cxDpy))) {
		x += w - mWidth;
		cxAssert(x >= 0);
	}
	if ((int)(y + h + mHeight) < DisplayHeight(cxDpy, DefaultScreen(cxDpy))) {
		y += h;
	} else {
		y -= mHeight;
	}
	Move(x, y);
	if (mScrolling) {
		x = cxxx::cxRTL ? 1 : mWidth - mScrollbar.GetWidth() - 1;
		mScrollbar.Move(x, 1);
		mScrollbar.Resize(mScrollbar.GetWidth(), mHeight - 2);
		mScrollbar.Show();
	}
}

void CDropDownPopup::DrawItem(unsigned idx, bool selected)
{
	unsigned y = 2 + 20 * idx - mOffset;
	CColor col;
	if (selected) {
		FillRectangle(2, y, GetWidth() - 4, 20, CColor(TC_TITLE));
		ClearArea(1, y, 1, 20);
		ClearArea(mWidth - 2, y, 1, 20);
		col = CColor(TC_TITLE_TEXT);
	} else {
		ClearArea(1, y, GetWidth() - 2, 20);
		col = CColor(TC_BACKGND_TEXT);
	}
	DrawString(5, y + (20 - GetTextHeight()) / 2, col, mList[idx], false);
}
