/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CEdit.cpp
  Info   : $Id: CEdit.cpp,v 1.40 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CCaret.h>
#include <cxxx/CEdit.h>
#include <cxxx/CError.h>

// X11 headers
#include <X11/cursorfont.h>

// namespace declarations
using cxxx::cxError;
using cxxx::cxDpy;
using cxxx::cxgettext;
using cxxx::cxWiden;
using std::string;
using std::wstring;
using std::isdigit;

CEdit::CEdit(CWindow* parent)
: CWindow(parent),
  mCaret(this),
  mSelMax(0),
  mSelMin(0),
  mStyle(LEFT | YCENTER),
  mTextWidth(0),
  mTimer(this, IDTIMER),
  mXOffset(0)
{
	mStyle = LEFT | YCENTER;
	CColor col(TC_BACKGND);
	mAttributes.background_pixel = col.GetPixel();
	ChangeAttributes(CWBackPixel);
	Resize(WIDTH, GetTextHeight() + 2 * BORDER_WIDTH);
	SetMargin(0, 0);
	SetFocusable(true);
	mCursor = XCreateFontCursor(cxDpy, XC_xterm);
	XDefineCursor(cxDpy, GetXID(), mCursor);
	mCaret.SetBlinkRate(500);
	mCaret.HideCaret();
	SetCaretPos(0);
}

/* virtual */ CEdit::~CEdit()
{
	XFreeCursor(cxDpy, mCursor);
}

/* virtual */ void CEdit::Recolor()
{
	CColor col;
	if (IsEnabled()) col = CColor(TC_BACKGND);
	else col = CColor(TC_FRAME);
	mAttributes.background_pixel = col.GetPixel();
	ChangeAttributes(CWBackPixel);
	Redraw();
}

/* virtual */ void CEdit::OnExpose(XExposeEvent& e)
{
	if (IsEnabled()) {
		if (mSelMin == mSelMax) {
			ClearArea(e);
			if (!mText.empty()) {
				if (cxxx::cxRTL) {
					std::wstring sav = mText;
					std::reverse(sav.begin(), sav.end());
					DrawUniString(GetXOffset(), GetYOffset(), CColor(TC_BACKGND_TEXT), sav);
				} else DrawUniString(GetXOffset(), GetYOffset(), CColor(TC_BACKGND_TEXT), mText);
			}
		} else {
			unsigned l1 = mSelMin;
			unsigned l2 = mSelMax - mSelMin;
			unsigned l3 = mText.size() - mSelMax;
			wstring s1 = mText.substr(0, l1);
			wstring s2 = mText.substr(mSelMin, l2);
			wstring s3 = mText.substr(mSelMax, l3);
			unsigned w1 = GetTextWidth(s1);
			unsigned w2 = GetTextWidth(s2);
			if (cxxx::cxRTL) {
				std::reverse(s1.begin(), s1.end());
				std::reverse(s2.begin(), s2.end());
				std::reverse(s3.begin(), s3.end());
			}
			DrawUniString(GetXOffset(), GetYOffset(), CColor(TC_BACKGND_TEXT), s1);
			FillRectangle(GetXOffset() + w1, GetYOffset(), w2, GetTextHeight(), CColor(TC_TITLE));
			DrawUniString(GetXOffset() + w1, GetYOffset(), CColor(TC_TITLE_TEXT), s2, false);
			DrawUniString(GetXOffset() + w1 + w2, GetYOffset(), CColor(TC_BACKGND_TEXT), s3);
		}
	} else {
		FillRectangle(e.x, e.y, e.width, e.height, CColor(TC_FRAME));
		if (cxxx::cxRTL) {
			std::wstring sav = mText;
			std::reverse(sav.begin(), sav.end());
			DrawUniString(GetXOffset(), GetYOffset(), CColor(TC_FRAME_TEXT), sav, false);
		} else DrawUniString(GetXOffset(), GetYOffset(), CColor(TC_FRAME_TEXT), mText, false);
	}
	DrawFrame(0, 0, GetWidth(), GetHeight(), FS_SUNKEN_THICK, false);
}

void CEdit::SetText(const string& text)
{
	wstring wtxt = cxWiden(text);
	if (cxxx::cxRTL) {
		wtxt = cxxx::cxBidi->Convert(wtxt);
		std::reverse(wtxt.begin(), wtxt.end());
	}
	if (wtxt != mText) {
		mText = wtxt;
		mSelMax = mSelMin;
		mTextWidth = GetTextWidth(mText);
		SetCaretPos(mText.size());
		RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
	}
}

void CEdit::SetCaretPos(unsigned pos)
{
	if (pos > mText.size()) return ;
	mCaretPos = pos;
	mSelMin = pos;
	mSelMax = pos;
	int xs = GetTextWidth(mText, pos);
	int x;
	if ((GetWidth() - mRightMargin - mLeftMargin) > (mTextWidth + 10)) x = xs + GetXOffset();
	else {
		x = xs - mXOffset + 5;
		if (x < (mLeftMargin + 5)) {
			mXOffset = xs - mLeftMargin;
			RedrawArea(mLeftMargin + BORDER_WIDTH, BORDER_WIDTH, mWidth - mLeftMargin - mRightMargin - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			x = 5 + mLeftMargin;
		} else if (x > ((int)GetWidth() - mRightMargin - 5)) {
			mXOffset = xs - GetWidth() + mRightMargin + 10;
			RedrawArea(mLeftMargin + BORDER_WIDTH, BORDER_WIDTH, mWidth - mLeftMargin - mRightMargin - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			x = GetWidth() - mRightMargin - 5;
		}
	}
	mCaret.Move(x, GetYOffset());
}

/* virtual */ void CEdit::OnKey(wchar_t key, unsigned mask __attribute__ ((unused)))
{
	switch (key) {
   	case CK_Left:
		if (cxxx::cxRTL) {
			if (mSelMin != mSelMax) {
				SetCaretPos(mSelMax);
				RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			} else if (mCaretPos < mText.size()) SetCaretPos(++mCaretPos);
		} else {
			if (mSelMin != mSelMax) {
				SetCaretPos(mSelMin);
				RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			} else if (mCaretPos) SetCaretPos(--mCaretPos);
		}
		break;
   	case CK_Right:
		if (cxxx::cxRTL) {
			if (mSelMin != mSelMax) {
				SetCaretPos(mSelMin);
				RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			} else if (mCaretPos) SetCaretPos(--mCaretPos);
		} else {
			if (mSelMin != mSelMax) {
				SetCaretPos(mSelMax);
				RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
			} else if (mCaretPos < mText.size()) SetCaretPos(++mCaretPos);
		}
   		return ;
   	case L'\b':
   		if (mSelMin != mSelMax) {
   			DeleteSelection();
   			SetCaretPos(mSelMin);
   			RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		} else if (mCaretPos) {
   			mTextWidth = GetTextWidth(mText);
               mText.erase(--mCaretPos, 1);
   			SetCaretPos(mCaretPos);
   			RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		}
   		return ;
   	case L'\x7f':
   		if (mSelMin != mSelMax) {
   			DeleteSelection();
   			SetCaretPos(mSelMin);
   			RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		} else if (mCaretPos < mText.size()) {
   			mTextWidth = GetTextWidth(mText);
   			mText.erase(mCaretPos, 1);
   			SetCaretPos(mCaretPos);
   			RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		}
   		return ;
   	case CK_Home:
   		SetCaretPos(0);
   		RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		return ;
   	case CK_End:
   		SetCaretPos(mText.size());
   		RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
   		return ;
   	case L'\n':
   		break;
   	default:
   		if (key < L'\xff00') {
   			if (mStyle & NUMERIC) {
   				if (isdigit(key)) InsertChar(key);
   			} else if (isprint(key)) InsertChar(key);
   		}
   		break;
   	}
}

/* virtual */ void CEdit::OnLButtonDown(XButtonEvent& e)
{
	unsigned i = GetPosFromXpos(e.x);
	SetCaretPos(i);
	RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
	XChangeActivePointerGrab(cxDpy, ButtonReleaseMask | Button1MotionMask, None, CurrentTime);
}

/* virtual */ void CEdit::OnDblClick(XButtonEvent& e)
{
	if (e.button != Button1) return;
	unsigned i = GetPosFromXpos(e.x);
	unsigned lo(i);
	while (lo && !isspace(mText[lo - 1])) --lo;
	unsigned hi(i);
	while (hi < mText.size() && !isspace(mText[hi])) ++hi;
	SetCaretPos(hi);
	Select(lo, hi);
	RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
}


/* virtual */ void CEdit::OnButtonUp(XButtonEvent& e)
{
	if (e.button != Button1) return;
	mAttributes.event_mask = ExposureMask | StructureNotifyMask | ButtonPressMask;
	ChangeAttributes(CWEventMask);
	mTimer.Stop();
}

/* virtual */ void CEdit::OnFocus(bool in)
{
	if (in) {
		if (IsEnabled()) mCaret.ShowCaret();
	} else mCaret.HideCaret();
}

/* virtual */ void CEdit::OnEvent(XEvent& e)
{
	if (e.xany.type == MotionNotify) {
		if ((e.xmotion.state & Button1Mask) == 0) return;
		ScrollText();
	}
}

/* virtual */ void CEdit::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
	switch (msgid) {
	case IDTIMER:
		switch (type) {
		case MT_TIMER:
			ScrollText();
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

void CEdit::ScrollText()
{
	int x, y;
	GetPointerPos(x, y, this);
	if (x < 5) {
		int d = x - 5;
		if (d < -20) d = -20;
		mXOffset += d;
		if (mXOffset < 0) mXOffset = 0;
		mTimer.SetTime(200);
	} else if (x > (int)GetWidth() - 5) {
		int d = x - GetWidth() + 5;
		if (d > 20) d = 20;
		mXOffset += d;
		if (mXOffset > (int)mTextWidth - (int)GetWidth()) mXOffset = mTextWidth - GetWidth();
		mTimer.SetTime(200);
	} else mTimer.Stop();
	unsigned i = GetPosFromXpos(x);
	if (i < mCaretPos) {
		mSelMin = i;
		mSelMax = mCaretPos;
	} else {
		mSelMin = mCaretPos;
		mSelMax = i;
	}
	RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
}

unsigned CEdit::GetPosFromXpos(int xpos)
{
	xpos -= GetXOffset();
	if (xpos < 0) return 0;
	if (mText.empty()) return 0;
	int twl = 0;
	int twh = 0;
	unsigned pos = 0;
	while (twh < xpos && pos < mText.size()) {
		twh += GetCharWidth(mText[pos++]);
	}
	if (twh <= xpos) return pos;
	twl = twh - GetCharWidth(mText[pos - 1]);
	if ((twh - xpos) > (xpos - twl)) return pos -1;
	else return pos;
}

void CEdit::InsertChar(wchar_t c)
{
	if (mSelMin != mSelMax) {
		DeleteSelection();
		mCaretPos = mSelMin;
	}
	if (cxxx::cxRTL) {
		mText.insert(mCaretPos, 1, c);
		mTextWidth = GetTextWidth(mText);
		// where should the caret move
		CBidi::TBidiType b = cxxx::cxBidi->GetBidiType(c);
		if (b == CBidi::LTR_NUMBER) b = CBidi::LTR;
		else if (b == CBidi::RTL_NUMBER) b = CBidi::RTL;
		else if (b != CBidi::LTR && b != CBidi::RTL) {
			// try harder: characters to the right
			CBidi::TBidiType tmp = CBidi::NONE;
			for (int i = mCaretPos; i >= 0; --i) {
				b = cxxx::cxBidi->GetBidiType(mText[i]);
				if (b == CBidi::LTR_NUMBER || b == CBidi::RTL_NUMBER) tmp = b;
				if (b == CBidi::LTR || b == CBidi::RTL) break;
			}
			if (b != CBidi::LTR && b != CBidi::RTL) {
				if (tmp == CBidi::LTR_NUMBER) b = CBidi::LTR;
				else if (tmp == CBidi::RTL_NUMBER) b = CBidi::RTL;
				else {
					// try even harder: characters to the left
					b = cxxx::cxBidi->GetLineDirection(&mText[mCaretPos]);
					if (b == CBidi::RTL || b == CBidi::RTL_NUMBER) b = CBidi::RTL;
					else if (b == CBidi::LTR || b == CBidi::LTR_NUMBER) b = CBidi::LTR;
					else b = CBidi::RTL;
				}
			}
		}
		if (b == CBidi::RTL) SetCaretPos(++mCaretPos);
		else SetCaretPos(mCaretPos);
	} else {
		mText.insert(mCaretPos, 1, c);
		mTextWidth = GetTextWidth(mText);
		SetCaretPos(++mCaretPos);
	}
	RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
}

/* virtual */ void CEdit::Enable(bool enable)
{
	if (enable == IsEnabled()) return;
	CWindow::Enable(enable);
	CColor col;
	if (enable) col = CColor(TC_BACKGND);
	else col = CColor(TC_FRAME);
	mAttributes.background_pixel = col.GetPixel();
	ChangeAttributes(CWBackPixel);
	SetFocusable(enable);
	RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
}
