/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CEdit.h
  Info   : $Id: CEdit.h,v 1.21 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CEdit_h
#define CEdit_h

// C++X headers
#include <cxxx/CCaret.h>
#include <cxxx/CError.h>

class CEdit : public CWindow
{
private:
	enum
	{
		// Geometry
		WIDTH = 100,
		BORDER_WIDTH = 2,

		// Message IDs
		IDTIMER = cxxx::msgid::MSGID_CCaret
	};

public:
	enum
	{
		// Styles
		TOP		= 1 << 0,
		BOTTOM	= 1 << 1,
		YCENTER = 1 << 2,
		LEFT	= 1 << 3,
		RIGHT	= 1 << 4,
		XCENTER	= 1 << 5,
		XJUSTIFY = LEFT | RIGHT | XCENTER,
		YJUSTIFY = TOP | BOTTOM | YCENTER,
		JUSTIFY = XJUSTIFY | YJUSTIFY,
		NUMERIC	= 1 << 6
	};

public:
	// Constructor
	CEdit(CWindow* parent);

	// Destructor
	virtual ~CEdit();

	// X event processing
	virtual void OnExpose(XExposeEvent& e);
	virtual void OnKey(wchar_t key, unsigned mask);
	virtual void OnLButtonDown(XButtonEvent& e);
	virtual void OnDblClick(XButtonEvent& e);
	virtual void OnButtonUp(XButtonEvent& e);
	virtual void OnFocus(bool in);
	virtual void OnEvent(XEvent& e);

	// C++X message processing
	virtual void OnMessage(unsigned msgid, TMessage type, unsigned value);

	// Enables or disables the window
	virtual void Enable(bool enable);

	// Updates the look of the window
	virtual void Recolor();
	
	// Sets the text of the field
	void SetText(const std::string& text);
	
	// Sets the style of the field
	void SetStyle(unsigned style)
	{
		mStyle = style;
		SetCaretPos(mCaretPos);
		RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
	}
		
	void SetMargin(int left, int right)
	{
		mLeftMargin = left;
		mRightMargin = right;
	}

	// Returns the character position of screen position x
	unsigned GetPosFromXpos(int xpos);

	// Returns the text of the field
	std::string GetText() const
	{
		if (cxxx::cxRTL) {
			std::wstring tmp = cxxx::cxBidi->Convert(mText);
			std::reverse(tmp.begin(), tmp.end());
			return cxxx::cxNarrow(tmp);
		} else return cxxx::cxNarrow(mText);
	}

	// Returns the selected text
	std::string GetSelection() const
	{
		if (mSelMin == mSelMax) return "";
		if (cxxx::cxRTL) {
			std::wstring tmp = cxxx::cxBidi->Convert(mText.substr(mSelMin, mSelMax - mSelMin));
			std::reverse(tmp.begin(), tmp.end());
			return cxxx::cxNarrow(tmp);
		} else return cxxx::cxNarrow(mText.substr(mSelMin, mSelMax - mSelMin));
	}

	// Scrolls the text
	void ScrollText();
		
	// Selects all of the text in the field
	void SelectAll()
	{
		Select(0, mText.size());
	}

protected:
	// Selects a range of characters
	void Select(unsigned min, unsigned max)
	{
		if (min > max) {
			unsigned tmp = min;
			min = max;
			max = tmp;
		}
		if (mSelMin != min || mSelMax != max) {
			mSelMin = min;
			mSelMax = max;
			RedrawArea(BORDER_WIDTH, BORDER_WIDTH, mWidth - 2 * BORDER_WIDTH, mHeight - 2 * BORDER_WIDTH);
		}
	}

		// Inserts a character at the cursor position
	void InsertChar(wchar_t c);

	// Deletes the selection
	void DeleteSelection()
	{
		mText.erase(mSelMin, mSelMax - mSelMin);
		mSelMax = mSelMin;
		mTextWidth = GetTextWidth(mText);
	}

	// Returns the offsets
	int GetXOffset()
	{
		if (GetWidth() > (mTextWidth + 10 + mRightMargin + mLeftMargin)) {
			mXOffset = 0;
			switch (mStyle & XJUSTIFY) {
			case LEFT:
				return 5 + mLeftMargin;
			case RIGHT:
				return GetWidth() - mTextWidth - mRightMargin - 5;
			case XCENTER:
				return (GetWidth() - mRightMargin + mLeftMargin - mTextWidth) / 2;
			default:
				cxxx::cxError << cxxx::lib::_("CEdit: unknown horizontal justification", 401) << CError::Error;
				mStyle &= ~XJUSTIFY;
				mStyle |= LEFT;
				return 5;
			}
		} else return 5 - mXOffset;
	}

	int GetYOffset()
	{
		switch (mStyle & YJUSTIFY) {
		case TOP:
			return 5;
		case BOTTOM:
			return GetHeight() - GetTextHeight() - 5;
		case YCENTER:
			return (GetHeight() - GetTextHeight()) / 2;
		default:
			cxxx::cxError << cxxx::lib::_("CEdit: unknown vertical justification", 400) << CError::Error;
			mStyle &= ~YJUSTIFY;
			mStyle |= YCENTER;
			return 5;
		}
	}

	// Sets the cursor position to character position pos
	void SetCaretPos(unsigned pos);

protected:
	CCaret mCaret;
	unsigned mCaretPos;
	Cursor mCursor;
	unsigned mSelMax;
	unsigned mSelMin;
	unsigned mStyle;
	std::wstring mText;
	unsigned mTextWidth;
	CTimer mTimer;
	int mXOffset;
	int mLeftMargin;
	int mRightMargin;
};

#endif // CEdit_h
