/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFileHandler.cpp
  Info   : $Id: CFileHandler.cpp,v 1.55 2012/09/16 17:49:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly s as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// cxxx headers
#include <cxxx/Cxxx.h>
#include <cxxx/CFileHandler.h>
#include <cxxx/CDupFNDialog.h>
#include <cxxx/CMessageBox.h>
#include <cxxx/CApplication.h>
#include <cxxx/CZStreambuf.h>

// C++ headers
#include <cerrno>
#include <fstream>
// #include <cstdio>
#include <csignal>
#include <cstdlib>
#include <sstream>

// C headers
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/wait.h>

// namespace declarations
using std::vector;
using std::ifstream;
using std::ofstream;
using std::ios;
using std::string;
using std::filebuf;
using std::ios_base;
using std::ostringstream;
using cxxx::cxApp;
using cxxx::cxError;
using cxxx::cxMainWin;
using cxxx::cxHomeDir;
using cxxx::lib::cxgettext;


bool CFileHandler::Init(const string& cxxxdir)
{
	if (cxxxdir.empty()) return false;
	mTrashDir = cxxxdir + "/trash";
	mCacheDir = cxxxdir + "/cache";
	if (access(mTrashDir.c_str(), R_OK | W_OK | X_OK) != 0) {
		if (errno == ENOENT) {
			if (mkdir(mTrashDir.c_str(), S_IRWXU)) {
				cxError << _("Failed to create trash directory:\n{1}", 43) << strerror(errno) << CError::Error;
				return false;
			}
		} else {
			cxError << _("Cannot access trash directory:\n{1}", 44) << strerror(errno) << CError::Error;
			return false;
		}
	}
	if (access(mCacheDir.c_str(), R_OK | W_OK | X_OK) != 0) {
		if (errno == ENOENT) {
			if (mkdir(mCacheDir.c_str(), S_IRWXU)) {
				cxError << _("Failed to create cache directory:\n{1}", 452) << strerror(errno) << CError::Error;
				return false;
			}
		} else {
			cxError << _("Cannot access cache directory:\n{1}", 448) << strerror(errno) << CError::Error;
			return false;
		}
	}
	mIndexFile = mTrashDir;
	mIndexFile += "/.trash";
	mAliasFile = mTrashDir;
	mAliasFile += "/.alias";
	mLockFile = mTrashDir;
	mLockFile += "/.lock";
	return true;
}

void CFileHandler::HandleFiles(const vector<string>& from, const string& todir, unsigned action)
{
	bool trash = false;
	string to = todir;
	if (to[to.size() - 1] != '/') to += '/';
	unsigned flag = 0;
	unsigned tolen = to.size();
	mContinue = true;
	for (vector<string>::const_iterator i = from.begin(); i != from.end(); ++i) {
		mMore = i + 1 != from.end();
		string name = GetFileName(*i);
		// trashed file: restore original name;
		if (GetFilePath(*i) == mTrashDir) {
			trash = true;
			name = GetFileName(GetOrigin(name));
			if (name.empty()) {
				ErrorBox(_("File {1} not found.", 45), *i);
				if (!mContinue) return;
				else continue;
			}
		}
		to.resize(tolen);
		to += name;
		if (action == MOVE && *i == to) break;
		bool doit = true;
		if (flag != CDupFNDialog::IDOVERALL) {
			if (access(to.c_str(), F_OK) == 0) {
				// file exists
				if (flag == CDupFNDialog::IDAUTOSKIP) {
					if (action == MOVE) TrashFile(*i);
					if (!mContinue) return;
					else continue;
				}
				CDupFNDialog dlg(todir, name, *i != to);
				flag = dlg.DoModal();
				if (flag == CDupFNDialog::IDCANCEL) return;
				if (flag == CDupFNDialog::IDRENAME) {
					to.resize(tolen);
					to += dlg.GetName();
					doit = true;
				} else if (flag == CDupFNDialog::IDSKIP || flag == CDupFNDialog::IDAUTOSKIP) continue;
			}
		}
		if (doit) {
			if (trash) {
				if (!CopyFile(*i, to) && !mContinue) return;
			} else {
				switch (action) {
				case MOVE:
					if (!MoveFile(*i, to) && !mContinue) return;
					break;
				case COPY:
					if (!CopyFile(*i, to) && !mContinue) return;
					break;
				case LINK:
					if (!LinkFile(*i, to) && !mContinue) return;
					break;
				default:
					cxAssert(0);
				}
			}
		}
	}
}

bool CFileHandler::MoveFile(const string& from, const string& to)
{
	if (rename(from.c_str(), to.c_str()) == 0) return true;
	if (errno == ENOTEMPTY) {
		if (DeleteFile(to) && MoveFile(from, to)) return true;
	}
	if (errno == EXDEV) {
		if (CopyFile(from, to) && DeleteFile(from)) return true;
	}
	ErrorBox(_("Failed to move file {1} to {2}:\n{3}", 46), from, to);
	return false;
}

bool CFileHandler::CopyFile(const string& from, const string& to)
{
	struct stat st;
	if (lstat(from.c_str(), &st) != 0) {
		ErrorBox(_("Failed to read file status of {1}: {3}", 47), from);
		return false;
	}
	// delegate to the appropriate subroutine
	if (S_ISLNK(st.st_mode)) return CopyLink(from, to);
	if (S_ISDIR(st.st_mode)) {
		if (!CopyDir(from, to)) return false;
	} else if (S_ISREG(st.st_mode)) {
		if (st.st_size == 0) {
			std::ofstream o(to.c_str());
			if (!o) {
				ErrorBox(_("Failed to create empty file {1}: {3}", 48), to);
				return false;
			}
		} else {
			// size > 0
			if (!CopyReg(from, to)) return false;
		}
	} else {
		// copying of special files is not (yet?) supported
		ErrorBox(_("Cannot copy special file {1} to {2}: {3}", 49), from, to);
		return false;
	}
	if (chown(to.c_str(), st.st_uid, st.st_gid) != 0) {
		cxError << _("Failed to change ownership of file {1}: {2}", 50) << to << strerror(errno) << CError::Warn;
		// wrong owner: do not allow suid and sgid
		st.st_mode &= ~(S_ISUID | S_ISGID);
	}
	if (chmod(to.c_str(), st.st_mode) != 0) {
		ErrorBox(_("Failed to set permission flags of file {1}: {3}", 51), to);
		return false;
	}
	return true;
}

bool CFileHandler::CopyLink(const string& from, const string& to)
{
	// get the destination of the symlink
	string buf(PATH_MAX, 0);
	int len = readlink(from.c_str(), &buf[0], PATH_MAX);
	if (len == -1) {
		ErrorBox(_("Failed to read symbolic link {1}:\n{3}", 52), from);
		return false;
	}
	cxAssert(len < PATH_MAX);
	buf.resize(len);
	// do not interpret the link target, because we do not really know whether
	// absolute or relative paths lead to better results
 	if (symlink(buf.c_str(), to.c_str()) != 0) {
		ErrorBox(_("Failed to create symbolic link {1}:\n{3}", 53), to);
		return false;
	}
	return true;
}

bool CFileHandler::CopyDir(const string& from, const string& to)
{
	// read source directory prior to creating destination directory to
	// avoid recursion
	vector<string> v;
	if (!ReadDirectory(from, v)) return false;
	// create an empty destination directory
	if (mkdir(to.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
		if (errno == EEXIST) {
			if (!DeleteFile(to)) return false;
			return CopyDir(from, to);
		}
		ErrorBox(_("Failed to create directory {1}:\n{3}", 54), to);
		return false;
	}
	string f = from;
	string t = to;
	if (f[f.size() - 1] != '/') f += '/';
	if (t[t.size() - 1] != '/') t += '/';
	unsigned flen = f.size();
	unsigned tlen = t.size();
	bool tmp = mMore;
	for (vector<string>::iterator i = v.begin(); i != v.end(); ++i) {
		mMore = tmp | ((i + 1) != v.end());
		f.resize(flen);
		t.resize(tlen);
		f += *i;
		t += *i;
		if (!CopyFile(f, t) && !mContinue) return false;
	}
	return true;
}

bool CFileHandler::CopyReg(const string& from, const string& to)
{
	// open files
	std::ifstream ff(from.c_str());
	std::ofstream tf(to.c_str());
	if (!ff) {
		ErrorBox(_("Failed to open source file {1}:\n{3}", 55), from);
		return false;
	}
	if (!tf) {
		ErrorBox(_("Failed to open destination file {1}:\n{3}", 56), to);
		return false;
	}
	//copy contents from source to destination
	tf << ff.rdbuf();
	if (ff.fail() || tf.fail()) {
		ErrorBox(_("Failed to copy file {1} to {2}:\n{3}", 57), from, to);
		return false;
	}
	return true;
}

bool CFileHandler::LinkFile(const string& from, const string& to)
{
	if (symlink(from.c_str(), to.c_str()) != 0) {
		ErrorBox(_("Failed to create symbolic link {1}:\n{3}", 53), to);
		return false;
	}
	return true;
}

bool CFileHandler::DeleteFile(const string& file)
{
	struct stat st;
	if (lstat(file.c_str(), &st) != 0) {
		ErrorBox(_("Failed to read file status of {1}: {3}", 47), file);
		return false;
	}
	if (S_ISDIR(st.st_mode)) return DeleteDir(file);
	if (unlink(file.c_str()) != 0) {
		ErrorBox(_("Failed to delete file {1}: {3}", 460), file);
		return false;
	}
	return true;
}

bool CFileHandler::DeleteDir(const string& dir)
{
	vector<string> list;
	if(!ReadDirectory(dir, list)) return false;
	string file = dir;
	if (file[file.size() - 1] != '/') file += '/';
	unsigned flen = file.size();
	bool tmp = mMore;
	for (vector<string>::iterator i = list.begin(); i != list.end(); ++i) {
		mMore = tmp | ((i + 1) != list.end());
		file.resize(flen);
		file += *i;
		if (!DeleteFile(file) && !mContinue) return false;
	}
	mMore = tmp;
	if (rmdir(dir.c_str()) != 0) {
		ErrorBox(_("Failed to delete directory\n{1}\n{3}", 59), dir);
		return false;
	}
	return true;
}

bool CFileHandler::ReadDirectory(const string& dir, vector<string>& ret)
{
	ret.clear();
	DIR * d = opendir(dir.c_str());
	if (!d) {
		ErrorBox(_("Failed to open directory {1}.\n{3}", 60), dir);
		return false;
	}
 	dirent * ent;
	while ((ent = readdir(d))) {
		if (memcmp(ent->d_name, ".", 2) == 0) continue;
		if (memcmp(ent->d_name, "..", 3) == 0)continue;
        ret.push_back(ent->d_name);
  	}
	if (closedir(d) != 0) cxError << _("Failed to close file\n{1}\n{2}", 492) << dir << strerror(errno) << CError::Fatal;
	return true;
}

void CFileHandler::ErrorBox(const string& s1, const string& s2, const string& s3)
{
	if (cxMainWin == NULL) {
		mContinue = false;
		return;
	}
	string text = s1;
	string::size_type pos = text.find("{1}");
	if (pos != string::npos) text.replace(pos, 3, s2);
	pos = text.find("{2}");
	if (pos != string::npos) text.replace(pos, 3, s3);
	pos = text.find("{3}");
	if (pos != string::npos) text.replace(pos, 3, strerror(errno));
	if (mMore) {
		CMessageBox mb(text, "C++X", CMessageBox::IGNORE | CMessageBox::CANCEL, CIcon::CRITICAL);
		mb.FocusButton(CMessageBox::CANCEL);
		mContinue = (mb.DoModal() == CMessageBox::IGNORE);
	} else {
		CMessageBox mb(text, "C++X", CMessageBox::OK, CIcon::WARNING);
		mb.FocusButton(CMessageBox::OK);
		mb.DoModal();
		mContinue = false;
	}
}

string CFileHandler::GetAbsolutePath(const string& file, const string& dir) const
{
	string ret;
	if (file[0] == '~') {
		ret = "/";
		ret += cxHomeDir;
		ret += file.substr(1);
	} else if (file[0] == '/') {
		ret = file;
	} else {
		ret = dir;
		ret += "/";
		ret += file;
	}
	// appending a slash eases globbing; we will remove it later
	ret += "/";
	string::size_type pos = ret.find("/./");
	while (pos != string::npos) {
		ret.erase(pos, 2);
		pos = ret.find("/./");
	}
	pos = ret.find("/../");
	while (pos != string::npos) {
		if (pos == 0) {
			ret = "/";
			return ret;
		}
		string::size_type pos1 = ret.rfind('/', pos - 1);
		if (pos1 == string::npos) {
			ret = "";
			return ret;
		}
		ret.erase(pos1, pos - pos1 + 3);
		pos = ret.find("/../");
	}
	pos = ret.find("//");
	while (pos != string::npos) {
		ret.erase(pos, 1);
		pos = ret.find("//");
	}
	cxAssert(!ret.empty() && ret[ret.size() - 1] == '/');
	if (ret.size() > 1) ret.resize(ret.size() - 1);
	return ret;
}

bool CFileHandler::TrashFile(const string& file)
{
	if (!Lock()) return false;
	string alias = GetAlias();
	if (alias.empty()) {
		Unlock();
		return false;
	}
	string to = mTrashDir;
	to += "/";
	to += alias;
	if (!MoveFile(file, to)) {
		Unlock();
		return false;
	}
	ofstream ost;
	ost.open(mIndexFile.c_str(), ios::out | ios::app);
	if (!ost) {
		Unlock();
		return false;
	}
	ost << alias << "=" << file << "\n";
	ost.close();
	Unlock();
	mLastTrashed.push_back(alias);
	return true;
}

bool CFileHandler::Lock()
{
	unsigned retry = 5;
	int fd = -1;
	timeval timeout;
	//gettimeofday(&now, NULL);
	timeout.tv_usec = 200000; // 200ms;
	timeout.tv_sec = 0;
	while (retry-- != 0) {
		fd = -1;
		do {
			// use low level I/O because streams do not fail to create an existing file
			fd = open(mLockFile.c_str(), O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
		} while (fd == -1 && errno == EINTR);
		if (errno != EEXIST) break;
		select(0, NULL, NULL, NULL, &timeout);
	}
	if (fd == -1) {
		cxError << _("Trash: failed to create lock file:\n{1}", 62) << strerror(errno) << CError::Error;
		return false;
	}
	while (close(fd) == -1 && errno == EINTR);
	return true;
}

string CFileHandler::GetAlias()
{
	string s;
	ifstream ist;
	ist.open(mAliasFile.c_str());
	if (!ist) {
		if (errno == ENOENT) {
			// does not exist: create;
			ist.close();
			ist.open(mIndexFile.c_str());
			if (!ist) {
				if (errno != ENOENT) {
					cxError << _("Trash: failed to read index:\n{1}", 63) << strerror(errno) << CError::Error;
					return s;
				}
			} else {
				string buf;
				while (ist) {
					getline(ist, buf);
					if (!buf.empty()) s = buf.substr(0, 8);
				}
			}
		} else {
			cxError << _("Trash: failed to read current alias:\n{1}", 64) << strerror(errno) << CError::Error;
			return s;
		}
	} else getline(ist, s);
	if (s.empty()) s = "AAAAAAAA";
	else if (s.size() != 8) {
		cxError << _("Trash: .last file corrupted.", 65) << CError::Error;
    	s = "";
     	return s;
  	}
	string::reverse_iterator i = s.rbegin();
	while (i != s.rend() && ++(*i) > 'Z') *i++ = 'A';
	cxAssert(i != s.rend());
	ofstream ost;
	ost.open(mAliasFile.c_str());
	if (!ost) {
		cxError << _("Trash: failed to write alias file:\n{1}", 66) << strerror(errno) << CError::Error;
		s = "";
		return s;
	}
	ost << s << '\n';
	return s;
}

string CFileHandler::GetOrigin(const string& alias)
{
	string s;
	ifstream ist;
	ist.open(mIndexFile.c_str());
	while (ist) {
		getline(ist, s);
		if (s.substr(0, 8) == alias) {
			if (s.size() < 10) break;
			if (s[8] != '=') break;
			return s.substr(9);
		}
	}
	s = "";
	return s;
}

CFileHandler::~CFileHandler()
{
	vector<string> allfiles;
	vector<string> oldfiles;
	string file = mCacheDir + "/";
	string::size_type sz = file.size();
	struct stat st;
	timeval age;
	gettimeofday(&age, NULL);
	age.tv_sec -= DAYS_IN_CACHE * 24 * 3600;
	if (ReadDirectory(mCacheDir, allfiles)) {
		unsigned dirsize = 0;
		for (vector<string>::iterator i = allfiles.begin(); i < allfiles.end(); ++i) {
			file.erase(sz);
			file += *i;
			if (stat(file.c_str(), &st) == 0) {
				dirsize += st.st_size;
				if (st.st_atime < age.tv_sec) oldfiles.push_back(file);
			}
		}
		if (dirsize > MAX_CACHE_SIZE * 1048576) Delete(oldfiles);
	}
}

void CFileHandler::ToUrl(string& file)
{
	string::size_type pos = file.find("://");
	if (pos == string::npos) {
		// local filesystem: add protocol
		file.insert(0, "file:///");
		// add hostname
		file.insert(6, cxApp->GetHostname());
	}
}


bool CFileHandler::ToLocalFS(string& url)
{
	if (url.size() < 8 || url.compare(0, 6, "file:/")) return false;
	if (url[6] != '/') {
		url = url.substr(5);
		return true;
	}
	url = url.substr(7);
	string::size_type pos = url.find('/');
	if (pos == string::npos) return false;
	if (pos) {
		if (url.compare(0, pos, cxApp->GetHostname()) && url.compare(0, pos, "localhost")) {
			url = "/" + url;
			return true;
		}
		url = url.substr(pos);
	}
	return true;
}

void CFileHandler::ReadSearchPath()
{
	const char* path = getenv("PATH");
	if (path == NULL || *path == 0) {
		// $PATH not set or empty
		mSearchPath.push_back("/usr/local/bin");
		mSearchPath.push_back("/usr/bin");
		mSearchPath.push_back("/bin");
		mSearchPath.push_back("/usr/bin/X11");
	} else {
		// $PATH is valid, read it
		string tmp = path;
		string::size_type beg = 0;
		string::size_type end = 0;
		while (end < tmp.size()) {
			end = tmp.find(':', beg);
			if (end > tmp.size()) end = tmp.size();
			mSearchPath.push_back(tmp.substr(beg, end - beg));
			beg = ++end;
		}
	}
}

// decompresses file and returns the name of the decompressed file in case of success or an
// empty string otherwise
string CFileHandler::Decompress(const string& file)
{
	if (file.empty()) return "";
	struct stat origstat;
	if (stat(file.c_str(), &origstat) != 0) origstat.st_mode = S_IRUSR;
	string cachefile = file;
	string::size_type pos = 1;
	while ((pos = cachefile.find('/', pos)) != string::npos) {
		cachefile.replace(pos, 1, "_$_");
	}
	cachefile = mCacheDir + cachefile;
	struct stat cachestat;
	if (stat(cachefile.c_str(), &cachestat) == 0) {
		// stat has been successful -> file exists
		if (origstat.st_mtime <= cachestat.st_mtime) return cachefile;
	}
	// file doesn't exist or is too old
	filebuf compin;
	if (compin.open(file.c_str(), ios_base::in) == NULL) {
		cxError << _("Failed to open {1} for decompression.", 450) << file << CError::Error;
		return "";
	}
	CZStreambuf bufin(compin, CZStreambuf::LARGE);
	ofstream out(cachefile.c_str());
	if (!out) return "";
	if (!bufin) return "";
	out << bufin;
	// restore permissions
	if (out) {
		chmod(cachefile.c_str(), origstat.st_mode);
		return cachefile;
	} else return "";
}

void CFileHandler::TestTrash()
{
	SyncTrash();
	vector<string> allfiles;
	vector<string> oldfiles;
	timeval now;
	gettimeofday(&now, NULL);
	string file;
	struct stat st;
	if (ReadDirectory(mTrashDir, allfiles)) {
		int age = 1;
		int tmpage;
		unsigned dirsize = 0;
		for (vector<string>::iterator i = allfiles.begin(); i < allfiles.end(); ++i) {
			file = mTrashDir + "/" + *i;
			if (stat(file.c_str(), &st) == 0) {
				dirsize += st.st_size;
				tmpage = now.tv_sec - st.st_ctime;
				tmpage /= 24 * 3600;
				if (tmpage >= age) {
					if (age < DAYS_IN_TRASH && tmpage > age) {
						oldfiles.clear();
						age = tmpage;
						if (age > DAYS_IN_TRASH) age = DAYS_IN_TRASH;
					}
					oldfiles.push_back(file);
				}
			}
		}
		if (dirsize >= MAX_TRASH_SIZE * 1048576 && !oldfiles.empty()) {
			string msg = _("Trashcan size exceeds {1} MB and {2} files are older than {3} days.\nShall these files be deleted?", 453);
			string::size_type pos = msg.find("{1}");
			if (pos != string::npos) {
				ostringstream os;
				os << MAX_TRASH_SIZE;
				msg.replace(pos, 3, os.str());
			}
			pos = msg.find("{2}");
			if (pos != string::npos) {
				ostringstream os;
				os << oldfiles.size();
				msg.replace(pos, 3, os.str());
			}
			pos = msg.find("{3}");
			if (pos != string::npos) {
				ostringstream os;
				os << age;
				msg.replace(pos, 3, os.str());
			}
			CMessageBox mb(msg, _("Trashcan full", 454), CMessageBox::YES | CMessageBox::NO, CIcon::WARNING);
			mb.FocusButton(CMessageBox::NO);
			if (mb.DoModal() == CMessageBox::YES) Delete(oldfiles);
		} 
	}
}

void CFileHandler::Untrash(const string& from, const string& todir)
{
	string src;
	string ofn = GetFileName(GetOrigin(from));
	src = GetAbsolutePath(from, GetTrashDir());
	string dest = todir;
	if (dest.empty()) dest = GetFilePath(GetOrigin(from));
	if (access(dest.c_str(), W_OK | X_OK) != 0) {
		cxError << _("Destination directory {1} does not exist or is not writeable.", 458) << dest << CError::Error;
		return;
	}
	dest = GetAbsolutePath(ofn, dest);
	if (access(dest.c_str(), F_OK) == 0) {
		string msg = _("A file {1} already exists. Overwrite it?", 459);
		string::size_type pos = msg.find("{1}");
		if (pos != string::npos) msg.replace(pos, 3, dest);
		CMessageBox mb(msg, _("File exists", 34), CMessageBox::YES | CMessageBox::NO, CIcon::WARNING);
		mb.FocusButton(CMessageBox::NO);
		if (mb.DoModal() == CMessageBox::YES) {
			if (!MoveFile(dest, dest + '~')) return;
		} else return;
	}
	if (MoveFile(src, dest)) SyncTrash();		
}

void CFileHandler::SyncTrash()
{
	ifstream in;
	ofstream out;
	string tmpidx = mIndexFile + ".bak";
	in.open(mIndexFile.c_str());
	out.open(tmpidx.c_str());
	if (!in) return;
	if (!out) return;
	string s;
	do {
		getline(in, s);
		if (in && s.size() > 10) {
			if (access(GetAbsolutePath(s.substr(0, 8), GetTrashDir()).c_str(), F_OK) == 0) {
				out << s << "\n";
			}
		}
	} while (in);
	in.close();
	out.close();
	MoveFile(tmpidx, mIndexFile);
}

bool CFileHandler::Execute(const string& cmd, const string& workdir)
{
	if (cmd.empty()) return false;
	if (cmd[0] == ' ') return false;
	// get options and remove escape characters
	string utfcmd = cmd;
	vector<char*> opt;
	string::size_type pos = 0;
	while ((pos = utfcmd.find(' ', pos)) < utfcmd.size()) {
		if (utfcmd[pos - 1] == '\\') utfcmd.erase(pos - 1, 1);
		else utfcmd[pos] = '\0';
	}
	utfcmd += '\0';
	// break down utfcmd
	pos = utfcmd.find('\0');
	string::size_type pos1 = utfcmd.rfind('/', pos);
	++pos1;
	// the path to the executable
	opt.push_back((char*)&utfcmd[0]);
	string wd = workdir;
	if (wd.empty()) wd = GetFilePath(opt.back());
	// it's name 
	opt.push_back((char*)&utfcmd[pos1]);
	// the options
	while (++pos < utfcmd.size()) {
		opt.push_back((char*)&utfcmd[pos]);
		pos = utfcmd.find('\0', pos);
	}
	opt.push_back(NULL);
	// cd to working directory
	chdir(wd.c_str());
	pid_t pid = fork();
	if (pid == 0) {
		if (execvp(opt[0], &opt[1]) != 0) exit(errno);
		exit(0);
	}
	if (pid < 0) {
		cxError << _("Failed to create child process", 58) << CError::Error;
		return false;
	}
	if (pid > 0) {
		std::signal(SIGCHLD, &cxxx::cxSigChildHandler);
	}
	return true;
}

void cxxx::cxSigChildHandler(int signum __attribute__ ((unused)))
{
	int save(errno);
	int status;
	waitpid(-1, &status, WNOHANG);
	errno = save;
}
