/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFileHandler.h
  Info   : $Id: CFileHandler.h,v 1.26 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CFileHandler_h
#define CFileHandler_h

// C++X headers
#include <cxxx/Cxxx.h>

// C++ headers
#include <vector>
#include <fstream>

// C headers
#include <csignal>

class CFileHandler
{
private:
	enum
	{
		// Function type
		MOVE,
		COPY,
		LINK
	};

	enum
	{
		// Memory management
		DAYS_IN_CACHE = 5,
		MAX_CACHE_SIZE = 25, // MByte
		DAYS_IN_TRASH = 30,
		MAX_TRASH_SIZE = 50  // MByte
	};

public:
	// Constructor
	CFileHandler()
	: mContinue(false),
	  mMore(false)
	{}
	
	// Destructor
	~CFileHandler();
    
	// Initialize file handler
	bool Init(const std::string& cxxxdir);

	// Move one file
	void Move(const std::string& from, const std::string& to)
	{
		mMore = false;
		MoveFile(from, to);
	}

	// Move multiple files
	void Move(const std::vector<std::string>& from, const std::string& todir)
	{
		HandleFiles(from, todir, MOVE);
	}

	// Copy one file 
	void Copy(const std::string& from, const std::string& to)
	{
		mMore = false;
		CopyFile(from, to);
	}

	// Copy multiple files
	void Copy(const std::vector<std::string>& from, const std::string& todir)
	{
		HandleFiles(from, todir, COPY);
	}

	// Link one file
	void Link(const std::string& from, const std::string& to)
	{
		mMore = false;
		LinkFile(from, to);
	}
	
	// Link multiple files
	void Link(const std::vector<std::string>& from, const std::string& todir)
	{
		HandleFiles(from, todir, LINK);
	}
	
	// Delete one file
	void Delete(const std::string& file)
	{
		mMore = false;
		DeleteFile(file);
	}
	
	// Delete multiple files
	void Delete(const std::vector<std::string>& files)
	{
		for (std::vector<std::string>::const_iterator i = files.begin(); i != files.end(); ++i) {
			mMore = ((i + 1) != files.end());
			if (!DeleteFile(*i) && !mContinue) return;
		}
	}
	
	// Restore one file from trash
	void Untrash(const std::string& from, const std::string& todir = "");
	
	// Empties a directory
	bool ClearDirectory(const std::string& dir)
	{
		std::vector<std::string> files;
		if (!ReadDirectory(dir, files)) return false;
		for (std::vector<std::string>::iterator i = files.begin(); i != files.end(); ++i) {
			*i = GetAbsolutePath(*i, dir);
		}
		Delete(files);
		return true;
	}
	
	// Empties trash directory
	bool ClearTrash()
	{
		if (!ClearDirectory(GetTrashDir())) return false;
		std::ofstream out;
		out.open(mIndexFile.c_str());
		bool ret = (bool)out;
		out.close();
		out.clear();
		out.open(mAliasFile.c_str());
		if (ret == true) ret = (bool)out;
		return ret;
	}
	
	// Moves one file to trash and returns the name of the trashed file
	const std::string& Trash(const std::string& file)
	{
		mMore = false;
		mLastTrashed.clear();
		TrashFile(file);
		if (!mLastTrashed.empty()) {
			TestTrash();
			return mLastTrashed[0];
		} else return cxxx::cxEmptyString;
	}
	
	// Moves multiple files to trash and returns the names of the trashed files
	const std::vector<std::string>& Trash(const std::vector<std::string>& files)
	{
		mMore = false;
		mLastTrashed.clear();
		TrashFiles(files);
		TestTrash();
		return mLastTrashed;
	}
	
	// Returns the file name part of path
	std::string GetFileName(const std::string& path)
	{
		std::string s;
		std::string::size_type pos = path.rfind('/');
		if (pos == std::string::npos) s = path;
		else if (++pos <= path.size()) s = path.substr(pos);
		return s;
	}
	
	// Returns the file path part of path
	std::string GetFilePath(const std::string& path)
	{
		std::string s;
		std::string::size_type pos = path.rfind('/');
		if (pos != std::string::npos && pos != 0) s = path.substr(0, pos);
		else s = "/";
		return s;
	}
	
	// Creates an URL of the file path file
	void ToUrl(std::string& file);
	
	// Converts an URL to a local file path
	bool ToLocalFS(std::string& url);
	
	// Returns the absolute path of file, with dir as current working directory
	std::string GetAbsolutePath(const std::string& file, const std::string& dir) const;
	
	// Reads a directory
	std::vector<std::string> ReadDirectory(const std::string& dir)
	{
		std::vector<std::string> v;
		mMore = false;
		ReadDirectory(dir, v);
		return v;
	}
	
	// Returns the trash directory
	const std::string& GetTrashDir() const
	{
		return mTrashDir;
	}
	
	// Returns the search path
	const std::vector<std::string>& GetSearchPath()
	{
		if (mSearchPath.empty()) ReadSearchPath();
		return mSearchPath;
	}
	
	// Decompresses file and returns the name of the decompressed file
	std::string Decompress(const std::string& file);
	
	// Returns the location from which file alias has been trashed
	std::string GetOrigin(const std::string& alias);
	
	// Ececutes cmd
	bool Execute(const std::string& cmd, const std::string& workdir = "");
	
protected:
	// Locks trash directory
	bool Lock();

	// Unlocks trash directory
	void Unlock()
	{
		Delete(mLockFile);
	}

	// Trash handling
	std::string GetAlias();
	void TestTrash();
	void SyncTrash();

	// Process multiple files
	void HandleFiles(const std::vector<std::string>& from, const std::string& todir, unsigned action);

	// Moves one file
	bool MoveFile(const std::string& from, const std::string& to);

	// Copies one file
	bool CopyFile(const std::string& from, const std::string& to);

	// Copies one link
	bool CopyLink(const std::string& from, const std::string& to);

	// Copies one directory
	bool CopyDir(const std::string& from, const std::string& to);
	
	// Copies one regular file
	bool CopyReg(const std::string& from, const std::string& to);
	
	// Links one file
	bool LinkFile(const std::string& from, const std::string& to);
	
	// Deletes one file
	bool DeleteFile(const std::string& file);
	
	// Deletes one directory
	bool DeleteDir(const std::string& file);
	
	// Trashes one file
	bool TrashFile(const std::string& file);
	
	// Trashes multiple files
	void TrashFiles(const std::vector<std::string>& files)
	{
		for (std::vector<std::string>::const_iterator i = files.begin(); i != files.end(); ++i) {
			mMore = ((i + 1) != files.end());
			if (!TrashFile(*i) && !mContinue) return;
		}
	}
	
	// Reads a directory and returns the files in ret
	bool ReadDirectory(const std::string& dir, std::vector<std::string>& ret);
	
	// Displays a message box to indicate an error
	void ErrorBox(const std::string& s1, const std::string& s2)
	{
		ErrorBox(s1, s2, s2);
	}

	void ErrorBox(const std::string& s1, const std::string& s2, const std::string& s3);
	
	// Acquires search path for executables
	void ReadSearchPath();

protected:
	bool mContinue;
	bool mMore;
	std::string mTrashDir;
	std::string mIndexFile;
	std::string mAliasFile;
	std::string mLockFile;
	std::vector<std::string> mLastTrashed;
	std::vector<std::string> mSearchPath;
	std::string mCacheDir;
};

// The SIGCHILD handler needed for Execute
namespace cxxx
{
	extern "C" void cxSigChildHandler(int signum);
}
	
#endif // CFileHandler_h
