/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFileListControl.cpp
  Info   : $Id: CFileListControl.cpp,v 1.64 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++ headers
#include <fstream>
#include <cerrno>

// cxxx headers
#include <cxxx/CFileListControl.h>
#include <cxxx/CDeskTopEntry.h>
#include <cxxx/CFileHandler.h>
#include <cxxx/CError.h>

// C headers
#include <unistd.h> // access

// namespace declarations
using std::string;
using std::wstring;
using std::vector;
using std::lower_bound;
using std::stable_sort;
using std::max;
using cxxx::cxError;
using cxxx::cxFileManager;
using cxxx::cxWiden;
using cxxx::filesort::HIDDEN;
using cxxx::filesort::NAME;
using cxxx::filesort::SIZE;
using cxxx::filesort::TIME;
using cxxx::filesort::TYPE;
using cxxx::filesort::ASCENDING;
using cxxx::filesort::DIRFIRST;

CFileListControl::CFileListControl(CWindow* parent, unsigned msgid)
: CListControl(parent, msgid),
  mFlags(0),
  mWriteAccess(false),
  mFileWatch(this, IDFILEWATCH),
  mIconSize(16)
{
	mLinkOverlay.Create(CIcon::LINK, mIconSize);
	mLockOverlay.Create(CIcon::LOCK, mIconSize);
	mZipOverlay.Create(CIcon::ZIP, mIconSize);
	mFileWatch.SetTime(5000);
}

/* virtual */ void CFileListControl::DrawItem(unsigned idx)
{
	cxAssert(idx < mList.size());
	CFileListItem item = GetItem(idx);
	unsigned row;
	unsigned col;
	// translate index to row and column
	if (mStyle == MULTI_COLUMN_VIEW) {
		col = idx / mRow.size();
		row = idx % mRow.size();
	} else {
		col = idx % mColumn.size();
		row = idx / mColumn.size();
	}
	int x;
	int y;
	if (col) x = mColumn[col - 1];
	else x = 0;
	if (row) y = mRow[row - 1];
	else y = 0;
	unsigned w = mColumn[col] - x;
	unsigned h = mRow[row] - y;
	WindowToScreen(x, y);
	ClearArea(x, y, w, h);
	int xi, yi;
	unsigned wi, hi;
	// Multi column view
	if (mStyle == MULTI_COLUMN_VIEW) {
		xi = x + ICON_BORDER + mIconWidth / 2;
		yi = y + h / 2;
		CColor col(TC_BACKGND_TEXT);
		if (item.GetIcon().IsValid()) DrawIcon(xi - item.GetIcon().GetWidth() / 2, yi - item.GetIcon().GetHeight() / 2, item.GetIcon());
		if (item.IsLink()) DrawIcon(xi - mLinkOverlay.GetWidth() / 2, yi - mLinkOverlay.GetHeight() / 2, mLinkOverlay);
		if (item.IsLocked()) DrawIcon(xi - mLockOverlay.GetWidth() / 2, yi - mLockOverlay.GetHeight() / 2, mLockOverlay);
		if ((item.GetType() & TMagic::COMPRESSED) != 0) DrawIcon(xi - mZipOverlay.GetWidth() / 2, yi - mZipOverlay.GetHeight() / 2, mZipOverlay);
		xi = x + mIconWidth + TEXT_BORDER;
		yi = y + (h - GetTextHeight()) / 2 -TEXT_BORDER;
		wi = item.GetWidth() - mIconWidth;
		hi = GetTextHeight() + 2 * TEXT_BORDER;
		if (item.IsSelected()) {
			if (IsFocused()) {
				FillRectangle(xi, yi, wi, hi, CColor(TC_TITLE));
				col = CColor(TC_TITLE_TEXT);
			} else {
				col = CColor(TC_BACKGND_TEXT);
				DrawDottedRectangle(xi, yi, wi, hi, col);
			}
		} else col = CColor(TC_BACKGND_TEXT);
		xi += TEXT_BORDER;
		yi += TEXT_BORDER;
		if (!item.GetText().empty()) DrawString(xi, yi, col, item.GetText(), !item.IsSelected());
	} else {
		// Icon view
		xi = x + w / 2;
		yi = y + ICON_BORDER + mIconHeight / 2;
		CColor col(TC_BACKGND_TEXT);
		if (item.GetIcon().IsValid()) DrawIcon(xi - item.GetIcon().GetWidth() / 2, yi - item.GetIcon().GetHeight() / 2, item.GetIcon());
		if (item.IsLink()) DrawIcon(xi - mLinkOverlay.GetWidth() / 2, yi - mLinkOverlay.GetHeight() / 2, mLinkOverlay);
		if ((item.GetType() & TMagic::COMPRESSED) != 0) DrawIcon(xi - mZipOverlay.GetWidth() / 2, yi - mZipOverlay.GetHeight() / 2, mZipOverlay);
		yi = y + mIconHeight;
		hi = GetTextHeight() + 2 * TEXT_BORDER;
		unsigned pos = 0;
		unsigned line = 0;
		CTextSizeHint hint = item.GetTextHint(0);
		wstring wtxt = cxWiden(item.GetText());
		while (hint.mLength != 0) {
			xi = x + (w - hint.mWidth) / 2;
			wi = hint.mWidth;
			if (item.IsSelected()) {
				if (IsFocused()) {
					FillRectangle(xi, yi, wi, hi, CColor(TC_TITLE));
					col = CColor(TC_TITLE_TEXT);
				} else {
					col = CColor(TC_BACKGND_TEXT);
					DrawDottedRectangle(xi, yi, wi, hi, col);
				}
			} else col = CColor(TC_BACKGND_TEXT);
			xi += TEXT_BORDER;
			yi += TEXT_BORDER;
			DrawString(xi, yi, col, wtxt.substr(pos, hint.mLength), !item.IsSelected());
			yi += GetTextHeight() + TEXT_BORDER;
			pos += hint.mLength;
			hint = item.GetTextHint(++line);
		}
	}
	// restore the border
	if ((x < BORDER_WIDTH) || (x + w + BORDER_WIDTH >= GetWidth()) || (y + h + BORDER_WIDTH >= GetHeight())) {
		DrawFrame();
	}		
}

void CFileListControl::Reload()
{
	vector<string> sel;
	if (!mSelection.empty()) {
		for (vector<unsigned>::iterator i = mSelection.begin(); i < mSelection.end(); ++i) sel.push_back(GetItem(*i).GetText());
	}
	ClearList();
	ReadDirectory();
	if (!sel.empty()) {
		for (std::vector<std::string>::iterator i = sel.begin(); i < sel.end(); ++i) {
			for (unsigned j = 0; j < mList.size(); ++j) {
				if (GetItem(j).GetText() == *i) {
					GetItem(j).Select(true);
					mSelection.push_back(j);
				}
			}
		}
	}
	Update();
}

/* virtual */ void CFileListControl::ReadDirectory()
{
	mFileWatch.SetTime(5000);
	mWriteAccess = false;
	cxAssert(mList.empty());
	if (mCurrentDir.empty()) return;
	mWriteAccess = (access(mCurrentDir.c_str(), W_OK | X_OK) == 0);
	struct stat st;
	// locked directory
	if (stat(mCurrentDir.c_str(), &st) != 0) {
		mFileTime = 0;
		return;
	}
	mFileTime = st.st_mtime;
	vector<string> dir = cxFileManager->ReadDirectory(mCurrentDir);
	if (dir.size() > 150) DisplayWaitCursor();
	string path = mCurrentDir;
	if (path.empty() || path[path.size() - 1] != '/') path += '/';
	unsigned pathlen = path.size();
	for (vector<string>::iterator i = dir.begin(); i != dir.end(); ++i) {
		if ((mFlags & HIDDEN) == 0 && (*i)[0] == '.') continue;
		path.resize(pathlen);
		path += *i;
		Append(path, *i);
	}
	Sort();
	UndisplayWaitCursor();
}

/* virtual */ void CFileListControl::Append(const string& path, const string& name)
{
	unsigned flags = 0;
	unsigned typ;
	wstring tmp = cxWiden(name);
	struct stat st;
	if (lstat(path.c_str(), &st)) {
		typ = TMagic::LOCKED;
		flags |= CFileListItem::LOCKED;
	} else {
		if (S_ISLNK(st.st_mode)) flags |= CFileListItem::LINKED;
		typ = mMagic.GetFileType(path);
		if (typ == TMagic::DIRECTORY) {
			flags |= CFileListItem::DIRECTORY;
			if (access(path.c_str(), X_OK)) flags |= CFileListItem::LOCKED;
		} else if (access(path.c_str(), R_OK)) flags |= CFileListItem::LOCKED;
	}
	// filter file types
	if (!mFilter.empty()) {
		vector<unsigned>::const_iterator it = std::find(mFilter.begin(), mFilter.end(), typ);
		if (it == mFilter.end()) return;
	}
	if (flags & CFileListItem::LOCKED) {
		st.st_size = 0;
		st.st_mtime = 0;
	}
	CIcon icon(CMagic::GetIcon(typ), mIconSize);
	unsigned h;
	unsigned w;
	vector<CTextSizeHint> hint;
	if (mStyle == MULTI_COLUMN_VIEW) {
		h = max(mIconHeight, (unsigned)GetTextHeight() + 2 * TEXT_BORDER);
		w = mIconWidth + GetTextWidth(tmp) + 2 * TEXT_BORDER;
	} else {
		// mStyle == ICON_VIEW
		w = ICON_ITEM_WIDTH;
		// Text may be multiline, do the calculation
		unsigned d;
		unsigned pos;
		while (!tmp.empty()) {
			d = tmp.size() + 1;
			pos = 0;
			do {
				d = (d + 1) / 2;
				if (pos + d <= tmp.size() && (unsigned)GetTextWidth(tmp, pos + d) <= w) pos += d;
			} while (d > 1);
			hint.push_back(CTextSizeHint(pos, GetTextWidth(tmp, pos) + 2 * TEXT_BORDER));
			tmp = tmp.substr(pos);
		}
		h = mIconHeight + (GetTextHeight() + 2 * TEXT_BORDER) * hint.size();
	}
	CFileListItem it(icon, name, st, typ, w, h);
	if (flags & CFileListItem::LOCKED) it.IsLocked(true);
	if (flags & CFileListItem::LINKED) it.IsLink(true);
	if (flags & CFileListItem::DIRECTORY) it.IsDirectory(true);
	for (vector<CTextSizeHint>::iterator i = hint.begin(); i != hint.end(); ++i) it.SetTextHint(i->mLength, i->mWidth);
	mList.push_back(it);
}

unsigned CFileListControl::GetIndex(const CFileListItem& item) const
{
	const CFileListItem* i = NULL;
	CFileListItem* beg = (CFileListItem*) &mList[0];
	CFileListItem* end = (CFileListItem*) &mList[mList.size()];
	if (mFlags & ASCENDING) {
		if (mFlags & NAME) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompNameAD);
			else i = lower_bound(beg, end, item, CompNameA);
		} else if (mFlags & TIME) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompTimeAD);
			else i = lower_bound(beg, end, item, CompTimeA);
		} else if (mFlags & SIZE) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompSizeAD);
			else i = lower_bound(beg, end, item, CompSizeA);
		} else i = lower_bound(beg, end, item, CompTypeA);
	} else {
		if (mFlags & NAME) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompNameDD);
			else i = lower_bound(beg, end, item, CompNameD);
		} else if (mFlags & TIME) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompTimeDD);
			else i = lower_bound(beg, end, item, CompTimeD);
		} else if (mFlags & SIZE) {
			if (mFlags & DIRFIRST) i = lower_bound(beg, end, item, CompSizeDD);
			else i = lower_bound(beg, end, item, CompSizeD);
		} else i = lower_bound(beg, end, item, CompTypeD);
	}
	if (i->GetText() == item.GetText()) return i - beg;
	else return -1u;
}

void CFileListControl::Sort()
{
	CFileListItem* beg = (CFileListItem*)&mList[0];
	CFileListItem* end = (CFileListItem*)&mList[mList.size()];
	if (mFlags & ASCENDING) {
		if (mFlags & NAME) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompNameAD);
			else stable_sort(beg, end, CompNameA);
		} else if (mFlags & TIME) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompTimeAD);
			else stable_sort(beg, end, CompTimeA);
		} else if (mFlags & SIZE) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompSizeAD);
			else stable_sort(beg, end, CompSizeA);
		} else stable_sort(beg, end, CompTypeA);
	} else {
		if (mFlags & NAME) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompNameDD);
			else stable_sort(beg, end, CompNameD);
		} else if (mFlags & TIME) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompTimeDD);
			else stable_sort(beg, end, CompTimeD);
		} else if (mFlags & SIZE) {
			if (mFlags & DIRFIRST) stable_sort(beg, end, CompSizeDD);
			else stable_sort(beg, end, CompSizeD);
		} else stable_sort(beg, end, CompTypeD);
	}
}

/* virtual */ void CFileListControl::OnMessage(unsigned msgid, TMessage type, unsigned value)
{
	switch (msgid) {
	case IDFILEWATCH:
		switch (type) {
		case MT_TIMER:
			mFileWatch.SetTime(5000);
			struct stat st;
			if (stat(mCurrentDir.c_str(), &st) != 0) st.st_mtime = 0;
			if (mFileTime != st.st_mtime) {
				Reload();
				mFileTime = st.st_mtime;
			}
			break;
		default:
			break;
		}
		break;
	default:
		CListControl::OnMessage(msgid, type, value);
		break;
	}
}

/* virtual */ void CFileListControl::OnDblClick(XButtonEvent& e)
{
	if (e.button == 1 && GetSelCount() == 1) {
		unsigned idx = GetSelection().front();
		cxAssert(idx != -1u);
		if (!GetItem(idx).IsLocked()) PostMessage(MT_CONFIRM_SELECTION);
	}
	CancelDrag();
}

/* virtual */ void CFileListControl::SetStyle(unsigned style, unsigned wico, unsigned hico)
{
	mIconSize = wico;
	mLinkOverlay.Create(CIcon::LINK, mIconSize);
	mLockOverlay.Create(CIcon::LOCK, mIconSize);
	mZipOverlay.Create(CIcon::ZIP, mIconSize);
	CListControl::SetStyle(style, wico, hico);
}
