/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFileListControl.h
  Info   : $Id: CFileListControl.h,v 1.31 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CFileListControl_h
#define CFileListControl_h

// C++X headers
#include <cxxx/CListControl.h>
#include <cxxx/CMagic.h>
#include <cxxx/CFileHandler.h>

// C headers
#include <sys/stat.h>

class CFileListControl : public CListControl
{
private:
	enum
	{
		// Message IDs
		IDFILEWATCH = cxxx::msgid::MSGID_CFileListControl
	};

public:
	// Constructor
	CFileListControl(CWindow* parent, unsigned msgid);

	// Set the directory to be displayed
	void SetDirectory(const std::string& dir)
	{
		if (dir.empty()) return;
		std::string tmp = cxxx::cxFileManager->GetAbsolutePath(dir, mCurrentDir);
		if (tmp != mCurrentDir) {
			mCurrentDir = tmp;
			Reload();
		}
	}

	// Returns the absolute file path of item at index idx
	std::string GetPath(unsigned idx) const
	{
		std::string ret(mCurrentDir);
		if (idx < mList.size())  {
			if (ret[ret.size() - 1] != '/') ret += '/';
			ret += GetItem(idx).GetText();
		}
		return ret;
	}

	// Returns the item at index idx
	CFileListItem& GetItem(unsigned idx) const
	{
		return (CFileListItem&)mList[idx];
	}

	// Reloads the current directory
	void Reload();

	// Sets the sort critereon
 	void SetSortKey(unsigned key)
	{
		mFlags = key;
		Reload();
	}

	// Returns the sort critereon
	unsigned GetSortKey() const
	{
		return mFlags;
	}

	// Returns the index of item 
	unsigned GetIndex(const CFileListItem& item) const;

	// C++X message processing
	virtual void OnMessage(unsigned msgid, TMessage type, unsigned value);

	// Sets the style; wico and hico are icon width and height
	virtual void SetStyle(unsigned style, unsigned wico, unsigned hico);

	// Adds a filter; only items that match filter are displayed
	// If no filters at all are added or if filter DELETE is added, any file is displayed
	void AddFilter(unsigned typ)
	{
		if (typ == TMagic::DELETE) mFilter.clear();
		else mFilter.push_back(typ);
	}

	// Comparison functions
	static bool CompNameA(const CFileListItem& i1, const CFileListItem& i2)
	{
		return cxxx::cxLocale(i1.GetText(), i2.GetText());
	}

	static bool CompTimeA(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i1.GetTime() - i2.GetTime();
		if (cmp) return cmp < 0;
		return CompNameA(i1, i2);
	}

	static bool CompTypeA(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		cmp = i1.GetType() - i2.GetType();
		if (cmp) return cmp < 0;
		return CompNameA(i1, i2);
	}

	static bool CompSizeA(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i1.GetSize() - i2.GetSize();
		if (cmp) return cmp < 0;
		return CompNameA(i1, i2);
	}

	static bool CompNameD(const CFileListItem& i1, const CFileListItem& i2)
	{
		return CompNameA(i2, i1);
	}

	static bool CompTimeD(const CFileListItem& i1, const CFileListItem& i2)
	{
		return CompTimeA(i2, i1);
	}

	static bool CompTypeD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i1.IsDirectory() - i2.IsDirectory();
		if (cmp) return cmp < 0;
		cmp = i1.GetType() - i2.GetType();
		if (cmp) return cmp < 0;
		return CompNameA(i2, i1);
	}

	static bool CompSizeD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i1.GetSize() - i2.GetSize();
		if (cmp) return cmp < 0;
		return CompNameA(i1, i2);
	}

	static bool CompNameAD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompNameA(i1, i2);
	}

	static bool CompTimeAD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompTimeA(i1, i2);
	}

	static bool CompSizeAD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompSizeA(i1, i2);
	}

	static bool CompNameDD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompNameD(i1, i2);
	}

	static bool CompTimeDD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompTimeD(i1, i2);
	}

	static bool CompSizeDD(const CFileListItem& i1, const CFileListItem& i2)
	{
		int cmp = i2.IsDirectory() - i1.IsDirectory();
		if (cmp) return cmp < 0;
		return CompSizeD(i1, i2);
	}
	
	// X event processing
	virtual void OnDblClick(XButtonEvent& e);
	
protected:
	// Reads the current directory
	virtual void ReadDirectory();

	// Draws the item with index idx
	virtual void DrawItem(unsigned idx);

	// Adds an item to the directory
	virtual void Append(const std::string& path, const std::string& name);
    
	// Sorts the items according to the sort critereon
	void Sort();
	
protected:
	unsigned mFlags;
	std::string mCurrentDir;
	CMagic mMagic;
	CIcon mLinkOverlay;
	CIcon mLockOverlay;
	CIcon mZipOverlay;
	bool mWriteAccess;
	time_t mFileTime;
	CTimer mFileWatch;
	std::vector<unsigned> mFilter;
	unsigned mIconSize;
};

#endif // CFileListControl_h
