/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFileTree.cpp
  Info   : $Id: CFileTree.cpp,v 1.55 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CFileTree.h>
#include <cxxx/CError.h>

// C++ headers
#include <cerrno>  // errno
#include <cstring> // strerror

// C headers
#include <dirent.h> // DIR
#include <sys/stat.h> // stat
#include <unistd.h>
#ifdef CXXX_HAS_MNTENT
#include <mntent.h> // setmntent, getmntent, endmntent
#endif // CXXX_HAS_MNTENT
#ifdef CXXX_HAS_MNTINFO
#include <sys/mount.h>
#endif // CXXX_HAS_MNTINFO

// namespace declarations
using std::string;
using std::wstring;
using std::vector;
using cxxx::cxError;
using cxxx::cxNarrow;
using cxxx::cxWiden;
using cxxx::filesort::HIDDEN;
using cxxx::lib::cxgettext;

// Constructor
// Creates a CFileTree object. Initially, the opened root folder is shown.
// Every 5 seconds the tree is reloaded to reflect changes in the file

CFileTree::CFileTree(CWindow *parent, unsigned msgid)
: CTreeControl(parent, msgid),
  mFileWatch(this, IDFILEWATCH),
  mIconSize(16)
{
	mFileTime = 0;
	mFlags = 0;
	mFileWatch.SetTime(5000);
	mLockOverlay.Create(CIcon::LOCK, mIconSize);
	mLinkOverlay.Create(CIcon::LINK, mIconSize);
	SetTree (CIcon::FOLDER, CIcon::FOLDER_OPEN, "/");
	mTree.Open();
	ReadDir(mTree);
}

// DrawItem
// Draws the item with index idx in the list

/* virtual */ void CFileTree::DrawItem(unsigned idx)
{
	if(!IsVisible()) return;
	CTreeItem& item(GetItem(idx));
	int x = 0;
	int y = idx * mIconHeight;
	unsigned w = item.GetIndentation() * INDENT_STEP + item.GetWidth() + INDENT_STEP;
	unsigned h = mIconHeight;
	WindowToScreen(x, y);
	ClearArea(x, y, w, h);
	CColor col(TC_BACKGND_TEXT);
	x += item.GetIndentation() * INDENT_STEP + (INDENT_STEP + 1) / 2;
	CTreeItem itmp(item);
	if (!item.IsTree()) itmp = itmp.GetParent();
	int x1 = x;
	while (!itmp.IsTree()) {
		x1 -= INDENT_STEP;
		if (itmp.GetParent().GetChildren().back() !=  itmp) DrawLine(x1, y, x1, y + h - 1, col);
		itmp = itmp.GetParent();
	}
	x1 = x + INDENT_STEP - (INDENT_STEP + 1) / 2;
	int y1 = y;
	y += mIconHeight / 2;
	if (item.GetChildren().empty() == false) {
		if (item.IsOpened()) DrawIcon(x - 4, y - 4, mMinus, col);
		else DrawIcon(x - 4, y - 4, mPlus, col);
		if (!item.IsTree()) DrawLine(x, y1, x, y - 5, col);
		DrawLine(x + 5, y, x1 - 1, y, col);
		if (!item.IsTree() && item.GetParent().GetChildren().back() !=  item) DrawLine(x, y + 5, x, y1 + h - 1, col);
	} else {
		if (!item.IsTree()) DrawLine(x, y1, x, y, col);
		DrawLine(x, y, x1 - 1, y, col);
		if (!item.IsTree() && item.GetParent().GetChildren().back() !=  item) DrawLine(x, y, x, y1 + h - 1, col);
	}
	DrawIcon(x1 + ICON_BORDER, y1 + ICON_BORDER, item.GetIcon());
	if (item.IsLocked()) DrawIcon(x1 + ICON_BORDER, y1 + ICON_BORDER, mLockOverlay);
	if (item.IsLink()) DrawIcon(x1 + ICON_BORDER, y1 + ICON_BORDER, mLinkOverlay);
	x = x1 + mIconWidth;
	y = y1 +  (mIconHeight - GetTextHeight()) / 2;
	if ( item.IsSelected()) {
		if (IsFocused()) {
			FillRectangle(x, y - TEXT_BORDER, item.GetWidth() - mIconWidth, GetTextHeight() + 2 * TEXT_BORDER, CColor(TC_TITLE));
			col = CColor(TC_TITLE_TEXT);
		} else {
			DrawDottedRectangle(x, y - TEXT_BORDER, item.GetWidth() - mIconWidth - 1, GetTextHeight() + 2 * TEXT_BORDER - 1, col);
		}
	}
	if (!item.GetText().empty()) DrawString(x + TEXT_BORDER, y, col, item.GetText(), false);
	x = 0;
	y = idx * mIconHeight;
	WindowToScreen(x, y);
}

// UpdateDir
// If 'update' is true, 'item' is synchronized with the file system.
// If 'item' is opened, its subdirectories are syncronized with the file system.
// The function returns true if 'item's branch changed in a way that is visible
// to the user, otherwise false.

bool CFileTree::UpdateDir(CTreeItem& item, bool update)
{
	string path = GetPath(item);
	bool retval = false;
	
	if (update) {
		vector<CTreeItem> tmpchild;
		tmpchild.swap(item.GetChildren());
		if (tmpchild.empty()) {
			// old status: no subdirectories
			if (GetStatus(path) == SUBDIRS) {
				Append(item, CIcon::FOLDER, CIcon::FOLDER, "");
				retval = true;
			} 
		} else if (tmpchild.front().GetText().empty() || !item.IsOpened()) {
			// old status: has subdirectories that are not yet read
			// or has subdirectories but the folder is closed, thus making
			// the child folders invisible
			if (GetStatus(path) == SUBDIRS) tmpchild.swap(item.GetChildren());
			else retval = true;
		} else {
			// the subdirectories are plainly visible to the user:
			// do the whole update process
				
			//open stream
			DIR *d = opendir(path.c_str());
			if (!d) {
				cxError << _("Failed to open directory:\n{1}\n{2}", 67) << path << strerror(errno) << CError::Error;
				item.Close();
				item.GetChildren().clear();
				item.IsLocked(true);
				return true;
			}
			//read directory entries
			dirent *ent;
			unsigned status;
			path += '/';
			unsigned s = path.size();
			bool lnk;
			struct stat buf;
			cxAssert(item.GetChildren().empty());
			while ((ent = readdir(d))) {
				if ((mFlags & HIDDEN) == 0 && ent->d_name[0] == '.') continue;
				if (!memcmp(ent->d_name, ".", 2)) continue;
				if (!memcmp(ent->d_name, "..", 3)) continue;
				path.resize(s);
				path += ent->d_name;
				lnk = false;
				//failed to read file status, most propably access denied.
				//anyway, we cannot list it, since we don't know it's type
				if (lstat(path.c_str(), &buf)) continue;
				else {
					if (S_ISLNK(buf.st_mode)) lnk = true;
					if (stat(path.c_str(), &buf)) continue;
				}
				if (S_ISDIR(buf.st_mode)) {
					// found a child, check whether it's new
					bool newchild = true;
					vector<CTreeItem>::iterator i;
					for (i = tmpchild.begin(); i != tmpchild.end(); ++i) {
						if (i->GetText() == ent->d_name) {
							newchild = false;
							break;
						}
					}
					if (newchild) {
						// new child
						status = GetStatus(path);
						Append(item, CIcon::FOLDER, CIcon::FOLDER_OPEN, ent->d_name);
						if (status == ISLOCKED) item.GetChildren().back().IsLocked(true);
						else if (status == SUBDIRS) Append(item.GetChildren().back(), CIcon::FOLDER, CIcon::FOLDER, "");
						if (lnk) item.GetChildren().back().IsLink(true);
						retval = true;
					} else {
						item.GetChildren().push_back(*i);
						if (UpdateDir(*i, (buf.st_mtime > mFileTime))) retval = true;
						tmpchild.erase(i);
					}
				}
			}
			if (!tmpchild.empty()) retval = true;
			if (closedir(d)) cxError << _("Failed to close directory\n{1}\n{2}", 68) << path << strerror(errno) << CError::Fatal;
			Sort(item);
		} 
	} else {
		if (item.IsOpened()) {
			// the subdirectories are plainly visible to the user:
			// update them all
			struct stat buf;
			path += '/';
			unsigned s = path.size();
			for (vector<CTreeItem>::iterator i = item.GetChildren().begin(); i != item.GetChildren().end(); ++i) {
				path.resize(s);
				path += i->GetText();
				if (stat(path.c_str(), &buf)) {
					item.GetChildren().erase(i--);
					if (item.GetChildren().empty()) item.Close();
					retval = true;
				} else if (UpdateDir(*i, (buf.st_mtime > mFileTime))) retval = true;
			}
		}
	}		
	return retval;
}

// ReadDir
// Discards any subdirectories in 'item', if any, and recreates them again.

void CFileTree::ReadDir(CTreeItem& item)
{
	item.GetChildren().clear();
	string path = GetPath(item);
	//open stream
	DIR *d = opendir(path.c_str());
	if (!d) {
		cxError << _("Failed to open directory:\n{1}\n{2}", 67) << path << strerror(errno) << CError::Error;
		item.Close();
		item.IsLocked(true);
		return ;
	}
	//read directory entries
	dirent *ent;
	struct stat buf;
	unsigned status;
	path += '/';
	unsigned s = path.size();
	cxAssert(item.GetChildren().empty());
	bool lnk;
	while ((ent = readdir(d))) {
		if ((mFlags & HIDDEN) == 0 && ent->d_name[0] == '.') continue;
		if (!memcmp(ent->d_name, ".", 2)) continue;
		if (!memcmp(ent->d_name, "..", 3)) continue;
        path.resize(s);
		path += ent->d_name;
		lnk = false;
		//failed to read file status, most propably access denied.
		//anyway, we cannot list it, since we don't know it's type
        if (lstat(path.c_str(), &buf)) {
			errno = 0;
			continue;
		} else {
			if (S_ISLNK(buf.st_mode)) lnk = true;
			if (stat(path.c_str(), &buf)) {
				errno = 0;
				continue;
			}
		}
		if (S_ISDIR(buf.st_mode)) {
			status = GetStatus(path);
			Append(item, CIcon::FOLDER, CIcon::FOLDER_OPEN, ent->d_name);
			if (status == ISLOCKED) item.GetChildren().back().IsLocked(true);
			else if (status == SUBDIRS) Append(item.GetChildren().back(), CIcon::FOLDER, CIcon::FOLDER_OPEN, "");
			if (lnk) item.GetChildren().back().IsLink(true);
		}
	}
	if (closedir(d)) cxError << _("Failed to close directory\n{1}\n{2}", 68) << path << strerror(errno) << CError::Fatal;
	Sort(item);
}

// GetStatus
// Checks whether a directory has subdirectories or not or if it is not
// accessible

unsigned CFileTree::GetStatus(const string& dir)
{
	unsigned status = NOSUBDIR;
	string tmp(dir);
	DIR *d = opendir(dir.c_str());
	if (!d) {
		if (errno != EACCES)
			cxError << _("Failed to open directory\n{1}\n{2}", 69) << dir << strerror(errno) << CError::Warn;
		errno = 0;
		return ISLOCKED;
	}
	tmp += '/';
	unsigned s = tmp.size();
	dirent *ent;
	struct stat buf;
	while ((ent = readdir(d))) {
		if ((mFlags & HIDDEN) == 0 && (ent->d_name[0] == '.')) continue;
		if (!memcmp(ent->d_name, ".", 2)) continue;
		if (!memcmp(ent->d_name, "..", 3)) continue;
		tmp.resize(s);
		tmp += ent->d_name;
		if (stat(tmp.c_str(), &buf)) {
			errno = 0;
			continue;
		}
		if (S_ISDIR(buf.st_mode)) {
			status = SUBDIRS;
			break;
		}
	}
	if (closedir(d)) cxError << _("Failed to close directory\n{1}\n{2}", 68) << dir << strerror(errno) << CError::Fatal;
	return status;
}

// SetDirectory
// Parses 'dir', openes all directories in path

void CFileTree::SetDirectory(const string& dir)
{
	if (GetSelCount() != 0) DeselectAll();
	CTreeItem item = mTree;
	if (item.GetChildren().empty() || item.GetChildren().front().GetText().empty()) ReadDir(item);
	if (dir.size() > 1) {
		string::size_type beg = 0;
		string::size_type end(beg);
		while (end != string::npos) {
			++beg;
			end = dir.find('/', beg);
			string cur = dir.substr(beg, end - beg);
			if (item.GetChildren().empty() || item.GetChildren().front().GetText().empty()) ReadDir(item);
			item.Open();
			vector<CTreeItem>::iterator i;
			for (i = item.GetChildren().begin(); i < item.GetChildren().end(); ++i) {
				if (i->GetText() == cur) break;
			}
			if (i == item.GetChildren().end()) break;
			item = *i;
			beg = end;
		}
	}
	mList.clear();
	mList.push_back(mTree);
	if (mTree.IsOpened()) Sync(mTree);
	unsigned idx = GetIndex(item);
	cxAssert(idx < mList.size());
	Update();
	Select(idx);
}

// OnMessage
// C++X message handling. Every 5 seconds, Reload() is called.

/* virtual*/ void CFileTree::OnMessage(unsigned msgid, TMessage type, unsigned value)
{
	if (msgid == IDFILEWATCH && type == MT_TIMER) {
		mFileWatch.SetTime(5000);
		Reload();
	} else CTreeControl::OnMessage(msgid, type, value);
}

// Open
// Openes the tree item with index 'idx'.

/* virtual */ void CFileTree::Open(unsigned idx)
{
	CTreeItem item(GetItem(idx));
	if (item.GetChildren().empty() || item.GetChildren().front().GetText().empty()) ReadDir(item);
	CTreeControl::Open(idx);
}

// Append
// Appends a subdirectory to 'parent'. 'closed' and 'opened' are file names to
// the icons of the respective state. 'text' is the name of the directory.

/* virtual */ void CFileTree::Append(CTreeItem& parent, CIcon::TIconName closed, CIcon::TIconName opened, const std::string& text)
{
	unsigned w = GetTextWidth(cxWiden(text)) + mIconWidth + 2 * TEXT_BORDER;
	parent.GetChildren().push_back(CTreeItem(parent, closed, opened, mIconSize, text, w, mIconWidth));
}

// Reload
// Brings the widget in sync with the file system.

void CFileTree::Reload()
{
	unsigned idx = 0;
	if (GetSelCount() != 0) {
		cxAssert(GetSelCount() == 1);
		idx = GetSelection().front();
		mList[idx].Select(false);
		mSelection.clear();
	}
	std::string sel;
	if (idx < GetItemCount()) sel = GetPath(GetItem(idx));
	else sel = "/";
	struct stat buf;
	if (stat("/", &buf)) cxError << _("Failed to read file status of root directory", 488) << CError::Fatal; 
	if ((UpdateMountPoints() | UpdateDir(mTree, (buf.st_mtime > mFileTime)))) {
		// something has changed, make it visible:
		// find the previously selected item
		bool selchange = false;
		CTreeItem item = mTree;
		if (sel.size() > 1) {
			string::size_type beg = 0;
			string::size_type end = beg;
			while (end != string::npos) {
				++beg;
				end = sel.find('/', beg);
				string cur = sel.substr(beg, end - beg);
				vector<CTreeItem>::iterator i;
				for (i = item.GetChildren().begin(); i < item.GetChildren().end(); ++i) {
					if (i->GetText() == cur) break;
				}
				if (i == item.GetChildren().end()) {
					selchange = true;
					break;
				}
				item = *i;
				beg = end;
			}
		}
		// 'item' is the previously selected item
		// syncronize list with tree
		mList.clear();
		mList.push_back(mTree);
		if (mTree.IsOpened()) Sync(mTree);
		// reselect previously selected item
		unsigned newidx = GetIndex(item);
		cxAssert(newidx < mList.size());
		cxAssert(!IsSelected(newidx));
		mList[newidx].Select(true);
		mSelection.push_back(newidx);
		if (idx != newidx) {
			newidx += mVScrollBar.GetPosition() - idx;
			Update();
			mVScrollBar.SetPosition(newidx);
		} else Update();
		if (selchange) PostMessage(MT_SELECTION_CHANGED); 
	} else {
		if (!mList.empty()) {
			mList[idx].Select(true);
			mSelection.push_back(idx);
		}
	}
	mFileTime = time(NULL);
}

// UpdateMountPoints
// Checks mtab for file system mount changes and synchronizes the tree.

bool CFileTree::UpdateMountPoints()
{
#ifndef CXXX_HAS_MNTENT
#ifndef CXXX_HAS_MNTINFO
	// No way to check mtab; pretend nothing has changed
	return false;
#endif // CXXX_HAS_MNTINFO
#endif // CXXX_HAS_MNTENT

	static CMountTable mpt;
	CMountTable tmp_mpt;

#ifdef CXXX_HAS_MNTENT
	static string path_mounted;
	CMountTableEntry tmp;
	// use /proc/mounts, if possible, because the automounter may not use /etc/mtab
	if (path_mounted.empty()) {
		if (access("/proc/mounts", R_OK)) {
			if (access("/etc/mtab", R_OK)) {
				if (access(_PATH_MOUNTED, R_OK)) path_mounted = "none";
				else path_mounted = _PATH_MOUNTED;
			} else path_mounted = "/etc/mtab";
		} else path_mounted = "/proc/mounts";
	}
	// mtab not found: give up
	if (path_mounted == "none") return false;
	// acquire mount points:
	FILE* mtab = setmntent (path_mounted.c_str(), "r");
	// not available: give up
	if (!mtab) return false;
	// iterate through the list
	struct mntent* entry;
	while ((entry = getmntent(mtab))) {
		tmp.mDevice = entry->mnt_fsname;
		tmp.mMountPoint = entry->mnt_dir;
		tmp.mValid = true;
		tmp_mpt.push_back(tmp);
	}
	// done, close the file
	endmntent(mtab);
#endif // CXXX_HAS_MNTENT

#ifdef CXXX_HAS_MNTINFO
	CMountTableEntry tmp;
	struct statfs* entries;
	// we do not want to wait, it's not that important we're up to date;
	// we'll get it next time
	int entrycount = getmntinfo(&entries, MNT_NOWAIT);
	for (int i = 0; i < entrycount; ++i) {
		tmp.mDevice = entries[i].f_mntfromname;
		tmp.mMountPoint = entries[i].f_mntonname;
		tmp.mValid = true;
		tmp_mpt.push_back(tmp);
	}
	// done
#endif // CXXX_HAS_MNTINFO

	// invalidate old mount point table
	for (CMountTable::iterator i = mpt.begin(); i != mpt.end(); ++i) i->mValid = false;
	std::vector<std::string> mp;
	// find new or altered entries
	CMountTable::iterator i;
	for (CMountTable::iterator j = tmp_mpt.begin(); j < tmp_mpt.end(); ++j) {
		// skip irrelevant entries
		if (j->mDevice[0] != '/') continue;
		// search in the old table
		i = mpt.begin();
		while (i != mpt.end()) {
			if (i->mMountPoint == j->mMountPoint) break;
			++i;
		}
		// 'i' is the mountpoint we searched for
		if (i == mpt.end()) {
			// this is a new mount point, add it to Table
			mpt.push_back(*j);
			// add it to the altered mount points
			mp.push_back(j->mMountPoint);
		} else {
			// this has been a mount point before
			if (i->mDevice != j->mDevice) {
				// add it to altered mount points if a fs has been mounted
				i->mDevice = j->mDevice;
				mp.push_back(i->mMountPoint);
			}
			i->mValid = true;
		}
	}
	// find and remove invalid entries in 'mpt'
	for (i = mpt.begin(); i != mpt.end(); ++i) {
		if (!i->mValid) {
			mp.push_back(i->mMountPoint);
			mpt.erase(i--);
			break;
		}
	}
	bool retval = false;
	for (std::vector<std::string>::iterator mpi = mp.begin(); mpi != mp.end(); ++mpi) {
		bool update = true;
		CTreeItem item = mTree;
		if (mpi->size() > 1) {
			string::size_type beg = 0;
			string::size_type end = beg;
			while (end != string::npos) {
				++beg;
				end = mpi->find('/', beg);
				string cur = mpi->substr(beg, end - beg);
				vector<CTreeItem>::iterator iter;
				for (iter = item.GetChildren().begin(); iter < item.GetChildren().end(); ++iter) {
					if (iter->GetText() == cur) break;
				}
				if (iter == item.GetChildren().end()) {
					update = false;
					break;
				}
				item = *iter;
				beg = end;
			}
		}
		if (update) {
			retval = true;
			ReadDir(item);
			if (item.GetChildren().empty()) item.Close();
		}
	}
	return retval;
}
