/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFont.cpp
  Info   : $Id: CFont.cpp,v 1.57 2012/07/16 11:49:42 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CFont.h>

// C++ headers
#include <sstream>

using cxxx::cxAlphaFormat;
using cxxx::cxBitSwap;
using cxxx::cxDpy;
using cxxx::cxEmptyString;
using cxxx::cxError;
using cxxx::cxLocale;
using cxxx::cxMainWin;
using cxxx::lib::cxgettext;
using std::copy;
using std::map;
using std::string;
using std::vector;
using std::wstring;

#ifdef CXXX_HAS_FREETYPE
FT_Library CFont::mFTLib = NULL;
#ifndef FC_WEIGHT_REGULAR
#define FC_WEIGHT_REGULAR FC_WEIGHT_MEDIUM
#endif // FC_WEIGHT_REGULAR
#endif // CXXX_HAS_FREETYPE

// Static variable initialization
GC CFont::mFontGC = (GC)(NULL);
GC CFont::mBackgroundGC = (GC)(NULL);
unsigned CFont::mPixmapSize = 0;
Pixmap CFont::mPixmap = None;
int CFont::mCount = 0;

void CFont::Clear()
{
	if (cxMainWin != NULL && mGlyphset != None) XRenderFreeGlyphSet(cxDpy, mGlyphset);
	mGlyphset = None;
#ifdef CXXX_HAS_FREETYPE
	if (mFace != NULL) FT_Done_Face(mFace);
	mFace = NULL;
#endif // CXXX_HAS_FREETYPE
	if (cxMainWin != NULL && !mCore.empty()) {
		for (unsigned i = 0; i < mCore.size(); ++i) {
			XFreeFont(cxDpy, mCore[i].mFontStruct);
			if (mCore[i].mConverter != (iconv_t)(-1)) iconv_close(mCore[i].mConverter);
		}
		mCore.clear();
	}
	if (cxMainWin != NULL && mCount == 0) {
		// release static data
		if (mFontGC != (GC)(NULL)) {
			XFreeGC(cxDpy, mFontGC);
			mFontGC = (GC)(NULL);
		}
		if (mBackgroundGC != (GC)(NULL)) {
			XFreeGC(cxDpy, mBackgroundGC);
			mBackgroundGC = (GC)(NULL);
		}
		if (mPixmap != None) {
			XFreePixmap(cxDpy, mPixmap);
			mPixmap = None;
		}
	} 		
	for (unsigned i = 0; i < mPages.size(); ++i) if (mPages[i]) delete[] mPages[i];
	mPages.clear();
	mCharCache.clear();
	mInfoCache.clear();
	mGlyphCache.clear();
}

CFontDescriptor CFont::ParseXlfd(const string& xlfd)
{
	CFontDescriptor fd;
	fd.mWeight = FW_REGULAR;
	fd.mSlant = FS_ROMAN;
	fd.mWidth = FW_NORMAL;
	fd.mSize = -1;
	if (xlfd.size() < 2) return fd;
	string::size_type start, end;
	if (xlfd[0] == '-') {
		start = 1;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		// foundry
		fd.mFoundry = xlfd.substr(start, end - start);
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		// family
		fd.mFamily = xlfd.substr(start, end - start);
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		string tmp = xlfd.substr(start, end - start);
		// turn to lowercase
		for (unsigned i = 0; i < tmp.size(); ++i) tmp[i] = tolower(tmp[i]);
		string::size_type pos;
		// remove blanks
		while ((pos = tmp.find(' ')) != string::npos) tmp.erase(pos, 1);
		// weight
		if (tmp ==  "thin") fd.mWeight = FW_THIN;
		else if (tmp == "extralight" || tmp == "ultralight") fd.mWeight = FW_EXTRALIGHT;
		else if (tmp == "light") fd.mWeight = FW_LIGHT;
		else if (tmp == "book") fd.mWeight = FW_BOOK;
		else if (tmp == "regular" || tmp == "normal") fd.mWeight = FW_REGULAR;
		else if (tmp == "medium") fd.mWeight = FW_MEDIUM;
		else if (tmp == "demibold" || tmp == "semibold") fd.mWeight = FW_DEMIBOLD;
		else if (tmp == "bold") fd.mWeight = FW_BOLD;
		else if (tmp == "extrabold" || tmp == "ultrabold") fd.mWeight = FW_EXTRABOLD;
		else if (tmp == "black" || tmp == "heavy") fd.mWeight = FW_BLACK;
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		tmp = xlfd.substr(start, end - start);
		// turn to lowercase
		for (unsigned i = 0; i < tmp.size(); ++i) tmp[i] = tolower(tmp[i]);
		// slant
		if (tmp.find('i') != string::npos) fd.mSlant = FS_ITALIC;
		else if (tmp.find('o') != string::npos) fd.mSlant = FS_OBLIQUE;
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		tmp = xlfd.substr(start, end - start);
		// turn to lowercase
		for (unsigned i = 0; i < tmp.size(); ++i) tmp[i] = tolower(tmp[i]);
		// remove blanks
		while ((pos = tmp.find(' ')) != string::npos) tmp.erase(pos, 1);
		// width
		if (tmp == "ultracondensed") fd.mWidth = FW_ULTRACONDENSED;
		else if (tmp == "extracondensed") fd.mWidth = FW_EXTRACONDENSED;
		else if (tmp == "condensed") fd.mWidth = FW_CONDENSED;
		else if (tmp == "semicondensed") fd.mWidth = FW_SEMICONDENSED;
		else if (tmp == "normal") fd.mWidth = FW_NORMAL;
		else if (tmp == "semiexpanded") fd.mWidth = FW_SEMIEXPANDED;
		else if (tmp == "expanded") fd.mWidth = FW_EXPANDED;
		else if (tmp == "extraexpanded") fd.mWidth = FW_EXTRAEXPANDED;
		else if (tmp == "ultraexpanded") fd.mWidth = FW_ULTRAEXPANDED;
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		// skip additional style information
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		tmp = xlfd.substr(start, end - start);
		int pixelsize = -1;
		int pointsize = 120;
		int resy = 72;
		if (!tmp.empty() && tmp != "*") pixelsize = atoi(tmp.c_str());
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		tmp = xlfd.substr(start, end - start);
		if (!tmp.empty() && tmp != "*") pointsize = atoi(tmp.c_str());
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		// skip x resolution
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		tmp = xlfd.substr(start, end - start);
		if (!tmp.empty() && tmp != "*") resy = atoi(tmp.c_str());
		if (pixelsize >= 0) fd.mSize = pixelsize;
		else {
			fd.mSize = pointsize * resy;
			fd.mSize += 360;
			fd.mSize /= 720;
		}
		// skip spacing
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		// skip avg width
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		fd.mEncoding = xlfd.substr(++end);
		// valid file descriptor, now do case conversion
		bool upper = true;
		for (unsigned i = 0; i < fd.mFoundry.size(); ++i) {
			if (upper) fd.mFoundry[i] = toupper(fd.mFoundry[i]);
			else fd.mFoundry[i] = tolower(fd.mFoundry[i]);
			upper = (fd.mFoundry[i] == ' ') ;
		}
		upper = true;
		for (unsigned i = 0; i < fd.mFamily.size(); ++i) {
			if (upper) fd.mFamily[i] = toupper(fd.mFamily[i]);
			else fd.mFamily[i] = tolower(fd.mFamily[i]);
			upper = (fd.mFamily[i] == ' ') ;
		}
	} else {
		start = 0;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		fd.mFamily = xlfd.substr(start, end - start);
		start = ++end;
		end = xlfd.find('-', start);
		if (end == string::npos) return fd;
		string tmp = xlfd.substr(start, end - start);
		// turn to lowercase
		for (unsigned i = 0; i < tmp.size(); ++i) tmp[i] = tolower(tmp[i]);
		// remove blanks
		string::size_type pos;
		while ((pos = tmp.find(' ')) != string::npos) tmp.erase(pos, 1);
		if ((pos = tmp.find("thin")) == 0) {
			fd.mWeight = FW_THIN;
			tmp.erase(pos, 4);
		} else if ((pos = tmp.find("extralight")) == 0) {
			fd.mWeight = FW_EXTRALIGHT;
			tmp.erase(pos, 10);
		} else if ((pos = tmp.find("ultralight")) == 0) {
			fd.mWeight = FW_EXTRALIGHT;
			tmp.erase(pos, 10);
		} else if ((pos = tmp.find("light")) == 0) {
			fd.mWeight = FW_LIGHT;
			tmp.erase(pos, 5);
		} else if ((pos = tmp.find("book")) == 0) {
			fd.mWeight = FW_BOOK;
			tmp.erase(pos, 4);
		} else if ((pos = tmp.find("regular")) == 0) {
			fd.mWeight = FW_REGULAR;
			tmp.erase(pos, 7);
		} else if ((pos = tmp.find("normal")) == 0) {
			tmp.erase(pos, 6);
			fd.mWeight = FW_REGULAR;
		} else if ((pos = tmp.find("medium")) == 0) {
			tmp.erase(pos, 6);
			fd.mWeight = FW_MEDIUM;
		} else if ((pos = tmp.find("demibold")) == 0) {
			tmp.erase(pos, 8);
			fd.mWeight = FW_DEMIBOLD;
		} else if ((pos = tmp.find("semibold")) == 0) {
			tmp.erase(pos, 8);
			fd.mWeight = FW_DEMIBOLD;
		} else if ((pos = tmp.find("extrabold")) == 0) {
			tmp.erase(pos, 9);
			fd.mWeight = FW_EXTRABOLD;
		} else if ((pos = tmp.find("ultrabold")) == 0) {
			tmp.erase(pos, 9);
			fd.mWeight = FW_EXTRABOLD;
		} else if ((pos = tmp.find("bold")) == 0) {
			tmp.erase(pos, 4);
			fd.mWeight = FW_BOLD;
		} else if ((pos = tmp.find("black")) == 0) {
			tmp.erase(pos, 5);
			fd.mWeight = FW_BLACK;
		} else if ((pos = tmp.find("heavy")) == 0) {
			tmp.erase(pos, 5);
			fd.mWeight = FW_BLACK;
		}
		if ((pos = tmp.find("roman")) == 0) {	
			tmp.erase(pos, 5);
			fd.mSlant = FS_ROMAN;
		} else if ((pos = tmp.find("italic")) == 0) {
			tmp.erase(pos, 6);
			fd.mSlant = FS_ITALIC;
		} else if ((pos = tmp.find("oblique")) == 0) {
			tmp.erase(pos, 7);
			fd.mSlant = FS_OBLIQUE;
		}
		if ((pos = tmp.find("ultracondensed")) == 0) fd.mWidth = FW_ULTRACONDENSED;
		else if ((pos = tmp.find("extracondensed")) == 0) fd.mWidth = FW_EXTRACONDENSED;
		else if ((pos = tmp.find("condensed")) == 0) fd.mWidth = FW_CONDENSED;
		else if ((pos = tmp.find("semicondensed")) == 0) fd.mWidth = FW_SEMICONDENSED;
		else if ((pos = tmp.find("normal")) == 0) fd.mWidth = FW_NORMAL;
		else if ((pos = tmp.find("semiexpanded")) == 0) fd.mWidth = FW_SEMIEXPANDED;
		else if ((pos = tmp.find("expanded")) == 0) fd.mWidth = FW_EXPANDED;
		else if ((pos = tmp.find("extraexpanded")) == 0) fd.mWidth = FW_EXTRAEXPANDED;
		else if ((pos = tmp.find("ultraexpanded")) == 0) fd.mWidth = FW_ULTRAEXPANDED;
		start = ++end;
		tmp = xlfd.substr(start);
		if (!tmp.empty()) fd.mSize = atoi(tmp.c_str());
	}
	return fd;
}

std::string CFont::MatchCoreFont(wchar_t c)
{
	if (mFontDescriptor.mSize <= 0) return cxEmptyString;
	int maxcount = 4096;
	int count;
	char** fonts = NULL;
	string xlfd = "-";
	if (c == '\0') {
		if (mFontDescriptor.mFoundry.empty()) xlfd += '*';
		else xlfd += mFontDescriptor.mFoundry;
		xlfd += '-';
		if (mFontDescriptor.mFamily.empty()) xlfd += '*';
		else xlfd += mFontDescriptor.mFamily;
		xlfd += "-*";
	} else {
		xlfd += '*';
	}
	// read font list
	while (fonts == NULL) {
		fonts = XListFonts(cxDpy, xlfd.c_str(), maxcount, &count);
		if(fonts == NULL) break;
		if (count == maxcount) {
			XFreeFontNames(fonts);
			fonts = NULL;
			maxcount *= 2;
		}			
	}
	if (fonts == NULL) return cxEmptyString;
	// find the closest match
	unsigned match = 100;
	unsigned tmpmatch;
	CFontDescriptor tmpfd;
	string matchfont;
	vector<string> ignoreencoding;
	for (int i = 0; i < count; ++i) {
		tmpmatch = 0;
		tmpfd = ParseXlfd(fonts[i]);
		// convert mEncoding to be understood by iconv
		PrepareEncoding(tmpfd.mEncoding);
		for (vector<string>::iterator ie = ignoreencoding.begin(); ie < ignoreencoding.end(); ++ie) {
			if (tmpfd.mEncoding == *ie) tmpfd.mSize = 0;
		}
		if (tmpfd.mSize <= 0) continue;
		if (!mFontDescriptor.mFoundry.empty() && strcasecmp(tmpfd.mFoundry.c_str(), mFontDescriptor.mFoundry.c_str())) tmpmatch += 5;
		if (!mFontDescriptor.mFamily.empty() && strcasecmp(tmpfd.mFamily.c_str(), mFontDescriptor.mFamily.c_str())) tmpmatch += 10;
		tmpmatch += abs(mFontDescriptor.mSize - tmpfd.mSize) * abs(mFontDescriptor.mSize - tmpfd.mSize);
		tmpmatch += abs(tmpfd.mWeight - mFontDescriptor.mWeight) / 20;
		if (tmpfd.mSlant != mFontDescriptor.mSlant) {
			if (tmpfd.mSlant != FS_ROMAN && mFontDescriptor.mSlant != FS_ROMAN) tmpmatch += 5;
			else tmpmatch += 10;
		}
		tmpmatch += abs(tmpfd.mWidth - mFontDescriptor.mWidth) / 20;
		// do not use iso10646, if possible
		if (tmpfd.mEncoding == "iso10646-1") tmpmatch += 2;
		if (tmpmatch < match) {
			// convert c to font's encoding
			iconv_t tmpconv;
			if (tmpfd.mEncoding == "iso10646-1") tmpconv = (iconv_t)(-1);
			else {
				tmpconv = iconv_open(tmpfd.mEncoding.c_str(), "wchar_t");
				if (tmpconv == (iconv_t)(-1)) {
					ignoreencoding.push_back(tmpfd.mEncoding); 
					continue;
				}
			}
			if (c == '\0') {
				match = tmpmatch;
				matchfont = fonts[i];
				if (tmpconv != (iconv_t)(-1)) iconv_close(tmpconv);
				continue;
			}				
			XChar2b tmpidx = GetCharIndex(c, tmpconv);
			if (tmpidx.byte1 == 0xff && tmpidx.byte2 == 0xff) {
				if (tmpconv != (iconv_t)(-1)) iconv_close(tmpconv);
				ignoreencoding.push_back(tmpfd.mEncoding); 
				continue;
			}
					
			// load the font
			XFontStruct* tmpfs = XLoadQueryFont(cxDpy, fonts[i]);
			if (tmpfs == NULL) {
				cxError << _("Failed to load font {1}.", 471) << fonts[i] <<CError::Warn;
				if (tmpconv != (iconv_t)(-1)) iconv_close(tmpconv);
				continue;
			}
			// check if char exists
			if (CharExists(tmpidx, tmpfs)) {
				tmpmatch += abs(tmpfs->ascent - (int)GetAscent()) * abs(tmpfs->ascent - (int)GetAscent());
				if (tmpmatch < match) {
					match = tmpmatch;
					matchfont = fonts[i];
				}
			}
			// clean up
			XFreeFont(cxDpy, tmpfs);
			if (tmpconv != (iconv_t)(-1)) iconv_close(tmpconv);
		}
	}
	XFreeFontNames(fonts);
	return matchfont;
}

#ifdef CXXX_HAS_FREETYPE
FcPattern* CFont::MatchAAFont()
{
	if (mFontDescriptor.mSize <= 0) return NULL;
	FcPattern* pat = FcPatternCreate();
	if (pat == NULL) {
		cxError << _("Fontconfig: Failed to create empty font pattern", 70) << CError::Fatal;
		return NULL;
	}
	if (!mFontDescriptor.mFoundry.empty()) {
		if (FcPatternAddString(pat, FC_FOUNDRY, (FcChar8*)mFontDescriptor.mFoundry.c_str()) != FcTrue) {
			FcPatternDestroy(pat);
			return NULL;
		}
	}
	if (!mFontDescriptor.mFamily.empty()) {
		if (FcPatternAddString(pat, FC_FAMILY, (FcChar8*)mFontDescriptor.mFamily.c_str()) != FcTrue) {
			FcPatternDestroy(pat);
			return NULL;
		}
	}
	if (FcPatternAddInteger(pat, FC_WEIGHT, mFontDescriptor.mWeight) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}
	if (FcPatternAddInteger(pat, FC_SLANT, mFontDescriptor.mSlant) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}
	if (FcPatternAddInteger(pat, FC_WIDTH, mFontDescriptor.mWidth) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}
	if (FcPatternAddDouble(pat, FC_PIXEL_SIZE, (double)mFontDescriptor.mSize) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}
	if (FcPatternAddBool(pat, FC_SCALABLE, FcTrue) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}
	string lang = cxLocale.name();
	if (lang == "C" || lang == "POSIX") lang = "en";
	else if (lang.size() >= 2) lang = lang.substr(0, 2);
	if (FcPatternAddString(pat, FC_LANG, reinterpret_cast<const FcChar8*>(lang.c_str())) != FcTrue) {
		FcPatternDestroy(pat);
		return NULL;
	}		
	// ok, so we got the pattern; let fontconfig do the matching
	FcDefaultSubstitute(pat);
	FcConfigSubstitute(NULL, pat, FcMatchPattern);
	FcResult res = FcResultMatch;
	FcPattern* pres = FcFontMatch(NULL, pat, &res);
	FcPatternDestroy (pat);
	FcBool aa = FcFalse;
	if (res == FcResultMatch) {
		FcChar8* family;
		res = FcPatternGetString(pres, FC_FAMILY, 0, &family);
		if (mFontDescriptor.mFamily == (char*)family && res == FcResultMatch) {
			res = FcPatternGetBool(pres, FC_ANTIALIAS, 0, &aa);
			if (aa == FcFalse || res != FcResultMatch) res = FcPatternGetBool(pres, FC_SCALABLE, 0, &aa);
		} else aa = FcFalse;
		if (aa == FcFalse || res != FcResultMatch) {
			// not an antialiased font
			FcPatternDestroy(pres);
			return NULL;
		}
	}
	FcLangSet* ls;
	res = FcPatternGetLangSet(pres, FC_LANG, 0, &ls);
	if (FcLangSetHasLang(ls, reinterpret_cast<const FcChar8*>(lang.c_str())) == FcLangDifferentLang) {
		// charset insufficient
		FcPatternDestroy(pres);
		return NULL;
	}	
	return pres;
}
#endif // CXXX_HAS_FREETYPE

bool CFont::LoadAAFont()
{
#ifdef CXXX_HAS_FREETYPE
	// initialize libfontconfig
	if(!FcInit()) {
		cxError << _("Failed to initialize libfontconfig.", 78) << CError::Error;
		return false;
	}
	FcPattern* pat = MatchAAFont();
	if (pat == NULL) return false;
	FcChar8* file = NULL;
	if (FcPatternGetString(pat, FC_FILE, 0, &file) != FcResultMatch) {
		cxError << _("Fontconfig: no filename.", 71) << CError::Error;
		FcPatternDestroy(pat);
		return false;
	}
	if (mFTLib == NULL && FT_Init_FreeType(&mFTLib) != 0) {
   		cxError << _("Failed to initialize FreeTypeLibrary.", 72) << CError::Error;
		FcPatternDestroy(pat);
		return false;
   	}
	//Clear();
	FT_Error err =  FT_New_Face(mFTLib, (char*)file, 0, &mFace);
	if (err ==  FT_Err_Unknown_File_Format) {
      	cxError << _("Font file format not recognized.", 74) << CError::Error;
		FcPatternDestroy(pat);
       	return false;
   	} else if (err != 0) {
       	cxError << _("Failed to open font file {1}.", 75) << (char*)file << CError::Error;
		FcPatternDestroy(pat);
       	return false;
   	}
	if (FT_IS_SCALABLE(mFace)) {
   		// do scaling
		double size;
		if (FcPatternGetDouble(pat, FC_PIXEL_SIZE, 0, &size) != FcResultMatch) {
			cxError << _("Fontconfig: no font size.", 76) << CError::Error;
			size = 12.0;
		}
   		if (FT_Set_Pixel_Sizes(mFace, 0, (unsigned)size) != 0) {
			cxError << _("Failed to scale font to requested size", 77) << CError::Error;
    	}
	}
	FcPatternDestroy(pat);
    // get the metrics:
    mAscent = GetCeil(mFace->size->metrics.ascender);
    mDescent = -GetFloor(mFace->size->metrics.descender);
	return true;
#else // CXXX_HAS_FREETYPE
	return false;
#endif // CXXX_HAS_FREETYPE
}

bool CFont::LoadCoreFont(wchar_t c)
{
	if (c < L'\x20' && c != '\0')  return false;
	else if (c >= L'\x80' && c < L'\xa0') return false;
	else if (c >= L'\xe000' && c <= L'\xffff') {
		if (c < L'\xf900') return false;
		else if (c >= L'\xfbb2' && c < L'\xfbd3') return false;
		else if (c >= L'\xfd40' && c < L'\xfd50') return false;
		else if (c >= L'\xfdc8' && c < L'\xfdf0') return false;
		else if (c >= L'\xfdfe' && c < L'\xfe00') return false;
	}
	string xlfd = MatchCoreFont(c);
	if (xlfd.empty()) return false;
	CFontDescriptor tmpfd = ParseXlfd(xlfd);
	// convert mEncoding to be understood by iconv
	PrepareEncoding(tmpfd.mEncoding);
	CCoreFont font;
	if (tmpfd.mEncoding == "iso10646-1") font.mConverter = (iconv_t)(-1);
	else font.mConverter = iconv_open(tmpfd.mEncoding.c_str(), "wchar_t");
	font.mFontStruct = XLoadQueryFont(cxDpy, xlfd.c_str());
	cxAssert(font.mFontStruct != NULL);
	mCore.push_back(font);
	if (mCore.size() == 1) {
		mAscent = mCore[0].mFontStruct->ascent;
		mDescent = mCore[0].mFontStruct->descent;
	}
	return true;
}

bool CFont::LoadFont(const string& xlfd)
{
	Clear();
	mFontDescriptor = ParseXlfd(xlfd);
	string failname = mFontDescriptor.mFamily;
	if (mFontDescriptor.mSize < 0) cxError << _("Failed to parse X Logical File Descriptor", 421) << CError::Warn;
	else {
		if (LoadAAFont() || LoadCoreFont()) return true;
		// SuSE configures fontconfig to use FOUNDRY FAMILY as family name
		if (mFontDescriptor.mFoundry.empty()) {
			string::size_type pos = mFontDescriptor.mFamily.find(' ');
			if (pos < mFontDescriptor.mFamily.size() + 1) {
				mFontDescriptor.mFoundry = mFontDescriptor.mFamily.substr(0, pos);
				mFontDescriptor.mFamily = mFontDescriptor.mFamily.substr(++pos);
				if (LoadAAFont() || LoadCoreFont()) return true;
			}
		}				
		// no matching font, try helvetica 
		mFontDescriptor.mFoundry = "";
		mFontDescriptor.mFamily = "Helvetica";
		if (LoadCoreFont()) {
			cxError << _("Failed to load {1} font. Using Helvetica instead.", 445) << failname << CError::Warn;
			return true;
		}
	}
	// we do not want to leave without a font; try everything
	mFontDescriptor.mFoundry = "";
	mFontDescriptor.mFamily = "";
	mFontDescriptor.mSlant = FS_ANY;
	mFontDescriptor.mSize = 12;
	if (LoadAAFont() || LoadCoreFont()) {
		cxError << _("Failed to load {1} font.", 443) << failname << CError::Warn;
		return true;
	}
	cxError << _("No usable font found!", 444) << CError::Fatal;
	return false;
}


#ifdef CXXX_HAS_FREETYPE
void CFont::LoadAAGlyph(wchar_t c, XGlyphInfo& gi, vector<char>& v)
{
	// make sure the font is loaded
	if (mFace == NULL) {
		cxError << _("Cannot load glyph: No font set", 79) << CError::Error;
		LoadDefaultGlyph(gi, v);
		return;
	}
	FT_UInt idx = FT_Get_Char_Index(mFace, c);
	FT_Error err = 0;
	err = FT_Load_Glyph(mFace, idx, FT_LOAD_DEFAULT);
	if (err != 0) {
		cxError << _("Failed to load glyph U + {1}", 80) << (unsigned)(c) << CError::Error;
		LoadDefaultGlyph(gi, v);
		return;
	} else {
		// get the metrics
		int left = GetFloor(mFace->glyph->metrics.horiBearingX);
		int right = GetCeil(mFace->glyph->metrics.horiBearingX + mFace->glyph->metrics.width);
		int top = GetCeil(mFace->glyph->metrics.horiBearingY);
		int bottom = GetFloor(mFace->glyph->metrics.horiBearingY - mFace->glyph->metrics.height);
		gi.width = right - left;
		gi.height = top - bottom;
		gi.x = -left;
		gi.y = top;
		gi.xOff = GetRound(mFace->glyph->metrics.horiAdvance);
		gi.yOff = 0;
		if (FT_Render_Glyph(mFace->glyph, ft_render_mode_normal) != 0) {
			cxError << "Failed to render glyph U + {1}" << (unsigned)(c) << CError::Error;
			LoadDefaultGlyph(gi, v);
			return;
		}
		// get the bitmap
		int x, y;
		unsigned char* tmp;
		int pad = (4 - mFace->glyph->bitmap.width) & 3;
		switch (mFace->glyph->bitmap.num_grays) {
		case 1:
			for (y = 0; y < mFace->glyph->bitmap.rows; ++y) {
				tmp = mFace->glyph->bitmap.buffer + y * mFace->glyph->bitmap.pitch;
				for (x = 0; x < mFace->glyph->bitmap.width; ++x) {
					if ((tmp[x >> 3] & (0x80 >> (x & 7))) != 0) v.push_back('\xff');
					else v.push_back(0);
				}
				for (x = 0; x < pad; ++x) v.push_back('\0');
			}
			break;
		case 256:
			for (y = 0; y < mFace->glyph->bitmap.rows; ++y) {
					tmp = mFace->glyph->bitmap.buffer + y * mFace->glyph->bitmap.pitch;
					for (x = 0; x < mFace->glyph->bitmap.width; ++x) {
							v.push_back(tmp[x]);
					}
					for (x = 0; x < pad; ++x) v.push_back(0);
			}
			break;
		default:
			cxError << _("Font renderer: bitmap depth of {1} bits not supported.", 81);
			cxError << mFace->glyph->bitmap.num_grays << CError::Error;
			LoadDefaultGlyph(gi, v);
			return;
		}
	}
}

void CFont::ListAAFonts(vector<CFontDescriptor>& l)
{
	// initialize libfontconfig
	if(!FcInit()) {
		cxError << _("Failed to initialize libfontconfig.", 78) << CError::Error;
		return;
	}
	// create a pattern for searching
	FcPattern* pat = FcPatternCreate();
	FcObjectSet* obj = FcObjectSetBuild(FC_FAMILY, FC_WEIGHT, FC_SLANT, FC_WIDTH, FC_PIXEL_SIZE, FC_SCALABLE, FC_LANG, (char*)NULL);
	// acquire the font list
	FcFontSet* fonts = FcFontList(NULL, pat, obj);
	// clean up
	FcPatternDestroy(pat);
	FcObjectSetDestroy(obj);
	// create Charset
	string lang = cxLocale.name();
	if (lang == "C" || lang == "POSIX") lang = "en";
	else if (lang.size() > 2) lang = lang.substr(0, 2);
	FcChar8* buf;
	double size;
	FcBool scale;
	int tmp;
	FcLangSet* ls;
	CFontDescriptor fd;
	for (int i = 0; i < fonts->nfont; ++i) {
		// extract the descriptions
		if (FcPatternGetString(fonts->fonts[i], FC_FAMILY, 0, &buf) != FcResultMatch) continue;
		fd.mFamily = (char*)buf;
		if (FcPatternGetLangSet(fonts->fonts[i], FC_LANG, 0, &ls) != FcResultMatch) continue;
    	if (FcLangSetHasLang(ls, reinterpret_cast<const FcChar8*>(lang.c_str())) == FcLangDifferentLang) continue;
		if (FcPatternGetInteger(fonts->fonts[i], FC_WEIGHT, 0, &tmp) != FcResultMatch) tmp = FW_REGULAR;
		if (tmp < (FW_THIN + FW_EXTRALIGHT) / 2) fd.mWeight = FW_THIN;
		else if (tmp < (FW_EXTRALIGHT + FW_LIGHT) / 2) fd.mWeight = FW_EXTRALIGHT;
		else if (tmp < (FW_LIGHT + FW_BOOK) / 2) fd.mWeight = FW_LIGHT;
		else if (tmp < (FW_BOOK + FW_REGULAR) / 2) fd.mWeight = FW_BOOK;
		else if (tmp < (FW_MEDIUM + FW_DEMIBOLD) / 2) fd.mWeight = FW_REGULAR;
		else if (tmp < (FW_DEMIBOLD + FW_BOLD) / 2) fd.mWeight = FW_DEMIBOLD;
		else if (tmp < (FW_BOLD + FW_EXTRABOLD) / 2) fd.mWeight = FW_BOLD;
		else if (tmp < (FW_EXTRABOLD + FW_BLACK) / 2) fd.mWeight = FW_EXTRABOLD;
		else fd.mWeight = FW_BLACK;
		if (FcPatternGetInteger(fonts->fonts[i], FC_SLANT, 0, &tmp) != FcResultMatch) tmp = FS_ROMAN;
		if (tmp == FS_ITALIC) fd.mSlant = FS_ITALIC;
		else if (tmp == FS_OBLIQUE) fd.mSlant = FS_OBLIQUE;
		else fd.mSlant = FS_ROMAN;
		if (FcPatternGetInteger(fonts->fonts[i], FC_WIDTH, 0, &tmp) != FcResultMatch) tmp = FW_NORMAL;
		if (tmp < (FW_ULTRACONDENSED + FW_EXTRACONDENSED) / 2) fd.mWidth = FW_ULTRACONDENSED;
		else if (tmp < (FW_EXTRACONDENSED + FW_CONDENSED) / 2) fd.mWidth = FW_EXTRACONDENSED;
		else if (tmp < (FW_CONDENSED + FW_SEMICONDENSED) / 2) fd.mWidth = FW_CONDENSED;
		else if (tmp < (FW_SEMICONDENSED + FW_NORMAL) / 2) fd.mWidth = FW_SEMICONDENSED;
		else if (tmp < (FW_NORMAL + FW_SEMIEXPANDED) / 2) fd.mWidth = FW_NORMAL;
		else if (tmp < (FW_SEMIEXPANDED + FW_EXPANDED) / 2) fd.mWidth = FW_SEMIEXPANDED;
		else if (tmp < (FW_EXPANDED + FW_EXTRAEXPANDED) / 2) fd.mWidth = FW_EXPANDED;
		else if (tmp < (FW_EXTRAEXPANDED + FW_ULTRAEXPANDED) / 2) fd.mWidth = FW_EXTRAEXPANDED;
		else fd.mWidth = FW_ULTRAEXPANDED;
		if (FcPatternGetBool(fonts->fonts[i], FC_SCALABLE, 0, &scale) != FcResultMatch) continue;
		if (scale == FcTrue) fd.mSize = 0;
		else {
			if (FcPatternGetDouble(fonts->fonts[i], FC_PIXEL_SIZE, 0, &size) != FcResultMatch) continue;
			fd.mSize = (int)(size + 0.5);
		}
		l.push_back(fd);
	}
	// release the font list
	FcFontSetDestroy(fonts);
}

#else // CXXX_HAS_FREETYPE
void CFont::LoadAAGlyph(wchar_t c __attribute__ ((unused)), XGlyphInfo& gi __attribute__ ((unused)), vector<char>& v __attribute__ ((unused)))
{
	cxAssert(0);
}

void CFont::ListAAFonts(vector<CFontDescriptor>& l __attribute__ ((unused)))
{
}
#endif // CXXX_HAS_FREETYPE


void CFont::ListCoreFonts(vector<CFontDescriptor>& l)
{
	int maxcount = 4096;
	int count = 4096;
	char** xfonts = NULL;
	// read font list
	while (xfonts == NULL) {
		xfonts = XListFonts(cxDpy, "-*", maxcount, &count);
		if(xfonts == NULL) cxError << _("No fonts found.", 73) << CError::Fatal;
		if (count == maxcount) {
			XFreeFontNames(xfonts);
			xfonts = NULL;
			maxcount *= 2;
		}			
	}
	CFontDescriptor fd;
	for (int i = 0; i < count; ++i) {
		fd = ParseXlfd(xfonts[i]);
		if (fd.mSize <= 0) continue;
		l.push_back(fd);
	}
	// release the font list
	XFreeFontNames(xfonts);
}

string CFont::GetStyle(const CFontDescriptor& fd)
{
	string ret;
	switch (fd.mWeight) {
	case FW_THIN:
		ret = "Thin ";
		break;
	case FW_EXTRALIGHT:
		ret = "Extralight ";
		break;
	case FW_LIGHT:
		ret = "Light ";
		break;
	case FW_BOOK:
		ret = "Book ";
		break;
	case FW_DEMIBOLD:
		ret = "Demibold ";
		break;
	case FW_BOLD:
		ret = "Bold ";
		break;
	case FW_EXTRABOLD:
		ret = "Extrabold ";
		break;
	case FW_BLACK:
		ret = "Black ";
		break;
	default:
		break;
	}
	switch (fd.mSlant) {
	case FS_ITALIC:
		ret += "Italic ";
		break;
	case FS_OBLIQUE:
		ret += "Oblique ";
		break;
	default:
		break;
	}
	switch (fd.mWidth) {
	case FW_ULTRACONDENSED:
		ret += "Ultracondensed";
		break;
	case FW_EXTRACONDENSED:
		ret += "Extracondensed";
		break;
	case FW_CONDENSED:
		ret += "Condensed";
		break;
	case FW_SEMICONDENSED:
		ret += "Semicondensed";
		break;
	case FW_SEMIEXPANDED:
		ret += "Semiexpanded";
		break;
	case FW_EXPANDED:
		ret += "Expanded";
		break;
	case FW_EXTRAEXPANDED:
		ret += "Extraexpanded";
		break;
	case FW_ULTRAEXPANDED:
		ret += "Ultraexpanded";
		break;
	default:
		break;
	}
	if (ret.empty()) ret = "Regular";
	else if (ret[ret.size() - 1] == ' ') ret.resize(ret.size() - 1);
	return ret;
}

void CFont::LoadCoreGlyph(wchar_t c, XGlyphInfo& gi, vector<char>& v)
{
	static XGCValues gcv;
	// first try one of the existing fonts
	for (vector<CCoreFont>::iterator it = mCore.begin(); it != mCore.end(); ++it) {
		// get character index
		XChar2b x2b = GetCharIndex(c, it->mConverter);
		if (CharExists(x2b, it->mFontStruct)) {
			// we found it!
			XCharStruct* cs = NULL;
			XFontStruct* fs = it->mFontStruct;
			unsigned row = x2b.byte1;
			unsigned col = x2b.byte2;
			if (fs->max_byte1 == 0) {
				col |= row << 8;
				if (col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) {
					if (fs->per_char == NULL) cs = &(fs->min_bounds);
					else cs = &(fs->per_char[col - fs->min_char_or_byte2]);
				}
			} else {
				if (row >= fs->min_byte1 && row <= fs->max_byte1) {
					if (col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) {
						if (fs->per_char == NULL) cs = &(fs->min_bounds);
						else cs = &(fs->per_char[(row - fs->min_byte1) * (fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1) + (col - fs->min_char_or_byte2)]);
					}
				}
			}
			cxAssert (cs != NULL);
			gi.x = -cs->lbearing;
			//gi.x = 0;
			gi.y = cs->ascent;
			gi.width = cs->rbearing - cs->lbearing;
			//gi.width = cs->rbearing;
			gi.height = cs->ascent + cs->descent;
			gi.xOff = cs->width;
			gi.yOff = 0;
			unsigned pad = (4u - gi.width) & 3u;
			if (gi.width > mPixmapSize || gi.height > mPixmapSize) {
				if (gi.width > gi.height) mPixmapSize = gi.width;
				else mPixmapSize = gi.height;
				if (mPixmap != None) {
					XFreePixmap(cxDpy, mPixmap);
					mPixmap = None;
				}
			}
			if (mPixmap == None) {
				mPixmap = XCreatePixmap(cxDpy, DefaultRootWindow(cxDpy), mPixmapSize, mPixmapSize, 1);
			}
			if (gi.width > 0 && gi.height > 0) {
				if (mFontGC == (GC)(NULL)) {
					gcv.font = fs->fid;
					gcv.foreground = 1;
					mFontGC = XCreateGC(cxDpy, mPixmap, GCFont | GCForeground, &gcv);
					mBackgroundGC = XCreateGC(cxDpy, mPixmap, 0, &gcv);
				} else {
					XGetGCValues(cxDpy, mFontGC, GCFont, &gcv);
					if (gcv.font != fs->fid) {
						gcv.font = fs->fid;
						XChangeGC(cxDpy, mFontGC, GCFont, &gcv);
					}
				}
				XFillRectangle(cxDpy, mPixmap, mBackgroundGC, 0, 0, gi.width, gi.height);
				if (fs->max_byte1 == 0 && fs->max_char_or_byte2 < 256) {
					XDrawString(cxDpy, mPixmap, mFontGC, gi.x, gi.y, (char*)&x2b.byte2, 1);
				} else {
					XDrawString16(cxDpy, mPixmap, mFontGC, gi.x, gi.y, &x2b, 1);
				}
				XImage* img = XGetImage(cxDpy, mPixmap, 0, 0, gi.width, gi.height, 1, ZPixmap);
				cxAssert(img);
				for (unsigned y = 0; y < gi.height; ++ y) {
					for (unsigned x = 0; x < gi.width; ++x) {
						if (XGetPixel(img, x, y)) v.push_back('\xff');
						else v.push_back('\0');
					}
					for (unsigned x = 0; x < pad; ++x) v.push_back('\0');
				}
				XDestroyImage(img);
			} else {
				gi.width = 1;
				gi.height = 1;
				for (unsigned x = 0; x < 4; ++x) v.push_back('\0');
			}
			return;
		}
	}
	// Character not availabe in any existing font
	if (LoadCoreFont(c)) LoadCoreGlyph(c, gi, v);
	else LoadDefaultGlyph(gi, v);
}

void CFont::LoadDefaultGlyph(XGlyphInfo& gi, vector<char>& v)
{
	gi.height = mAscent;
	if (gi.height == 0) gi.height = 10;
	gi.width = gi.height / 2;
	gi.x = -1;
	gi.y = gi.height;
	gi.xOff = gi.width + 2;
	gi.yOff = 0;
	unsigned pad = (4u - gi.width) & 3u;
	for (unsigned y = 0; y < gi.height; ++y) {
		for (unsigned x = 0; x < gi.width; ++x) {
			if (y == 0 || y == gi.height - 1u || x == 0 || x == gi.width - 1u) v.push_back('\xff');
			else v.push_back('\0');
		}
		for (unsigned x = 0; x < pad; ++x) v.push_back('\0');
	}
}

XChar2b CFont::GetCharIndex(wchar_t c, iconv_t conv)
{
	XChar2b ret;
	ret.byte1 = 0xff;
	ret.byte2 = 0xff;
	if (conv != (iconv_t)(-1)) {
		union
		{
			char c[8];
			unsigned short u[4];
		} buf;
		buf.u[0] = 0;
		size_t bufsz = 8;
		size_t wcharsz = 4;
		
		cxxx::iconvstring_t in = (cxxx::iconvstring_t)(&c);
		char* out = buf.c;
		if (iconv(conv, &in, &wcharsz, &out, &bufsz) != (size_t)(-1)) {
			ret.byte1 = buf.u[0] / 256;
			ret.byte2 = buf.u[0] % 256;
		}
	} else {
		ret.byte1 = (unsigned)(c) / 256;
		ret.byte2 = (unsigned)(c) % 256;
	}		
	return ret;
}

bool CFont::CharExists(XChar2b idx, XFontStruct* fs)
{
	if (idx.byte1 == 0xff && idx.byte2 == 0xff) return false;
	XCharStruct* cs = NULL;
	unsigned row = (unsigned) idx.byte1;
	unsigned col = (unsigned) idx.byte2;
	if (fs->max_byte1 == 0) {
		col |= row << 8;
		if (col < fs->min_char_or_byte2 || col > fs->max_char_or_byte2) return false; 
		if (fs->per_char == NULL) return true; 
		cs = &fs->per_char[col - fs->min_char_or_byte2]; 
	} else {
		if (row < fs->min_byte1 || row > fs->max_byte1) return false;
		if (col < fs->min_char_or_byte2 || col > fs->max_char_or_byte2) return false;
		if (fs->per_char == NULL) return true;
		cs = &fs->per_char[(row - fs->min_byte1) * (fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1) + (col - fs->min_char_or_byte2)];
	}
	return ((cs->width | cs->rbearing | cs->lbearing | cs->ascent | cs->descent) != 0);
}

void CFont::AddGlyphs(const wstring& c, const vector<XGlyphInfo>& gi, const vector<char>& v)
{
	if (mGlyphset == None) mGlyphset = XRenderCreateGlyphSet(cxDpy, cxAlphaFormat);
	std::vector<Glyph> glyphs;
	for (unsigned i = 0; i < c.size(); ++i) {
		glyphs.push_back(static_cast<Glyph>(c[i]));
	}
	XRenderAddGlyphs (cxDpy, mGlyphset, &glyphs[0], &gi[0], gi.size(), &v[0], v.size());
}
