/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFont.h
  Info   : $Id: CFont.h,v 1.33 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CFont_h
#define CFont_h

// C++X headers
#include <cxxx/Cxxx.h>

// C++ headers
#include <vector>
#include <map>
#include <cstring>

// Optional headers
#ifdef CXXX_HAS_FREETYPE
  #include <ft2build.h>
  #include FT_FREETYPE_H
  #include FT_GLYPH_H
  #include <fontconfig/fontconfig.h>
#endif // CXXX_HAS_FREETYPE

#include <X11/extensions/Xrender.h>

typedef enum
{
	// Font weight
	FW_THIN       = 0,
	FW_EXTRALIGHT = 40,
	FW_LIGHT      = 50,
	FW_BOOK       = 75,
	FW_REGULAR    = 80,
	FW_MEDIUM     = 100,
	FW_DEMIBOLD   = 180,
	FW_BOLD       = 200,
	FW_EXTRABOLD  = 205,
	FW_BLACK      = 210
} TFontWeight;

typedef enum
{
	// Font width
	FW_ULTRACONDENSED = 50,
	FW_EXTRACONDENSED = 63,
	FW_CONDENSED      = 75,
	FW_SEMICONDENSED  = 87,
	FW_NORMAL         = 100,
	FW_SEMIEXPANDED   = 113,
	FW_EXPANDED       = 125,
	FW_EXTRAEXPANDED  = 150,
	FW_ULTRAEXPANDED  = 200
} TFontWidth;

typedef enum
{
	// Font slant
	FS_ANY,
	FS_ROMAN,
	FS_ITALIC,
	FS_OBLIQUE
} TFontSlant;

struct CFontDescriptor
{
	CFontDescriptor()
	: mWeight(FW_MEDIUM),
	  mSlant(FS_ANY),
	  mWidth(FW_NORMAL),
	  mSize(0)
	{}

	std::string mFoundry;
	std::string mFamily;
	TFontWeight mWeight;
	TFontSlant mSlant;
	TFontWidth mWidth;
	int mSize;
	std::string mEncoding;
};

class CFont
{
private:
	enum
	{
		PAGE_SIZE = 0x80
	};

public:
	struct CCoreFont
	{
		XFontStruct* mFontStruct;
		iconv_t mConverter;
	};

	// Constructor
	CFont()
	: mFace (NULL),
	  mGlyphset (None),
	  mAscent(0),
	  mDescent(0)
	{
		++mCount;
	}

	// Destructor
	~CFont()
	{
		--mCount;
		Clear();
	}
	
	// Loads the font described in xlfd
	bool LoadFont(const std::string& xlfd);

	// Returns the font metrics for glyph c
	unsigned GetCharWidth(wchar_t c)
	{
		unsigned page = (unsigned)(c) >> 7;
		unsigned idx = (unsigned)(c) & 0x7f;
		if (page >= mPages.size()) {
			mPages.resize(page + 1, NULL);
		}
		if (mPages[page] == NULL) {
			mPages[page] = new unsigned short[PAGE_SIZE];
			memset(mPages[page], 0xff, PAGE_SIZE * sizeof(unsigned short));
		}
		unsigned short* usp = mPages[page] + idx;
		if (*usp != (unsigned short)(-1u)) return *usp;
		XGlyphInfo gi;
		LoadGlyph(c, gi, mGlyphCache);
		mCharCache.push_back((Glyph)(c));
		mInfoCache.push_back(gi);
		*usp = gi.xOff;
		return gi.xOff;
	}

	// Returns the glyph set for this font
	GlyphSet GetGlyphset()
	{
		return mGlyphset;
	}

	// Returns the font height
	unsigned GetHeight()
	{
		return mAscent + mDescent;
	}

	// Returns the font ascent
	unsigned GetAscent()
	{
		return mAscent;
	}
	
	// Returns the font descent
	unsigned GetDescent()
	{
		return mDescent;
	}
	
	// Returns false if this is a core font 
	bool IsAntialiased()
	{
		return mFace != NULL;
	}
	
	// Returns a list of font descriptors
	static void ListFonts(std::vector<CFontDescriptor>& list)
	{
		list.clear();
		ListAAFonts(list);
		ListCoreFonts(list);
	}
	
	// Unloads the font and all associated information
	void Clear();
	
	// Creates a CFontDescriptor from xlfd
	static CFontDescriptor ParseXlfd(const std::string& xlfd);
	
	// Extracts the style information from fd
	static std::string GetStyle(const CFontDescriptor& fd);
	// Adds missing glyphs of s
	void AddMissingGlyphs(const std::wstring& s) 
	{
		XGlyphInfo gi;
		for (unsigned i = 0; i < s.size(); ++i) {
			if (!IsValid(s[i])) {
				LoadGlyph(s[i], gi, mGlyphCache);
				mCharCache += s[i];
				mInfoCache.push_back(gi);
			}					
		}
		if (!mCharCache.empty()) {
			AddGlyphs(mCharCache, mInfoCache, mGlyphCache);
			mCharCache.clear();
			mInfoCache.clear();
			mGlyphCache.clear();
		}
	}
	
private:
	// Returns the xlfd of the core font that matches closest and includes the glyph of c
	std::string MatchCoreFont(wchar_t c);

	// Returns a list of font descriptors of non-core fonts
	static void ListAAFonts(std::vector<CFontDescriptor>& l);

	// Returns a list of font descriptors of core fonts
	static void ListCoreFonts(std::vector<CFontDescriptor>& l);

	// Loads a non-core font
	bool LoadAAFont();

	// Loads a core font which includes the glyph of c 
	bool LoadCoreFont(wchar_t c = L'\0');

	// Loads a glyph and returns its metrics
	void LoadGlyph (wchar_t c, XGlyphInfo& gi, std::vector<char>& v)
	{
		if (IsAntialiased()) LoadAAGlyph(c, gi, v);
		else LoadCoreGlyph(c, gi, v);
	}

	// Loads the glyph of c from a core font
	void LoadCoreGlyph(wchar_t c, XGlyphInfo& gi, std::vector<char>& v);

	// Loads the glyph of c from a non-core font
	void LoadAAGlyph(wchar_t c, XGlyphInfo& gi, std::vector<char>& v);

	// Loads the default glyph instead of the glyph for c
	void LoadDefaultGlyph(XGlyphInfo& gi, std::vector<char>& v);

	// Returns the character index of c
	// conv points to a converter converting to the font's encoding
	XChar2b GetCharIndex(wchar_t c, iconv_t conv);
	
	// Returns true if the glyph with index idx exists in the font described by fs
	bool CharExists(XChar2b idx, XFontStruct* fs);
	
	// Returns true if info is valid
	bool IsValid(wchar_t c) const
	{
		unsigned page = (unsigned)(c) >> 7;
		unsigned idx = (unsigned)(c) & 0x7f;
		if (page >= mPages.size()) return false;
		unsigned short* gw = mPages[page];
		if (!gw) return false;
		return gw[idx] != (unsigned short)(-1u);
	}

	// Adds all glyph relevant information for multiple glyphs
	void AddGlyphs(const std::wstring& c, const std::vector<XGlyphInfo>& gi, const std::vector<char>& v);

	// Prepares the encoding of a xlfd to be understood by iconv
	void PrepareEncoding(std::string& e)
	{
		if (e.find("microsoft-") == 0) e = e.substr(10);
		else if (e.find("ibm-") == 0) e = e.substr(4);
		else {
			std::string::size_type pos = e.find('.');
			if (pos != std::string::npos) e = e.substr(0, pos);
			if (e == "iso646") e = "ascii";
		}
	}

#ifdef CXXX_HAS_FREETYPE
	// Returns a FcPattern of the non-core font that matches closest
	FcPattern* MatchAAFont();
	
	// FT_F26Dot6 rounding routines
	int GetFloor(FT_F26Dot6 i)
	{
		return i >> 6;
	}
	
	int GetRound(FT_F26Dot6 i)
	{
		return (i + 32) >> 6;
	}
	
	int GetCeil(FT_F26Dot6 i)
	{
		return (i + 63) >> 6;
	}
#endif

private:
#ifdef CXXX_HAS_FREETYPE  
	static FT_Library mFTLib;
	FT_Face mFace;
#else 
	void* mFace;
#endif // CXXX_HAS_FREETYPE
	GlyphSet mGlyphset;
	std::vector<CCoreFont> mCore;
	int mAscent;
	int mDescent;
	CFontDescriptor mFontDescriptor;
	std::wstring mCharCache;
	std::vector<XGlyphInfo> mInfoCache;
	std::vector<char> mGlyphCache;
	std::vector<unsigned short*> mPages;
	static GC mFontGC;
	static GC mBackgroundGC;
	static unsigned mPixmapSize;
	static Pixmap mPixmap;
	static int mCount;
};

#endif // CFont_h
