/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CFontSheet.cpp
  Info   : $Id: CFontSheet.cpp,v 1.39 2012/06/26 08:05:52 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2005 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CFontSheet.h>
#include <cxxx/CProfile.h>

// C++ headers
#include <sstream>

// Namespace declarations
using std::min;
using std::max;
using std::string;
using std::map;
using std::vector;
using std::ostringstream;
using cxxx::lib::cxgettext;
using cxxx::lib::cxProfile;
using cxxx::cxError;
using cxxx::cxDpy;
using cxxx::cxTextFont;
using cxxx::cxTitleFont;

// Static variables 
const CFontStyleNames CFontSheet::mWeightNames[WEIGHTCOUNT] = 
{
	{"Thin", cxxx::noop::_("Thin", 423)},			{"Extralight", cxxx::noop::_("Extralight", 439)},
	{"Light", cxxx::noop::_("Light", 418)},			{"Book", cxxx::noop::_("Book", 419)},
	{"Demibold", cxxx::noop::_("Demibold", 428)},	{"Extrabold", cxxx::noop::_("Extrabold", 430)},
	{"Bold", cxxx::noop::_("Bold", 433)},			{"Black", cxxx::noop::_("Black", 434)}
};

const CFontStyleNames CFontSheet::mSlantNames[SLANTCOUNT] =
{
	{"Italic", cxxx::noop::_("Italic", 435)}, 		{"Oblique", cxxx::noop::_("Oblique", 436)}
};

const CFontStyleNames CFontSheet::mWidthNames[WIDTHCOUNT] = 
{
	{"Ultracondensed", cxxx::noop::_("Ultracondensed", 426)},	{"Extracondensed", cxxx::noop::_("Extracondensed", 424)},
	{"Semicondensed", cxxx::noop::_("Semicondensed", 432)},		{"Condensed", cxxx::noop::_("Condensed", 420)},
	{"Semiexpanded", cxxx::noop::_("Semiexpanded", 422)},		{"Extraexpanded", cxxx::noop::_("Extraexpanded", 437)},
	{"Ultraexpanded", cxxx::noop::_("Ultraexpanded", 438)},		{"Expanded", cxxx::noop::_("Expanded", 425)}
};

CFontSheet::CFontSheet(CWindow* parent)
: CWindow(parent),
  mTextFontFamily(this, IDTEXTFAMILY),
  mTitleFontFamily(this, IDTITLEFAMILY),
  mTextFontStyle(this, IDTEXTSTYLE),
  mTitleFontStyle(this, IDTITLESTYLE),
  mTextFontSize(this, IDTEXTSIZE),
  mTitleFontSize(this, IDTITLESIZE),
  mTextFontPreview(this),
  mTitleFontPreview(this)
{
	string textfont = cxProfile->Read("Standard Font", "-adobe-helvetica-medium-r-*-*-12-*-*-*-*-*-iso8859-1");
 	string titlefont = cxProfile->Read("Title Font", textfont);
	CFontDescriptor textfd = CFont::ParseXlfd(textfont);
	CFontDescriptor titlefd = CFont::ParseXlfd(titlefont);
	textfont = textfd.mFamily;
	titlefont = titlefd.mFamily;
	CreateFontList();
	mTextFontFamily.Resize(210, 20);
	mTitleFontFamily.Resize(210, 20);
	unsigned textidx = 0;
	unsigned titleidx = 0;
	unsigned idx = 0;
	string tmp;
	for (CFontList::iterator i = mFontList.begin(); i != mFontList.end(); ++i, ++idx) {
		tmp = i->first;
		mTextFontFamily.AppendItem(i->first);
		mTitleFontFamily.AppendItem(i->first);
		if (tmp == textfont) textidx = idx;
		if (tmp == titlefont) titleidx = idx;
	}
	mTextFontFamily.Select(textidx);
	mTitleFontFamily.Select(titleidx);
	textfont = CFont::GetStyle(textfd);
	titlefont = CFont::GetStyle(titlefd);
	mTextFontStyle.Resize(160, 20);
	mTitleFontStyle.Resize(160, 20);
	map<string, unsigned> textstyle = mFontList[mTextFontFamily.GetText()];
	map<string, unsigned> titlestyle = mFontList[mTitleFontFamily.GetText()];
	FillStyle(mTextFontStyle, textstyle, textfont);
	FillStyle(mTitleFontStyle, titlestyle, titlefont);
	mTextFontSize.Resize(45, 20);
	mTitleFontSize.Resize(45, 20);
	FillSize(mTextFontSize, textstyle[textfont], textfd.mSize);
	FillSize(mTitleFontSize, titlestyle[titlefont], titlefd.mSize);
	if (mTextFont.LoadFont(ComposeXlfd(false))) mTextFontPreview.SetFont(mTextFont);
	if (mTitleFont.LoadFont(ComposeXlfd(true))) mTitleFontPreview.SetFont(mTitleFont);
	mTextFontPreview.Resize(106, 41);
	mTextFontPreview.SetText(_("The quick brown fox jumps over the lazy dog", 442));
	mTitleFontPreview.Resize(106, 41);
	mTitleFontPreview.SetText(_("The quick brown fox jumps over the lazy dog", 442));
	mTextFontFamily.Show();
	mTextFontStyle.Show();
	mTextFontSize.Show();
	mTextFontPreview.Show();
	mTitleFontFamily.Show();
	mTitleFontStyle.Show();
	mTitleFontSize.Show();
	mTitleFontPreview.Show();
}	

/* virtual */ void CFontSheet::OnExpose(XExposeEvent& e __attribute__ ((unused)))
{
	CColor col(TC_FRAME_TEXT);
	DrawString(10, 20 - GetTextHeight() / 2, col, _("Text Font", 431));
	DrawString(10, 110 - GetTextHeight() / 2, col, _("Title Font", 429));
	DrawFrame(0, 0, GetWidth(), GetHeight(), FS_SHEET, false);
	DrawFrame(230, 30, 110, 45, FS_SUNKEN, false);
	DrawFrame(230, 120, 110, 45, FS_SUNKEN, false);
}

/* virtual */ void CFontSheet::OnConfigure(XConfigureEvent& e __attribute__ ((unused)))
{
	mTextFontFamily.Move(10, 30);
	mTitleFontFamily.Move(10, 120);
	mTextFontStyle.Move(10, 55);
	mTitleFontStyle.Move(10, 145);
	mTextFontSize.Move(175, 55);
	mTitleFontSize.Move(175, 145);
	mTextFontPreview.Move(232, 32);
	mTitleFontPreview.Move(232, 122);
}

/* virtual */ void CFontSheet::OnMessage(unsigned msgid, TMessage type __attribute__ ((unused)), unsigned value __attribute__ ((unused)))
{
	const map<string, unsigned>& family = msgid < IDTITLE ? mFontList[mTextFontFamily.GetText()] : mFontList[mTitleFontFamily.GetText()];
	map<string, unsigned>::const_iterator style = family.begin();
	switch (msgid) {
	case IDTEXTFAMILY:
		FillStyle(mTextFontStyle, family, "");
		style = family.find(DeLocalize(mTextFontStyle.GetText()));
	case IDTEXTSTYLE:
		FillSize(mTextFontSize, style->second, 0);
		break;
	case IDTITLEFAMILY:
		FillStyle(mTitleFontStyle, family, "");
		style = family.find(DeLocalize(mTitleFontStyle.GetText()));
	case IDTITLESTYLE:
		FillSize(mTitleFontSize, style->second, 0);
	case IDTEXTSIZE:
	case IDTITLESIZE:
		break;
	default:
		return;
	}
	if (msgid < IDTITLE) {
		if (mTextFont.LoadFont(ComposeXlfd(false))) {
			mTextFontPreview.SetFont(mTextFont);
			mTextFontPreview.Redraw();
		}
	} else {
		if (mTitleFont.LoadFont(ComposeXlfd(true))) {
			mTitleFontPreview.SetFont(mTitleFont);
			mTitleFontPreview.Redraw();
		}
	}		
}

/* virtual */ bool CFontSheet::DoOK()
{
	string fnt = ComposeXlfd(false);
	cxProfile->Write("Standard Font", fnt);
	cxTextFont->Clear();
	cxTextFont->LoadFont(fnt);
	fnt = ComposeXlfd(true);
	cxProfile->Write("Title Font", fnt);
	cxTitleFont->Clear();
	cxTitleFont->LoadFont(fnt);
	return true;
}

void CFontSheet::FillStyle(CDropDownList& d, const map<string, unsigned>& m, const string& preset)
{
	string pres = Localize(preset);
	if (d.GetItemCount() > d.GetSelection()) pres = d.GetText();
	d.Clear();
	unsigned idx = 0;
	unsigned sel = 0;
	string tmp;
	for (map<string, unsigned>::const_iterator i = m.begin(); i != m.end(); ++i) {
		tmp = Localize(i->first);
		d.AppendItem(tmp);
		if (tmp == pres) sel = idx;
		else ++idx;		
	}
	d.Select(sel);
}

void CFontSheet::FillSize(CDropDownList&d, unsigned s, unsigned preset)
{
	if (d.GetItemCount() > d.GetSelection()) preset = atoi(d.GetText().c_str());
	unsigned delta = 36;
	unsigned idx = 0;
	unsigned sel = 0;
	d.Clear();
	for (unsigned i = 1; i < 32; ++i) {
		if (s & (1 << i)) {
			if ((unsigned)abs((int)preset - (int)i - 5) < delta) {
				delta = abs((int)preset - (int)i - 5);
				sel = idx;
			}
			++idx;
			ostringstream os;
			os << i + 5;
			d.AppendItem(os.str());
		}
	}
	if (d.GetItemCount() > sel) d.Select(sel);
}

void CFontSheet::CreateFontList()
{
	vector<CFontDescriptor> list;
	CFont::ListFonts(list);
	if(list.empty()) cxError << _("No fonts found.", 73) << CError::Fatal;
	unsigned sizes;
	string style;
	string family;
	for (vector<CFontDescriptor>::iterator i = list.begin(); i < list.end(); ++i) {
		style = CFont::GetStyle(*i);
		sizes = 0;
		if (i->mSize >= 6 && i->mSize < 36) sizes = 1 << (i->mSize - 5);
		else if (i->mSize == 0) sizes = -1u;
		CFontList::iterator it;
		if (i->mFoundry.empty()) it = mFontList.find(i->mFamily);
		else {
			family = i->mFoundry;
			family += ' ';
			family += i->mFamily;
			for (it = mFontList.begin(); it != mFontList.end(); ++it) {
				if (strcasecmp(it->first.c_str(), i->mFamily.c_str()) == 0) break;
				if (strcasecmp(it->first.c_str(), family.c_str()) == 0) break;
			}
		}
		if (it == mFontList.end()) {
			map<string, unsigned> tmp;
			tmp.insert(map<string, unsigned>::value_type(style, sizes)); 
			mFontList.insert(CFontList::value_type(i->mFamily, tmp));
		} else {
			map<string, unsigned>::iterator itmp = it->second.find(style);
			if (itmp == it->second.end()) {
				it->second.insert(map<string, unsigned>::value_type(style, sizes));
			} else {
				itmp->second |= sizes;
			}
		}
	}
}

string CFontSheet::Localize(const string& s)
{
	string ret = s;
	string::size_type pos = ret.find("Regular");
	if (pos != string::npos) ret.replace(pos, 7, _("Regular", 427));
	else {
		for (unsigned i = 0; i < WEIGHTCOUNT; ++i) {
			pos = ret.find(mWeightNames[i].mOriginal);
			if (pos != string::npos) {
				ret.replace(pos, strlen(mWeightNames[i].mOriginal), cxgettext(mWeightNames[i].mTranslation));
				break;
			}
		}
		for (unsigned i = 0; i < SLANTCOUNT; ++i) {
			pos = ret.find(mSlantNames[i].mOriginal);
			if (pos != string::npos) {
				ret.replace(pos, strlen(mSlantNames[i].mOriginal), cxgettext(mSlantNames[i].mTranslation));
				break;
			}
		}
		for (unsigned i = 0; i < WIDTHCOUNT; ++i) {
			pos = ret.find(mWidthNames[i].mOriginal);
			if (pos != string::npos) {
				ret.replace(pos, strlen(mWidthNames[i].mOriginal), cxgettext(mWidthNames[i].mTranslation));
				break;
			}
		}
	}
	return ret;
}
	
string CFontSheet::DeLocalize(const string& s)
{
	string ret = s;
	string::size_type pos = ret.find(_("Regular", 427));
	if (pos != string::npos) ret.replace(pos, strlen(_("Regular", 427)), "Regular");
	else {
		for (unsigned i = 0; i < WEIGHTCOUNT; ++i) {
			pos = ret.find(cxgettext(mWeightNames[i].mTranslation));
			if (pos != string::npos) {
				ret.replace(pos, strlen(cxgettext(mWeightNames[i].mTranslation)), mWeightNames[i].mOriginal);
				break;
			}
		}
		for (unsigned i = 0; i < SLANTCOUNT; ++i) {
			pos = ret.find(cxgettext(mSlantNames[i].mTranslation));
			if (pos != string::npos) {
				ret.replace(pos, strlen(cxgettext(mSlantNames[i].mTranslation)), mSlantNames[i].mOriginal);
				break;
			}
		}
		for (unsigned i = 0; i < WIDTHCOUNT; ++i) {
			pos = ret.find(cxgettext(mWidthNames[i].mTranslation));
			if (pos != string::npos) {
				ret.replace(pos, strlen(cxgettext(mWidthNames[i].mTranslation)), mWidthNames[i].mOriginal);
				break;
			}
		}
	}
	return ret;
}
