/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CGZStreambuf.cpp
  Info   : $Id: CGZStreambuf.cpp,v 1.3 2012/06/26 08:05:52 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/CGZStreambuf.h>

// C++ headers
#include <cstdio>
#include <algorithm>

// Namespace declarations
using std::streambuf;
using std::streamoff;
using std::streampos;
using std::streamsize;
using std::ios_base;
using std::min;
using std::copy;

CGZStreambuf::CGZStreambuf(std::streambuf& in, unsigned size)
: mIn(in),
  mBufferSize(size)
{
	mState = UNINITIALIZED;
	mInBuffer = new char[mBufferSize];
	mOutBuffer = new char[mBufferSize];
	mHeaderOff = 0;
	if (!DecompressInit()) DecompressEnd();
}

/* virtual */ CGZStreambuf::~CGZStreambuf()
{
	DecompressEnd();
	delete[] mInBuffer;
	delete[] mOutBuffer;
}

/* virtual */ int CGZStreambuf::overflow(int c __attribute__ ((unused)))
{
	// write function disabled
	return EOF;
}

/* virtual */ streampos CGZStreambuf::seekoff(streamoff off, ios_base::seekdir way, ios_base::openmode which)
{
	streampos pos;
	if (way == ios_base::beg) pos = 0;
	else if (way == ios_base::cur) pos = mGZ.total_out;
	else if (way == ios_base::end) return -1; // way too expensive
	pos += off;
	return seekpos(pos, which);
}

/* virtual */ streampos CGZStreambuf::seekpos(streampos pos, ios_base::openmode which)
{
	uLong upos = static_cast<uLong>(pos);
	if ((which & ios_base::out) != 0) return -1;
	if ((which & ios_base::in) == 0) return -1;
	if (mGZ.total_out >= upos) {
		// pos has already been read
		if (static_cast<int>(mGZ.total_out - upos) <= egptr() - eback()) {
			setg(eback(), egptr() - (mGZ.total_out - upos), egptr());
			return pos;
		}
		if (!DecompressInit()) {
			DecompressEnd();
			return -1;
		}
	}
	// decompress until position
	while (mGZ.total_out < upos) {
		if (!Decompress(upos - mGZ.total_out)) break;
	}
	if (mGZ.total_out == upos) {
		setg(eback(), egptr(), egptr());
		return pos;
	} else return -1;
}

/* virtual */ streamsize CGZStreambuf::showmanyc()
{
	return egptr() - gptr();
}

/* virtual */ int CGZStreambuf::underflow()
{
	if (mState != GOOD) return EOF;
	if (gptr() < egptr()) return *gptr();
	while (egptr() == gptr()) {
		if (!Decompress(-1)) break;
	}
	if (gptr() == egptr()) return EOF;
	else return *gptr() & 0xff;
}

/* virtual */ streamsize CGZStreambuf::xsgetn(char* s, streamsize n)
{
	if (mState != GOOD || n == 0) return 0;
	unsigned ret = min(showmanyc(), n);
	if (ret != 0) {
		copy(gptr(), gptr() + ret, s);
		setg(eback(), gptr() + ret, egptr());
		if (ret != n) ret += xsgetn(s + ret, n - ret);	return ret;
	}
	while (egptr() == gptr()) {
		if (!Decompress(n)) break;
	}
	ret = showmanyc();
	cxAssert(ret <= n);
	if (ret == 0) return 0;
	copy(gptr(), egptr(), s);
	setg(eback(), egptr(), egptr());
	if (ret != n) ret += xsgetn(s + ret, n - ret);
	return ret;
}

// Initializes the compression library, returns true if successfull
bool CGZStreambuf::DecompressInit()
{
	if (mState != UNINITIALIZED) DecompressEnd();
	if (!mHeaderOff) {
		// read header
		if (mIn.pubseekpos(0, ios_base::in) == -1) {
			// failed to rewind mIn
			return false;
		}
		if (mIn.sgetn(mInBuffer, 10) != 10) return false;
		char flags = mInBuffer[3];
		if (flags & GZ_RESERVED) return false;
		if (flags & GZ_EXTRA_FIELD) {
			unsigned len = (unsigned)(mInBuffer[9] & 0xff);
			len += (unsigned)(mInBuffer[10] & 0xff) << 8;
			if (mIn.pubseekoff(len, ios_base::cur, ios_base::in) == -1) {
				// failed to position mIn
				return false;
			}
		}
		int in;
		if (flags & GZ_ORIG_NAME) {
			while ((in = mIn.sbumpc()) && in != EOF);
		}
		if (flags & GZ_COMMENT) {
			while ((in = mIn.sbumpc()) && in != EOF);
		}
		mHeaderOff = mIn.pubseekoff(0, ios_base::cur, ios_base::in);
		if (mHeaderOff == -1u) return false;
		if (flags & GZ_HEAD_CRC) mHeaderOff += 2;
	} 
	if (mIn.pubseekpos(mHeaderOff, ios_base::in) == -1) {
		// failed to position mIn
		return false;
	}
	// set buffer pointers
	setg(mOutBuffer + 1, mOutBuffer + 1, mOutBuffer + 1);
	// initialize ZStream pointers
	mGZ.zalloc = NULL;
	mGZ.zfree = NULL;
	mGZ.opaque = NULL;
	mGZ.next_in = reinterpret_cast<Bytef*>(mInBuffer);
	mGZ.avail_in = 0;
	mGZ.total_out = 0;
	mGZ.next_out = reinterpret_cast<Bytef*>(egptr());
	mGZ.avail_out = mBufferSize - 1;
	if (inflateInit2(&mGZ, -MAX_WBITS) == Z_OK) {
		mState = GOOD;
		return true;
	} else return false;
}

// Decompresses a maximum of size characters, returns true if successfull
bool CGZStreambuf::Decompress(unsigned size)
{
	if (mState != GOOD) return false;
	// prepare input buffer
	if (!mGZ.avail_in) {
		mGZ.next_in = (Bytef*)mInBuffer;
		mGZ.avail_in = mIn.sgetn(mInBuffer, mBufferSize);
	}
	// prepare output buffer
	if (!mGZ.avail_out) {
		mOutBuffer[0] = *(egptr() - 1);
		mGZ.next_out = (Bytef*)eback();
		mGZ.avail_out = mBufferSize - 1;
		setg(eback(), eback(), eback());
	}
	// honor size parameter
	if (mGZ.avail_out > size) mGZ.avail_out = size;
	// do the actual decompression
	int ret;
	ret = inflate(&mGZ, Z_SYNC_FLUSH);
	if (ret != Z_OK) {
		if (ret == Z_STREAM_END) mState = END;
		else mState = FAULT;
	}
	// set streambuf pointers
	setg(eback(), gptr(), reinterpret_cast<char*>(mGZ.next_out));
	return mState != FAULT;
}
