/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CIcon.cpp
  Info   : $Id: CIcon.cpp,v 1.41 2012/09/13 14:41:28 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CIcon.h>
#include <cxxx/CApplication.h>
#include <cxxx/CImage.h>
#include <cxxx/CMagic.h>
#include <cxxx/CFileHandler.h>

// C++ headers
#include <cerrno>
#include <cstring>
#include <locale>

// C headers
#include <unistd.h>
#include <dirent.h>

// Namespace declarations
using std::string;
using std::vector;
using std::ostringstream;
using std::ifstream;
using std::ofstream;
using std::stable_sort;
using std::lower_bound;
using std::binary_search;
using std::endl;
using std::locale;
using std::wstring;
using cxxx::cxApp;
using cxxx::cxSysIconDir;
using cxxx::cxHomeDir;
using cxxx::cxDpy;
using cxxx::cxError;
using cxxx::cxMainWin;
using cxxx::cxFileManager;
using cxxx::cxWiden;
using cxxx::lib::cxgettext;

// static variables
bool CIcon::mIsImported = false;
string CIcon::mIconTheme;
CIcon CIcon::mIconCache[SIZE_COUNT*ICON_COUNT + INTERNAL_COUNT];
string CIcon::mFileNameCache[SIZE_COUNT*ICON_COUNT];
vector<string> CIcon::mThemeDirs;
vector<string> CIcon::mAppIconLookup;
vector<string> CIcon::mAppPixmapLookup;

void CIcon::Create(TIconName name, unsigned size)
{
	unsigned idx;
	if (!size) idx = SIZE_COUNT * ICON_COUNT;
	else idx = (size / 10 - 1) * ICON_COUNT;
	idx += name;
	// mIconCache is the icon cache. The file names are directly mapped to
	// the icons.
	CIcon ico = mIconCache[idx];
	if (ico.IsValid()) *this = ico;
	else {
		Detach();
		// not cached, load it
		string fn;
		if (mIconTheme.empty() || size == 0) fn = DefaultFileName(name, size);
		else fn = mFileNameCache[idx];
		// Create the icon
        CImage im;
		im.ReadFile(fn);
		if (im.IsValid()) {
			mInfo = new CIconInfo();
			mInfo->mPicture = im.GetPicture();
			mInfo->mWidth = im.GetWidth();
			mInfo->mHeight = im.GetHeight();
			mInfo->mType = im.GetType();
		} else if (!mIsImported) {
			ImportTheme(mIconTheme);
			fn = mFileNameCache[idx];
			im.ReadFile(fn);
			if (im.IsValid()) {
				mInfo = new CIconInfo();
				mInfo->mPicture = im.GetPicture();
				mInfo->mWidth = im.GetWidth();
				mInfo->mHeight = im.GetHeight();
				mInfo->mType = im.GetType();
			}
			ForceTheme(mIconTheme);
		}			
		Attach();
		mIconCache[idx] = *this;
	}
}

void CIcon::Create(const string& name, unsigned size)
{
	string fn = FindFileName(name, size);
	Create(fn, size, size, CImage::SCALE_UP | CImage::SCALE_DOWN);
}

void CIcon::Create(const string& fn, unsigned w, unsigned h, unsigned flag)
{
	Detach();
	// only absolute file names are allowed
	if (fn.empty() || fn[0] != '/') {
		cxError << _("CIcon::Create: Invalid file name {1}.", 485) << fn << CError::Warn;
		return;
	}
	CImage im;
	im.ScaleFile(fn, w, h, flag);
	if (im.IsValid()) {
		mInfo = new CIconInfo();
		mInfo->mPicture = im.GetPicture();
		mInfo->mWidth = im.GetWidth();
		mInfo->mHeight = im.GetHeight();
		mInfo->mType = im.GetType();
	}
	Attach();
}

vector<CIcon> CIcon::PreviewCreate(const string& theme, const TIconName* name, const unsigned* size)
{
	vector<CIcon> ret;
	// do not mess with mFileNameCache
	string namecache[NAME_CACHE_SIZE];
	// sanity check
	if (theme.empty()) return ret;
	// import the theme even if it has been imported before
	mThemeDirs.clear();
	if (!ImportTheme(theme)) {
		mThemeDirs.clear();
		return ret;
	}
	// open the theme file
    string themefile = cxHomeDir + "/.cxxx/icons/" + ToLower(theme) + ".theme";
	ifstream in(themefile.c_str());
	if (!in) {
		mThemeDirs.clear();
		return ret;
	}
	// fill the name cache
	for (unsigned i = 0; i < NAME_CACHE_SIZE; ++i) {
		getline(in, namecache[i]);
		if (!in) {
			mThemeDirs.clear();
			ret.clear();
			return ret;
		}
	}
	// size 0 indicates the end of the list
	while (*size) {
		unsigned idx = (*size / 10 - 1) * ICON_COUNT + *name;
		CIcon ico;
		ico.Create(namecache[idx], *size, *size, CImage::SCALE_NONE);
		ret.push_back(ico);
		++name;
		++size;
	}
	mThemeDirs.clear();
	return ret;
}

string CIcon::FindFileName(TIconName name, unsigned size)
{
	unsigned idx;
	if (!size) idx = SIZE_COUNT * ICON_COUNT;
	else idx = (size / 10 - 1) * ICON_COUNT;
	idx += name;
	return mFileNameCache[idx];
}

string CIcon::FindFileName(const string& name, unsigned size)
{
	if (name.empty() || name[0] == '/') return name;
	string fn;
	// remove suffix if present
	string::size_type pos = name.rfind('.');
	if (pos == string::npos) fn = name;
	else {
		if (++pos < name.size()) {
			fn = name.substr(pos--);
			if (fn == "png" || fn == "svg" || fn == "pam" || fn == "xpm" || fn == "pbm") fn = name.substr(0, pos);
			else fn = name;
		} else fn = name;
	}
	string tmp;
	// search in icon themes first
	if (mAppIconLookup.empty()) InitAppIconLookup();
	ostringstream os;
	os << size;
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		for (unsigned j = 0; j < APP_CATEGORY_COUNT; ++j) {
			tmp = *i + "/" + os.str() + "x" + os.str() + "/" + mAppCategoryTable[j] + "/" + fn + ".";
			vector<string>::iterator it = lower_bound(mAppIconLookup.begin(), mAppIconLookup.end(), tmp);
			if (it->substr(0, it->size() - 3) == tmp) return *it;
			// Gnome 3
			tmp = *i + "/" + mAppCategoryTable[j] + "/" + os.str() + "/" + fn + ".";
			it = lower_bound(mAppIconLookup.begin(), mAppIconLookup.end(), tmp);
			if (it->substr(0, it->size() - 3) == tmp) return *it;
		}
	}
	if (size != 48) return FindFileName(name, 48);
	// not found: search pixmap directories
	if (mAppPixmapLookup.empty()) InitAppIconLookup();
	for (unsigned i = 0; i < PIXDIR_COUNT; ++i) {
		tmp = mPixmapDirectories[i];
		tmp += "/" + fn + ".";
		vector<string>::iterator j = lower_bound(mAppPixmapLookup.begin(), mAppPixmapLookup.end(), tmp);
		if (j->substr(0, j->size() - 3) == tmp) return *j;
	}
	// not found: give up
	return "";
}

void CIcon::ForceTheme(const std::string& theme)
{
	mIsImported = false;
	mIconTheme = theme;
	mThemeDirs.clear();
	mAppIconLookup.clear();
	mAppPixmapLookup.clear();
	if (!ImportTheme(theme) || !ReadTheme(theme)) mIconTheme = "";
		
	CIcon* i = mIconCache;
	for (unsigned sz = 0; sz < SIZE_COUNT; ++sz)
	{
		for (unsigned name = 0; name < ICON_COUNT; ++name) {
			if (i->IsValid()) {
				string fn = FindFileName((TIconName)name, mSizeTable[sz]);
				CImage im;
				im.ReadFile(fn);
				if (im.IsValid()) {
					// destroy the old Picture
					XRenderFreePicture(cxxx::cxDpy, i->mInfo->mPicture);
					// replace it with the new one
					i->mInfo->mPicture = im.GetPicture();
					i->mInfo->mWidth = im.GetWidth();
					i->mInfo->mHeight = im.GetHeight();
					i->mInfo->mType = im.GetType();
				}
			}
			++i;
		}
	}
}
		
bool CIcon::ImportTheme(const std::string& theme)
{
	if (theme.empty()) return false;
	ofstream out;
	string themefile = cxHomeDir + "/.cxxx/icons";
	struct stat st;
	if (stat(themefile.c_str(), &st) != 0) {
		if (errno == ENOENT) {
			if (mkdir(themefile.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
				cxError << _("Failed to create directory {1}.", 484) << themefile << CError::Error;
				return false;
			}
		} else {
			cxError << _("Could not access directory {1}.", 491) << themefile << CError::Error;
			return false;
		}
	}
	themefile += "/" + ToLower(theme) + ".theme";
	out.open(themefile.c_str());
	
	// create a lookup table for all icons in the theme path
	vector<string> lookup;
	string fn;
	if (theme == "C++X") mThemeDirs.clear();
	else InitTheme(theme);
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		for (unsigned j = 0; j < SIZE_COUNT; ++j) {
			for (unsigned k = 0; k < CATEGORY_COUNT; ++k) {
				fn = *i + "/" + mSizeNameTable[j] + "/" + mCategoryTable[k];
				AddDirectory(fn, lookup);
			}
		}
	}
	// Gnome 3 uses a different naming scheme
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		for (unsigned j = 0; j < SIZE_COUNT; ++j) {
			for (unsigned k = 0; k < CATEGORY_COUNT; ++k) {
				fn = *i + "/" + mCategoryTable[k] + "/" + mGnome3SizeNameTable[j];
				AddDirectory(fn, lookup);
			}
		}
	}
	stable_sort(lookup.begin(), lookup.end());
	lookup.push_back("end.");

	// do the actual import
	for (unsigned i = 0; i < SIZE_COUNT; ++i) {
		for (unsigned j = 0; j < ICON_COUNT; ++j) {
			fn = LookupFileName(static_cast<TIconName>(j), i, lookup);
			if (access(fn.c_str(), R_OK)) fn = DefaultFileName(static_cast<TIconName>(j), mSizeTable[i]);
			out << fn << endl;
		}
	}

	mThemeDirs.clear();
	return true;
}

bool CIcon::ReadTheme(const string& theme)
{
	ifstream in;
	if (!theme.empty()) {
		string themefile = cxHomeDir + "/.cxxx/icons/" + ToLower(theme) + ".theme";
		in.open(themefile.c_str());
		if (!in) {
			in.close();
			in.clear();
			if (mIsImported) return false;
			mIsImported = true;
			if (!ImportTheme(theme)) return false;
			in.open(themefile.c_str());
			if (!in) return false;
		}
		for (unsigned i = 0; i < NAME_CACHE_SIZE; ++i) {
			getline(in, mFileNameCache[i]);
			if (!in) {
				in.close();
				if (mIsImported) return false;
				mIsImported = true;
				if (!ImportTheme(theme)) return false;
				return ReadTheme(theme);
			}
		}
	}
	SetAppIcon(cxApp->GetAppname());
	return true;
}

string CIcon::DefaultFileName(TIconName name, unsigned size)
{
	ostringstream os;
	os << size;
	string sz = os.str() + "x" + os.str();
	return DefaultFileName(name, sz.c_str());
}
	
string CIcon::DefaultFileName(TIconName name, const char* const size)
{
	string fn = cxSysIconDir + "/" + size + "/";
	if (*size == '0') fn += mInternalTable[name];
	else {
		fn += mAliasTable[name];
		fn += ".pam";
	}
	return fn;
}
	
string CIcon::LookupFileName(TIconName name, unsigned sizeidx, const vector<string>& lookup)
{
	const char* alias;
	string fn;
	string tmp;
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		alias = mAliasTable[name];
		alias += strlen(alias) + 1;
		while (*alias) {
			fn = *i + "/" + mSizeNameTable[sizeidx] + "/" + alias + ".";
			vector<string>::const_iterator ret = std::lower_bound(lookup.begin(), lookup.end(), fn);
			if (ret->substr(0, ret->size() - 3) == fn) return *ret;
			// Now for Gnome 3
			tmp = alias;
			string::size_type pos = tmp.rfind('/');
			fn = *i + "/" + tmp.substr(0, pos) + "/" + mGnome3SizeNameTable[sizeidx] + "/" + tmp.substr(pos + 1) + ".";
			ret = std::lower_bound(lookup.begin(), lookup.end(), fn);
			if (ret->substr(0, ret->size() - 3) == fn) return *ret;
			alias += strlen(alias) + 1;
		}
	}
	return DefaultFileName(name, mSizeTable[sizeidx]);
}

void CIcon::InitTheme(const std::string& theme)
{
	if (!theme.empty()) AddTheme(theme);
	AddTheme("hicolor");
}
	
void CIcon::AddTheme(const string& theme)
{
	wstring ws = cxxx::cxWiden(theme);
	const std::ctype<wchar_t>& ct = std::use_facet<std::ctype<wchar_t> >(locale::classic());
	ct.tolower(&ws[0], &ws[0] + ws.size());
	string lotheme = cxxx::cxNarrow(ws);
	
	bool up = true;
	for (unsigned i = 0; i < ws.size(); ++i) {
		if (!ct.is(std::ctype_base::alnum, ws[i])) up = true;
		else if (up) {
			ws[i] = ct.toupper(ws[i]);
			up = false;
		}
	}
	string hitheme = cxxx::cxNarrow(ws);
	string dir = cxHomeDir + "/.icons";
	AddThemeDir(dir, lotheme);
	AddThemeDir(dir, hitheme);
	AddThemeDir("/usr/local/share/icons", lotheme);
	AddThemeDir("/usr/local/share/icons", hitheme);
	AddThemeDir("/usr/share/icons", lotheme);
	AddThemeDir("/usr/share/icons", hitheme);
}

void CIcon::AddThemeDir(const string& dir, const string& theme)
{
	string themepath = dir + "/" + theme;

	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		// this directory has already been processed, nothing to do
		if (themepath == *i) return;
	}
	
	// themepath is not accessible: give up
	if (access(themepath.c_str(), X_OK)) return;
	
	mThemeDirs.push_back(themepath);
	
	string index = themepath + "/index.theme";
	string parent;
	ifstream in;
	in.open(index.c_str());
	while (in) {
		getline(in, parent);
		if (parent.substr(0, 8) == "Inherits") {
			string::size_type pos = 8;
			while (pos < parent.size() && parent[pos] == ' ') ++pos;
			if (++pos > parent.size()) return;
			while (pos < parent.size() && parent[pos] == ' ') ++pos;
			parent = parent.substr(pos);
			while ((pos = parent.find(',')) != string::npos) {
				AddTheme(parent.substr(0, pos));
				if (++pos < parent.size()) parent = parent.substr(pos);
				else break;
			}
			AddTheme(parent);
			break;
		}
	}
}

void CIcon::InitAppIconLookup()
{
	string tmp;
	if (mThemeDirs.empty()) InitTheme(mIconTheme);
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		for (unsigned j = 0; j < SIZE_COUNT; ++j) {
			for (unsigned k = 0; k < APP_CATEGORY_COUNT; ++k) {
				tmp = *i + "/" + mSizeNameTable[j] + "/" + mAppCategoryTable[k];
				AddDirectory(tmp, mAppIconLookup);
			}
		}
	}
	// Gnome 3 uses a different naming scheme
	string size;
	for (vector<string>::iterator i = mThemeDirs.begin(); i != mThemeDirs.end(); ++i) {
		for (unsigned j = 0; j < SIZE_COUNT; ++j) {
			for (unsigned k = 0; k < APP_CATEGORY_COUNT; ++k) {
				tmp = *i + "/" + mAppCategoryTable[k] + "/" + mGnome3SizeNameTable[j];
				AddDirectory(tmp, mAppIconLookup);
			}
		}
	}	
	stable_sort(mAppIconLookup.begin(), mAppIconLookup.end());
	mAppIconLookup.push_back("end.");
	for (unsigned i = 0; i < PIXDIR_COUNT; ++i) {
		AddDirectory(mPixmapDirectories[i], mAppPixmapLookup);
	}
	stable_sort(mAppPixmapLookup.begin(), mAppPixmapLookup.end());
	mAppPixmapLookup.push_back("end.");
}

void CIcon::AddDirectory(const string& dir, vector<string>& cont)
{
	string tmp;
	DIR * d = opendir(dir.c_str());
	if (d) {
	 	dirent * ent;
		while ((ent = readdir(d))) {
			// the icons must follow the naming scheme name.suffix
			// with suffix being 3 letters
			tmp = ent->d_name;
			if (tmp.size() < 5) continue;
			if (tmp[tmp.size() - 4] != '.') continue;
			tmp = tmp.substr(tmp.size() - 3);
			if (tmp != "pam" && tmp != "png" && tmp != "pbm" && tmp != "xpm") continue;
			string path = dir + "/" + ent->d_name;
			cont.push_back(path);
	  	}
		if (closedir(d) != 0) cxError << _("Failed to close directory {1}", 61) << dir << CError::Fatal;
	}
}

void CIcon::SetAppIcon(const string& name)
{
	if (name.empty()) return;
	string fn = cxHomeDir + "/.cxxx/icons/" + name + ".icon";
	string tmp;
	ifstream in(fn.c_str());
	if (in) {
		getline(in, tmp);
		if (!in || tmp != cxSysIconDir) tmp.clear();
	}
	if (tmp.empty()) {
		// file does not exist, or cxSysIconDir has changed, create new file
		ofstream out(fn.c_str());
		// file creation failed: use cxxx icons
		if (!out) return;
		out << cxSysIconDir << endl;
		for (unsigned i = 0; i < SIZE_COUNT; ++i) {
			tmp = FindFileName(name, mSizeTable[i]);
			if (tmp.empty()) tmp = DefaultFileName(CXXX, mSizeTable[i]);
			out << tmp << endl;
			mFileNameCache[i * ICON_COUNT + APP] = tmp;
		}
		return;
	}
	for (unsigned i = 0; i < SIZE_COUNT; ++i) {
		getline(in, tmp);
		if (in) mFileNameCache[i * ICON_COUNT + APP] = tmp;
		else {
			// something is wrong with this file; delete it
			in.close();
			unlink(fn.c_str());
			SetAppIcon(name);
			return;
		}
	}
}

string CIcon::ToLower(const string& s)
{
	wstring ws = cxxx::cxWiden(s);
	std::use_facet<std::ctype<wchar_t> >(locale::classic()).tolower(&ws[0], &ws[0] + ws.size());
	string ret = cxxx::cxNarrow(ws);
	return ret;
}

const unsigned CIcon::mSizeTable[SIZE_COUNT] =
{
	/* 0 */ 16,
	/* 1 */ 22,
	/* 2 */ 32,
	/* 3 */ 48
};

const char* const CIcon::mSizeNameTable[SIZE_COUNT] =
{
	/* 0 */ "16x16",
	/* 1 */ "22x22",
	/* 2 */ "32x32",
	/* 3 */ "48x48"
};

const char* const CIcon::mGnome3SizeNameTable[SIZE_COUNT] =
{
	/* 0 */ "16",
	/* 1 */ "22",
	/* 2 */ "32",
	/* 3 */ "48"
};

const char* const CIcon::mCategoryTable[CATEGORY_COUNT] =
{
	"mimetypes",
	"actions",
	"filesystems",
	"devices",
	"places",
	"status",
	"stock"
};

const char* const CIcon::mAppCategoryTable[APP_CATEGORY_COUNT] =
{
	"apps",
	"places",
	"categories",
	"devices"
};

const char* const CIcon::mPixmapDirectories[PIXDIR_COUNT] =
{
	"/usr/share/pixmaps",
	"/usr/local/share/pixmaps",
	"/usr/share/icons",
	"/usr/local/share/icons"
};
	
const char* const CIcon::mAliasTable[ICON_COUNT] = 
{
	"abiword\0mimetypes/abiword_abi\0mimetypes/application-x-abiword\0mimetypes/gnome-mime-application-x-abiword\0",
	"active_exec\0",
	"applix\0mimetypes/applix\0mimetypes/application-x-applix-word\0mimetypes/gnome-mime-application-x-applix-word\0",
	"archive\0mimetypes/package\0mimetypes/application-x-archive\0mimetypes/gnome-mime-application-x-archive\0mimetypes/gnome-package\0mimetypes/package-x-generic\0",
	"back\0actions/back\0actions/go-previous\0stock/gtk-go-back-ltr\0",
	"binary\0mimetypes/binary\0mimetypes/application-octet-stream\0",
	"blockdevice\0devices/blockdevice\0filesystems/gnome-fs-blockdev\0",
	"browse\0",
	"cdbo_list\0mimetypes/cdbo_list\0",
	"cdimage\0mimetypes/cdimage\0mimetypes/application-x-cd-image\0mimetypes/gnome-mime-application-x-cd-image\0mimetypes/misc-cd-image\0",
	"chardevice\0filesystems/chardevice\0filesystems/gnome-fs-chardev\0",
	"colorpicker\0actions/colorpicker\0actions/color-picker\0",
	"compress\0mimetypes/application-x-compress\0mimetypes/gnome-mime-application-x-compress\0",
	"config\0",
	"configure\0actions/configure\0stock/gtk-configure\0",
	"core\0mimetypes/core\0mimetypes/gnome-mime-application-x-core\0",
	"critical\0actions/messagebox_critical\0status/dialog-error\0stock/gtk-dialog-error\0stock/stock_dialog_error\0",
	"cxxx\0",
	"database\0",
	"deb\0mimetypes/deb\0mimetypes/application-x-deb/mimetypes/gnome-mime-application-x-deb\0mimetypes/package-x-generic\0",
	"desktop\0filesystems/desktop\0mimetypes/application-x-desktop\0mimetypes/gnome-mime-application-x-desktop\0",
	"document\0mimetypes/document\0",
	"dvi\0mimetypes/dvi\0mimetypes/gnome-mime-application-x-dvi\0",
	"editcopy\0actions/editcopy\0actions/edit-copy\0stock/gtk-copy\0",
	"editcut\0actions/editcut\0actions/edit-cut\0stock/gtk-cut\0",
	"editdelete\0actions/editdelete\0actions/edit-delete\0stock/gtk-delete\0",
	"editpaste\0actions/editpaste\0actions/edit-paste\0stock/gtk-paste\0",
	"empty\0mimetypes/empty\0mimetypes/application-x-zerosize\0",
	"exec\0filesystems/exec\0mimetypes/application-x-executable\0mimetypes/gnome-mime-application-x-executable\0filesystems/gnome-mime-executable\0mimetypes/gnome-fs-executable\0",
	"exec_wine\0mimetypes/exec_wine\0",
	"exit\0actions/exit\0actions/application-exit\0stock/gtk-quit\0",
	"file_important\0filesystems/file_important\0",
	"file_locked\0filesystems/file_locked\0",
	"filenew\0actions/filenew\0actions/document-new\0",
	"fileopen\0actions/fileopen\0actions/document-open\0stock/gtk-open\0",
	"folder\0filesystems/folder\0filesystems/gnome-fs-directory\0places/folder\0",
	"folder_open\0filesystems/folder_open\0filesystems/gnome-fs-directory-visiting\0status/folder-open\0places/folder-open\0",
	"folder_tar\0filesystems/folder_tar\0places/folder-tar\0",
	"foldernew\0filesystems/folder_new\0actions/folder_new\0actions/folder-new\0",
	"font\0mimetypes/font\0mimetypes/gnome-mime-application-x-font-pcf\0mimetypes/application/font-x-pcf\0mimetypes/font-x-generic\0",
	"font_bitmap\0mimetypes/font_bitmap\0mimetypes/application-x-font-pcf\0mimetypes/gnome-mime-application-x-font-pcf\0mimetypes/font-x-generic\0",
	"font_truetype\0mimetypes/font_truetype\0mimetypes/application-x-font-ttf\0mimetypes/gnome-mime-application-x-font-ttf\0mimetypes/font-x-generic\0",
	"font_type1\0mimetypes/font_type1\0mimetypes/application-x-font-type1\0mimetypes/gnome-mime-application-x-font-type1\0mimetypes/font-x-generic\0",
	"forward\0actions/forward\0actions/go-next\0stock/gtk-go-forward-ltr\0",
	"gettext\0mimetypes/gettext\0",
	"gohome\0actions/gohome\0actions/go-home\0stock/gtk-home\0",
	"html\0mimetypes/html\0mimetypes/text-html\0mimetypes/gnome-mime-text-html\0",
	"image\0mimetypes/image\0mimetypes/gnome-mime-image\0mimetypes/image-x-generic\0",
	"info\0mimetypes/info\0mimetypes/text-x-texinfo\0mimetypes/gnome-mime-application-x-gnome-app-info\0",
	"install\0mimetypes/install\0mimetypes/text-x-install\0mimetypes/gnome-mime-text-x-install\0",
	"java_jar\0mimetypes/java_jar\0mimetypes/application-x-java-archive\0mimetypes/gnome-mime-application-x-jar\0",
	"karbon\0mimetypes/karbon_karbon\0",
	"kchart\0mimetypes/kchart_chrt\0",
	"kformula\0mimetypes/kformula_kfo\0mimetypes/application-x-kformula\0",
	"kivio\0mimetypes/kivio_flw\0",
	"koffice\0mimetypes/koffice\0",
	"kpovmodeler\0mimetypes/kpovmodeler_doc\0",
	"kpresenter\0mimetypes/kpresenter_kpr\0mimetypes/gnome-mime-application-x-kpresenter\0",
	"krita\0mimetypes/krita_kra\0mimetypes/application-x-krita\0",
	"kspread\0mimetypes/kspread_ksp\0mimetypes/application-x-kspread\0mimetypes/gnome-mime-application-x-kspread\0",
	"kugardata\0mimetypes/kugardata\0",
	"kword\0mimetypes/kword_kwd\0mimetypes/application-x-kword\0mimetypes/gnome-mime-application-x-kword\0",
	"lib\0mimetypes/application-x-sharedlib\0mimetypes/gnome-mime-application-x-sharedlib\0",
	"lib_wine\0",
	"link\0",
	"linux\0",
	"lock\0",
	"log\0mimetypes/log\0mimetypes/text-x-log\0",
	"lyx\0mimetypes/mime_lyx\0mimetypes/application-x-lyx\0mimetypes/gnome-mime-text-x-lyx\0",
	"make\0mimetypes/make\0mimetypes/text-x-makefile\0mimetypes/gnome-mime-text-x-makefile\0",
	"man\0mimetypes/man\0mimetypes/application-x-troff-man\0mimetypes/gnome-mime-text-x-troff-man\0",
	"message\0mimetypes/message\0",
	"midi\0mimetypes/midi\0mimetypes/audio-midi\0mimetypes/gnome-mime-audio-midi\0",
	"msdos\0mimetypes/application-x-ms-dos-executable\0mimetypes/gnome-mime-application-x-ms-dos-executable\0",
	"ooo_calc\0mimetypes/ooo_calc\0mimetypes/application-vnd.oasis.opendocument.spreadsheet\0mimetypes/x-office-spreadsheet\0",
	"ooo_calc_tpl\0mimetypes/ooo_calc_tpl\0mimetypes/application-vnd.oasis.opendocument.spreadsheet-template\0mimetypes/x-office-spreadsheet-template\0",
	"ooo_draw\0mimetypes/ooo_draw\0mimetypes/x-office-drawing\0",
	"ooo_draw_tpl\0mimetypes/ooo_draw_tpl\0mimetypes/x-office-drawing-template\0",
	"ooo_impress\0mimetypes/ooo_impress\0mime/application-vnd.oasis.opendocument.presentation\0mimetypes/x-office-presentation\0",
	"ooo_impress_tpl\0mimetypes/ooo_impress_tpl\0mime/application-vnd.oasis.opendocument.presentation-template\0mimetypes/x-office-presentation-template\0",
	"ooo_writer\0mimetypes/ooo_writer\0mimetypes/application-vnd.oasis.opendocument.text\0mimetypes/x-office-document\0",
	"ooo_writer_tpl\0mimetypes/ooo_writer_tpl\0mimetypes/application-vnd.oasis.opendocument.text-template\0mimetypes/x-office-document-template\0",
	"openterm\0",
	"patch\0mimetypes/patch\0mimetypes/text-x-patch\0mimetypes/gnome-mime-text-x-patch\0",
	"pdf\0mimetypes/pdf\0mimetypes/application-pdf\0mimetypes/gnome-mime-application-pdf\0",
	"pipe\0filesystems/pipe\0filesystems/gnome-fs-fifo\0",
	"postscript\0mimetypes/postscript\0mimetypes/application-postscript\0mimetypes/gnome-mime-application-postscript\0",
	"quicktime\0mimetypes/quicktime\0mimetypes/video-x-generic\0",
	"real\0mimetypes/real\0mimetypes/gnome-mime-audio-x-pn-realaudio\0",
	"recycled\0mimetypes/recycled\0mimetypes/application-x-trash\0mimetypes/gnome-mime-application-x-trash\0",
	"reload\0actions/reload\0actions/view-refresh\0stock/gtk-refresh\0stock/gtk-reload\0",
	"rpm\0mimetypes/rpm\0mimetypes/gnome-mime-application-x-rpm\0mimetypes/package-x-generic\0",
	"shellscript\0mimetypes/shellscript\0mimetypes/application-x-executable-script\0mimetypes/gnome-mime-application-x-shellscript\0mimetypes/gnome-mime-text-x-sh\0mimetypes/text-x-script\0",
	"socket\0filesystems/socket\0filesystems/gnome-fs-socket\0",
	"soffice\0mimetypes/soffice\0",
	"sound\0mimetypes/sound\0mimetypes/gnome-mime-audio\0mimetypes/audio-x-generic\0",
	"source\0mimetypes/source\0",
	"source_c\0mimetypes/source_c\0mimetypes/text-x-csrc\0mimetypes/gnome-mime-text-x-csrc\0",
	"source_cpp\0mimetypes/source_cpp\0mimetypes/text-x-c++src\0mimetypes/gnome-mime-text-x-c++src\0",
	"source_css\0mimetypes/source_css\0mimetypes/gnome-mime-text-css\0",
	"source_f\0mimetypes/source_f\0",
	"source_h\0mimetypes/source_h\0mimetypes/text-x-chdr\0mimetypes/gnome-mime-text-x-chdr\0",
	"source_java\0mimetypes/source_java\0mimetypes/text-x-java\0mimetypes/gnome-mime-text-x-java\0",
	"source_l\0mimetypes/source_l\0",
	"source_moc\0mimetypes/source_moc\0",
	"source_o\0mimetypes/source_o\0mimetypes/application-x-object\0mimetypes/gnome-mime-application-x-object\0",
	"source_p\0mimetypes/source_p\0",
	"source_php\0mimetypes/source_php\0mimetypes/gnome-mime-text-x-php\0",
	"source_pl\0mimetypes/source_pl\0",
	"source_py\0mimetypes/source_py\0mimetypes/text-x-python\0mimetypes/gnome-mime-text-x-python\0",
	"spreadsheet\0mimetypes/spreadsheet\0mimetypes/x-office-spreadsheet\0",
	"tar\0mimetypes/tar\0mimetypes/application-x-tar\0mimetypes/gnome-mime-application-x-tar\0mimetypes/package-x-generic\0",
	"termconfig\0",
	"tex\0mimetypes/tex\0mimetypes/text-x-tex\0mimetypes/gnome-mime-application-x-tex\0",
	"trashcan_empty\0filesystems/trashcan_empty\0filesystems/gnome-fs-trash-empty\0places/user-trash\0",
	"trashcan_full\0filesystems/trashcan_full\0filesystems/gnome-fs-trash-full\0status/user-trash-full\0",
	"txt\0mimetypes/txt\0mimetypes/gnome-mime-text\0mimetypes/ascii\0mimetypes/mime_ascii\0mimetypes/text-x-generic\0",
	"undo\0actions/undo\0actions/edit-undo\0stock/gtk-undo-ltr\0",
	"unknown\0mimetypes/unknown\0",
	"up\0actions/up\0actions/go-up\0stock/gtk-go-up\0",
	"vectorgfx\0mimetypes/vectorgfx\0",
	"video\0mimetypes/video\0mimetypes/gnome-mime-video\0mimetypes/video-x-generic\0",
	"warning\0actions/messagebox_warning\0status/dialog-warning\0stock/gtk-dialog-warning\0",
	"widget_doc\0mimetypes/widget_doc\0",
	"wordprocessing\0mimetypes/wordprocessing\0mimetypes/x-office-document\0",
	"zip\0",
	"cxxx\0"
};

const char* const CIcon::mInternalTable[INTERNAL_COUNT] =
{
	"arrow_down.pbm",
	"arrow_left.pbm",
	"arrow_right.pbm",
	"arrow_up.pbm",
	"bullet4.pbm",
	"bullet8.pbm",
	"check.pbm",
	"radioblack.pbm",
	"radiobright.pbm",
	"radiodark.pbm",
	"radioframe.pbm",
	"treeminus.pbm",
	"treeplus.pbm"
};
