/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CIcon.h
  Info   : $Id: CIcon.h,v 1.22 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CIcon_h
#define CIcon_h

// C++X headers
#include <cxxx/Cxxx.h>

// C++ headers
#include <map>
#include <cstring>

// X11 headers
#include <X11/extensions/Xrender.h>

class CIconInfo
{
public:

	// Constructor
	CIconInfo()
	: mPicture(None),
	  mWidth(0),
	  mHeight(0),
	  mRefCount(0),
	  mType(0)
	{}

	// Destructor
	~CIconInfo()
	{
		if (mRefCount != 0) cxxx::cxError << cxxx::lib::_("Deleting a CIconInfo that is still referenced", 82) << CError::Error;
		if (cxxx::cxMainWin == NULL) return;
		if (mPicture != None) XRenderFreePicture(cxxx::cxDpy, mPicture);
	}

	Picture mPicture;
	uint16_t mWidth;
	uint16_t mHeight;
	uint16_t mRefCount;
	uint16_t mType;
};

class CIcon
{
public:
	enum TIconName
	{
		ARROW_DOWN = 0,
		ARROW_LEFT,
		ARROW_RIGHT,
		ARROW_UP,
		BULLET4,
		BULLET8,
		CHECK,
		RADIOBLACK,
		RADIOBRIGHT,
		RADIODARK,
		RADIOFRAME,
		TREEMINUS,
		TREEPLUS,
		INTERNAL_COUNT,
		ABIWORD = 0,
		ACTIVE_EXEC,
		APPLIX,
		ARCHIVE,
		BACK,
		BINARY,
		BLOCKDEVICE,
		BROWSE,
		CDBO_LIST,
		CDIMAGE,
		CHARDEVICE,
		COLORPICKER,
		COMPRESS,
		CONFIG,
		CONFIGURE,
		CORE,
		CRITICAL,
		CXXX,
		DATABASE,
		DEB,
		DESKTOP,
		DOCUMENT,
		DVI,
		EDITCOPY,
		EDITCUT,
		EDITDELETE,
		EDITPASTE,
		EMPTY,
		EXEC,
		EXEC_WINE,
		EXIT,
		FILE_IMPORTANT,
		FILE_LOCKED,
		FILENEW,
		FILEOPEN,
		FOLDER,
		FOLDER_OPEN,
		FOLDER_TAR,
		FOLDERNEW,
		FONT,
		FONT_BITMAP,
		FONT_TRUETYPE,
		FONT_TYPE1,
		FORWARD,
		GETTEXT,
		GOHOME,
		HTML,
		IMAGE,
		INFO,
		INSTALL,
		JAVA_JAR,
		KARBON,
		KCHART,
		KFORMULA,
		KIVIO,
		KOFFICE,
		KPOVMODELER,
		KPRESENTER,
		KRITA,
		KSPREAD,
		KUGARDATA,
		KWORD,
		LIB,
		LIB_WINE,
		LINK,
		LINUX,
		LOCK,
		LOG,
		LYX,
		MAKE,
		MAN,
		MESSAGE,
		MIDI,
		MSDOS,
		OOO_CALC,
		OOO_CALC_TPL,
		OOO_DRAW,
		OOO_DRAW_TPL,
		OOO_IMPRESS,
		OOO_IMPRESS_TPL,
		OOO_WRITER,
		OOO_WRITER_TPL,
		OPENTERM,
		PATCH,
		PDF,
		PIPE,
		POSTSCRIPT,
		QUICKTIME,
		REAL,
		RECYCLED,
		RELOAD,
		RPM,
		SHELLSCRIPT,
		SOCKET,
		SOFFICE,
		SOUND,
		SOURCE,
		SOURCE_C,
		SOURCE_CPP,
		SOURCE_CSS,
		SOURCE_F,
		SOURCE_H,
		SOURCE_JAVA,
		SOURCE_L,
		SOURCE_MOC,
		SOURCE_O,
		SOURCE_P,
		SOURCE_PHP,
		SOURCE_PL,
		SOURCE_PY,
		SPREADSHEET,
		TAR,
		TERMCONFIG,
		TEX,
		TRASHCAN_EMPTY,
		TRASHCAN_FULL,
		TXT,
		UNDO,
		UNKNOWN,
		UP,
		VECTORGFX,
		VIDEO,
		WARNING,
		WIDGET_DOC,
		WORDPROCESSING,
		ZIP,
		APP,
		ICON_COUNT,

		// CMenuItem
		MENU_BULLET,
		MENU_CHECK,
		MENU_NO_ICON		
	};
	
private:
	enum 
	{
		// Table sizes
		CATEGORY_COUNT = 7,
		APP_CATEGORY_COUNT = 4,
		SIZE_COUNT = 4,
		PIXDIR_COUNT = 4,
		NAME_CACHE_SIZE = SIZE_COUNT * ICON_COUNT,
		ICON_CACHE_SIZE = NAME_CACHE_SIZE + INTERNAL_COUNT
	};
	
public:	
	// Constructor
	CIcon()
	: mInfo(NULL)
	{}

	CIcon(TIconName fn, unsigned size)
	: mInfo(NULL)
	{
		Create(fn, size);
	}
	
	// Copy constructor
	CIcon(const CIcon& ico)
	: mInfo(ico.mInfo)
	{
	Attach();
	}

	// Destructor
	~CIcon()
	{
	Detach();
	}

	// Assignment
	CIcon& operator=(const CIcon& ico)
	{
		if (mInfo == ico.mInfo) return *this;
		Detach();
		mInfo = ico.mInfo;
		Attach();
		return *this;
	}

	// creates a stock icon with size x size pixels
	void Create(TIconName fn, unsigned size);

	// creates an icon from file fn with size x size pixels
	// if fn is not an absolute file path, the icon theme and standard
	// application icon directories are searched
	void Create(const std::string& fn, unsigned size);

	// creates an icon from file fn and scales it to w, h
	// if fn is not an absolute file path, the icon theme and standard
	// application icon directories are searched
	void Create(const std::string& fn, unsigned w, unsigned h, unsigned flag);

	// creates the stock icons reffered to by names using theme theme
	// and returns them
	static std::vector<CIcon> PreviewCreate(const std::string& theme, const TIconName* name, const unsigned* size);

	// clears the icon
	void Clear()
	{
		Detach();
	}

	// returns true if the icon is valid
	bool IsValid() const
	{
		return mInfo != NULL;
	}

	// returns the width of the icon
	unsigned GetWidth() const
	{
		if (IsValid()) return mInfo->mWidth;
		else return 0;
	}

	// returns the height of the icon
	unsigned GetHeight() const
	{
		if (IsValid()) return mInfo->mHeight;
		else return 0;
	}

	// access functions
	Picture GetPicture() const
	{
		if (IsValid()) return mInfo->mPicture;
		else return None;
	}
	
	unsigned GetType() const
	{
		if (IsValid()) return mInfo->mType;
		else return 0;
	}

	// Returns the cached FileName for name
	static std::string FindFileName(TIconName name, unsigned size);

	// Searches the application icon search path for name
	// size x size icons will be returned if the icon is found in the
	// icon theme, size is ignored for icons in the standard icon directories.
	static std::string FindFileName(const std::string& name, unsigned size);


	// Loads theme and replaces all icons
	static void ForceTheme(const std::string& theme);

	// Initializes the icon cache
	static void Init(const std::string& theme)
	{
		if (theme == "C++X") mIconTheme = "";
		else mIconTheme = theme;
		ReadTheme(mIconTheme);
	}

	// Returns the current icon theme
	static const std::string& GetTheme()
	{
		return mIconTheme;
	}

private:
	// Attach a CIconInfo 
	void Attach()
	{
		if (mInfo != NULL) ++mInfo->mRefCount;
	}

	// Detach a CIconInfo
	void Detach()
	{
		if (mInfo != NULL)
		{
			cxAssert(mInfo->mRefCount != 0);
			if (--mInfo->mRefCount == 0) delete mInfo;		
			mInfo = NULL;
		}
	}

	// Imports the icon theme theme, returns true if successfull
	// Prepares the theme for ReadTheme()
	static bool ImportTheme(const std::string& theme);

	// Reads the icon file list of theme, returns true if successfull
	static bool ReadTheme(const std::string& theme);
	
	// Sets the application icon of this application
	static void SetAppIcon(const std::string& name);
	
	// Returns the file name of the icon in cxSysIconDir
	static std::string DefaultFileName(TIconName name, const char* const size);
	static std::string DefaultFileName(TIconName name, unsigned size);
	
	// Returns the cached file name of the stock icon
	static std::string LookupFileName(TIconName name, unsigned sizeidx, const std::vector<std::string>& lookup);

	// Icon themes
	static void InitTheme(const std::string& theme);
	static void AddTheme(const std::string& theme);
	static void AddThemeDir(const std::string& dir, const std::string& theme);

	// App icons
	static void InitAppIconLookup();
	
	// Adds dir to container cont; used by Init*IconLookup
	static void AddDirectory(const std::string& dir, std::vector<std::string>& cont);

	// Returns alternate names for fn
	static std::vector<const char*> GetAlias(const std::string& fn);

	// Compare function for const char*
	static bool Less(const char* s1, const char* s2)
	{
		return std::strcmp(s1, s2) < 0;
	}

	// Returns lowercase of s
	static std::string ToLower(const std::string& s);
	
private:
	
	// All information is stored in this object
	CIconInfo* mInfo;

	// These static members are used to provide caching and theme import
	static bool mIsImported;
	static std::string mIconTheme;
	static CIcon mIconCache[ICON_CACHE_SIZE];
	static std::string mFileNameCache[NAME_CACHE_SIZE];
	static std::vector<std::string> mThemeDirs;
	static std::vector<std::string> mAppIconLookup;
	static std::vector<std::string> mAppPixmapLookup;
	static const char* const mSizeNameTable[SIZE_COUNT];
	static const char* const mGnome3SizeNameTable[SIZE_COUNT];
	static const unsigned mSizeTable[SIZE_COUNT];
	static const char* const mAliasTable[ICON_COUNT];
	static const char* const mCategoryTable[CATEGORY_COUNT];
	static const char* const mAppCategoryTable[APP_CATEGORY_COUNT];
	static const char* const mPixmapDirectories[PIXDIR_COUNT];
	static const char* const mInternalTable[INTERNAL_COUNT];
};

#endif // CIcon_h
