/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CImage.cpp
  Info   : $Id: CImage.cpp,v 1.84 2012/09/12 13:14:46 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CImage.h>
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/pnglite.h>

// C++ headers
#include <fstream>
#include <sstream>
#include <cctype>
#include <cstring>
#include <stdexcept>

#ifdef CXXX_HAS_GIF
  #ifdef CXXX_C_ONLY_GIF
    extern "C"
    {
      #include <gif_lib.h>
    }
  #else  // CXXX_C_ONLY_GIF
    #include <gif_lib.h>
  #endif // CXXX_C_ONLY_GIF
#endif // CXXX_HAS_GIF

#ifdef CXXX_HAS_JPEG
  #ifdef CXXX_C_ONLY_JPEG
    extern "C"
    {
      #include <jpeglib.h>
    }
  #else  // CXXX_C_ONLY_JPEG
    #include <jpeglib.h>
  #endif // CXXX_C_ONLY_JPEG
#endif // CXXX_HAS_JPEG

#ifdef CXXX_HAS_TIFF
  namespace libtiff
  {
    #include <tiffio.h>
  }
#endif // CXXX_HAS_TIFF

// X11 headers
#include <X11/Xutil.h>
#include <X11/extensions/Xrender.h>

// Namespace declarations
using std::string;
using std::ifstream;
using std::ofstream;
using std::hex;
using std::vector;
using std::map;
using std::abs;
using std::runtime_error;
using cxxx::cxARGBFormat;
using cxxx::cxError;
using cxxx::cxDpy;
using cxxx::cxMainWin;
using cxxx::cxBitSwap;
using cxxx::lib::cxgettext;

void CImage::PutImage(const string& file)
{
	Clear();
	mFile.open(file.c_str());
	if (!mFile) {
		cxError << _("Failed to open image file {1}", 83) << file << CError::Warn;
		return ;
	}
	bool success = false;
	switch ((char)mFile.peek()) {
	case 'P':
		success = ReadPNM();
		break;
	case '\x89':
		success = ReadPNG();
		break;
	case '/':
		success = ReadXPM();
		break;
	case 'G':
		success = ReadGIF();
		break;
	case 'i':
	case '\xff':
		success = ReadJPEG();
		break;
	case 'I':
	case 'M':
		mFile.close();
		success = ReadTIFF(file);
		break;
	case 'B':
		success = ReadBMP();
		break;
	case 'X':
		success = ReadXCUR();
	default:
		break;
	}
	mFile.close();
	if (!success) {
		Clear();
		cxError << _("Failed to read image file {1}", 84) << file << CError::Warn;
	}
}


bool CImage::PutPAM()
{
	if (mChannels == 4) {
		// RGB with alpha
		unsigned i = mWidth * mHeight;
		if (mColors == 255) {
			// = 8 bpp: special case
			unsigned char buf[4];
			while (i--) {
				if (!mFile.read((char*)buf, 4)) return false;
				mImage.push_back(CColor(buf[0], buf[1], buf[2], buf[3]));
			}
		} else if (mColors > 255) {
			// > 8 bpp
			char buf[8];
			unsigned char r, g, b, a;
			while (i--) {
				if (!mFile.read(buf, 8)) return false;
				r = GetPAMColor2(buf);
				g = GetPAMColor2(buf + 2);
				b = GetPAMColor2(buf + 4);
				a = GetPAMColor2(buf + 6);
				mImage.push_back(CColor(r, g, b, a));
			}
		} else {
			// < 8 bpp
			char buf[4];
			unsigned char r, g, b, a;
			while (i--) {
				if (!mFile.read(buf, 4)) return false;
				r = GetPAMColor1(buf);
				g = GetPAMColor1(buf + 1);
				b = GetPAMColor1(buf + 2);
				a = GetPAMColor1(buf + 3);
				mImage.push_back(CColor(r, g, b, a));
			}
		}
		mType = ARGB;
	} else if (mChannels == 3) {
		// RGB 
		return PutRawPPM();
	} else if (mChannels == 2) {
		// Grayscale with alpha
		unsigned i = mWidth * mHeight;
		unsigned char val, a;
		if (mColors > 255) {
			// > 8 bpp
			char buf[4];
			while (i--) {
				if (!mFile.read(buf, 4)) return false;
				val = GetPAMColor2(buf);
				a = GetPAMColor2(buf +2);
				mImage.push_back(CColor(val, val, val, a));
			}
		} else {
			// <= 8 bpp
			char buf[2];
			while (i--) {
				if (!mFile.read(buf, 2)) return false;
				val = GetPAMColor1(buf);
				a = GetPAMColor1(buf + 1);
				mImage.push_back(CColor(val, val, val, a));
			}
		}
		mType = ARGB;
	} else {
		// Grayscale
		return PutRawPGM();
	}
	return mFile;
}

bool CImage::PutPBM()
{
	unsigned g;
	unsigned i = mHeight * mWidth;
	while (i--) {
		mFile >> g;
		if (g == 1) g = 0;
		else if (g) {
			cxError << _("CImage::PutPBM: Index out of range.", 392) << CError::Warn;
			return false;
		} else g = 0xff;					
		mImage.push_back(CColor(g, g, g, g));
	}
	mType = ALPHA;
	return mFile;
}

bool CImage::PutPGM()
{
	unsigned g;
	unsigned i = mHeight * mWidth;
	while (i--) {
		mFile >> g;
		g = (g * 0xff) / mColors;
		mImage.push_back(CColor(g, g, g, g));
	}
	mType = ALPHA;
	return mFile;
}

bool CImage::PutPPM()
{
	unsigned i = mHeight * mWidth;
	unsigned r, g, b;
	while (i--) {
		mFile >> r >> g >> b;
		r = (r * 0xff) / mColors;
		g = (g * 0xff) / mColors;
		b = (b * 0xff) / mColors;
		mImage.push_back(CColor(r, b, g));
	}
	mType = ARGB;
	return mFile;
}

bool CImage::PutRawPBM()
{
	unsigned w1 = (mWidth) / 8;
	unsigned w2 = (mWidth) % 8;
	unsigned char c = 0;
	unsigned char pix;
	for (unsigned y = 0; y < mHeight; ++y) {
		for (unsigned x = 0; x < w1; ++x) {
			if (!mFile.get((char&)c)) return false; 
			for (unsigned i = 0x80; i != 0; i >>= 1) {
				if (c & i) pix = 0;
				else pix = 0xff;
				mImage.push_back(CColor(pix, pix, pix, pix));
			}
		}
		if (w2) {
			if (!mFile.get((char&)c)) return false; 
			for (unsigned i = 0x80; i != (unsigned)(0x80 >> w2); i >>= 1) {
				if (c & i) pix = 0;
				else pix = 0xff;
				mImage.push_back(CColor(pix, pix, pix, pix));
			}
		}
	}
	mType = ALPHA;
	return mFile;
}

bool CImage::PutRawPGM()
{
	char buf[2];
	unsigned char g;
	unsigned i = mHeight * mWidth;
	if (mColors == 0xff) {
		// = 8 bpp
		while (i--) {
			if (!mFile.get((char&)g)) return false;
			mImage.push_back(CColor(g, g, g, g));
		}
	} else if (mColors > 0xff) {
		// > 8 bpp
		while (i--) {
			if (!mFile.read(buf, 2)) return false;
			g = GetPAMColor2(buf);
			mImage.push_back(CColor(g, g, g, g));
		}
	} else {
		// <= 8bpp
		while (i--) {
			if (!mFile.get(*buf)) return false;
			g = GetPAMColor1(buf);
			mImage.push_back(CColor(g, g, g, g));
		}
	}
	mType = ALPHA;
	return true;
}

bool CImage::PutRawPPM()
{
	unsigned char r, g, b;
	unsigned i = mHeight * mWidth;
	if (mColors == 0xff) {
		unsigned char buf[3];
		while (i--) {
			if (!mFile.read((char*)buf, 3)) return false;
			mImage.push_back(CColor(buf[0], buf[1], buf[2]));
		}
	} else if (mColors > 0xff) {
		// > 8 bpp
		char buf[6];
		while (i--) {
			if (!mFile.read(buf, 6)) return false;
			r = GetPAMColor2(buf);
			g = GetPAMColor2(buf + 2);
			b = GetPAMColor2(buf + 4);
			mImage.push_back(CColor(r, g, b));
		}
	} else {
		// <= 8 bpp
		char buf[3];
		while (i--) {
			if (!mFile.read(buf, 3)) return false;
			r = GetPAMColor1(buf);
			g = GetPAMColor1(buf + 1);
			b = GetPAMColor1(buf + 2);
			mImage.push_back(CColor(r, g, b));
		}
	}
	mType = ARGB;
	return mFile;
}

bool CImage::ReadBMP()
{
	// Read BMP header
	char buf[54];
	unsigned char& b = (unsigned char&)buf[0];
	unsigned char& g = (unsigned char&)buf[1];
	unsigned char& r = (unsigned char&)buf[2];
	if (!mFile.read(buf, 54)) return false;
	if (buf[0] != 'B' || buf[1] != 'M') {
		cxError << _("CImage::ReadBMP: Wrong magic.", 85) << CError::Warn;
		return false;
	}
	// buf[2]...buf[5] contain file size
	if (GetUint32LE(buf + 6) != 0) {
		cxError << _("CImage::ReadBMP: Reserved Bytes not zero.", 387) << CError::Warn;
		return false;
	}
	if (GetUint32LE(buf + 14) != 40) {
		cxError << _("CImage::ReadBMP: Wrong bitmap info header size.", 380) << CError::Warn;
		return false;
	}
	if (GetUint32LE(buf + 30) != 0) {
		cxError << _("C++X cannot (yet) handle compressed BMP images.", 382) << CError::Warn;
		return false;
	}
	mWidth = abs(static_cast<int>(GetUint32LE(buf + 18)));
	mHeight = GetUint32LE(buf + 22);
	if (static_cast<int>(mHeight) < 0) {
		cxError << _("CImage::ReadBMP: Upside down image not supported", 481) << CError::Warn;
		return false;
	}
	unsigned bpp = buf[28] + 256 * buf[29];
	if (bpp <= 8) {
		// these are indexed
		mColors = GetUint32LE(buf + 46);
		if (mColors == 0) mColors = 1 << bpp;
	} else mColors = 0;
	// Read bmp colormap; mColors = # of colormap entries
	for (unsigned i = 0; i < mColors; ++i) {
		mFile.read(buf, 4);
		mColormap.push_back(CColor(r, g, b));
	}
	if (!mFile) {
		cxError << _("CImage::ReadBMP: premature end of file.", 379) << CError::Warn;
		return false;
	}
	// Read image data
	mImage.resize(mWidth * mHeight);
	vector<CColor>::iterator iter = mImage.begin() + (mHeight - 1) * mWidth;
	// 1 bit per pixel
	if (bpp == 1) {
		unsigned w = mWidth / 8;
		unsigned bytepad = ((mWidth + 31) / 32) * 4 - (mWidth + 7) / 8;
		unsigned i;
		for (unsigned y = 0; y < mHeight; ++y) {
			for (i = 0; i < w; ++i) {
				mFile.get((char&)r);
				for (unsigned char mask = 0x80; mask != 0; mask >>= 1) {
					if ((r & mask) == 0) *iter++ = mColormap[0];
					else *iter++ = mColormap[1];
				}
			}
			if ((mWidth & 7) != 0) {
				mFile.get((char&)r);
				for (i = 0; i < (mWidth & 7); ++i) {
					if (r & 0x80) *iter++ = mColormap[1];
					else *iter++ = mColormap[0];
					r <<= 1;
				}
			}
			if (bytepad != 0) mFile.read(buf, bytepad);
			if (!mFile) {
				cxError << _("CImage::ReadBMP: premature end of file.", 379) << CError::Warn;
				return false;
			}
			iter -= 2 * mWidth;
		}
	// 4 bits per pixel
	} else if (bpp == 4) {		
		unsigned w = mWidth / 2;
		unsigned bytepad = ((mWidth + 7) / 8) * 4 - (mWidth + 1) / 2;
		unsigned i;
		for (unsigned y = 0; y < mHeight; ++y) {
			for (i = 0; i < w; ++i) {
				mFile.get((char&)r);
				g = (r >> 4);
				r &= 0x0f;
				if (r >= mColors || g >= mColors) {
					r = 0;
					g = 0;
					cxError << _("CImage::ReadBMP: index out of range.", 384) << CError::Warn;
				}
				*iter++ = mColormap[g];
				*iter++ = mColormap[r];
			}
			if ((mWidth & 1) != 0) {
				mFile.get((char&)r);			
				r >>= 4;
				if (r >= mColors) {
					r = 0;
					cxError << _("CImage::ReadBMP: index out of range.", 384) << CError::Warn;
				}
				*iter++ = mColormap[r];
			}
			if (bytepad != 0) mFile.read(buf, bytepad);
			if (!mFile) {
				cxError << _("CImage::ReadBMP: premature end of file.", 379) << CError::Warn;
				return false;
			}
			iter -= 2 * mWidth;
		}
	// 8 bits per pixel
	} else if (bpp == 8) {		
		unsigned bytepad = ((mWidth + 3) / 4) * 4 - mWidth;
		unsigned i;
		for (unsigned y = 0; y < mHeight; ++y) {
			for (i = 0; i < mWidth; ++i) {
				mFile.get((char&)r);
				if (r >= mColors) {
					r = 0;
					cxError << _("CImage::ReadBMP: index out of range.", 384) << CError::Warn;
				}
				*iter++ = mColormap[r];
			}
			if (bytepad != 0) mFile.read(buf, bytepad);
			if (!mFile) {
				cxError << _("CImage::ReadBMP: premature end of file.", 379) << CError::Warn;
				return false;
			}
			iter -= 2 * mWidth;
		}
	// 24 bits per pixel
	} else if (bpp == 24) {		
		unsigned i;
		for (unsigned y = 0; y < mHeight; ++y) {
			for (i = 0; i < mWidth; ++i) {
				mFile.read(buf, 3);
				*iter++ = CColor(r, g, b);
			}
			if (!mFile) {
				cxError << _("CImage::ReadBMP: premature end of file.", 379) << CError::Warn;
				return false;
			}
			iter -= 2 * mWidth;
		}
	} else {
		cxError << _("CImage::ReadBMP: color depth of {1} bits not supported.", 381) << bpp << CError::Warn;
		return false;
	}
	mType = ARGB;
	return true;
}

extern "C" {
	unsigned cxxx_PngRead(void* output, size_t size, size_t numel, void* user_pointer)
	{
		ifstream* ist = (ifstream*)user_pointer;
		if (output) ist->read((char*)output, size * numel);
		else ist->ignore(size * numel);
		if (*ist) return numel;
		else return 0;
	}
}

bool CImage::ReadPNG()
{
	png_t spng;
	if (png_open_read(&spng, &cxxx_PngRead, &mFile) != PNG_NO_ERROR) return false;
	mWidth = spng.width;
	mHeight = spng.height;
	vector<unsigned char>array(mWidth * mHeight * spng.bpp);
	if (png_get_data(&spng, &array[0]) != PNG_NO_ERROR) {
		png_end_read(&spng);
		return false;
	}
	vector<unsigned char>::iterator i = array.begin();
	unsigned char* tmp;
	unsigned char r;
	unsigned char g;
	unsigned char b;
	unsigned char a;

	switch (spng.color_type) {
	case PNG_GREYSCALE:
		if (spng.bpp == 1) {
			// 8 bit grayscale
			while (i < array.end()) {
				g = *i++;
				if (spng.transparency && g == spng.png_trns[0]) mImage.push_back(CColor(0, 0, 0, 0));
				else mImage.push_back(CColor(g, g, g, g));
			}
		} else if (spng.bpp == 2) {
			// 16 bit grayscale
			while (i < array.end()) {
				tmp = &(*i);
				g = *i++;
				++i;
				if (spng.transparency && memcmp(tmp, spng.png_trns, 2) == 0) mImage.push_back(CColor(0, 0, 0, 0));
				else mImage.push_back(CColor(g, g, g, g));
			}
		} else return false;
		mType = ALPHA;
		break;

	case PNG_TRUECOLOR:
		if (spng.bpp == 3) {
			// 8 bit RGB
			while (i < array.end()) {
				tmp = &(*i);
				r = *i++;
				g = *i++;
				b = *i++;
				if (spng.transparency && memcmp(tmp, spng.png_trns, 3) == 0) mImage.push_back(CColor(0, 0, 0, 0));
				else mImage.push_back(CColor(r, g, b));
			}
		} else if (spng.bpp == 6) {
			// 16 bit RGB
			while (i < array.end()) {
				tmp = &(*i);
				r = *i++;
				++i;
				g = *i++;
				++i;
				b = *i++;
				++i;
				if (spng.transparency && memcmp(tmp, spng.png_trns, 6) == 0) mImage.push_back(CColor(0, 0, 0, 0));
				else mImage.push_back(CColor(r, g, b));
			}
		} else return false;
		mType = ARGB;
		break;
			
	case PNG_GREYSCALE_ALPHA:
		if (spng.bpp == 2) {
			// 8 bit grayscale alpha
			while (i < array.end()) {
				g = *i++;
				a = *i++;
				mImage.push_back(CColor(g, g, g, a));
			}
		} else if (spng.bpp == 4) {
			// 16 bit grayscale alpha
			while (i < array.end()) {
				g = *i++;
				++i;
				a = *i++;
				++i;
				mImage.push_back(CColor(g, g, g, a));
			}
		} else return false;
		mType = ARGB;
		break;

	case PNG_TRUECOLOR_ALPHA:
		if (spng.bpp == 4) {
			// 8 bit RGBA
			while (i < array.end()) {
				r = *i++;
				g = *i++;
				b = *i++;
				a = *i++;
				mImage.push_back(CColor(r, g, b, a));
			}
		} else if (spng.bpp == 8) {
			// 16 bit RGBA
			while (i < array.end()) {
				//i++;
				r = *i++;
				++i;
				g = *i++;
				++i;
				b = *i++;
				++i;
				a = *i++;
				++i;
				mImage.push_back(CColor(r, g, b, a));
			}
		} else return false;
		mType = ARGB;
		break;

		case PNG_INDEXED:
			// 8 bit RGBA
			unsigned idx;
			unsigned char* pal;
			while (i < array.end()) {
				idx = *i++;
				pal = spng.png_plte + 4 * idx;
				r = *pal++;
				g = *pal++;
				b = *pal++;
				a = *pal;
				mImage.push_back(CColor(r, g, b, a));
			}
			mType = ARGB;
			break;
			
		default:
			png_end_read(&spng);
			return false;
	}
	
	png_end_read(&spng);
	return true;
}

#ifdef CXXX_HAS_GIF
int cxxx_GifRead(GifFileType* gif, GifPixelType* buf, int len)
{
	ifstream& file = *(ifstream*)gif->UserData;
	if (file.read((char*)buf, len)) return len;
	else return 0;
}
/* CImage::ReadGIF()
This function is partly based on libungif 4.1.3, file gif2x11.c,
written by Gershon Elber. */
bool CImage::ReadGIF()
{
    static int sInterlacedOffset[] = { 0, 4, 2, 1 }; /* The way Interlaced image should. */
    static int sInterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */
	GifFileType* gif = DGifOpen(&mFile, cxxx_GifRead);
	if (gif == NULL) return false;
	mWidth = gif->SWidth;
	mHeight = gif->SHeight;
	vector<GifPixelType> array(mWidth * mHeight, gif->SBackGroundColor);
    GifRecordType rec = UNDEFINED_RECORD_TYPE;
	int idx;
	int extcode;
	GifByteType* extension;
	unsigned char alphaidx = 255;
	bool has_alpha = false;
	int t, l, w, h;
	do {
		if (DGifGetRecordType(gif, &rec) == GIF_ERROR) {
			rec = UNDEFINED_RECORD_TYPE;
			break;
		}
		switch (rec) {
		case IMAGE_DESC_RECORD_TYPE:
			if (DGifGetImageDesc(gif) == GIF_ERROR) {
				rec = UNDEFINED_RECORD_TYPE;
				break;
			}
			t = gif->Image.Top;
			l = gif->Image.Left;
			w = gif->Image.Width;
			h = gif->Image.Height;
			if ((l + w) > (int)mWidth || (t + h) > (int)mHeight) {
				rec = UNDEFINED_RECORD_TYPE;
				break;
			}
			if (gif->Image.Interlace) {
				/* Need to perform 4 passes on the images: */
				for (int i = 0; i < 4; i++) {
					for (int j = sInterlacedOffset[i]; j < h; j += sInterlacedJumps[i]) {
						idx = (j + t) * mWidth + l;
						if (DGifGetLine(gif, &array[idx], w) == GIF_ERROR) {
							rec = UNDEFINED_RECORD_TYPE;
							i = 3;
							break;
						}
					}
				}
			} else {
				for (int i = 0; i < h; ++i) {
					idx = (i + t) * mWidth + l;
					if (DGifGetLine(gif, &array[idx], w) == GIF_ERROR) {
						rec = UNDEFINED_RECORD_TYPE;
						break;
					}
				}
			}
			if (rec != UNDEFINED_RECORD_TYPE) rec = TERMINATE_RECORD_TYPE;
			break;
		case EXTENSION_RECORD_TYPE:
			if (DGifGetExtension(gif, &extcode, &extension) == GIF_ERROR) {
				rec = UNDEFINED_RECORD_TYPE;
				break;
			}
			if (extcode == 0xf9) {
				// Graphics Control Block
				if (extension[0] == 4 && (extension[1] & 1) == 1) {
					has_alpha = true;
					alphaidx = extension[4];
				}
			}					
			while (extension != NULL) {
				if (DGifGetExtensionNext(gif, &extension) == GIF_ERROR) {
					rec = UNDEFINED_RECORD_TYPE;
					break;
				}
			}
			break;
		case TERMINATE_RECORD_TYPE:
			break;
		default:
			cxError << _("CImage::ReadGif: Unknown Record Type.", 393) << CError::Warn;
			break;
		}
	} while (rec != TERMINATE_RECORD_TYPE && rec != UNDEFINED_RECORD_TYPE);
	ColorMapObject* cmap;
	if (rec == TERMINATE_RECORD_TYPE) {
		if (gif->Image.ColorMap != NULL) cmap = gif->Image.ColorMap;
		else cmap = gif->SColorMap;
		cxAssert(cmap != NULL);
		GifColorType color;
		for (unsigned i = 0; i < array.size(); ++i) {
			if (array[i] == alphaidx && has_alpha) mImage.push_back(CColor(0, 0, 0, 0));
			else {
				if (array[i] >= cmap->ColorCount) {
					rec = UNDEFINED_RECORD_TYPE;
					break;
				}
				color = cmap->Colors[array[i]];
				mImage.push_back(CColor(color.Red, color.Green, color.Blue));
			}
		}
	}
	if (DGifCloseFile(gif) == GIF_OK) {
		mType = ARGB;
		return rec == TERMINATE_RECORD_TYPE;
	}
	else return false;
}
#else // CXXX_HAS_GIF
bool CImage::ReadGIF()
{
	cxError << _("libgif required to read GIF images.", 386) << CError::Warn;
	return false;
}
#endif // CXXX_HAS_GIF

#ifdef CXXX_HAS_JPEG
#define JPEGBUFFER 256
extern "C" {
	struct cxxx_jpeg_src
	{ 
		cxxx_jpeg_src(ifstream& in, vector<JOCTET>& buf)
		: infile(in),
		  buffer(buf)
		{};		
		jpeg_source_mgr pub;	/* public fields */
		ifstream& infile;		/* source stream */
		vector<JOCTET>& buffer;		/*buffer */
	};

	void cxxx_init_source(j_decompress_ptr jpeg __attribute__ ((unused)))
	{
	}

	boolean cxxx_fill_input_buffer(j_decompress_ptr jpeg)
	{
		cxxx_jpeg_src& src = (cxxx_jpeg_src&)*jpeg->src;
		if (!src.infile.read((char*)&src.buffer[0], JPEGBUFFER)) {
			if ((src.pub.bytes_in_buffer = src.infile.gcount()) == 0) {
				src.buffer[0] = (JOCTET) 0xFF;
				src.buffer[1] = (JOCTET) JPEG_EOI;
				src.pub.bytes_in_buffer = 2;
			}
		} else src.pub.bytes_in_buffer = JPEGBUFFER;
		src.pub.next_input_byte = &src.buffer[0];
		return TRUE;
	}			

	void cxxx_skip_input_data (j_decompress_ptr jpeg, long count)
	{
		cxxx_jpeg_src& src = (cxxx_jpeg_src&)*jpeg->src;
		if (count > 0) {
			while (count > (long)src.pub.bytes_in_buffer) {
			count -= src.pub.bytes_in_buffer;
			cxxx_fill_input_buffer(jpeg);
		}
		src.pub.next_input_byte += count;
		src.pub.bytes_in_buffer -= count;
	  }
	}

	void cxxx_term_source (j_decompress_ptr jpeg __attribute__ ((unused)))
	{
	  /* no work necessary here */
	}

	void cxxx_error_exit(j_common_ptr jpeg)
	{
		(*jpeg->err->output_message)(jpeg);
		/* Let the memory manager delete any temp files before we die */
		jpeg_destroy(jpeg);
		throw runtime_error("Libjpeg");
	}

	void cxxx_output_message(j_common_ptr jpeg)
	{
		char buffer[JMSG_LENGTH_MAX];
		(*jpeg->err->format_message) (jpeg, buffer);
		cxError << "Libjpeg: " << buffer << CError::Warn;
	}		
}

bool CImage::ReadJPEG()
{
	struct jpeg_decompress_struct jpeg;
	struct jpeg_error_mgr jerr;
	jpeg.err = jpeg_std_error(&jerr);
	jpeg.err->error_exit = &cxxx_error_exit;
	jpeg.err->output_message=&cxxx_output_message;
	jpeg_create_decompress(&jpeg);
	vector<JOCTET> inbuf(JPEGBUFFER);
	cxxx_jpeg_src srcmgr (mFile, inbuf);
	jpeg.src = (jpeg_source_mgr*) &srcmgr;
	srcmgr.pub.init_source = cxxx_init_source;
	srcmgr.pub.fill_input_buffer = cxxx_fill_input_buffer;
	srcmgr.pub.skip_input_data = cxxx_skip_input_data;
	srcmgr.pub.resync_to_restart = jpeg_resync_to_restart;
	srcmgr.pub.term_source = cxxx_term_source;
	srcmgr.pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	srcmgr.pub.next_input_byte = NULL; /* until buffer loaded */
	try {
		jpeg_read_header(&jpeg, TRUE);
		mWidth = jpeg.image_width;
		mHeight = jpeg.image_height;
		jpeg_start_decompress(&jpeg);
		unsigned sz = jpeg.output_width * jpeg.output_components;
		JSAMPROW buf;
		vector<JSAMPLE> samples(sz);
		buf = &samples[0];
		if (jpeg.output_components == 1) {
			while (jpeg.output_scanline < jpeg.output_height) {
				jpeg_read_scanlines(&jpeg, &buf, 1);
				for (unsigned i = 0; i < sz; ++i) {
					mImage.push_back(CColor(buf[i], buf[i], buf[i]));
				}
			}
		} else if (jpeg.output_components == 3) {
			while (jpeg.output_scanline < jpeg.output_height) {
				jpeg_read_scanlines(&jpeg, &buf, 1);
				for (unsigned i = 0; i < sz; i += 3) {
					mImage.push_back(CColor(buf[i], buf[i + 1], buf[i + 2]));
				}
			}
		} else {
			jpeg_destroy_decompress(&jpeg);
			return false;
		}
		jpeg_finish_decompress(&jpeg);
		jpeg_destroy_decompress(&jpeg);
		mType = ARGB;
		return true;
	}
	catch (const runtime_error& e) {
		return false;
	}
}
#else // CXXX_HAS_JPEG
bool CImage::ReadJPEG()
{
	cxError << _("libjpeg required to read jpeg images.", 385) << CError::Warn;
	return false;
}
#endif // CXXX_HAS_JPEG

#ifdef CXXX_HAS_TIFF
using libtiff::TIFFOpen;
using libtiff::TIFFGetField;
using libtiff::TIFFReadRGBAImage;
using libtiff::TIFFClose;
using libtiff::uint32;

bool CImage::ReadTIFF(const string& file)
{
	libtiff::TIFF* tif = TIFFOpen(file.c_str(), "r");
	if (!tif) return false;
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &mWidth);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &mHeight);
	vector<uint32> array(mWidth * mHeight);
	bool ret = TIFFReadRGBAImage(tif, mWidth, mHeight, &array[0], 0);
	TIFFClose(tif);
	if (!ret) return false;
	uint32 pixel;
	for (unsigned i = mHeight; i > 0;) {
		--i;
		for (unsigned j = 0; j < mWidth; ++j) {
			pixel = array[i * mWidth + j];
			mImage.push_back(CColor(TIFFGetR(pixel), TIFFGetG(pixel), TIFFGetB(pixel), TIFFGetA(pixel)));
		}
	}
	mType = ARGB;
	return true;
}
#else // CXXX_HAS_TIFF
bool CImage::ReadTIFF(const string& file __attribute__ ((unused)))
{
	cxError << _("libtiff required to read TIFF images.", 86) << CError::Warn;
	return false;
}
#endif // CXXX_HAS_TIFF

bool CImage::ReadXPM()
{
	unsigned bpp = ReadXPMHeader();
	if (bpp == 0 || bpp > 15) return false;
	if (!ReadXPMColormap(bpp)) return false;
	char buf[16];
	memset(buf, 0, 16);
	unsigned x, y;
	map<string, CColor>::iterator idx;
	for (y = 0; y < mHeight; ++y) {
		mFile.ignore(1000, '\"');
		for (x = 0; x < mWidth; ++x) {
			mFile.read(buf, bpp);
			idx = mColortoken.find(buf);
			if (idx == mColortoken.end()) return false;
			mImage.push_back(idx->second);
		}
		mFile.ignore(1000, '\"');
	}
	mType = ARGB;
	return mFile;
}

bool CImage::ReadXCUR()
{
	char buf[4];
	mFile.read(buf, 4);
	if (!mFile) return false;
	// file magic
	if (memcmp(buf, "Xcur", 3)) return false;
	// bytes in this header and version number - ignored for now
	mFile.ignore(8);
	// entries in table of contents
	unsigned ntoc = ReadUint32LE();
	// this should be sufficient, beware of animated cursors
	if (ntoc > 100) return false;
	int delta = 1000;
	unsigned pos = -1u;
	// iterate through the list
	for (unsigned i = 0; i < ntoc; ++i) {
		// entry type (image or text)
		unsigned typ = ReadUint32LE();
		if (typ == 0xfffd0002) {
			// the nominal size
			int sz = ReadUint32LE();
			if (sz == -1) return false;
			if (abs(sz - 20) < delta) {
				// closer to 32; use this image
				delta = abs(sz - 20);
				// the position in this file
				pos = ReadUint32LE();
			} else {
				// discard this one
				ReadUint32LE();
			}
		}
	}
	if (pos == -1u) return false;
	// move to the image
	mFile.seekg(pos);
	if (!mFile) return false;
	// header size has to be 36
	if (ReadUint32LE() != 36) return false;
	// type is 0xfffd0002
	if (ReadUint32LE() != 0xfffd0002) return false;
	// nominal size
	if (ReadUint32LE() == -1u) return false;
	// version - only version 1 is supported
	if (ReadUint32LE() != 1) return false;
	// width of the image
	mWidth = ReadUint32LE();
	if (mWidth > MAXWIDTH) return false;
	// height of the image
	mHeight = ReadUint32LE();
	if (mHeight > MAXHEIGHT) return false;
	// hot spot x position - must be equal or less than width
	mXHot = ReadUint32LE();
	if (mXHot > mWidth) return false;
	// hot spot y position - must be equal or less than height
	mYHot = ReadUint32LE();
	if (mYHot > mHeight) return false;
	// delay
	if (ReadUint32LE() == -1u) return false;
	// finally the array of ARGB pixels
	for (unsigned i = 0; i < mWidth * mHeight; ++i) {
		mFile.read(buf, 4);
		mImage.push_back(CColor(buf[2], buf[1], buf[0], buf[3]));
	}
	mType = ARGB;
	return mFile;
}

bool CImage::ReadPNMHeader(bool bw)
{
	char c;
	if (mFile.peek() == '#') mFile.ignore(1000, '\n');
	mFile >> mWidth;
	do {
		mFile.get(c);
	} while (mFile && !isspace(c));
	if (mFile.peek() == '#') mFile.ignore(1000, '\n');
	mFile >> mHeight;
	do {
		mFile.get(c);
	} while (mFile && !isspace(c));
	// sanity check
	if (mWidth == 0) return false;
	if (mHeight == 0) return false;
	if(bw) return mFile;
	if (mFile.peek() == '#') mFile.ignore(1000, '\n');
	mFile >> mColors;
	do {
		mFile.get(c);
	} while (mFile && !isspace(c));
	// sanity check
	if (!mColors) return false;
	if (mColors > (0xffff)) {
		cxError << _("CImage::ReadPNMHeader: More than 16 bit per channel.", 394) << CError::Warn;
		return false;
	}
	return mFile;
}


unsigned CImage::ReadPAMHeader()
{
	string token, tupltyp;
	char c;
	while (mFile) {
		mFile.get(c);
		token = "";
		// skip comment
		while (mFile && c == '#') {
			mFile.ignore(1000, '\n');
			mFile.get(c);
		}
		// get token
		while (mFile && !isspace(c)) {
			token += c;
			mFile.get(c);
		}
		if (token == "WIDTH") mFile >> mWidth;
		else if (token == "HEIGHT") mFile >> mHeight;
		else if (token == "DEPTH") mFile >> mChannels;
		else if (token == "MAXVAL") mFile >> mColors;
		else if (token == "TUPLTYPE") {
			mFile.get(c);
			while (mFile && !isspace(c)) {
				tupltyp += c;
				mFile.get(c);
			}
			if (c == '\n') continue;
		} else if (token == "ENDHDR") break;
		else cxError << _("CImage::ReadPAMHeader: Unknown header token {1}.", 391) << token << CError::Warn;
		do {
			mFile.get(c);
		} while (mFile && !isspace(c));	
	}
	// sanity check
	if (mWidth == 0 || mWidth > MAXWIDTH) return false;
	if (mHeight == 0 || mHeight > MAXHEIGHT) return false;
	if (mChannels == 0 || mChannels > 4) {
		cxError << _("CImage::ReadPAMHeader: Unsupported channel depth.", 388) << CError::Warn;
		return false;
	}
	if (!mColors) return false;
	if (mColors > (0xffff)) {
		cxError << _("CImage::ReadPAMHeader: More than 16 bit per channel.", 389) << CError::Warn;
		return false;
	}
	// check TUPLTYPE:
	if (mChannels == 4) {
		if (tupltyp != "RGB_ALPHA" && tupltyp != "RGBA") tupltyp = "?";
	} else if (mChannels == 3) {
		if (tupltyp != "RGB") tupltyp = "?";
	} else if (mChannels == 1) {
		if (mColors == 2) {
			if (tupltyp != "BLACKANDWHITE") tupltyp = "?";
		} else {
			if (tupltyp != "GRAYSCALE" && tupltyp != "GREYSCALE") tupltyp = "?";
		}
	} else if (mChannels == 2) {
		if (mColors == 2) {
			if (tupltyp != "BLACKANDWHITE_ALPHA") tupltyp ="?";
		} else {
			if (tupltyp != "GRAYSCALE_ALPHA" && tupltyp != "GREYSCALE_ALPHA") tupltyp = "?";
		}
	} else {
		cxError << _("CImage::ReadPAMHeader: Unsupported channel depth.", 388) << CError::Warn;
		return false;
	}
	if (tupltyp == "?") cxError << _("CImage::ReadPAMHeader: Wrong tupletype", 390) << CError::Warn;
	return true;
}

/*** ReadXPMHeader ***
return value = bytes per pixel */

unsigned CImage::ReadXPMHeader()
{
	char buf[10];
	mFile.read(buf, 10);
	if (strncmp(buf, "/* XPM */", 9) != 0) return 0;
	if (!isspace(buf[9])) return 0;
	if (!SkipXPMComment()) return 0;
	unsigned bpp = 0;
	mFile >> mWidth;
	mFile >> mHeight;
	mFile >> mColors;
	mFile >> bpp;
	mFile.ignore(1000, '\"');
	if (!mFile) return 0;
	if ((mWidth == 0) || (mHeight == 0) || (mColors == 0) || (bpp == 0)) return 0;
	if (mWidth > MAXWIDTH || mHeight > MAXHEIGHT) return 0;
	else return bpp;
}

bool CImage::ReadXPMColormap(unsigned bpp)
{
	char token[16];
	memset(token, 0, 16);
	char c;
	string value;
	CColor color;
	mColortoken.clear();
	for (unsigned i = 0; i < mColors; ++i) {
		SkipXPMComment();
		mFile.read(token, bpp);
		while (mFile) {
			do {
				mFile.get(c); 
			} while (mFile && (isspace(c)));
			if (c == 'c' ) {
				mFile.get(c);
				if (isspace(c)) break;
			} else if (c == 'g') {
				mFile.get(c);
				if (c == '4') mFile.get(c);
				if (isspace(c)) break;
			} else while (mFile && (!isspace(c)) && c != '\"') mFile.get(c);
			if (c == '\"') return false;
		}
		while (mFile && isspace(mFile.peek())) mFile.ignore();
		if (!mFile) return false;
		getline(mFile, value, '\"');
		string::size_type pos = value.find(' ');
		if (pos != string::npos) value.resize(pos);
		if (value[0] == '#') {
			unsigned char r = 0;
			unsigned char g = 0;
			unsigned char b = 0;
			for (unsigned i = 1; i < value.size(); ++i) {
				if (value[i] >= 'a' && value[i] <= 'f') value[i] += 10 - 'a';
				else if (value[i] >= 'A' && value[i] <= 'F') value[i] += 10 - 'A';
				else if (value[i] >= '0' && value[i] <= '9') value[i] -= '0';
				else if (isspace(c)) {
					value.resize(i);
					break;
				} else return false;
			}
			switch(value.size()) {
			case 4:
				r = (value[1] << 4) + value[1];
				g = (value[2] << 4) + value[2];
				b = (value[3] << 4) + value[3];
				break;
			case 7:
				r = (value[1] << 4) + value[2];
				g = (value[3] << 4) + value[4];
				b = (value[5] << 4) + value[6];
				break;
			case 13:
				r = (value[1] << 4) + value[2];
				g = (value[5] << 4) + value[6];
				b = (value[9] << 4) + value[10];
				break;
			default:
				return false;
			}
			color = CColor(r, g, b);
		} else {
			if (strncasecmp(value.c_str(), "none", 4) == 0) color = CColor(0, 0, 0, 0);
			else color = CColor(value);
		}
		mColortoken.insert(map<string, CColor>::value_type(token, color));
	}
	return mFile;
}

bool CImage::CreatePicture()
{
	cxAssert(cxMainWin != NULL);
	if (mPicture != None) return false;
	if (!IsValid()) return false;
	if (mImage.empty()) return false;
	XImage* img = XCreateImage(cxDpy, DefaultVisual(cxDpy, DefaultScreen(cxDpy)), 32, ZPixmap, 0, NULL, mWidth, mHeight, 32, 0);
	if (img == NULL) cxError << _("X-Server failed to create image", 5) << CError::Fatal;
	// malloc used, because that's the way X does
	img->data = (char*) malloc(img->bytes_per_line * mHeight);
	if (img->data == NULL) {
		XDestroyImage(img);
		cxError << _("Out of memory", 87) << CError::Fatal;
	}
	vector<CColor>::iterator i = mImage.begin();
	// copy data to image using ARGB format
	unsigned long col;
	unsigned alpha;
	for (unsigned y = 0; y < mHeight; ++y) {
		for (unsigned x = 0; x < mWidth; ++i, ++x) {
			alpha = i->GetAlpha();
			col = alpha;
			col <<= 8;
			col += (i->GetRed() * alpha) / 255;
			col <<= 8;
			col += (i->GetGreen() * alpha) / 255;
			col <<= 8;
			col += (i->GetBlue() * alpha) / 255;
			XPutPixel(img, x, y, col);
		}
	}
	Pixmap p = XCreatePixmap(cxDpy, cxMainWin->GetXID(), mWidth, mHeight, 32);
	if (p == None) {
		XDestroyImage(img);
		cxError << _("X-Server failed to create pixmap", 375) << CError::Fatal;
	}
	// copy image to pixmap	
	XPutImage(cxDpy, p, cxxx::cxGC32, img, 0, 0, 0, 0, mWidth, mHeight);
	XDestroyImage(img);
	// copy pixmap to picture
	mPicture = XRenderCreatePicture(cxDpy, p, cxARGBFormat, 0, NULL);
	XFreePixmap(cxDpy, p);
	return true;
}

void CImage::DoScale(unsigned w, unsigned h)
{
	vector<CColor> nvec;
	unsigned wstep = mWidth / (w * 4) + 1;
	unsigned hstep = mHeight / (h * 4) + 1;
	if (wstep > 1 || hstep > 1) {
		unsigned nwidth = mWidth / wstep;
		unsigned nheight = mHeight / hstep;
		// crop down image
		for (unsigned ih = 0; ih < nheight; ++ih) {
			for (unsigned iw = 0; iw < nwidth; ++iw) nvec.push_back(mImage[ih * mWidth * hstep + iw * wstep]);
		}
		mImage.clear();
		swap(nvec, mImage);
		mWidth = nwidth;
		mHeight = nheight;
	}
	unsigned a(0);
	unsigned r(0);
	unsigned g(0);
	unsigned b(0);
	if (mWidth != w) {
		unsigned in(w);
		unsigned out(mWidth);
		vector<CColor>::iterator i = mImage.begin();
		while (i != mImage.end()) {
			if (in >= out) {
				a += out * i->GetAlpha();
				r += out * i->GetRed();
				g += out * i->GetGreen();
				b += out * i->GetBlue();
				a /= mWidth;
				r /= mWidth;
				g /= mWidth;
				b /= mWidth;
				nvec.push_back(CColor(r, g, b, a));
				in -= out;
				out = mWidth;
				a = 0;
				r = 0;
				g = 0;
				b = 0;
			} else {
				a += in * i->GetAlpha();
				r += in * i->GetRed();
				g += in * i->GetGreen();
				b += in * i->GetBlue();
				out -= in;
				in = 0;
			}
			if (in == 0) {
				++i;
				in = w;
			}
		}
	} else nvec = mImage;
	mImage.clear();
	mImage.resize(w * h);
	if (mHeight != h) {
		unsigned oidx = 0;
		unsigned in =h;
		unsigned out = mHeight;
		vector<CColor>::iterator i;
		for (unsigned idx = 0; idx < w * mHeight;) {
			i = nvec.begin() + w * (idx % mHeight) + idx / mHeight;
			if (in >= out) {
				a += out * i->GetAlpha();
				r += out * i->GetRed();
				g += out * i->GetGreen();
				b += out * i->GetBlue();
				a /= mHeight;
				r /= mHeight;
				g /= mHeight;
				b /= mHeight;
				mImage[oidx] = CColor(r, g, b, a);
				oidx += w;
				if (oidx >= mImage.size()) {
					++oidx;
					oidx %= w;
				}
				a = 0;
				r = 0;
				g = 0;
				b = 0;
				in -= out;
				out = mHeight;
			} else {
				a += in * i->GetAlpha();
				r += in * i->GetRed();
				g += in * i->GetGreen();
				b += in * i->GetBlue();
				out -= in;
				in = 0;
			}
			if (in == 0) {
				++idx;
				in = h;
			}
		}
	} else mImage = nvec;
	mWidth = w;
	mHeight = h;
}

bool CImage::WriteFile(const std::string& file)
{
	if (!IsValid()) return false;
	ofstream out(file.c_str());
	if (!out) return false;
	vector<CColor>::iterator i;
	out << "P7\nWIDTH " << mWidth << "\nHEIGHT " << mHeight << "\nMAXVAL 255\n";
	if (mType == ARGB) {
		out << "DEPTH 4\nTUPLTYPE RGB_ALPHA\nENDHDR\n";
		for (i = mImage.begin(); i != mImage.end(); ++i) {
			out.put(i->GetRed());
			out.put(i->GetGreen());
			out.put(i->GetBlue());
			out.put(i->GetAlpha());
		}
	} else {
		// mType == ALPHA
		out << "DEPTH 1\nTUPLTYPE GRAYSCALE\nENDHDR\n";
		for (i = mImage.begin(); i != mImage.end(); ++i) out.put(i->GetAlpha());
	}
	return out;
}
		

