/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CImage.h
  Info   : $Id: CImage.h,v 1.41 2012/09/14 18:31:01 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CImage_h
#define CImage_h

// C++X headers
#include <cxxx/CColor.h>

// C++ headers
#include <fstream>
#include <map>
#include <vector>
#include <cctype>

class CImage
{
	enum
	{
		// Limits
		MAXWIDTH = 10000,
		MAXHEIGHT = 10000
	};

public:
	enum
    {
		// Scaling
		SCALE_NONE = 0,
		SCALE_ASPECT = 1 << 0,
		SCALE_UP = 1 << 1,
		SCALE_DOWN = 1 << 2
    };

	enum
	{
		INVALID,
		ARGB,
		ALPHA
	};

public:
	// Constructor
    CImage()
	: mWidth(0),
	  mHeight(0),
	  mColors(0),
	  mChannels(0),
	  mType(INVALID),
	  mXHot(0),
	  mYHot(0),
	  mPicture(None)
	{
	}

	// Destructor
	~CImage()
	{
		Clear();
	}

	// Access functions
	unsigned GetWidth() const
	{
		return mWidth;
	}

	unsigned GetHeight() const
	{
		return mHeight;
	}

	unsigned GetType() const
	{
		return mType;
	}

	CColor GetPixel(unsigned idx)
	{
		if (idx < mImage.size()) return mImage[idx];
		else return CColor(0, 0, 0);
	}

	Picture GetPicture()
	{
		if (!IsValid()) return None;
		if (mPicture != None) return mPicture;
		if(CreatePicture()) return mPicture;
		return None;
	}

	unsigned GetXHot()
	{
		return mXHot;
	}
	

	unsigned GetYHot()
	{
		return mYHot;
	}
	
	// Loads file
	void ReadFile(const std::string& file)
	{
		PutImage(file);
	}

	// Writes a pam image file_size
	// returns true if successfull
	bool WriteFile(const std::string& file);

	// Loads file and scales it to w, h
	void ScaleFile(const std::string& file, unsigned w, unsigned h, unsigned flag)
	{
		PutImage(file);
		Scale(w, h, flag);
	}

	// Returns true if the image is valid
	bool IsValid()
	{
		return (mType && mWidth && mHeight);
	}

	// Scales the image to w, h
	void Scale(unsigned w, unsigned h, unsigned flag)
	{
		if (IsValid()) {
			if (w != mWidth || h != mHeight) {
				if ((flag & SCALE_ASPECT) != 0) {
					if (w * mHeight > h * mWidth) w = (h * mWidth) / mHeight;
					else h = (w * mHeight) / mWidth;
				}
				if ((flag & SCALE_UP) == 0) {
					if (w > mWidth) w = mWidth;
					if (h > mHeight) h = mHeight;
				}
				if ((flag & SCALE_DOWN) == 0) {
					if (w < mWidth) w = mWidth;
					if (h < mHeight) h = mHeight;
				}
				DoScale(w, h);
			}
		}
	}

private:
	// Reads file and fills the image
	void PutImage(const std::string& file);

	// Clears the image
	void Clear()
	{
		mImage.clear();
		mWidth = 0;
		mHeight = 0;
		mColors = 0;
		mChannels = 0;
		mType = INVALID;
		mXHot = 0;
		mYHot = 0;
		mColormap.clear();
		mColortoken.clear();
		mPicture = None;
		mFile.close();
		mFile.clear();
	}

	// Reads a netbpm file
	bool ReadPNM()
	{
		char buf[3];
		mFile.read(buf, 3);
		if (!mFile) return false;
		if (!isspace(buf[2])) return false;
		switch (buf[1]) {
		case '1':
			if (!ReadPNMHeader(true)) return false;
			return PutPBM();
		case '2':
			if (!ReadPNMHeader(false)) return false;
			return PutPGM();
		case '3':
			if (!ReadPNMHeader(false)) return false;
			return PutPPM();
		case '4':
			if (!ReadPNMHeader(true)) return false;
			return PutRawPBM();
		case '5':
			if (!ReadPNMHeader(false)) return false;
			return PutRawPGM();
		case '6':
			if (!ReadPNMHeader(false)) return false;
			return PutRawPPM();
		case '7':
			if (!ReadPAMHeader()) return false;
			return PutPAM();
		default:
			return false;
		}
	}

	// Reads a (windows) bitmap file
	bool ReadBMP();

	// Reads a xpm file
	bool ReadXPM();

	// Reads a png file
	bool ReadPNG();

	// Reads a gif file
	bool ReadGIF();

	// Reads a jpeg file
 	bool ReadJPEG();
	
	// Reads a tiff file
 	bool ReadTIFF(const std::string& file);

	// Reads a Xcursor file
	bool ReadXCUR();
	
	// Reads a pbm, pgm or ppm header
	bool ReadPNMHeader(bool bw);
	
	// Reads a pam header
	unsigned ReadPAMHeader();
	
	// Reads a xpm header
	unsigned ReadXPMHeader();
	
	// Skips comments in xpm files
	bool SkipXPMComment()
	{
		char c = 0;
		unsigned comment = 0;
		while (mFile) {
			mFile.get(c);
			if (c == '\"') {
				if (comment == 0) return mFile;
				else if (comment == 1 || comment == 3) --comment;
			} else if (c == '/') {
				if (comment == 0) ++comment;
				else if (comment == 3) comment = 0;
			} else if (c == '*') {
				if (comment == 1 || comment == 2) ++ comment;
			} else {
				if (comment == 1 || comment == 3) --comment;
			}
		}
		return false;
	}
	
	// Reads the colormap of xpm files
	bool ReadXPMColormap(unsigned bpp);
	
	// Creates the picture
	bool CreatePicture();
	
	// Filling routines for different formats
	bool PutPAM();
	bool PutPBM();
	bool PutPGM();
	bool PutPPM();
	bool PutRawPBM();
	bool PutRawPGM();
	bool PutRawPPM();
	
	// Scale image
	void DoScale(unsigned w, unsigned h);
	
	// Pam routines
	unsigned char GetPAMColor1(char* value)
	{
		unsigned ret = value[0] & 0xff;
		ret *= 0xff;
		ret /= mColors;
		return ret;
	}

	unsigned char GetPAMColor2(char* value)
	{
		unsigned ret = value[0] & 0xff;
		ret <<= 8;
		ret += value[1] & 0xff;
		ret *= 0xff;
		ret /= mColors;
		return ret;
	}
	
	unsigned GetUint32LE(const char* val)
	{
		unsigned ret = *(val + 3) & 0xff;
		ret <<= 8;
		ret += *(val + 2) & 0xff;
		ret <<= 8;
		ret += *(val + 1) & 0xff;
		ret <<= 8;
		ret += *val & 0xff;
		return ret;
	}

	unsigned ReadUint32LE()
	{
		char buf[4];
		mFile.read(buf, 4);
		if (!mFile) return -1u;
		return GetUint32LE(buf);
	}

private:
	std::vector<CColor> mImage;
	std::map<std::string, CColor> mColortoken;
	std::vector<CColor> mColormap;
	unsigned mWidth;
	unsigned mHeight;
	unsigned mColors;
	unsigned mChannels;
	unsigned mType;
	unsigned mXHot;
	unsigned mYHot;
	Picture mPicture;
	std::ifstream mFile;
};

#endif // CImage_h
