/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CListItem.h
  Info   : $Id: CListItem.h,v 1.35 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CListItem_h
#define CListItem_h

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/TMagic.h>
#include <cxxx/CIcon.h>
#include <cxxx/CControlItem.h>

// C headers
#include <sys/stat.h>

struct CTextSizeHint
{
	// Constructor
	CTextSizeHint()
	: mLength(0),
	  mWidth(0)
	{}
		
	CTextSizeHint(unsigned len, unsigned width)
	: mLength(len),
	  mWidth(width)
	{}
		
	unsigned mLength;
	unsigned mWidth;
};

class CListItem : public CControlItem
{
protected:
	class CListItemInfo : public CControlItemInfo
	{
	public:
		// Constructor
		CListItemInfo(const CIcon& icon, const std::string& text, unsigned w, unsigned h)
		: CControlItemInfo(text, false),
		  mWidth(w),
		  mHeight(h),
		  mIcon(icon)
		{
		}
	
		unsigned mWidth;
		unsigned mHeight;
		CIcon mIcon;
		std::vector<CTextSizeHint> mHints;
	};

protected:
	// Constructor
	CListItem()
	{
		mInfo = NULL;
	}

public:
	// Constructor
	CListItem(const CIcon& icon, const std::string& text, unsigned w, unsigned h)
	{
		mInfo = new CListItemInfo(icon, text, w, h);
		//mInfo->mRefCount = 1;
	}
	
	// Copy constructor
	CListItem(const CListItem& it)
	: CControlItem()
	{
		mInfo = it.mInfo;
		if (mInfo != NULL) ++mInfo->mRefCount;
	}
	
	// Assignment
	CListItem& operator=(const CListItem& it)
	{
		if (mInfo != it.mInfo) {
			if (mInfo != NULL && --mInfo->mRefCount == 0) delete mInfo;
			mInfo = it.mInfo;
			if (mInfo != NULL) ++mInfo->mRefCount;
		}
		return *this;
	}

	// Comparisons
	bool operator==(const CListItem& it) const
	{
		return mInfo == it.mInfo;
	}

	bool operator!=(const CListItem& it) const
	{
		return mInfo != it.mInfo;
	}

	// Acess functions
	unsigned GetWidth() const
	{
		return GetInfo()->mWidth;
	}

	unsigned GetHeight() const
	{
		return GetInfo()->mHeight;
	}

	const CIcon& GetIcon() const
	{
		return GetInfo()->mIcon;
	}

	void SetIcon(CIcon& icon)
	{
		GetInfo()->mIcon = icon;
	}

	void SetTextHint(unsigned len, unsigned width)
	{
		if (len) GetInfo()->mHints.push_back(CTextSizeHint(len, width));
		else GetInfo()->mHints.clear();
	}

	CTextSizeHint GetTextHint(unsigned idx)
	{
		if (idx < GetInfo()->mHints.size()) return GetInfo()->mHints[idx];
		else return CTextSizeHint(0, 0);
	}

private:
	CListItemInfo* GetInfo() const
	{
		cxAssert (mInfo != NULL);
		return static_cast<CListItemInfo*>(mInfo);
	}
};


class CFileListItem : public CListItem
{
public:
	enum
	{
		LINKED = CControlItem::SELECTED << 2,
		LOCKED = LINKED << 1,
		DIRECTORY = LOCKED << 1
	};
		
protected:
	class CFileListItemInfo : public CListItemInfo
	{
	public:
		// Constructor
		CFileListItemInfo(const CIcon& icon, const std::string& text, const struct stat& st, unsigned typ, unsigned w, unsigned h)
		: CListItemInfo(icon, text, w, h),
		  mSize(st.st_size),
		  mTime(st.st_mtime),
		  mType(typ)
		{
			if (mType == TMagic::DIRECTORY) mState |= DIRECTORY;
		}
		
		off_t mSize;
		time_t mTime;
		unsigned mType;
	};

public:
	// Constructor
	CFileListItem(const CIcon& icon, const std::string& text, const struct stat& st, unsigned typ, unsigned w, unsigned h)
	{
		mInfo = new CFileListItemInfo(icon, text, st, typ, w, h);
		mInfo->mRefCount = 1;
	}

	// Assignment
	CFileListItem& operator=(const CFileListItem& it)
	{
		if (mInfo != it.mInfo) {
			if (mInfo != NULL && --mInfo->mRefCount == 0) delete mInfo;
			mInfo = it.mInfo;
			if (mInfo != NULL) ++mInfo->mRefCount;
		}
		return *this;
	}
	
	// Access functions

	off_t GetSize() const
	{
		return GetInfo()->mSize;
	}

	time_t GetTime() const
	{
		return GetInfo()->mTime;
	}

	unsigned GetType() const
	{
		return GetInfo()->mType;
	}

	void SetType(unsigned typ)
	{
		GetInfo()->mType = typ;
	}

	// Returns whether the item is a directory
	bool IsDirectory() const
	{
		return GetInfo()->mState & DIRECTORY;
	}
	
	// Returns whether the item is a link
	bool IsLink() const
	{
		return GetInfo()->mState & LINKED;
	}
	
	// Returns whether the item is locked
	bool IsLocked() const
	{
		return GetInfo()->mState & LOCKED;
	}
	
	// Marks the item as being a directory
	void IsDirectory(bool d)
	{
		if (d) GetInfo()->mState |= DIRECTORY;
		else GetInfo()->mState &= ~DIRECTORY;
	}
	
	// Marks the item as being a link
	void IsLink(bool l)
	{
		if (l) GetInfo()->mState |= LINKED;
		else GetInfo()->mState &= ~LINKED;
	}

	// Marks the item as being unreadable
	void IsLocked(bool l)
	{
		if (l) GetInfo()->mState |= LOCKED;
		else GetInfo()->mState &= ~LOCKED;
	}

private:
	CFileListItemInfo* GetInfo() const
	{
		cxAssert (mInfo != NULL);
		return static_cast<CFileListItemInfo*>(mInfo);
	}
};

#endif // CListItem_h
