/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CMagic.h
  Info   : $Id: CMagic.h,v 1.42 2012/09/12 13:14:46 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CMagic_h
#define CMagic_h

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/CIcon.h>
#include <cxxx/TMagic.h>

// C++ headers
#include <istream>
#include <cstring>
#include <algorithm>
#include <fstream>

// C headers
#include <sys/types.h>
#include <sys/stat.h>

class CMagic
{
public:
	struct CMagInfo
	{
		CIcon::TIconName mIcon;
		const char* mMime;
		int mDescription;
	};

	struct CGlob
	{
		const char* mPattern;
		unsigned mType;
	};

public:
	// Constructor
	CMagic()
	: mStream(NULL),
	  mStreamSize(0),
	  mCompressed(false)
	{
#ifdef CXXX_DEBUG
		for (unsigned i = 0; i < ENDINGSIZE - 1; ++i) {
			if (!Less(mEnding[i], mEnding[i + 1])) {
				cxxx::cxError << "CMagic::mEnding: {1} >= {2}!" << mEnding[i].mPattern << mEnding[i + 1].mPattern << CError::Error;
			}
		}
		for (unsigned i = 0; i < NAMESIZE - 1; ++i) {
			if (!Less(mName[i], mName[i + 1])) {
				cxxx::cxError << "CMagic::mName: {1} >= {2}!" << mName[i].mPattern << mName[i + 1].mPattern << CError::Error;
			}
		}
		for (unsigned i = 0; i < MAGIC0SIZE - 1; ++i) {
			if (!LessMag(mMagic0[i], mMagic0[i + 1])) {
				cxxx::cxError << "CMagic::mMagic0: {1} >= {2}!" << mMagic0[i].mPattern + 1 << mMagic0[i + 1].mPattern + 1 << CError::Error;
			}
			const char* tmp = mMagic0[i].mPattern;
			switch (tmp[tmp[0] + 1]) {
			case '\x00':
			case '\xff':
			case '\xfe':
			case '\xfd':
			case '\xfc':
			case '\xfb':
			case '\xfa':
			case '\xf9':
			case '\xf8':
			case '\xf7':
			case '\xf6':
			case '\xf5':
			case '\xf4':
			case '\xf3':
				break;
			default:
				cxxx::cxError << "CMagic::mMagic0: synthax error in {1}" << mMagic0[i].mPattern + 1 << CError::Error;
				break;
			}
		}
#endif //CXXX_DEBUG
	}

	// Returns the file type of file
	unsigned GetFileType(const std::string& file);

	// Returns the file type of the open filebuf buf
	unsigned GetFileType(std::filebuf& fbuf, const struct stat& st, const std::string& fn);
		
	// Returns the description of type
	static const char* GetDescription(unsigned type)
	{
		type = type & TMagic::UNCOMPRESSED;
		return cxxx::lib::cxgettext(mMagInfo[type].mDescription);
	}
	
	// Returns the MIME type of type
    static const char* GetMimeType(unsigned type)
	{
		type = type & TMagic::UNCOMPRESSED;
		return mMagInfo[type].mMime;
	}
	
	// Returns the icon file name for type
	static CIcon::TIconName GetIcon(unsigned type)
	{	
		type = type & TMagic::UNCOMPRESSED;
		return mMagInfo[type].mIcon;
	}
	
private:
	enum
	{
		// Sizes of the tables 
		ENDINGSIZE = 121,
		NAMESIZE = 58,
		MAGIC0SIZE = 243,
		KOFFICESIZE = 10,
		OPENOFFICESIZE = 5,
		OPENDOCSIZE = 12
	};

	// Comparisons
	static bool Less(const CGlob& g1, const CGlob& g2)
	{
		return strcmp(g1.mPattern, g2.mPattern) < 0;
	}

	static bool LessMag(const CGlob& g1, const CGlob& g2)
	{
		return memcmp(g1.mPattern + 1, g2.mPattern + 1, std::min(g1.mPattern[0], g2.mPattern[0])) < 0;
	}

	// Tries to detect file type by magic
	unsigned GetTypeByMagic();

	// Tries to detect file type by file name
	unsigned GetTypeByGlobbing();

	// Checks whether the file is a special type
	unsigned GetSpecialType(mode_t m)
	{
		if (S_ISREG(m))	return TMagic::UNKNOWN;
		if (S_ISDIR(m)) return TMagic::DIRECTORY;
		if (S_ISCHR(m)) return TMagic::CHARDEV;
		if (S_ISBLK(m)) return TMagic::BLOCKDEV;
		if (S_ISFIFO(m)) return TMagic::PIPE;
		if (S_ISSOCK(m)) return TMagic::SOCKET;
		return TMagic::UNKNOWN;
	}

	// Detects whether file is ASCII, UTF-8, iso-8859 or other 
	unsigned GetTextType();

	// Tries to reset modification and access times of the file
	void ResetFileTimes();

	// Returns the suffix of the file
	std::string GetSuffix()
	{
		std::string ret;
		std::string::size_type pos = mPath.rfind('.');
		if (pos != 0 && pos != std::string::npos && mPath[pos - 1] != '/') {
			while (++pos < mPath.size()) ret += ((char)tolower(mPath[pos]));
		}
		return ret;
	}
	
private:
	std::string mPath;
	std::istream mStream;
	std::filebuf mFilebuffer;
	std::ios::pos_type mOffset;
	unsigned mStreamSize;
	struct stat mStatus;
	bool mCompressed;
	static const CGlob mEnding[ENDINGSIZE];
	static const CGlob mName[NAMESIZE];
	static const CGlob mMagic0[MAGIC0SIZE];
	static const CGlob mOpenOffice[OPENOFFICESIZE];
	static const CGlob mKOffice[KOFFICESIZE];
	static const CGlob mOpenDoc[OPENDOCSIZE];
	static const CMagInfo mMagInfo[TMagic::MAGINFOCOUNT];
};

inline bool operator==(const CMagic::CGlob& g, const char* s)
{
	const char* t = g.mPattern;
	while (*t) if (*s++ != *t++) return false;
	return true;
}

#endif // CMagic_h
