/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CMenuItem.h
  Info   : $Id: CMenuItem.h,v 1.19 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CMenuItem_h
#define CMenuItem_h

// C++X headers
#include <cxxx/CxxxFeatures.h>
#include <cxxx/CIcon.h>
#include <cxxx/CWindow.h>

// C++ headers
#include <vector>
#include <bitset>
#include <cstddef>

// Forward declarations
class CWindow;
class CIcon;
class CPopup;
class CMenu;

class CMenuItem
{
private:
	enum
	{
		// Types
		CHECK,
		BULLET,
		ICON,
		SEPERATOR,
		ISCHECKED,
		FLAG_COUNT,
	};

	// Constructor is private, CMenu has access
	CMenuItem(CWindow* owner, unsigned msgid, const std::string& text, CIcon::TIconName icon, CMenu* child);
	
public:

    // Returns true if the item has a submenu
	bool HasSubmenu() const
	{
		return mChild != NULL;
	}

	// Returns true if the item has an icon
	bool HasIcon() const
	{
		return mFlags.test(ICON);
	}

	// Returns true if the item is a seperator
	bool IsSeperator() const
	{
		return mFlags.test(SEPERATOR);
	}

	// Returns true if this is a radio type item
	bool IsRadioType() const
	{
		return mFlags.test(BULLET);
	}

	// Returns true if this is a check type item
	bool IsCheckType() const
	{
		return mFlags.test(CHECK);
	}

	// Returns true if the item is checked
	bool IsChecked() const
	{
		if (!IsCheckType() && !IsRadioType()) return false;
		else return mFlags.test(ISCHECKED);
	}

	// Returns true if the item is enabled
	bool IsEnabled() const;

	// Access functions
	const std::string& GetText() const
	{
		return mText;
	}

	const std::string& GetScutText() const
	{
		return mScutText;
	}

	unsigned GetTextWidth() const
	{
		return mTextWidth;
	}

	unsigned GetScutTextWidth() const
	{
		return mScutTextWidth;
	}

	const CHotCharPos& GetHotCharPos() const
	{
		return mHotCharPos;
	}

	wchar_t GetAccelerator() const
	{
		return mAccelerator;
	}

	unsigned GetMessageID() const
	{
		return mMsgID;
	}

	CMenu& GetSubmenu()
	{
		cxAssert(HasSubmenu());
		return *mChild;
	}

	const CIcon& GetIcon() const
	{
		return mIcon;
	}

	void Check()
	{
		cxAssert(IsCheckType() || IsRadioType());
		Check(true);
	}

	void Uncheck()
	{
		cxAssert(IsCheckType() || IsRadioType());
		Check(false);
	}

private:
	void Check(bool check)
	{
		mFlags[ISCHECKED] = check;
	}
		
private:
	CWindow* mOwner;
	CMenu* mChild;
	CIcon mIcon;
	std::bitset<FLAG_COUNT> mFlags;
	unsigned mTextWidth;
	unsigned mScutTextWidth;
	std::string mText;
	std::string mScutText;
	wchar_t mAccelerator;
	CHotCharPos mHotCharPos;
	unsigned mMsgID;
	friend class CMenu;
};

class CMenu
{
public:
	// Constructor
	CMenu (CWindow* owner, const std::string& name);
	
	// Access operator
	CMenuItem& operator[](unsigned idx)
	{
		cxAssert(idx < GetSize());
		return mMenu[idx];
	}

	// Creates a menu entry
	void Entry(unsigned msgid, const std::string& text, CIcon::TIconName icon = CIcon::MENU_NO_ICON)
	{
		mMenu.push_back(CMenuItem(mOwner, msgid, text, icon, NULL));
	}

	void Entry(const std::string& text, CIcon::TIconName icon = CIcon::MENU_NO_ICON)
	{
		mMenu.push_back(CMenuItem(mOwner, 0, text, icon, NULL));
	}

	void Entry(CMenu& menu, const std::string& text, CIcon::TIconName icon = CIcon::MENU_NO_ICON)
	{
		mMenu.push_back(CMenuItem(mOwner, 0, text, icon, &menu));
	}

	// Creates a seperator
	void Seperator()
	{
		mMenu.push_back(CMenuItem(mOwner, 0, "", CIcon::MENU_NO_ICON, NULL));
	}

	// Returns the entry count
	unsigned GetSize() const
	{
		return mMenu.size();
	}

	// Returns the accelerator for the menu
	wchar_t GetAccelerator() const
	{
		return mAccelerator;
	}

	// Returns the name of the menu
	const std::string& GetText() const
	{
		return mText;
	}
	
	// Returns the text width of the name
	unsigned GetTextWidth() const
	{
		return mTextWidth;
	}
	
	// Returns the hot char position
	const CHotCharPos& GetHotCharPos() const
	{
		return mHotCharPos;
	}
	
	// Returns true if the menu is enabled
	bool IsEnabled() const
	{
		for (unsigned i = 0; i < GetSize(); ++i) if (mMenu[i].IsEnabled()) return true;
		return false;
	}
	
	// Returns true if the menu is empty
	bool IsEmpty() const
	{
		return mMenu.empty();
	}
	
	// Checks item with message ID msgid
	void Check(unsigned msgid)
	{
		for (std::vector<CMenuItem>::iterator i = mMenu.begin(); i < mMenu.end(); ++i) {
			if (i->mMsgID == msgid) i->Check();
		}
	}
	
	// Unchecks item with message ID msgid	
	void Uncheck(unsigned msgid)
	{
		for (std::vector<CMenuItem>::iterator i = mMenu.begin(); i < mMenu.end(); ++i) {
			if (i->mMsgID == msgid) i->Uncheck();
		}
	}
	
	// Removes all entries
	void Clear()
	{
		mMenu.clear();
	}
	
	// Returns the owner of the menu
	CWindow* GetOwner() const
	{
		return mOwner;
	}
	
protected:
	std::vector<CMenuItem> mMenu;
	CWindow* mOwner;
	wchar_t mAccelerator;
	std::string mText;
	unsigned mTextWidth;
	CHotCharPos mHotCharPos;
};

#endif // CMenuItem_h
