/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CPopup.cpp
  Info   : $Id: CPopup.cpp,v 1.26 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CPopup.h>
#include <cxxx/CMenuItem.h>
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/CTimer.h>
#include <cxxx/CApplication.h>

// X11 headers
#include <X11/cursorfont.h>
#include <X11/Xutil.h>

// Namespace declarations
using cxxx::cxApp;
using cxxx::cxDpy;
using std::string;
using std::vector;

CPopup::CPopup(CWindow* w, CMenu& m)
: CWindow(NULL, false),
  mMenu(m),
  mSelected(-1u),
  mTextWidth(0),
  mTextYPos(0),
  mPosHint(DOWN),
  mCursor(None),
  mTimer(this, IDTIMER),
  mChild(NULL)
{
	SetMessageTarget(mMenu.GetOwner());
	mWindowState.set(ISPOPUP);
	Reparent(w);
	if (!IsChildOf(cxApp->GetPopupWindow())) cxApp->SetPopupWindow(NULL);
	if (cxxx::cxRTL) mArrow.Create(CIcon::ARROW_LEFT, 0);
	else mArrow.Create(CIcon::ARROW_RIGHT, 0);
	mCheck.Create(CIcon::CHECK, 0);
	mBullet.Create(CIcon::BULLET4, 0);
	mAttributes.event_mask |= PointerMotionMask | ButtonPressMask | ButtonReleaseMask | LeaveWindowMask | EnterWindowMask;
	mAttributes.override_redirect = true;
	mAttributes.save_under = true;
	ChangeAttributes(CWOverrideRedirect | CWSaveUnder | CWEventMask);
	cxAssert(!mMenu.IsEmpty());
	mTextYPos = (20 - GetTextHeight()) / 2;
	int height = 2 * BORDER_HEIGHT;
	mCursor = XCreateFontCursor(cxDpy, XC_left_ptr);
	XDefineCursor(cxDpy, GetXID(), mCursor);
	mTextWidth = 0;
	unsigned sw = 0;
	for (unsigned i = 0; i < mMenu.GetSize(); ++i) {
		CMenuItem& it = mMenu[i];
		if (it.IsSeperator()) height += 4;
		else {
			if (it.GetAccelerator()) GetTopLevelWindow()->BindKey(it.GetAccelerator(), this);
			if (it.GetTextWidth() > mTextWidth) mTextWidth = it.GetTextWidth();
			if (it.GetScutTextWidth() > sw) sw = it.GetScutTextWidth();
			height += 20;
		}
	}
	int width = mTextWidth + 2 * BORDER_WIDTH;
	if (sw) width += sw + BORDER_WIDTH;
	Resize(width, height);
	cxApp->SetPopupWindow(this);
}

/* virtual */ CPopup::~CPopup()
{
	if (mChild)	delete mChild;
	XFreeCursor(cxDpy, mCursor);
	GetTopLevelWindow()->ReleaseKey(this);
}

/* virtual */ void CPopup::OnExpose(XExposeEvent& e __attribute__ ((unused)))
{
	DrawFrame(0, 0, mWidth, mHeight, FS_RAISED_THICK, true);
	for (unsigned i = 0; i < mMenu.GetSize(); ++i) DrawMenu(i, (i == mSelected));
}

/* virtual */ void CPopup::OnLButtonDown(XButtonEvent& e __attribute__ ((unused)))
{
	if (mSelected == -1u) {
		GetTopLevelWindow()->SetFocusWindow(NULL);
		return;
	}
	CMenuItem& it = mMenu[mSelected];
	if (!it.IsEnabled()) {
		GetTopLevelWindow()->SetFocusWindow(NULL);
		return;
	}
	if (it.IsRadioType()) {
		if (it.IsChecked()) return ;
		unsigned prev = mSelected;
		while (prev && mMenu[prev - 1].IsRadioType()) --prev;
		while (mMenu[prev].IsRadioType()) {
			if (mMenu[prev].IsChecked()) {
				mMenu[prev].Uncheck();
				break;
			} else ++prev;
		}
		it.Check();
		SetMessageID(it.GetMessageID());
		PostMessage(MT_CHECK);
	} else if (it.IsCheckType()) {
		if (it.IsChecked()) {
			it.Uncheck();
			SetMessageID(it.GetMessageID());
			PostMessage(MT_UNCHECK);
		} else {
			it.Check();
			SetMessageID(it.GetMessageID());
			PostMessage(MT_CHECK);
		}
	}
}

/* virtual */ void CPopup::OnButtonUp(XButtonEvent& e)
{
	CTopLevelWindow* w = GetTopLevelWindow();
	if ((e.button == 1) && mSelected != -1u) {
		CMenuItem& it = mMenu[mSelected];
		if (it.HasSubmenu()) return;
		if (!it.IsCheckType() && !it.IsRadioType()) {
			SetMessageID(it.GetMessageID());
			PostMessage(MT_CONFIRM_SELECTION);
		}
	}
	w->SetFocusWindow(NULL);
}

/* virtual */ void CPopup::OnEvent(XEvent& e)
{
	switch (e.xany.type) {
	case MotionNotify: {
			unsigned i = GetItemFromPos(e.xmotion.y);
			if (i == mSelected) return ;
			else Select(i);
		}
		break;
	case LeaveNotify:
		if (mSelected != -1u && !mMenu[mSelected].HasSubmenu()) Select(-1u);
		break;
	default:
		break;
	}
}

/* virtual */ void CPopup::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
	int x, y;
	switch (msgid) {
	case IDTIMER:
		switch (type) {
		case MT_TIMER:
			if (mSelected == -1u)	return ;
			if (!mMenu[mSelected].IsEnabled()) return ;
			if (!mMenu[mSelected].HasSubmenu()) return ;
			x = 4;
			y = CalculatePosition(mSelected);
			TranslateCoordinates(x, y, NULL);
			cxAssert(mChild == NULL);
			mChild = new CPopup(this, mMenu[mSelected].GetSubmenu());
			mChild->Position(x, y, GetWidth() - 8, 20, mPosHint);
			mChild->Show();
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

void CPopup::Position(int x, int y, int w, int h, unsigned wp)
{
	switch (wp) {
	case DOWN:
		if ((int)(x + mWidth) < DisplayWidth(cxDpy, DefaultScreen(cxDpy))) {
			mPosHint = RIGHT;
		} else {
			x += w - mWidth;
			cxAssert(x >= 0);
			mPosHint = LEFT;
		}
		// set y-position
		if ((int)(y + h + mHeight) < DisplayHeight(cxDpy, DefaultScreen(cxDpy))) {
			y += h;
		} else {
			y -= mHeight;
			cxAssert(y >= 0);
		}
		Move(x, y);
		return ;
	case RIGHT:
		if ((int)(x + w + mWidth) < DisplayWidth(cxDpy, DefaultScreen(cxDpy))) {
			x += w;
			mPosHint = RIGHT;
		} else {
			x -= mWidth;
			cxAssert(x >= 0);
			mPosHint = LEFT;
		}
		if ((int)(y + mHeight) >= DisplayHeight(cxDpy, DefaultScreen(cxDpy))) {
			y += h - mHeight;
			cxAssert(y >= 0);
		}
		Move(x, y);
		return ;
	case LEFT:
		if (x >= (int)mWidth) {
			x -= mWidth;
			mPosHint = LEFT;
		} else {
			x += w;
			cxAssert((int)(x + mWidth) < DisplayWidth(cxDpy, DefaultScreen(cxDpy)));
			mPosHint = RIGHT;
		}
		if ((int)(y + mHeight) >= DisplayHeight(cxDpy, DefaultScreen(cxDpy))) {
			y += h - mHeight;
			cxAssert(y >= 0);
		}
		Move(x, y);
		return ;
	default:
		break;
	}
	cxAssert(0);
}

void CPopup::DrawMenu(unsigned idx, bool selected)
{
	cxAssert(idx < mMenu.GetSize());
	CMenuItem& it = mMenu[idx];
	int y = CalculatePosition(idx);
	if (it.IsSeperator()) {
		y ++;
		DrawLine(1, y, mWidth - 3, y, CColor(TC_DARK));
		y ++;
		DrawLine(2, y, mWidth - 2, y, CColor(TC_BRIGHT));
		return ;
	}
	bool en = it.IsEnabled();
	if (en && selected) {
		CColor col(TC_TITLE_TEXT);
		if (it.HasIcon()) {
			ClearArea(4, y + 2, it.GetIcon().GetWidth(), it.GetIcon().GetHeight());
			DrawFrame(2, y, 20, 20, FS_RAISED, true);
			DrawIcon(4, y + 2, it.GetIcon());
			FillRectangle(22, y, mWidth - 24, 20, CColor(TC_TITLE));
		} else {
			FillRectangle(2, y, mWidth - 4, 20, CColor(TC_TITLE));
			if (it.IsChecked()) {
				if (it.IsRadioType()) DrawIcon(12, y + 8, mBullet, col);
				else if (it.IsCheckType()) DrawIcon(10, y + 6, mCheck, col);
			}
		}
		DrawHotString(BORDER_WIDTH, y + mTextYPos, col, it.GetText(), it.GetHotCharPos(), false, false);
		if (!it.GetScutText().empty()) DrawString(mTextWidth + 2 * BORDER_WIDTH, y + mTextYPos, col, it.GetScutText(), false);
		if (it.HasSubmenu()) DrawIcon(mWidth - ARROW_WIDTH, y + (20 - mArrow.GetHeight()) / 2, mArrow, col);
	} else {
		CColor col(TC_FRAME_TEXT);
		ClearArea(2, y, mWidth - 4, 20);
		if (it.HasIcon()) DrawIcon(4, y + 2, it.GetIcon(), !en);
		else if (it.IsChecked()) {
			if (it.IsRadioType()) DrawIcon(12, y + 8, mBullet, col, !en);
			if (it.IsCheckType()) DrawIcon(10, y + 6, mCheck, col, !en);
		}
		DrawHotString(BORDER_WIDTH, y + mTextYPos, col, it.GetText(), it.GetHotCharPos(), !en);
		CHotCharPos hc;
		if (!it.GetScutText().empty()) DrawHotString(mTextWidth + 2 * BORDER_WIDTH, y + mTextYPos, col, it.GetScutText(), hc, !en);
		if (it.HasSubmenu()) DrawIcon(mWidth - ARROW_WIDTH, y + (20 - mArrow.GetHeight()) / 2, mArrow, col, !en);
	}
}

/* virtual */ void CPopup::OnAccelerator(wchar_t key)
{
	for (unsigned i = 0; i < mMenu.GetSize(); ++i) {
		CMenuItem& it = mMenu[i];
		if (it.GetAccelerator() == key) {
			if (it.HasSubmenu()) Select(i);
			else {
				if (it.IsRadioType()) {
					it.Check();
					SetMessageID(it.GetMessageID());
					PostMessage(MT_CHECK);
				} else if (it.IsCheckType()) {
					if (it.IsChecked()) {
						it.Uncheck();
						SetMessageID(it.GetMessageID());
						PostMessage(MT_UNCHECK);
					} else {
						it.Check();
						SetMessageID(it.GetMessageID());
						PostMessage(MT_CHECK);
					}
				} else {
					SetMessageID(it.GetMessageID());
					PostMessage(MT_CONFIRM_SELECTION);
				}
				GetTopLevelWindow()->SetFocusWindow(NULL);
			}
			break ;
		}
	}
}
