/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CScrollBar.h
  Info   : $Id: CScrollBar.h,v 1.13 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CScrollBar_h
#define CScrollBar_h

// C++X headers
#include <cxxx/CWindow.h>

class CScrollBar : public CWindow
{
	enum
	{
		// Message IDs
		IDTIMER = cxxx::msgid::MSGID_CScrollBar
	};

public:
	// Constructor
	CScrollBar(CWindow* parent, unsigned msgid);
	
	// X event processing
	virtual void OnExpose(XExposeEvent& e) = 0;
	virtual void OnLButtonDown(XButtonEvent& e) = 0;
	virtual void OnButtonUp(XButtonEvent& e) = 0;

	virtual void OnMouseWheelUp(XButtonEvent& e __attribute__ ((unused)))
	{
		PosDec();
	}

	virtual void OnMouseWheelDown(XButtonEvent& e __attribute__ ((unused)))
	{
		PosInc();
	}

	virtual void OnEvent(XEvent& e) = 0;

	// C++X message processing
	virtual void OnMessage(unsigned msgid, TMessage type, unsigned value);

	// Sets page size to page
	void SetPage(int page)
	{
		mPage = page;
	}

	// Sets range to range, the lower bound is always 0
	void SetRange(int range)
	{
		mRange = range;
	}

	// Sets position of track bar to pos
	void SetPosition(int pos)
	{
		// perform some rangecheck
		if (pos + mPage > mRange) pos = mRange - mPage;
		if (pos < 0) pos = 0;
		if (mPosition == pos) return;
		mPosition = pos;
		if (IsEnabled()) PostMessage(MT_SCROLL, mPosition);
		DrawTrackBar();
	}

	// Returns the position of the track bar
	int GetPosition() const
	{
		return mPosition;
	}

	// Enables or disables the scroll bar
	void Enable(bool enable)
	{
		if (IsEnabled() == enable) return ;
		CWindow::Enable(enable);
		if (enable) Redraw();
		else mPosition = 0;
	}

protected:
	// Calculates the position of the track bar from screen position x_or_y
	virtual int CalculatePosition(int x_or_y) = 0;

	// Draws the track bar and the background
	virtual void DrawTrackBar() = 0;

	// Draws a scroll bar button
	void DrawButton(int x, int y, int wh, bool pressed, const CIcon& icon);

	// Decrements position by one
	void PosDec()
	{
		if (mPosition > 0) SetPosition(mPosition - 1);
	}

	// Increments position by one
	void PosInc()
	{
	    if (mPosition < mRange) SetPosition(mPosition + 1);
	}

	// Decrements position by one page
	void PageDec()
	{
		int pos = mPosition - mPage;
		if (pos < 0) pos = 0;
   		if (pos != mPosition) SetPosition(pos);
	}

	// Increments position by one page
	void PageInc()
	{
		int pos = mPosition + mPage;
		if (pos > mRange) pos = mRange;
		if (pos != mPosition) SetPosition(pos);
	}

protected:
	int mPage;
	int mRange;
	int mPosition;
	int mTBStart;
	int mTBEnd;
	CIcon mArrow1;
	CIcon mArrow2;
	CTimer mTimer;
	bool mState1;
	bool mState2;
	bool mTracking;
	bool mEnabled;
	int mOffset;
};

class CHScrollBar : public CScrollBar
{
public:
	// Constructor
	CHScrollBar(CWindow* parent, unsigned msgid)
	: CScrollBar(parent, msgid)
	{
		if (cxxx::cxRTL) {
			mArrow2.Create(CIcon::ARROW_LEFT, 0);
			mArrow1.Create(CIcon::ARROW_RIGHT, 0);
		} else {			
			mArrow1.Create(CIcon::ARROW_LEFT, 0);
			mArrow2.Create(CIcon::ARROW_RIGHT, 0);
		}
		Resize(GetParent()->GetWidth(), 15);
	}

	// X event processing
	virtual void OnExpose(XExposeEvent& e);
	virtual void OnLButtonDown(XButtonEvent& e);
	virtual void OnButtonUp(XButtonEvent& e);
	virtual void OnEvent(XEvent& e);

protected:
	// Calculates the position of the track bar from screen position x
	virtual int CalculatePosition(int x)
	{
		int width = mWidth - 2 * mHeight + mTBStart - mTBEnd;
		if (!width) return 0;
		int pos = (x * (mRange - mPage) + width / 2) / width;
		if (pos < 0) pos = 0;
		if (pos > mRange - mPage) pos = mRange - mPage;
		return pos;
	}
 
	// Draws the track bar and the background
	 virtual void DrawTrackBar();
};

class CVScrollBar : public CScrollBar
{
public:
	// Constructor
	CVScrollBar(CWindow* parent, unsigned msgid)
	: CScrollBar(parent, msgid)
	{
		mArrow1.Create(CIcon::ARROW_UP, 0);
		mArrow2.Create(CIcon::ARROW_DOWN, 0);
		Resize(15, GetParent()->GetHeight());
	}

	// X event processing
	virtual void OnExpose(XExposeEvent& e);
	virtual void OnLButtonDown(XButtonEvent& e);
	virtual void OnButtonUp(XButtonEvent& e);
	virtual void OnEvent(XEvent& e);

protected:
	// Calculates the position of the track bar from screen position y
	virtual int CalculatePosition(int y)
	{
		int height = mHeight - 2 * mWidth + mTBStart - mTBEnd;
		if (!height) return 0;
		int pos = (y * (mRange - mPage) + height / 2) / height;
		if (pos < 0) pos = 0;
		if (pos > mRange - mPage) pos = mRange - mPage;
		return pos;
	}
 
	// Draws the track bar and the background
	virtual void DrawTrackBar();
};

#endif // CScrollBar_h
