/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CSelection.cpp
  Info   : $Id: CSelection.cpp,v 1.18 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// cxxx headers
#include <cxxx/CWindow.h>
#include <cxxx/CApplication.h>

// C++ headers
#include <cerrno>

// C headers
#include <unistd.h>

// X11 headers
#include <X11/Xatom.h>

// namespace declarations
using std::vector;
using std::string;
using cxxx::cxDpy;
using cxxx::cxApp;
using cxxx::cxServerTime;

// static members
Atom CSelection::TARGETS = None;
Atom CSelection::INCR = None;
Atom CSelection::MULTIPLE = None;
Atom CSelection::TIMESTAMP = None;
Atom CSelection::CLIPBOARD = None;


CSelection::CSelection()
: CMessageTransceiver(cxApp, CApplication::IDSEL),
  mTimer(this, IDTIMER),
  mTimeout(false),
  mLocked(false)
{
	mSelectionNotify.selection = None;
}

bool CSelection::Init(const string& cxxxdir)
{
	const char* names[] = {"TARGETS", "INCR", "MULTIPLE", "TIMESTAMP", "CLIPBOARD"};
	Atom atoms[5];
	if (XInternAtoms(cxDpy, const_cast<char**>(names), 5, False, atoms) == 0) return false;
	TARGETS = atoms[0];
	INCR = atoms[1];
	MULTIPLE = atoms[2];
	TIMESTAMP = atoms[3];
	CLIPBOARD = atoms[4];
	mClipSelDir = cxxxdir + "/clipboard";
	mDnDSelDir = cxxxdir + "/dnd";
	if (access(mClipSelDir.c_str(), R_OK | W_OK | X_OK) != 0) {
		if (errno != ENOENT || mkdir(mClipSelDir.c_str(), S_IRWXU) != 0) return false;
	}
	if (access(mDnDSelDir.c_str(), R_OK | W_OK | X_OK) != 0) {
		if (errno != ENOENT || mkdir(mDnDSelDir.c_str(), S_IRWXU) != 0) return false;
	}
	return true;
}


/* virtual */ void CSelection::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
	if (msgid == IDTIMER && type == MT_TIMER) mTimeout = true;
}

bool CSelection::Grab(CWindow* owner, Atom sel, Time time)
{
	if (time == CurrentTime) time = cxServerTime;
	vector<CSelectionItem> tmp;
	for (vector<CSelectionItem>::iterator i = mSelectionList.begin(); i != mSelectionList.end(); ++i) {
		if (i->mName != sel) tmp.push_back(*i);
	}
	swap(mSelectionList, tmp);
	if (XGetSelectionOwner(cxDpy, sel) != owner->GetXID())
	{
		XSetSelectionOwner(cxDpy, sel, owner->GetXID(), time);
		if (XGetSelectionOwner(cxDpy, sel) != owner->GetXID()) return false;
	}
	CSelectionItem it;
	it.mName = sel;
	it.mOwner = owner;
	it.mStartTime = time;
	it.mEndTime = time + 24 * 3600 * 1000;
	mSelectionList.push_back(it);
	return true;
}

void CSelection::HandleRequest(XSelectionRequestEvent& e)
{
	cxAssert(e.time != CurrentTime);
	if (e.requestor == None) return;
	CWindow* owner = GetOwner(e.selection, e.time);
	if (e.selection == None || e.target == None || owner == NULL) {
		CancelRequest(e);
		return;
	}
	if (e.target == TIMESTAMP) {
		vector<CSelectionItem>::iterator i = mSelectionList.begin();
		while (i->mName != e.selection) ++i;
		XChangeProperty(cxDpy, e.requestor, e.property, TIMESTAMP, 32, PropModeReplace, (unsigned char*)&(i->mStartTime), 1);
		return;
	}
	if (e.target == MULTIPLE) {
		Atom type;
		int fmt;
		unsigned long count;
		unsigned long rem;
		union {unsigned char* in; Atom* data;} buf;
		XGetWindowProperty(cxDpy, e.requestor, e.property, 0, -1u, True, XA_ATOM, &type, &fmt, &count, &rem, &buf.in);
		// is there really a target named ATOM_PAIRS as stated in icccm???
		// i guess they are pairs of type atom
		if(type != XA_ATOM || fmt != 32 || count == 0 || (count & 1) != 0) {
			CancelRequest(e);
			return;
		}
		Atom property = e.property;
		for (unsigned i = 0; i < count; ++i) {
			e.target = buf.data[i++];
			e.property = buf.data[i];
			if (!OnRequest(owner, e)) buf.data[i] = None;
		}
		e.property = property;
		XChangeProperty(cxDpy, e.requestor, e.property, MULTIPLE, 32, PropModeReplace, buf.in, count);
		XFree(buf.data);
		ConfirmRequest(e);
	        return;
	}
	// not MULTIPLE, not TIMESTAMP
	if (OnRequest(owner, e)) ConfirmRequest(e);
	else CancelRequest(e);
}

bool CSelection::OnRequest(CWindow* owner, XSelectionRequestEvent& e)
{
	if (e.property == None) e.property = e.target;
	vector<unsigned char> data;
	if (!owner->ConvertSelection(e.selection, e.target, data)) return false;
	if (mLocked) return false;
	else mLocked = true;
	// maximum request size in bytes
	unsigned req = 4 * XMaxRequestSize(cxDpy) - 128;
	unsigned sz = data.size();
	if (sz > req) {
		// use INCR protocol
		mTimer.SetTime(2000);
		mTimeout = false;
		XSelectInput(cxDpy, e.requestor, PropertyChangeMask);
		// Set INCR property
		mPropertyNotify.window = None;
		XChangeProperty(cxDpy, e.requestor, e.property, INCR, 32, PropModeReplace, (unsigned char*)&sz, 1);
		Atom tmp = e.target;
		e.target = INCR;
		ConfirmRequest(e);
		e.target = tmp;
		// wait for reply
		while (!mTimeout && (mPropertyNotify.window != e.requestor || mPropertyNotify.atom != e.property || mPropertyNotify.state != PropertyDelete))
			cxApp->MainLoop();
		if (mTimeout) {
			XSelectInput(cxDpy, e.requestor, NoEventMask);
			mLocked = false;
			return false;
		}
		unsigned beg = 0;
		// append data chunk by chunk
		while (beg < data.size()) {
			mTimer.SetTime(2000);
			mTimeout = false;
			mPropertyNotify.window = None;
			XChangeProperty(cxDpy, e.requestor, e.property, e.target, 8, PropModeAppend, &data[beg], req);
			// wait for reply
			while (!mTimeout && (mPropertyNotify.window != e.requestor || mPropertyNotify.atom != e.property || mPropertyNotify.state != PropertyDelete)) cxApp->MainLoop();
			if (mTimeout) {
				XSelectInput(cxDpy, e.requestor, NoEventMask);
				mLocked = false;
				return false;
			}
			beg += req;
			if ((beg + req) > data.size()) req = data.size() - beg;
		}
		// append zero length data
		XSelectInput(cxDpy, e.requestor, NoEventMask);
		XChangeProperty(cxDpy, e.requestor, e.property, e.target, 8, PropModeAppend, &data[0], 0);
		mLocked = false;
		return true;
	}
	// no INCR protocol: a whole lot easier
	XChangeProperty(cxDpy, e.requestor, e.property, e.target, 8, PropModeReplace, &data[0], data.size());
	mLocked = false;
	return true;
}

bool CSelection::Request(CWindow* req, Atom sel, Atom typ, Atom prop, vector<unsigned char>& ret)
{
	ret.clear();
	CWindow* owner = GetOwner(sel, CurrentTime);
	if (owner == NULL) {
		// invoke the server
		Atom type;
		int fmt;
		unsigned long count;
		unsigned long rem;
		unsigned char* data;
		if (mLocked) return false;
		else mLocked = true;
		mTimer.SetTime(2000);
		mTimeout = false;
		XDeleteProperty(cxDpy, req->GetXID(), prop);
		mSelectionNotify.requestor = None;
		XConvertSelection(cxDpy, sel, typ, prop, req->GetXID(), cxServerTime);
		while (!mTimeout && (mSelectionNotify.selection != sel || mSelectionNotify.requestor != req->GetXID())) cxApp->MainLoop();
		if (mTimeout) {
			mLocked = false;
			return false;
		}
		// transfer canceled
		if (mSelectionNotify.property == None) {
			mLocked = false;
			return false;
		}
		if (mSelectionNotify.target == INCR) {
			// handle INCR protocol
			mTimer.SetTime(2000);
			mTimeout = false;
			mPropertyNotify.window = None;
			XDeleteProperty(cxDpy, req->GetXID(), mSelectionNotify.property);
			while (!mTimeout) {
				mTimer.SetTime(2000);
				while (!mTimeout && (mPropertyNotify.window != req->GetXID() || mPropertyNotify.state != PropertyNewValue || mPropertyNotify.atom != prop)) cxApp->MainLoop();
				if(mTimeout) return false;
				mPropertyNotify.window = None;
				XGetWindowProperty(cxDpy, req->GetXID(), prop, 0, -1u, True, AnyPropertyType, &type, &fmt, &count, &rem, &data);
				count *= fmt / 8;
				copy(data, data + count, back_inserter(ret));
				XFree(data);
				if (count == 0) {
					mLocked = false;
					return true;
				}
			}
			mLocked = false;
			return false;
		}
		// normal protocol
		XGetWindowProperty(cxDpy, req->GetXID(), mSelectionNotify.property, 0, -1u, False, AnyPropertyType, &type, &fmt, &count, &rem, &data);
		count *= fmt / 8;
		copy(data, data + count, back_inserter(ret));
		XFree(data);
		XDeleteProperty(cxDpy, req->GetXID(), mSelectionNotify.property);
		mLocked = false;
		return true;
	} else {
		// in application: dont bother the server
		if (typ == TIMESTAMP) {
			vector<CSelectionItem>::iterator i = mSelectionList.begin();
			while (i->mName != sel) ++i;
			unsigned char* s = (unsigned char*)&(i->mStartTime);
			copy(s, s + sizeof(Time), back_inserter(ret));
			return true;
		}
		return (owner->ConvertSelection(sel, typ, ret));
	}
}
