/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CTopLevelWindow.cpp
  Info   : $Id: CTopLevelWindow.cpp,v 1.58 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/CDnDWindow.h>
#include <cxxx/CMessageBox.h>
#include <cxxx/CImage.h>

// C++ headers
#include <cctype>
#include <algorithm>
#include <iostream>

// X headers
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h> // for IsModifierKey, IsFunctionKey

// namespace declarations
using std::string;
using std::wstring;
using std::vector;
using std::copy;
using std::cerr;
using std::endl;
using std::map;
using cxxx::cxApp;
using cxxx::cxError;
using cxxx::cxDnDManager;
using cxxx::cxDpy;
using cxxx::cxWiden;
using cxxx::cxWindowFormat;
using cxxx::cxStartupID;
using cxxx::lib::cxgettext;

// static members
Atom CTopLevelWindow::WM_PROTOCOLS_ATOM = None;
Atom CTopLevelWindow::WM_DELETE_ATOM = None;
Atom CTopLevelWindow::WM_ICON_SIZE_ATOM = None;
Atom CTopLevelWindow::WM_TAKE_FOCUS_ATOM = None;
Atom CTopLevelWindow::NET_SUPPORTED_ATOM = None;
Atom CTopLevelWindow::NET_WM_NAME_ATOM = None;
Atom CTopLevelWindow::NET_WM_ICON_NAME_ATOM = None;
Atom CTopLevelWindow::NET_WM_ICON_ATOM = None;
Atom CTopLevelWindow::NET_WM_PID_ATOM = None;
Atom CTopLevelWindow::NET_WM_PING_ATOM = None;
Atom CTopLevelWindow::NET_SUPPORTING_WM_CHECK_ATOM = None;
Pixmap CTopLevelWindow::mWMIcon = None;
Pixmap CTopLevelWindow::mWMIconMask = None;
unsigned CTopLevelWindow::mWMCaps = 0;

void CTopLevelWindow::Init()
{
	if (WM_PROTOCOLS_ATOM == None) WM_PROTOCOLS_ATOM = GetAtomFromName("WM_PROTOCOLS");
	if (WM_PROTOCOLS_ATOM == None) cxError << _("Failed to register Atom WM_PROTOCOLS.", 97) << CError::Error;
	if (WM_DELETE_ATOM == None) WM_DELETE_ATOM = GetAtomFromName("WM_DELETE_WINDOW");
	if (WM_TAKE_FOCUS_ATOM == None) WM_TAKE_FOCUS_ATOM = GetAtomFromName("WM_TAKE_FOCUS");
	Atom wmproto[2] = {WM_DELETE_ATOM, WM_TAKE_FOCUS_ATOM};
	if (!XSetWMProtocols(cxDpy, GetXID(), wmproto, 2)) cxError << _("Failed to set WM_PROTOCOLS.", 397) << CError::Error;
	if (WM_ICON_SIZE_ATOM == None) WM_ICON_SIZE_ATOM = GetAtomFromName("WM_ICON_SIZE");
	if (WM_ICON_SIZE_ATOM == None) cxError << _("Failed to register Atom WM_ICON_SIZE.", 98) << CError::Error;
	if (NET_SUPPORTED_ATOM == None) NET_SUPPORTED_ATOM = GetAtomFromName("_NET_SUPPORTED");
	if (NET_SUPPORTED_ATOM == None) cxError << _("Failed to register Atom _NET_SUPPORTED.", 395) << CError::Error;
	if (mWMCaps == 0) mWMCaps = GetWMCaps();
	XWMHints* hints = XAllocWMHints();
	if (hints == 0) {
		cerr << "XAllocWMHints: out of Memory" << endl;
		exit( -1);
	}
	hints->flags = InputHint | StateHint;
	hints->input = True;
	hints->initial_state = NormalState;
	if (CreateWMPixmaps() && mWMIcon != None) {
		hints->icon_pixmap = mWMIcon;
		hints->icon_mask = mWMIconMask;
		hints->flags |= IconPixmapHint | IconMaskHint;
	}
	XSetWMHints(cxDpy, GetXID(), hints);
	XFree(hints);
	mAttributes.event_mask |= KeyPressMask | FocusChangeMask;
	ChangeAttributes(CWEventMask);
	for (unsigned i = 0; i <= 'Z' - 'A'; i++) mAccelTable[i] = NULL;
	mTransient = NULL;
	mTransientFor = NULL;
}

/* virtual */ CTopLevelWindow::~CTopLevelWindow()
{
	if (mTransientFor) mTransientFor->mTransient = mTransient;
	if (mTransient)	mTransient->mTransientFor = mTransientFor;
}

/* virtual */ void CTopLevelWindow::OnEvent(XEvent& e)
{
	switch (e.xany.type) {
	case FocusIn:
		/* fall through */
	case FocusOut:
		OnFocusEvent(e.xfocus);
		break;
	default:
		break;
	}
}

/* virtual */ void CTopLevelWindow::OnClientMessage(XClientMessageEvent& e)
{
    if (e.message_type == WM_PROTOCOLS_ATOM) {
        if (static_cast<Atom>(e.data.l[0]) == WM_DELETE_ATOM) OnQuit();
		else if (static_cast<Atom>(e.data.l[0]) == WM_TAKE_FOCUS_ATOM) {
			XSetInputFocus(cxDpy, GetXID(), RevertToParent, (Time)(e.data.l[1]));
			XFocusChangeEvent e;
			e.window = GetXID();
			e.type = FocusIn;
			OnFocusEvent(e);
		}
        return;
    }
    if (cxDnDManager->HandleClientMessage(e)) return;
}

/* virtual */ void CTopLevelWindow::OnFocusEvent(XFocusChangeEvent& e)
{
	if (e.type == FocusIn) {
		CTopLevelWindow* w = this;
		if (mTransient && mTransient->IsVisible()) {
			while (w->mTransient && w->mTransient->IsVisible()) {
				w->mWindowState.reset(ISFOCUSED);
				w = w->mTransient;
				XSetInputFocus(cxDpy, w->GetXID(), RevertToParent, CurrentTime);
			}
		} else {
			GetTopLevelWindow()->mWindowState.set(ISFOCUSED);
			SetFocusWindow(mFocusWindow);
		}
	} else {
		GetTopLevelWindow()->mWindowState.reset(ISFOCUSED);
		SetFocusWindow(NULL);
	}
}

void CTopLevelWindow::TransientFor(CTopLevelWindow* w)
{
	cxAssert (w != this);
	XSetTransientForHint(cxxx::cxDpy, GetXID(), w->GetXID());
	mTransientFor = w;
	w->mTransient = this;
}

bool CTopLevelWindow::CreateIcccmPixmaps()
{
	if (WM_ICON_SIZE_ATOM == None) return false;
	// Get supported icon sizes
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	union {unsigned char* in; XIconSize* data;} buf;
	// Read WM_ICON_SIZE property on root window
	if (XGetWindowProperty(cxDpy, DefaultRootWindow(cxDpy), WM_ICON_SIZE_ATOM, 0, -1u, False, WM_ICON_SIZE_ATOM, &type,
		&fmt, &count, &rem, &buf.in) != Success) {
		// Request failed
		cxError << _("Failed to read WM_ICON_SIZE property on root window", 396) << CError::Warn;
		XFree(buf.in);
		return false;
	}
	count = (count * 4) / sizeof(XIconSize);
	if (count == 0 || fmt != 32) {
		cxError << _("Failed to read WM_ICON_SIZE property on root window", 396) << CError::Warn;
		XFree(buf.in);
		return false;
	}
	// C++X supports 16x16, 22x22, 32x32 and 48x48 let's see if the WM can cope with it
	int w = buf.data->min_width;
	int h = buf.data->min_height;
	if (buf.data->width_inc > 0) {
		while (w < 16) w += buf.data->width_inc;
		if (w != 16) {
			while (w < 22) w += buf.data->width_inc;
			if (w != 22) {
				while (w < 32) w += buf.data->width_inc;
				if (w != 32) {
					while (w < 48) w += buf.data->width_inc;
				}
			}
		}
	}
	if (w > buf.data->max_width) w = buf.data->max_width;
	if (h == 0) h = w; 
	else {
		if (buf.data->height_inc > 0) while (h < w) h += buf.data->height_inc;
		if (h > buf.data->max_height) h = buf.data->max_height;
	}
	bool scale = false;
	XFree(buf.in);			
	if (w == 0 || h == 0) return false;
	string fn;
	if (w != h) scale = true;
	if (w != 16 && w != 22 && w != 32 && w != 48) scale = true;
	unsigned size = w;
	if (scale) size = 48;
	fn = CIcon::FindFileName(CIcon::APP, size);
	CImage im;
	im.ReadFile(fn);
	if (scale) im.Scale(w, h, CImage::SCALE_UP | CImage::SCALE_DOWN);
	if (im.IsValid()) {
		Picture src = im.GetPicture();
		Pixmap px = XCreatePixmap(cxDpy, GetXID(), w, h, DefaultDepth(cxDpy, DefaultScreen(cxDpy)));
		Picture dst = XRenderCreatePicture(cxDpy, px, cxWindowFormat, 0, NULL);
		XRenderColor rcol = CColor(TC_FRAME);
		XRenderFillRectangle(cxDpy, PictOpSrc, dst, &rcol, 0, 0, w, h);
		XRenderComposite(cxDpy, PictOpOver, src, src, dst, 0, 0, 0, 0, 0, 0, w, h);
		mWMIcon = px;
		mWMIconMask = None;
		XRenderFreePicture(cxDpy, src);
		XRenderFreePicture(cxDpy, dst);
	}
	return true;
}

bool CTopLevelWindow::CreateNetWMPixmaps()
{
	cxAssert(mWMCaps & NET_WM_ICON);
	unsigned long pix[258];
	string fn16 = CIcon::FindFileName(CIcon::APP, 16);
	unsigned long* ptr = pix;
	if (!fn16.empty()) {
		CImage im;
		im.ReadFile(fn16);
		if (im.IsValid()) {
			*ptr++ = 16;
			*ptr++ = 16;
			for (unsigned i = 0; i < 256; ++i) *ptr++ = im.GetPixel(i).GetARGB();
		} else return false;
	}
	XChangeProperty(cxDpy, GetXID(), NET_WM_ICON_ATOM, XA_CARDINAL, 32, PropModeReplace, (unsigned char*)(pix), 258);
	return true;
}	

unsigned CTopLevelWindow::GetWMCaps()
{
	if (NET_SUPPORTED_ATOM == None) return 1;
	const char* names[] = {"_NET_WM_NAME", "_NET_WM_ICON_NAME", "_NET_WM_ICON", "_NET_WM_PID", "_NET_WM_PING", "_NET_SUPPORTING_WM_CHECK"};
	unsigned protocount = 6;
	Atom atoms[6];
	if (XInternAtoms(cxDpy, const_cast<char**>(names), protocount, False, atoms) == 0) return NO_NET;
	NET_WM_NAME_ATOM = atoms[0];
	NET_WM_ICON_NAME_ATOM = atoms[1];
	NET_WM_ICON_ATOM = atoms[2];
	NET_WM_PID_ATOM = atoms[3];
	NET_WM_PING_ATOM = atoms[4];
	NET_SUPPORTING_WM_CHECK_ATOM = atoms[5];
	Atom type;
	int fmt;
	unsigned long count;
	unsigned long rem;
	union {unsigned char* in; Atom* proto; Window* win;} buf;
	// Read _NET_SUPPORTED property on root window
	if (XGetWindowProperty(cxDpy, DefaultRootWindow(cxDpy), NET_SUPPORTED_ATOM, 0, -1u, False, XA_ATOM, &type,
		&fmt, &count, &rem, &buf.in) != Success) {
		// Request failed, no net support
		XFree(buf.proto);
		return NO_NET;
	}
	if (type != XA_ATOM || count == 0 || fmt != 32) {
		// Wrong format, no net support
		XFree(buf.proto);
		return NO_NET;
	}
	// Get supported protocols
	unsigned ret = 0;
	for (unsigned i = 0; i < protocount; ++i) {
		if (atoms[i] == None) continue;
		for (unsigned j = 0; j < count; ++j) {
			if (atoms[i] == buf.proto[j]) {
				ret += 1 << i;
				break;
			}
		}
	}
	ret <<= 1;
	XFree(buf.proto);
	// The NET_SUPPORTING_WM_CHECK has to be set
	if (!(ret & NET_SUPPORTING_WM_CHECK)) return NO_NET;
	// Read this property
	if (XGetWindowProperty(cxDpy, DefaultRootWindow(cxDpy), NET_SUPPORTING_WM_CHECK_ATOM, 0, -1u, False, XA_WINDOW,
		&type, &fmt, &count, &rem, &buf.in) != Success) {
		// Request failed, no net support
		XFree(buf.win);
		return NO_NET;
	}
	if (type != XA_WINDOW || count == 0 || fmt != 32) {
		// Wrong format, no net support
		XFree(buf.win);
		return NO_NET;
	}
	Window checkwin = *buf.win;
	XFree(buf.win);
	// This window must have NET_SUPPORTING_WM_CHECK, too
	if (XGetWindowProperty(cxDpy, checkwin, NET_SUPPORTING_WM_CHECK_ATOM, 0, -1u, False, XA_WINDOW,
		&type, &fmt, &count, &rem, &buf.in) != Success) {
		// Request failed, no net support
		XFree(buf.win);
		return NO_NET;
	}
	if (type != XA_WINDOW || count == 0 || fmt != 32) {
		// Wrong format, no net support
		XFree(buf.win);
		return NO_NET;
	}
	if (checkwin != *buf.win) {
		// Wrong window, no net support
		XFree(buf.win);
		return NO_NET;
	}
	return ret;
}

void CTopLevelWindow::SetFocusWindow(CWindow* w)
{
	if (!w) {
		cxxx::cxApp->SetPopupWindow(NULL);
		if (mFocusWindow) {
			// unfocus previous focus window
			mFocusWindow->mWindowState.reset(ISFOCUSED);
			mFocusWindow->OnFocus(false);
		}
		return;
	}
	if (w->IsFocusable()) {
		// focus window changes
		if (mFocusWindow != w) SetFocusWindow(NULL);
		mFocusWindow = w;
		mFocusWindow->mWindowState.set(ISFOCUSED);
		w->OnFocus(true);
	}
}

wchar_t CTopLevelWindow::KeysymToWchar(KeySym sym)
{
	static const wchar_t tab2[] =
	{
		L'\x1A0', L'\x104', L'\x2D8', L'\x141', L'\x1A4', L'\x13D', L'\x15A', L'\x1A7',
		L'\x1A8', L'\x160', L'\x15E', L'\x164', L'\x179', L'\x1ad', L'\x17D', L'\x17B',
		L'\x1B0', L'\x105', L'\x2DB', L'\x142', L'\x1B4', L'\x13E', L'\x15B', L'\x2C7',
		L'\x1B8', L'\x161', L'\x15F', L'\x165', L'\x17A', L'\x2DD', L'\x17E', L'\x17C',
		L'\x154', L'\x1C1', L'\x1C2', L'\x102', L'\x1C4', L'\x139', L'\x106', L'\x1C7',
		L'\x10C', L'\x1C9', L'\x118', L'\x1CB', L'\x11A', L'\x1CD', L'\x1CE', L'\x10E',
		L'\x110', L'\x143', L'\x147', L'\x1D3', L'\x1D4', L'\x150', L'\x1D6', L'\x1D7',
		L'\x158', L'\x16E', L'\x1DA', L'\x170', L'\x1DC', L'\x1DD', L'\x162', L'\x1DF',
		L'\x155', L'\x1E1', L'\x1E2', L'\x103', L'\x1E4', L'\x13A', L'\x107', L'\x1E7',
		L'\x10D', L'\x1E9', L'\x119', L'\x1EB', L'\x11B', L'\x1ED', L'\x1EE', L'\x10F',
		L'\x111', L'\x144', L'\x148', L'\x1F3', L'\x1F4', L'\x151', L'\x171', L'\x1F7',
		L'\x159', L'\x16F', L'\x1FA', L'\x1FB', L'\x1FC', L'\x1FD', L'\x163', L'\x2D9'
	};
	static const wchar_t tab3[] =
	{
		L'\x2A0', L'\x126', L'\x2A2', L'\x2A3', L'\x2A4', L'\x2A5', L'\x124', L'\x2A7',
		L'\x2A8', L'\x130', L'\x2AA', L'\x11E', L'\x134', L'\x2AD', L'\x2AE', L'\x2AF',
		L'\x2B0', L'\x127', L'\x2B2', L'\x2B3', L'\x2B4', L'\x2B5', L'\x125', L'\x2B7',
		L'\x2B8', L'\x131', L'\x2BA', L'\x11F', L'\x135', L'\x2BD', L'\x2BE', L'\x2BF',
		L'\x2C0', L'\x2C1', L'\x2C2', L'\x2C3', L'\x2C4', L'\x10A', L'\x108', L'\x2C7',
		L'\x2C8', L'\x2C9', L'\x2CA', L'\x2CB', L'\x2CC', L'\x2CD', L'\x2CE', L'\x2CF',
		L'\x2D0', L'\x2D1', L'\x2D2', L'\x2D3', L'\x2D4', L'\x120', L'\x2D6', L'\x2D7',
		L'\x11C', L'\x2D9', L'\x2DA', L'\x2DB', L'\x2DC', L'\x16C', L'\x15C', L'\x2DF',
		L'\x2E0', L'\x2E1', L'\x2E2', L'\x2E3', L'\x2E4', L'\x10B', L'\x109', L'\x2E7',
		L'\x2E8', L'\x2E9', L'\x2EA', L'\x2EB', L'\x2EC', L'\x2ED', L'\x2EE', L'\x2EF',
		L'\x2F0', L'\x2F1', L'\x2F2', L'\x2F3', L'\x2F4', L'\x121', L'\x2F6', L'\x2F7',
		L'\x11D', L'\x2F9', L'\x2FA', L'\x2FB', L'\x2FC', L'\x16D', L'\x15D', L'\x2FF'
	};
	static const wchar_t tab4[] =
	{
		L'\x3A0', L'\x3A1', L'\x138', L'\x156', L'\x3A4', L'\x128', L'\x13B', L'\x3A7',
		L'\x3A8', L'\x3A9', L'\x112', L'\x122', L'\x166', L'\x3AD', L'\x3AE', L'\x3AF',
		L'\x3B0', L'\x3B1', L'\x3B2', L'\x157', L'\x3B4', L'\x129', L'\x13C', L'\x3B7',
		L'\x3B8', L'\x3B9', L'\x113', L'\x123', L'\x167', L'\x14A', L'\x3BE', L'\x14B',
		L'\x100', L'\x3C1', L'\x3C2', L'\x3C3', L'\x3C4', L'\x3C5', L'\x3C6', L'\x12E',
		L'\x3C8', L'\x3C9', L'\x3CA', L'\x3CB', L'\x116', L'\x3CD', L'\x3CE', L'\x12A',
		L'\x3D0', L'\x145', L'\x14C', L'\x136', L'\x3D4', L'\x3D5', L'\x3D6', L'\x3D7',
		L'\x3D8', L'\x172', L'\x3DA', L'\x3DB', L'\x3DC', L'\x168', L'\x16A', L'\x3DF',
		L'\x101', L'\x3E1', L'\x3E2', L'\x3E3', L'\x3E4', L'\x3E5', L'\x3E6', L'\x12F',
		L'\x3E8', L'\x3E9', L'\x3EA', L'\x3EB', L'\x117', L'\x3ED', L'\x3EE', L'\x12B',
		L'\x3F0', L'\x146', L'\x14D', L'\x137', L'\x3F4', L'\x3F5', L'\x3F6', L'\x3F7',
		L'\x3F8', L'\x173', L'\x3FA', L'\x3FB', L'\x3FC', L'\x169', L'\x16B', L'\x3FF'
	};
	static const wchar_t tab5[] =
	{
		L'\x04A0', L'\x3002', L'\x300C', L'\x300D', L'\x3001', L'\x30FB', L'\x30F2', L'\x30A1',
		L'\x30A3', L'\x30A5', L'\x30A7', L'\x30A9', L'\x30E3', L'\x30E5', L'\x30E7', L'\x30C3',
		L'\x30FC', L'\x30A2', L'\x30A4', L'\x30A6', L'\x30A8', L'\x30AA', L'\x30AB', L'\x30AD',
		L'\x30AF', L'\x30B1', L'\x30B3', L'\x30B5', L'\x30B7', L'\x30B9', L'\x30BB', L'\x30BD',
		L'\x30BF', L'\x30C1', L'\x30C4', L'\x30C6', L'\x30C8', L'\x30CA', L'\x30CB', L'\x30CC',
		L'\x30CD', L'\x30CE', L'\x30CF', L'\x30D2', L'\x30D5', L'\x30D8', L'\x30DB', L'\x30DE',
		L'\x30DF', L'\x30E0', L'\x30E1', L'\x30E2', L'\x30E4', L'\x30E6', L'\x30E8', L'\x30E9',
		L'\x30EA', L'\x30EB', L'\x30EC', L'\x30ED', L'\x30EF', L'\x30F3', L'\x309B', L'\x309C',
		L'\x04E0', L'\x04E1', L'\x04E2', L'\x04E3', L'\x04E4', L'\x04E5', L'\x04E6', L'\x04E7',
		L'\x04E8', L'\x04E9', L'\x04EA', L'\x04EB',	L'\x04EC', L'\x04ED', L'\x04EE', L'\x04EF',
		L'\x04F0', L'\x04F1', L'\x04F2', L'\x04F3', L'\x04F4', L'\x04F5', L'\x04F6', L'\x04F7',
		L'\x04F8', L'\x04F9', L'\x04FA', L'\x04FB',	L'\x04FC', L'\x04FD', L'\x04FE', L'\x04FF'
	};
	static const wchar_t tab6[] =
	{
		L'\x6A0', L'\x452', L'\x453', L'\x451', L'\x454', L'\x455', L'\x456', L'\x457',
		L'\x458', L'\x459', L'\x45A', L'\x45B', L'\x45C', L'\x491', L'\x45E', L'\x45F',
		L'\x2116', L'\x402', L'\x403', L'\x401', L'\x404', L'\x405', L'\x406', L'\x407',
		L'\x408', L'\x409', L'\x40A', L'\x40B', L'\x40C', L'\x490', L'\x40E', L'\x40E' + 1,
		L'\x44E', L'\x430', L'\x431', L'\x446', L'\x434', L'\x435', L'\x444', L'\x433',
		L'\x445', L'\x438', L'\x439', L'\x43A', L'\x43B', L'\x43C', L'\x43D', L'\x43E',
		L'\x43F', L'\x44F', L'\x440', L'\x441', L'\x442', L'\x443', L'\x436', L'\x432',
		L'\x44C', L'\x44B', L'\x437', L'\x448', L'\x44D', L'\x449', L'\x447', L'\x44A',
		L'\x42E', L'\x410', L'\x411', L'\x426', L'\x414', L'\x415', L'\x424', L'\x413',
		L'\x425', L'\x418', L'\x419', L'\x41A', L'\x41B', L'\x41C', L'\x41D', L'\x41E',
		L'\x41F', L'\x42F', L'\x420', L'\x421', L'\x422', L'\x423', L'\x416', L'\x412',
		L'\x42C', L'\x42B', L'\x417', L'\x428', L'\x42D', L'\x429', L'\x427', L'\x42A'
	};
	static const wchar_t tab7[] =
	{
		L'\x7A0', L'\x386', L'\x388', L'\x389', L'\x38A', L'\x3AA', L'\x7A6', L'\x38C',
		L'\x38E', L'\x3AB', L'\x7AA', L'\x38F', L'\x7AC', L'\x7AD', L'\x385', L'\x2015',
		L'\x7B0', L'\x3AC', L'\x3AD', L'\x3AE', L'\x3AF', L'\x3CA', L'\x390', L'\x3CC',
		L'\x3CD', L'\x3CB', L'\x3B0', L'\x3CE', L'\x7BC', L'\x7BD', L'\x7BE', L'\x7BF',
		L'\x7C0', L'\x391', L'\x392', L'\x393', L'\x394', L'\x395', L'\x396', L'\x397',
		L'\x398', L'\x399', L'\x39A', L'\x39B', L'\x39C', L'\x39D', L'\x39E', L'\x39F',
		L'\x3A0', L'\x3A1', L'\x3A3', L'\x7D3', L'\x3A4', L'\x3A5', L'\x3A6', L'\x3A7',
		L'\x3A8', L'\x3A9', L'\x7DA', L'\x7DB', L'\x7DC', L'\x7DD', L'\x7DE', L'\x7DF',
		L'\x7E0', L'\x3B1', L'\x3B2', L'\x3B3', L'\x3B4', L'\x3B5', L'\x3B6', L'\x3B7',
		L'\x3B8', L'\x3B9', L'\x3BA', L'\x3BB', L'\x3BC', L'\x3BD', L'\x3BE', L'\x3BF',
		L'\x3C0', L'\x3C1', L'\x3C3', L'\x3C2', L'\x3C4', L'\x3C5', L'\x3C6', L'\x3C7',
		L'\x3C8', L'\x3C9', L'\x7FA', L'\x7FB', L'\x7FC', L'\x7FD', L'\x7FE', L'\x7FF'
	};
	static const wchar_t tab8[] =
	{
		L'\x08A0', L'\x23B7', L'\x250C', L'\x2500', L'\x2320', L'\x2321', L'\x2502', L'\x23A1',
		L'\x23A3', L'\x23A4', L'\x23A6', L'\x239B', L'\x239D', L'\x239E', L'\x23A0', L'\x23A8',
		L'\x23AC', L'\x08B1', L'\x08B2', L'\x08B3', L'\x08B4', L'\x08B5', L'\x08B6', L'\x08B7',
		L'\x08B8', L'\x08B9', L'\x08BA', L'\x08BB', L'\x2264', L'\x2260', L'\x2265', L'\x222B',
		L'\x2234', L'\x221D', L'\x221E', L'\x08C3', L'\x08C4', L'\x2207', L'\x08C6', L'\x08C7',
		L'\x223C', L'\x2243', L'\x08CA', L'\x08CB', L'\x08CC', L'\x21D4', L'\x21D2', L'\x2261',
		L'\x08D0', L'\x08D1', L'\x08D2', L'\x08D3', L'\x08D4', L'\x08D5', L'\x221A', L'\x08D7',
		L'\x08D8', L'\x08D9', L'\x2282', L'\x2283', L'\x2229', L'\x222A', L'\x2227', L'\x2228',
		L'\x08E0', L'\x08E1', L'\x08E2', L'\x08E3', L'\x08E4', L'\x08E5', L'\x08E6', L'\x08E7',
		L'\x08E8', L'\x08E9', L'\x08EA', L'\x08EB', L'\x08EC', L'\x08ED', L'\x08EE', L'\x2202',
		L'\x08F0', L'\x08F1', L'\x08F2', L'\x08F3', L'\x08F4', L'\x08F5', L'\x0192', L'\x08F7',
		L'\x08F8', L'\x08F9', L'\x08FA', L'\x2190', L'\x2191', L'\x2192', L'\x2193', L'\x08FF'
	};
	static const wchar_t tab9[] =
	{
		L'\x25C6', L'\x2592', L'\x2409', L'\x240C', L'\x240D', L'\x240A', L'\x09E6', L'\x09E7',
		L'\x2424', L'\x240B', L'\x2518', L'\x2510', L'\x250C', L'\x2514', L'\x253C', L'\x23BA',
		L'\x23BB', L'\x2500', L'\x23BC', L'\x23BD', L'\x251C', L'\x2524', L'\x2534', L'\x252C',
		L'\x2502', L'\x09F9', L'\x09FA', L'\x09FB', L'\x09FC', L'\x09FD', L'\x09FE', L'\x09FF'
	};
	static const wchar_t tab10[] =
	{
		L'\x0AA0', L'\x2003', L'\x2002', L'\x2004', L'\x2005', L'\x2007', L'\x2008', L'\x2009',
		L'\x200A', L'\x2014', L'\x2013', L'\x0AAB', L'\x2423', L'\x0AAD', L'\x2026', L'\x2025',
		L'\x2153', L'\x2154', L'\x2155', L'\x2156', L'\x2157', L'\x2158', L'\x2159', L'\x215A',
		L'\x2105', L'\x0AB9', L'\x0ABA', L'\x2012', L'\x27E8', L'\x002E', L'\x27E9', L'\x0ABF',
		L'\x0AC0', L'\x0AC1', L'\x0AC2', L'\x215B', L'\x215C', L'\x215D', L'\x215E', L'\x0AC7',
		L'\x0AC8', L'\x2122', L'\x2613', L'\x0ACB', L'\x25C1', L'\x25B7', L'\x25CB', L'\x25AF',
		L'\x2018', L'\x2019', L'\x201C', L'\x201D', L'\x211E', L'\x0AD5', L'\x2032', L'\x2033',
		L'\x0AD8', L'\x271D', L'\x0ADA', L'\x25AC', L'\x25C0', L'\x25B6', L'\x25CF', L'\x25AE',
		L'\x25E6', L'\x25AB', L'\x25AD', L'\x25B3', L'\x25BD', L'\x2606', L'\x2022', L'\x25AA',
		L'\x25B2', L'\x25BC', L'\x261C', L'\x261E', L'\x2663', L'\x2666', L'\x2665', L'\x0AEF',
		L'\x2720', L'\x2020', L'\x2021', L'\x2713', L'\x2717', L'\x266F', L'\x266D', L'\x2642',
		L'\x2640', L'\x260E', L'\x2315', L'\x2117', L'\x2038', L'\x201A', L'\x201E', L'\x0AFF'
	};
	static const wchar_t tab11[] =
	{
		L'\x0BA0', L'\x0BA1', L'\x0BA2', L'\x003C', L'\x0BA4', L'\x0BA5', L'\x003E', L'\x0BA7',
		L'\x2228', L'\x2227', L'\x0BAA', L'\x0BAB', L'\x0BAC', L'\x0BAD', L'\x0BAE', L'\x0BAF',
		L'\x0BB0', L'\x0BB1', L'\x0BB2', L'\x0BB3', L'\x0BB4', L'\x0BB5', L'\x0BB6', L'\x0BB7',
		L'\x0BB8', L'\x0BB9', L'\x0BBA', L'\x0BBB', L'\x0BBC', L'\x0BBD', L'\x0BBE', L'\x0BBF',
		L'\x00AF', L'\x0BC1', L'\x22A5', L'\x2229', L'\x230A', L'\x0BC5', L'\x005F', L'\x0BC7',
		L'\x0BC8', L'\x0BC9', L'\x2218', L'\x0BCB', L'\x2395', L'\x0BCD', L'\x22A4', L'\x25CB',
		L'\x0BD0', L'\x0BD1', L'\x0BD2', L'\x2308', L'\x0BD4', L'\x0BD5', L'\x222A', L'\x0BD7',
		L'\x2283', L'\x0BD9', L'\x2282', L'\x0BDB', L'\x22A2', L'\x0BDD', L'\x0BDE', L'\x0BDF',
		L'\x0BE0', L'\x0BE1', L'\x0BE2', L'\x0BE3', L'\x0BE4', L'\x0BE5', L'\x0BE6', L'\x0BE7',
		L'\x0BE8', L'\x0BE9', L'\x0BEA', L'\x0BEB', L'\x0BEC', L'\x0BED', L'\x0BEE', L'\x0BEF',
		L'\x0BF0', L'\x0BF1', L'\x0BF2', L'\x0BF3', L'\x0BF4', L'\x0BF5', L'\x0BF6', L'\x0BF7',
		L'\x0BF8', L'\x0BF9', L'\x0BFA', L'\x0BFB', L'\x22A3', L'\x0BFD', L'\x0BFE', L'\x0BFF'
	};
	static const wchar_t tab12[] =
	{
		L' ',      L'\xFF81', L'\xFF82', L'\xFF83', L'\xFF84', L'\xFF85', L'\xFF86', L'\xFF87',
		L'\xFF88', L'\t',     L'\xFF8A', L'\xFF8B', L'\xFF8C', L'\n'    , L'\xFF8E', L'\xFF8F',
		L'\xFF90', L'\xFFBE', L'\xFFBF', L'\xFFC0', L'\xFFC1', L'\xFF50', L'\xFF51', L'\xFF52',
		L'\xFF53', L'\xFF54', L'\xFF55', L'\xFF56', L'\xFF57', L'\xFF58', L'\xFF63', L'\x7F',
		L'\xFFA0', L'\xFFA1', L'\xFFA2', L'\xFFA3', L'\xFFA4', L'\xFFA5', L'\xFFA6', L'\xFFA7',
		L'\xFFA8', L'\xFFA9', L'*',      L'+',      L'\xFFAC', L'-',      L'\xFFAE', L'/',
		L'0',      L'1',      L'2',      L'3',      L'4',      L'5',      L'6',      L'7',
		L'8',      L'9',      L'\xFFBA', L'\xFFBB', L'=',      L'\xFFBD', L'\xFFBE', L'\xFFBF'
	};

	if (sym < 0x1a0) return static_cast<wchar_t>(sym);
	if (sym > 0xffffff) return static_cast<wchar_t>(sym & 0xffffff);
	if (sym == 0x47e) return L'\x203E';
	if ((sym & 0xff) < 0xa0) {
		if ((sym & 0xffff00) != 0x00ff00) return static_cast<wchar_t>(sym);
		if (sym < 0xff20) {
			switch (sym) {
			case 0xff08:
				return L'\b';
			case 0xff09:
				return L'\t';
			case 0xff0d:
				return L'\n';
			case 0xff1b:
				return L'\x1b';
			default:
				return static_cast<wchar_t>(sym);
			}
		} else if (sym >= 0xff80) return tab12[sym - 0xff80];
		else return static_cast<wchar_t>(sym);
	}
	switch (sym & 0x00ff00) {
	case 0x100:
		return tab2[sym - 0x1a0];
	case 0x200:
		return tab3[sym - 0x2a0];
	case 0x300:
		return tab4[sym - 0x3a0];
	case 0x400:
		return tab5[sym - 0x4a0];
	case 0x500:
		return static_cast<wchar_t>(sym + 0x60);
	case 0x600:
		return tab6[sym - 0x6a0];
	case 0x700:
		return tab7[sym - 0x7a0];
	case 0x800:
		return tab8[sym - 0x8a0];
	case 0x900:
		if (sym < 0x9e0) return static_cast<wchar_t>(sym);
		else return tab9[sym - 0x9e0];
	case 0xa00:
		return tab10[sym - 0xaa0];
	case 0xb00:
		return tab11[sym - 0xba0];
	case 0xc00:
		if (sym == 0xcdf) return L'\x2017';
		else if (sym < 0xce0) return static_cast<wchar_t>(sym);
		else return static_cast<wchar_t>(sym - 0x710);
	case 0xd00:
		return static_cast<wchar_t>(sym + 0x60);
	case 0x1300:
		switch (sym) {
		case 0x13bc:
			return L'\x152';
		case 0x13bd:
			return L'\x153';
		case 0x13be:
			return L'\x178';
		default:
			return static_cast<wchar_t>(sym);
		}
	case 0xff00:
		if (sym == 0xffff) return L'\x7f';
		else if (sym < 0xffc0) return tab12[sym - 0xff80];
		else return static_cast<wchar_t>(sym);			
	default:
		return static_cast<wchar_t>(sym);
	}
}
		
/* virtual */ void CTopLevelWindow::HandleKeyPress(XKeyEvent& e)
{
	KeySym sym = XLookupKeysym(&e, 0);
	if (IsModifierKey(sym)) return;
	if (e.state & Mod1Mask) {
		// accelerator
		XLookupKeysym(&e, 1);
		wchar_t c = KeysymToWchar(XLookupKeysym(&e, 1));
		map<wchar_t, CWindow*>::const_iterator ac = mAccelTable.find(c);
		if (ac != mAccelTable.end()) {
			CWindow* w = ac->second;
			if (w != NULL && w->IsEnabled()) w->OnAccelerator(c);
		}
	} else if (IsFunctionKey(sym) || e.state & ControlMask) {
		// shortcut
		OnShortcut(KeysymToWchar(XLookupKeysym(&e, 1)), e.state);
		return ;
	} else {
		XLookupString(&e, NULL, 0, &sym, NULL);
		OnKey(KeysymToWchar(sym), e.state);
	}
}

/* virtual */ void CTopLevelWindow::OnKey(wchar_t key, unsigned state)
{
	if (mFocusWindow) {
		mFocusWindow->OnKey(key, state);
		if (key == L'\n') mFocusWindow->OnAccelerator(key);
	}
}


void CTopLevelWindow::ResetStartupNotification()
{
	cxAssert(!cxStartupID.empty());
	// Try to register Atoms
	Atom atom = GetAtomFromName("_NET_STARTUP_INFO");
	Atom begatom = GetAtomFromName("_NET_STARTUP_INFO_BEGIN");
	if (atom == None || begatom == None) {
		cxError << _("Failed to register Startup atoms", 447) << CError::Error;
		return;
	}
	// Create message string
	string msg("remove: ID=");
	msg += cxStartupID;
	msg += '\0';
	XClientMessageEvent e;
	e.type = ClientMessage;
    e.message_type = begatom;
	CWindow w(this, false);
    e.window = w.GetXID();
    e.format = 8;
	while (!msg.empty()) {
		unsigned len = msg.size();
		if (len > 20) len = 20;
		copy(&msg[0], &msg[0] + len, &e.data.b[0]);
		msg = msg.substr(len);
		XSendEvent(cxDpy, DefaultRootWindow(cxDpy), False, PropertyChangeMask, (XEvent*)&e);
		e.message_type = atom;
	}
}
