/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CTopLevelWindow.h
  Info   : $Id: CTopLevelWindow.h,v 1.31 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CTopLevelWindow_h
#define CTopLevelWindow_h

// C++X headers
#include <cxxx/CWindow.h>
#include <cxxx/CApplication.h>

// Forward declarations
class CDnDWindow;

class CTopLevelWindow : public CWindow
{
protected:
	enum
	{
		// Net support
		NO_NET = 1 << 0,
		NET_WM_NAME = 1 << 1,
		NET_WM_ICON_NAME = 1 << 2,
		NET_WM_ICON = 1 << 3,
		NET_WM_PID = 1 << 4,
		NET_WM_PING = 1 << 5,
		NET_SUPPORTING_WM_CHECK = 1 << 6
	};

public:
	// Constructor
	CTopLevelWindow()
	: CWindow(NULL),
	  mReturn(0),
	  mFocusWindow(NULL)
	{
		Init();
	}

	CTopLevelWindow(unsigned msgid)
	: CWindow(NULL, msgid),
	  mReturn(0),
	  mFocusWindow(NULL)
	{
		Init();
	}

	// Initialization
	void Init();

	// Destructor
	virtual ~CTopLevelWindow();

	// X event processing
	virtual void OnEvent(XEvent& e);
	virtual void OnClientMessage(XClientMessageEvent& e);
	virtual void HandleKeyPress(XKeyEvent& e);
	virtual void OnKey(wchar_t key, unsigned state);

	virtual void OnShortcut(wchar_t key __attribute__ ((unused)), unsigned state __attribute__ ((unused)))
	{
	}
	
	// Exit
	virtual void OnQuit()
	{
		cxxx::cxApp->Quit();
	}

	virtual void OnFocusEvent(XFocusChangeEvent& e);

	// Sets the input focus window to w
	void SetFocusWindow(CWindow* w);

	// Returns the window that has the focus
	CWindow* GetFocusWindow() const
	{
		return mFocusWindow;
	}

	// Sets the window title
	void SetCaption(const std::string& caption)
	{
		XTextProperty cap;
		std::wstring tmp = cxxx::cxWiden(caption);
		wchar_t* tmp1 = (wchar_t*)(tmp.c_str());
		XwcTextListToTextProperty(cxxx::cxDpy, &tmp1, 1, XStdICCTextStyle, &cap);
		XSetWMName(cxxx::cxDpy, GetXID(), &cap);
		XFree(cap.value);
	}

	// Fixes the size of the window
	void FixSize()
	{
		XSizeHints hint;
		hint.flags = PBaseSize | PMinSize | PMaxSize;
		hint.base_width = mWidth;
		hint.min_width = mWidth;
		hint.max_width = mWidth;
		hint.base_height = mHeight;
		hint.min_height = mHeight;
		hint.max_height = mHeight;
		XSetNormalHints(cxxx::cxDpy, GetXID(), &hint);
	}

	// Sets the minimal size of the window
	void SetMinSize(int w, int h)
	{
		XSizeHints hint;
		hint.flags = PMinSize;
		hint.min_width = w;
		hint.min_height = h;
		XSetNormalHints(cxxx::cxDpy, GetXID( ), &hint);
	}

	// Sets the maximal size of the window
	void SetMaxSize(int w, int h)
	{
		XSizeHints hint;
		hint.flags = PMaxSize;
		hint.max_width = w;
		hint.max_height = h;
		XSetNormalHints(cxxx::cxDpy, GetXID( ), &hint);
	}

	// Makes this a transient window for w
	void TransientFor(CTopLevelWindow* w);
	
	// Associates ALT + key to w
	void BindKey(wchar_t key, CWindow* w)
	{
		std::map<wchar_t, CWindow*>::iterator ac = mAccelTable.find(key);
		if (w != NULL && ac != mAccelTable.end() && ac->second != NULL) {
			std::ostringstream os;
			os.setf(std::ios::hex | std::ios::uppercase);
			os << std::hex << static_cast<unsigned>(key);
			cxxx::cxError << cxxx::lib::_("Duplicate Accelerator U+{1}", 3) << os.str() << CError::Warn;
		}
		if (ac == mAccelTable.end()) mAccelTable[key] = w;
		else ac->second = w;
	}
	
	// Releases binding of key
	void ReleaseKey(wchar_t key)
	{
		BindKey(key, NULL);
	}
	
	// Releases binding of w
	void ReleaseKey(CWindow* w)
	{
		for (std::map<wchar_t, CWindow*>::iterator i = mAccelTable.begin(); i != mAccelTable.end(); ++i) {
			if (i->second == w) i->second = NULL;
		}
	}
	
	// Makes this a modal window and returns mReturn
	unsigned DoModal()
	{
		mShutDown = false;
		CTopLevelWindow* p = cxxx::cxApp->SetModalWindow(this);
		while (!mShutDown) cxxx::cxApp->MainLoop();
		cxxx::cxApp->SetModalWindow(p);
		return mReturn;
	}
	
	// Returns true if window is running in modal mode
	bool IsRunning() const
	{
		return !mShutDown;
	}
	
	// Windows unregister here
	void OnDestroy(const CWindow* w)
	{
		if (mFocusWindow == w) {
			SetFocusWindow(NULL);
			mFocusWindow = NULL;
		}
	}
	
	// Sends a XClientMessageEvent
	void SendClientEvent(XClientMessageEvent& e, Window target = None)
	{
		e.type = ClientMessage;
		e.display = cxxx::cxDpy;
		if (target == None) target = e.window;
		cxAssert(target != None);
		if (target == GetXID()) OnClientMessage(e);
		else XSendEvent(cxxx::cxDpy, target, False, NoEventMask, (XEvent*)&e);
	}

	// Resets the startup notification
	void ResetStartupNotification();

protected:
	// Creates the pixmaps for the window manager
	bool CreateWMPixmaps()
	{
		if ((mWMCaps & NET_WM_ICON) && CreateNetWMPixmaps()) return true;
		if (mWMIcon != None) return true;
		return CreateIcccmPixmaps();
	}	

	// Creates the pixmaps for a window manager that has net support
	bool CreateNetWMPixmaps();

	// Creates the pixmaps for a window manager that has no net support
	bool CreateIcccmPixmaps();

	// Returns the window manager capabilities
	unsigned GetWMCaps();

	// Translates Keysyms to UCS wide characters
	wchar_t KeysymToWchar(KeySym sym);

protected:
	std::map<wchar_t, CWindow*> mAccelTable;
	unsigned mReturn;
	bool mShutDown;
	CWindow* mFocusWindow;
	CTopLevelWindow* mTransient;
	CTopLevelWindow* mTransientFor;
	static Atom WM_PROTOCOLS_ATOM;
	static Atom WM_DELETE_ATOM;
	static Atom WM_ICON_SIZE_ATOM;
	static Atom WM_TAKE_FOCUS_ATOM;
	static Atom NET_SUPPORTED_ATOM;
	static Atom NET_WM_NAME_ATOM;
	static Atom NET_WM_ICON_NAME_ATOM;
	static Atom NET_WM_ICON_ATOM;
	static Atom NET_WM_PID_ATOM;
	static Atom NET_WM_PING_ATOM;
	static Atom NET_SUPPORTING_WM_CHECK_ATOM;
	static Pixmap mWMIcon;
	static Pixmap mWMIconMask;
	static unsigned mWMCaps;
};

#endif // CTopLevelWindow_h
