/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CTreeItem.h
  Info   : $Id: CTreeItem.h,v 1.22 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CTreeItem_h
#define CTreeItem_h

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/CIcon.h>
#include <cxxx/CControlItem.h>

// C++ headers
#include <vector>

class CTreeItem : public CControlItem
{
protected:
	enum
	{
		OPENED = CControlItem::SELECTED << 1,
		LINKED = OPENED << 1,
		LOCKED = LINKED << 1
	};
	
protected:
	class CTreeItemInfo : public CControlItemInfo
	{
	public:
		// Constructor
		CTreeItemInfo(CTreeItemInfo* parent, CIcon::TIconName closed, CIcon::TIconName opened, unsigned iconsize, const std::string& text, unsigned w, unsigned h)
		: CControlItemInfo(text, false),
		  mParent(parent),
		  mWidth(w),
		  mHeight(h)
		{
			mIconClose.Create(closed, iconsize);
			mIconOpen.Create(opened, iconsize);
		}
	public:
		CTreeItemInfo* mParent;
		unsigned mWidth;
		unsigned mHeight;
		CIcon mIconOpen;
		CIcon mIconClose;
		std::vector<CTreeItem> mChildren;
	};
public:
	// Constructor
	CTreeItem()
	{
	}

	CTreeItem(CIcon::TIconName closed, CIcon::TIconName opened, unsigned iconsize, const std::string& text, unsigned w, unsigned h)
	{
		mInfo = new CTreeItemInfo(NULL, closed, opened, iconsize, text, w, h);
		//mInfo->mRefCount = 1;
	}

	CTreeItem(CTreeItem& parent, CIcon::TIconName closed, CIcon::TIconName opened, unsigned iconsize, const std::string& text, unsigned w, unsigned h)
	{
		mInfo = new CTreeItemInfo(parent.GetInfo(), closed, opened, iconsize, text, w, h);
		//mInfo->mRefCount = 1;
	}
		
	CTreeItem(const CTreeItem& item)
	: CControlItem(item)
	{
	}

	CTreeItem(CTreeItemInfo* info)
	{
		mInfo = info;
		if (mInfo) ++mInfo->mRefCount;
	}

	// Returns item width
	unsigned GetWidth() const
	{
		return GetInfo()->mWidth;
	}

	// Returns item height
	unsigned GetHeight() const
	{
		return GetInfo()->mHeight;
	}

	// Returns whether the item is opened 
	bool IsOpened() const
	{
		return GetInfo()->mState & OPENED;
	}

	// Returns whether the item is a link (for CFileTree)
	bool IsLink() const
	{
		return GetInfo()->mState & LINKED;
	}

	// Returns whether the item is unreadable (for CFileTree)
	bool IsLocked() const
	{
		return GetInfo()->mState & LOCKED;
	}

	// Returns true if this is the root
	bool IsTree() const
	{
		return GetInfo()->mParent == NULL;
	}
	
	// Returns the icon of the item
	const CIcon& GetIcon() const
	{
		if (IsSelected() || IsOpened()) return GetInfo()->mIconOpen;
		else return GetInfo()->mIconClose;
	}
	
	// Sets the icon of the item
	void SetIcon(CIcon& icon)
	{
		if (IsSelected() || IsOpened()) GetInfo()->mIconOpen = icon;
		else GetInfo()->mIconClose = icon;
	}
	
	// Returns the children of this item
	std::vector<CTreeItem>& GetChildren() const
	{
		return GetInfo()->mChildren;
	}
	
	// Returns the parent of this item
	CTreeItem GetParent() const
	{
		return CTreeItem(GetInfo()->mParent);
	}
	
	// Opens the item
	void Open()
	{
		GetInfo()->mState |= OPENED;
	}
	
	// Closes the item
	void Close()
	{
		GetInfo()->mState &= ~OPENED;
	}

	// Marks the item as being a link (for CFileTree)
	void IsLink(bool l)
	{
		if (l) GetInfo()->mState |= LINKED;
		else GetInfo()->mState &= ~LINKED;
	}

	// Marks the item as being unreadable (for CFileTree)
	void IsLocked(bool l)
	{
		if (l) GetInfo()->mState |= LOCKED;
		else GetInfo()->mState &= ~LOCKED;
	}
	
	// Assignment
	CTreeItem& operator =(const CTreeItem& item)
	{
		if (mInfo != item.mInfo) {
			if (mInfo != NULL && --mInfo->mRefCount == 0) delete mInfo;
			mInfo = item.mInfo;
			if (mInfo != NULL) ++mInfo->mRefCount;
		}
		return *this;
	}
	
	// Comparisons
	bool operator ==(const CTreeItem& item) const
	{
		return mInfo == item.mInfo;
	}

	bool operator !=(const CTreeItem& item) const
	{
		return mInfo != item.mInfo;
	}
	
	// Returns the indentation of the item
	unsigned GetIndentation() const
	{
		unsigned ret = 0;
		CTreeItemInfo* i = GetInfo();
		while ((i = i->mParent) != NULL) ++ret;
		return ret;
	}
	
private:
	// Returns the underlying information
	CTreeItemInfo* GetInfo() const
	{
		cxAssert(mInfo != NULL);
		return static_cast<CTreeItemInfo*>(mInfo);
	}
};

#endif // CTreeItem_h
