/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CWindow.cpp
  Info   : $Id: CWindow.cpp,v 1.79 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CTopLevelWindow.h>
#include <cxxx/CPopup.h>
#include <cxxx/CApplication.h>
#include <cxxx/CImage.h>

// C++  headers
#include <cctype>
#include <cmath>

// X11 headers
#include <X11/cursorfont.h>

// Namespace declarations
using cxxx::cxAlphaFormat;
using cxxx::cxApp;
using cxxx::cxARGBFormat;
using cxxx::cxDefaultPic;
using cxxx::cxDpy;
using cxxx::cxError;
using cxxx::cxGrayPic;
using cxxx::cxMainWin;
using cxxx::cxSelManager;
using cxxx::cxStipplePic;
using cxxx::cxTextFont;
using cxxx::cxWiden;
using cxxx::cxWindowFormat;
using cxxx::cxStartupID;
using cxxx::lib::cxgettext;
using std::string;
using std::wstring;
using std::vector;
using std::min;
using std::abs;

// static members
Cursor CWindow::mWaitCursor(None);
CWindow::TWindowMap CWindow::mWindowMap;

void CWindow::Init()
{
	bool first = false;
	// the first window is the main window
	if (mWindowMap.empty()) {
		cxAssert(cxMainWin == NULL);
		cxAssert(mParent == NULL);
		cxAssert(cxApp != NULL);
		first = true;
	}
	// set parent window
	Window pid;
	if (mParent) pid = mParent->GetXID();
	else pid = DefaultRootWindow(cxDpy);
	// window attributes
	CColor col(TC_FRAME);
	mAttributes.background_pixel = col.GetPixel();
	mAttributes.event_mask = CXXXEVENTS;
	// try to create window
	mXID = XCreateWindow(cxDpy, pid, mX, mY, mWidth, mHeight, 0, DefaultDepth(cxDpy, DefaultScreen(cxDpy)), InputOutput, DefaultVisual(cxDpy, DefaultScreen(cxDpy)), CWBackPixel | CWEventMask, &mAttributes);
	// if we failed print message and exit
	if (mXID == None) cxError << _("Failed to create Window", 99) << CError::Fatal;
	// register window
	mWindowMap.insert(TWindowMap::value_type(mXID, this));
	if (first) {
		cxMainWin = static_cast<CTopLevelWindow*>(this);
		// create the wait cursor
		mWaitCursor = XCreateFontCursor(cxDpy, XC_watch);
		if (mWaitCursor == None) cxError << _("Failed to create wait cursor", 100) << CError::Error;
		if (!cxStartupID.empty()) cxMainWin->ResetStartupNotification();
	}
	mRID = XRenderCreatePicture(cxDpy, mXID, cxWindowFormat, 0, NULL);
}

// destructor
/* virtual */ CWindow::~CWindow()
{
	if (cxMainWin == this) {
		cxAssert(mWindowMap.size() == 1);
		XFreeCursor(cxDpy, mWaitCursor );
		cxMainWin = NULL;
		cxApp->Quit();
	}
	XUnmapWindow(cxDpy, GetXID());
	XRenderFreePicture(cxDpy, GetRID());
	XDestroyWindow(cxDpy, GetXID());
	GetTopLevelWindow()->OnDestroy(this);
	cxSelManager->Release(this);
	// remove window from array
	TWindowMap::iterator pos = mWindowMap.find(GetXID());
	cxAssert(pos != mWindowMap.end());
	mWindowMap.erase(pos);
}

// functions
void CWindow::HandleEvent(XEvent& e)
{
	cxAssert(e.xany.window == GetXID());
	switch (e.xany.type)
	{
	case Expose:
	case GraphicsExpose:
		if (IsVisible()) {
			if (cxxx::cxRTL) e.xexpose.x = mWidth - e.xexpose.width - e.xexpose.x;
			ClearArea(e.xexpose);
			OnExpose(e.xexpose);
		}
		return;
	case ConfigureNotify:
		if (cxxx::cxRTL) {
			if (!IsPopup() && mParent) e.xconfigure.x = mParent->mWidth - e.xconfigure.width - e.xconfigure.x;
			else e.xconfigure.x = DisplayWidth(cxxx::cxDpy, DefaultScreen(cxxx::cxDpy)) - e.xconfigure.width - e.xconfigure.x;
		}
		mX = e.xconfigure.x;
		mY = e.xconfigure.y;
		mWidth = e.xconfigure.width;
		mHeight = e.xconfigure.height;
		OnConfigure(e.xconfigure);
		return ;
	case MapNotify:
		/* fall through */
	case UnmapNotify:
		OnMap(e.xmap);
		return ;
	case ButtonPress:
		if (!IsEnabled()) return;
		if (cxxx::cxRTL) e.xbutton.x = mWidth - e.xbutton.x;
		SetFocus();
		switch (e.xbutton.button) {
		case 1:
			OnLButtonDown(e.xbutton);
			break;
		case 2:
			OnMButtonDown(e.xbutton);
			break;
		case 3:
			OnRButtonDown(e.xbutton);
			break;
		case 4:
			OnMouseWheelUp(e.xbutton);
			break;
		case 5:
			OnMouseWheelDown(e.xbutton);
			break;
		default:
			break;
		}
		return;
	case ButtonDblClick:
		if (IsEnabled()) {
			if (cxxx::cxRTL) e.xbutton.x = mWidth - e.xbutton.x;
			OnDblClick(e.xbutton);
		}
		return;
	case ButtonRelease:
		if (!IsEnabled()) return;
		if (cxxx::cxRTL) e.xbutton.x = mWidth - e.xbutton.x;
		if (e.xbutton.button == 4) OnMouseWheelUp(e.xbutton);
		else if (e.xbutton.button == 5) OnMouseWheelDown(e.xbutton);
		else OnButtonUp(e.xbutton);
		return;
	case KeyPress:
		if (!IsEnabled()) return;
		GetTopLevelWindow()->HandleKeyPress(e.xkey);
		return;
	case ClientMessage:
		GetTopLevelWindow()->OnClientMessage(e.xclient);
		return;
	case CxxxMessage:
		OnMessage(e.xclient.data.l[0], (TMessage)e.xclient.message_type, e.xclient.data.l[1]);
		break;
	case MotionNotify:
		if (cxxx::cxRTL) e.xmotion.x = mWidth - e.xmotion.x;
	default:
		OnEvent(e);
		break;
	}
}


/* virtual */ void CWindow::Recolor()
{
	CColor col(TC_FRAME);
	mAttributes.background_pixel = col.GetPixel();
	ChangeAttributes(CWBackPixel);
	Redraw();
}

void CWindow::DrawFrame(int x, int y, int w, int h, TFrameStyle style, bool erase)
{
	if (cxxx::cxRTL) {
		x = mWidth - w - x;
		if (style == FS_TAB_LEFT) style = FS_TAB_RIGHT;
		else if (style == FS_TAB_RIGHT) style = FS_TAB_LEFT;
	}
	static XRectangle rects[4];
	XRenderColor col;
	switch (style)
	{
	case FS_SUNKEN:
		HLine(rects[0], x, y, w -1);
		VLine(rects[1], x, y + 1, h - 2);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);

		HLine(rects[0], x, y + h - 1, w);
		VLine(rects[1], x + w - 1, y, h);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);

		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x + 1, y + 1, w - 2, h - 2);
		}
		return;
			
	case FS_SUNKEN_THICK:
		HLine(rects[0], x, y, w);
		VLine(rects[1], x, y, h);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 1, y + 1, w - 2);
		VLine(rects[1], x + 1, y + 1, h - 2);
		col = CColor(TC_BLACK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 1, y + h - 1, w - 1);
		VLine(rects[1], x + w - 1, y + 1, h - 1);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 2, y + h - 2, w - 3);
		VLine(rects[1], x + w - 2, y + 2,  h - 3);
		col = CColor(TC_FRAME);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x + 2, y + 2, w - 4, h - 4);
		}
		return;
			
	case FS_RAISED:
		HLine(rects[0], x, y, w);
		VLine(rects[1], x, y, h - 1);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x, y + h - 1, w);
		VLine(rects[1], x + w - 1, y + 1, h - 1);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		if (erase) ClearArea(x + 1, y + 1, w - 2, h - 2);
		return;
			
	case FS_RAISED_THICK:
		HLine(rects[0], x, y, w);
		VLine(rects[1], x, y, h);
		col = CColor(TC_FRAME);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 2, y + h - 2,  w - 3);
		VLine(rects[1], x + w - 2, y + 1, h - 2);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 1, y + h - 1, w - 1);
		VLine(rects[1], x + w - 1, y + 1, h - 1);
		col = CColor(TC_BLACK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 1, y + 1, w - 3);
		VLine(rects[1], x + 1, y + 1, h - 2);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x + 2, y + 2, w - 4, h - 4);
		}
		return;
		
    case FS_ENGRAVED:
		HLine(rects[0], x, y, w - 2);
		VLine(rects[1], x + w - 2, y, h - 2);
		HLine(rects[2], x, y + h - 2, w - 1);
		VLine(rects[3], x, y + 1, h - 3);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 4);
		
		HLine(rects[0], x + 1, y + 1, w - 3);
		VLine(rects[1], x + 1, y + 2, h - 4);
		VLine(rects[2], x + w - 1, y, h - 1);
		HLine(rects[3], x, y + h - 1, w);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 4);
		
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x + 2, y + 2, w - 4, h - 4);
		}
		return;
			
    case FS_FLAT:
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x, y, w, h);
		}
		return;
			
    case FS_SHEET:
		col = CColor(TC_FRAME);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x, y, 1, h);

		HLine(rects[0], x + 2, y + h - 2, w - 3);
		VLine(rects[1], x + w - 2, y, h - 1);
		col = CColor(TC_DARK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		HLine(rects[0], x + 1, y + h - 1, w - 1);
		VLine(rects[1], x + w - 1, y, h);
		col = CColor(TC_BLACK);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		col = CColor(TC_BRIGHT);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x + 1, y, 1, h - 1);
			
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x + 2, y, w - 4, h - 4);
		}
		return;
	
	case FS_TAB_SEL:
		HLine(rects[0], x, y, w);
		VLine(rects[1], x, y, h - 1);
		HLine(rects[2], 0, y + h - 2, x);
		HLine(rects[3], x + w - 2, y + h - 2, mWidth - w - x - 1);
		col = CColor(TC_FRAME);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 4);

		HLine(rects[0], x + 1, y + 1, w - 2);
		VLine(rects[1], x + 1, y + 1, h - 1);
		HLine(rects[2], 1, y + h - 1, x);
		HLine(rects[3], x + w - 1, y + h - 1, mWidth - w - x - 1);
		col = CColor(TC_BRIGHT);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 4);
			
		col = CColor(TC_DARK);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x + w - 2, y + 2, 1, h - 2);

		col = CColor(TC_BLACK);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x + w - 1, y + 1, 1, h - 1);

		// ignore erase
		return;

	case FS_TAB_LEFT:
		col = CColor(TC_BRIGHT);
		HLine(rects[0], x + 1, y, w - 1);
		if (x > 0) VLine(rects[1], x, y, h - 2);
		else VLine(rects[1], x + 1, y, h);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
		
		// ignore erase
		return;
		
	case FS_TAB_RIGHT:
		col = CColor(TC_BRIGHT);
		HLine(rects[0], x + 1, y, w - 1);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 1);
		
		col = CColor(TC_DARK);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x + w - 1, y, 1, h - 2);
		
		// ignore erase
		return;

	case FS_TAB_NORMAL:
		col = CColor(TC_BRIGHT);
		HLine(rects[0], x + 1, y, w - 1);
		if (x > 0) VLine(rects[1], x, y, h - 2);
		else VLine(rects[1], x + 1, y, h);
		XRenderFillRectangles(cxDpy, PictOpSrc, GetRID(), &col, rects, 2);
			
		col = CColor(TC_DARK);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &col, x + w - 1, y, 1, h - 2);
		
		// ignore erase
		return;
		
	default:
		cxError << _("CWindow::DrawFrame:\nUnknown Framestyle", 101) << CError::Error;
		if (erase) {
			if (cxxx::cxRTL) x = mWidth - w - x;
			ClearArea(x, y, w, h);
		}
		break;
	}
}

void CWindow::DrawRectangle(int x, int y, int w, int h, CColor col)
{
	if (w > 1 && h > 1) {
		--h;
		XRectangle rects[4];
		rects[0].x = x;
		rects[0].y = y;
		rects[0].width = w;
		rects[0].height = 1;
		rects[1].x = x;
		rects[1].y = y + h;
		rects[1].width = w;
		rects[1].height = 1;
		--w;
		rects[2].x = x;
		rects[2].y = ++y;
		rects[2].width = 1;
		rects[2].height = --h;
		rects[3].x = x + w;
		rects[3].y = y;
		rects[3].width = 1;
		rects[3].height = h;
		XRenderColor rcol = col;
		XRenderFillRectangles(cxxx::cxDpy, PictOpSrc, GetRID(), &rcol, rects, 4);
	}
}

void CWindow::DrawDottedRectangle(int x, int y, int w, int h, CColor col)
{
	if (w > 0 && h > 0) {
		if (cxxx::cxRTL) x = mWidth - w - x;
		Pixmap pix = XCreatePixmap(cxDpy, mXID, w, h, 32);
		Picture pic = XRenderCreatePicture(cxDpy, pix, cxARGBFormat, 0, NULL);
		XFreePixmap(cxDpy, pix);
		XRenderColor rcol = col;
		XRenderFillRectangle(cxDpy, PictOpSrc, pic, &rcol, 0, 0, w, h);
		if (w > 2 && h > 2) {
			rcol = CColor(0, 0, 0, 0);
			XRenderFillRectangle(cxDpy, PictOpSrc, pic, &rcol, 1, 1, w - 2, h - 2);
		}
		XRenderComposite(cxDpy, PictOpOver, pic, cxStipplePic, GetRID(), 0, 0, 0, 0, x, y, w, h);
		XRenderFreePicture(cxDpy, pic);
	}
}

void CWindow::DrawHotString(int x, int y, CColor col, const wstring& s, const CHotCharPos& hot, bool grayed, bool erase)
{
	if (grayed) {
		DrawString(x + 1, y + 1, CColor(TC_BRIGHT), s, erase);
		col = CColor(TC_DARK);
		erase = false;
	}
	DrawString(x, y, col, s, erase);
	if (hot.mBegin == hot.mEnd) return;
	int x1, x2;
	if (cxxx::cxRTL) {
		x1 = GetTextWidth(s) + x - hot.mBegin;
		x2 = GetTextWidth(s) + x - hot.mEnd;
	} else {
		x1 = x + hot.mBegin;
		x2 = x + hot.mEnd;
	}
	y += mFont->GetAscent() + 1;
	if (grayed) DrawLine(x1 + 1, y + 1, x2 + 1, y + 1, CColor(TC_BRIGHT));
	DrawLine(x1, y, x2, y, col);
}

void CWindow::DrawMultiLineString(int x, int y, int w, CColor col, const string& s)
{
	if (s.empty()) return;
	wstring ws = cxWiden(s);
	if (w < 3 * GetAvgCharWidth()) return ;
	wstring::size_type start = 0;
	wstring::size_type tmp, end;
	std::wstring stmp;
	int len;
	while (start < ws.size()) {
		len = 0;
		end = start;
		while ((len < w) && (end < ws.size()) && (ws[end] != L'\n')) len += GetCharWidth(ws[end++]);
		if (len >= w) {
			tmp = ws.rfind(L' ', end - 1);
			if((tmp == wstring::npos) || (tmp <= start)) while(len > w) len -= GetCharWidth(ws[--end]);
			else end = tmp;
		}
		stmp = ws.substr(start, end - start);
		DrawString(x, y, col, stmp);
		start = end;
		if ((start < ws.size()) && (isspace(ws[start]))) ++start;
		y += GetTextHeight();
	}
}

void CWindow::DrawIcon(int x, int y, const CIcon& ico, CColor col, bool grayed)
{
	if (!ico.IsValid()) {
		cxError << _("Invalid icon in CWindow::DrawIcon", 102) << CError::Warn;
		return;
	}
	if (cxxx::cxRTL) x = mWidth - ico.GetWidth() - x;
	XRenderColor rcol = col;
	if (grayed) {
		rcol = CColor(TC_BRIGHT);
		XRenderFillRectangle(cxDpy, PictOpSrc, cxDefaultPic, &rcol, 0, 0, 1, 1);
		XRenderComposite(cxDpy, PictOpOver, cxDefaultPic, ico.GetPicture(), GetRID(), 0, 0, 0, 0, x + 1, y + 1, ico.GetWidth(), ico.GetHeight());

		rcol = CColor(TC_DARK);
		XRenderFillRectangle(cxDpy, PictOpSrc, cxDefaultPic, &rcol, 0, 0, 1, 1);
		XRenderComposite(cxDpy, PictOpOver, cxDefaultPic, ico.GetPicture(), GetRID(), 0, 0, 0, 0, x - 1, y - 1, ico.GetWidth(), ico.GetHeight());

		rcol = CColor(TC_FRAME);
	}
	XRenderFillRectangle(cxDpy, PictOpSrc, cxDefaultPic, &rcol, 0, 0, 1, 1);
	XRenderComposite(cxDpy, PictOpOver, cxDefaultPic, ico.GetPicture(), GetRID(), 0, 0, 0, 0, x, y, ico.GetWidth(), ico.GetHeight());
}

void CWindow::DrawIcon(int x, int y, const CIcon& ico, bool grayed)
{
	if (!ico.IsValid()) {
		cxError << _("Invalid icon in CWindow::DrawIcon", 102) << CError::Warn;
		return;
	}
	if (cxxx::cxRTL) x = mWidth - ico.GetWidth() - x;
	Picture mask = grayed ? cxGrayPic : None;
	int op = ico.GetType() == CImage::ALPHA ? PictOpSrc : PictOpOver;
	XRenderComposite(cxDpy, op, ico.GetPicture(), mask, GetRID(), 0, 0, 0, 0, x, y, ico.GetWidth(), ico.GetHeight());
}

void CWindow::DrawLine(int x1, int y1, int x2, int y2, CColor col)
{
	if (cxxx::cxRTL) {
		x1 = mWidth - x1;
		x2 = mWidth - x2;
	}		
	XRenderColor rcol = col;
	if (x1 == x2 || y1 == y2) {
		// special case: horizontal or vertical lines
		int x = min(x1, x2);
		int y = min(y2, y1);
		int w = std::abs(x2 - x1);
		int h = std::abs(y2 - y1);
		XRenderFillRectangle(cxDpy, PictOpSrc, GetRID(), &rcol, x, y, ++w, ++h); 
	} else {
		// arbitrary lines
		
		// conversion to XFixed 16.16
		x1 <<= 16;
		x2 <<= 16;
		y1 <<= 16;
		y2 <<= 16;
		
		// add half a pixel offset to y values
		y1 += 1 << 15;
		y2 += 1 << 15;
		
		XTrapezoid trap;
		
		if (y1 < y2) {
			trap.top = y1;
			trap.bottom = y2;
		} else {
			trap.top = y2;
			trap.bottom = y1;
		}
		
		trap.left.p1.x = x1;
		trap.left.p1.y = y1;
		trap.left.p2.x = x2 ;
		trap.left.p2.y = y2;
		x1 += 1 << 16;
		x2 += 1 << 16;
		trap.right.p1.x = x1;
		trap.right.p1.y = y1;
		trap.right.p2.x = x2;
		trap.right.p2.y = y2;
		
		XRenderFillRectangle(cxDpy, PictOpSrc, cxDefaultPic, &rcol, 0, 0, 1, 1);
 		XRenderCompositeTrapezoids (cxDpy, PictOpOver, cxDefaultPic, GetRID(), cxAlphaFormat, 0, 0, &trap, 1);
	}
}

/* Scrolls window contents by x, y pixels */
void CWindow::Scroll(int x, int y)
{
	int x1 = 0;
	int y1 = 0;
	int w = mWidth;
	int h = mHeight;

	if (x < 0) {
		x1 = -x;
		x = 0;
	} 
	if (y < 0) {
		y1 = -y;
		y = 0;
	}
	w -= x + x1;
	h -= y + y1;
	if ((w <= 0) || (h <= 0)) Redraw();
	else {
		// XRender does not generate Graphics Exposure Events, so we have to use the core protocol
		if (cxxx::cxRTL) XCopyArea(cxDpy, mXID, mXID, XDefaultGC(cxDpy, XDefaultScreen(cxDpy)), mWidth - w - x, y, w, h, mWidth - w - x1, y1); 
		else XCopyArea(cxDpy, mXID, mXID, XDefaultGC(cxDpy, XDefaultScreen(cxDpy)), x, y, w, h, x1, y1);
		if (x1) RedrawArea(0, 0, x1, mHeight);
		else RedrawArea(w, 0, x, mHeight);
		if (y1) RedrawArea(0, 0, mWidth, y1);
		else RedrawArea(0, h, mWidth, y);
	}
}

/* virtual */ bool CWindow::ConvertSelection(Atom sel __attribute__ ((unused)), Atom type __attribute__ ((unused)), vector<unsigned char>& ret __attribute__ ((unused)))
{
#ifdef CXXX_DEBUG
	cxError << "CWindow::ConvertSelection not implemented" << CError::Error;
#endif // CXXX_DEBUG
	return false;
}

void CWindow::DrawUniString(int x, int y, CColor col, const wstring& s, bool erase)
{
	XRenderColor rcol = col;
	if (erase) ClearArea(x, y, GetTextWidth(s), GetTextHeight());
	mFont->AddMissingGlyphs(s);
	XRenderFillRectangle(cxDpy, PictOpSrc, cxDefaultPic, &rcol, 0, 0, 1, 1);
	if (cxxx::cxRTL) x = mWidth - GetTextWidth(s) - x;
	XRenderCompositeString(cxDpy, PictOpOver, cxDefaultPic, GetRID(), cxAlphaFormat, mFont->GetGlyphset(), 0, 0, x, y + mFont->GetAscent(), s.c_str(), s.size());
}

int CWindow::GetCharWidth(wchar_t c) const
{
	return mFont->GetCharWidth(c);
}

void CWindow::SetFocus()
{
	if (IsEnabled()) GetTopLevelWindow()->SetFocusWindow(this);
}

void CWindow::Popup(CMenu& mnu)
{
	int x, y;
	GetPointerPos(x, y, NULL);
	CPopup* p = new CPopup(this, mnu);
	p->Position(x - 15, y - 15, 30, 30, CPopup::DOWN);
	p->Show();
}

void CWindow::SetFocusable(bool b)
{
	if (IsFocused()) GetTopLevelWindow()->SetFocusWindow(NULL);
	mWindowState[ISFOCUSABLE] = b;
}
