/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CWindow.h
  Info   : $Id: CWindow.h,v 1.46 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#ifndef CWindow_h
#define CWindow_h

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/CBidi.h>
#include <cxxx/CColor.h>
#include <cxxx/CSelection.h>
#include <cxxx/CIcon.h>
#include <cxxx/CFont.h>

// C++ headers
#include <map>
#include <bitset>

//Forward declarations
class CIcon;
class CTopLevelWindow;
class CMenu;
class CPopup;

class CWindow : public CMessageTransceiver
{
private:
    enum
    {
		// Geometry
        WIDTH = 400,
        HEIGHT = 100,
        MAX_WIDTH = 0x80000000,
        MAX_HEIGHT = 0x80000000
    };

protected:
    enum
    {
		// Standard event mask
        CXXXEVENTS   = ExposureMask | StructureNotifyMask | ButtonPressMask
    };

    typedef enum
    {
		// State flags
        ISVISIBLE,
        ISWAITING,
        ISFOCUSABLE,
        ISPOPUP,
        ISDNDCAPABLE,
		ISENABLED,
		ISFOCUSED,
		FLAG_COUNT
    } TWindowState;

public:
	
    typedef enum
    {
		// Frame styles
        FS_FLAT,
        FS_SUNKEN,
        FS_RAISED,
        FS_SUNKEN_THICK,
        FS_RAISED_THICK,
        FS_ENGRAVED,
        FS_SHEET,
		FS_TAB_SEL,
		FS_TAB_LEFT,
		FS_TAB_RIGHT,
		FS_TAB_NORMAL
    } TFrameStyle;

protected:
	// Type definitions
    typedef std::map<Window, CWindow*> TWindowMap;

public:
	// Constructor
	CWindow(CWindow* parent)
	: CMessageTransceiver(parent, 0),
	  mWindowState(0),
	  mX(0),
	  mY(0),
	  mWidth(WIDTH),
	  mHeight(HEIGHT),
	  mFont(cxxx::cxTextFont),
	  mParent(parent)
	{
		Enable(true);
		Init();
	}

	CWindow(CWindow* parent, unsigned msgid)
	: CMessageTransceiver(parent, msgid),
	  mWindowState(0),
	  mX(0),
	  mY(0),
	  mWidth(WIDTH),
	  mHeight(HEIGHT),
	  mFont(cxxx::cxTextFont),
	  mParent(parent)
	{
		Enable(true);
		Init();
	}

	// Destructor
	virtual ~CWindow();

	// Returns the top level window of this window
	CTopLevelWindow* GetTopLevelWindow()
	{
		CWindow* w = this;
		while (w->mParent != NULL) w = w->mParent;
		return reinterpret_cast<CTopLevelWindow*>(w);
	}

	// Makes the window visible
	void Show()
	{
		XMapWindow(cxxx::cxDpy, mXID);
	}

	// Makes the window invisible
	void Hide()
	{
		XUnmapWindow(cxxx::cxDpy, mXID);
	}

	// Sets the input focus to this window
	void SetFocus();

	// Returns the window from the window ID
    static CWindow* GetWindowByID(Window id)
    {
        TWindowMap::iterator it = mWindowMap.find(id);
        if(it == mWindowMap.end()) return NULL;
        else return it->second;
    }

	// Returns the width of the window
    unsigned GetWidth() const
	{
		return mWidth;
	}

	// Returns the height of the window
	unsigned GetHeight() const
	{
		return mHeight;
	}

	// Returns the x position of the window
	int GetXPos() const
	{
		if (cxxx::cxRTL) {
			if (!IsPopup() && mParent) return mParent->mWidth - mX;
			else return DisplayWidth(cxxx::cxDpy, DefaultScreen(cxxx::cxDpy)) - mWidth -mX;
		} else return mX;
	}

	// Returns the y position of the window
	int GetYPos() const
	{
		return mY;
	}

	// Returns true if the window is in waiting mode
	bool IsWaiting() const
	{
		return mWindowState.test(ISWAITING);
	}

	// Returns true if the window accepts the input focus
	bool IsFocusable() const
	{
		return mWindowState.test(ISFOCUSABLE);
	}

	// Returns true if the window is mapped by the X server
	bool IsVisible() const
	{
		return mWindowState.test(ISVISIBLE);
	}
	
	// Returns true if this is a popup window
	bool IsPopup() const
	{
		return mWindowState.test(ISPOPUP);
	}
	
	// Returns true if this window is drag 'n' drop enabled
	bool IsDnDAware() const
	{
		return mWindowState.test(ISDNDCAPABLE);
	}
	
	// Returns true if this window is enabled
	bool IsEnabled() const
	{
		return mWindowState.test(ISENABLED);
	}
	
	// Returns true if this window has the input focus
	bool IsFocused() const
	{
		return mWindowState.test(ISFOCUSED);
	}
	
	// Returns true if this is a child window of w
	bool IsChildOf(const CWindow* w) const
	{
		const CWindow* i = this;
		while ((i = i->mParent) != NULL) if (i == w) return true;
		return false;
	}
	
	// Returns the parent window of this window
	CWindow* GetParent() const
	{
		return mParent;
	}
	
	// Sets the font for this window
	void SetFont(CFont& font)
	{
		mFont = &font;
	}

	// Redraws a rectangular area
	void RedrawArea(int x, int y, unsigned w, unsigned h)
	{
		if ((w == 0) || (h == 0)) return;
		XExposeEvent e;
		e.x = x;
		e.y = y;
		e.width = w;
		e.height = h;
		e.window = mXID;
		e.type = Expose;
		ClearArea(e);
		OnExpose(e);
	}

	// Refreshes the window
    virtual void Redraw()
    {
	    RedrawArea(0, 0, GetWidth(), GetHeight());
    }
    // X event processing
	void HandleEvent(XEvent& e);

	virtual void OnExpose(XExposeEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnConfigure(XConfigureEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnMap(XMapEvent& e)
	{
		if (e.type == MapNotify) mWindowState.set(ISVISIBLE);
		else mWindowState.reset(ISVISIBLE);
	}

	virtual void OnLButtonDown(XButtonEvent& e)
	{
		OnButtonDown(e);
	}

	virtual void OnMButtonDown(XButtonEvent& e)
	{
		OnButtonDown(e);
	}

	virtual void OnRButtonDown(XButtonEvent& e)
	{
		OnButtonDown(e);
	}

	virtual void OnMouseWheelUp(XButtonEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnMouseWheelDown(XButtonEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnButtonDown(XButtonEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnDblClick(XButtonEvent& e)
	{
		if (e.button == 1)
			OnLButtonDown(e);
		else if (e.button == 3)
			OnRButtonDown(e);
		else
			OnButtonDown(e);
	}

	virtual void OnButtonUp(XButtonEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnEvent(XEvent& e __attribute__ ((unused)))
	{
	}
	
	virtual void OnFocus(bool in __attribute__ ((unused)))
	{
	}
	
	virtual void OnKey(wchar_t key __attribute__ ((unused)), unsigned mask __attribute__ ((unused)))
	{
	}
	
	virtual void OnAccelerator(wchar_t key __attribute__ ((unused)))
	{
	}
		
	// Draws a line from x1, y1 to x2, y2
	void DrawLine(int x1, int y1, int x2, int y2, CColor col);

	// Clears a rectangular area
	void ClearArea(int x, int y, int w, int h)
	{
		if (cxxx::cxRTL) x = mWidth - w - x;
		if (w > 0 && h > 0)	XClearArea(cxxx::cxDpy, mXID, x, y, w, h, false);
	}
	
	// Clears the area specified in e
	void ClearArea(XExposeEvent& e)
	{
		ClearArea(e.x, e.y, e.width, e.height);
	}
	
	// Clears the whole window
	void ClearWindow()
	{
		XClearWindow(cxxx::cxDpy, mXID);
	}
	
	// Draws a rectangle
	void DrawRectangle(int x, int y, int w, int h, CColor col);
	
	void DrawDottedRectangle(int x, int y, int w, int h, CColor col);
	
	// Fills a rectangle
	void FillRectangle(int x, int y, int w, int h, CColor col)
	{
		XRenderColor rcol = col;
		if (cxxx::cxRTL) x = mWidth - w - x;
		if (w > 0 && h > 0) XRenderFillRectangle(cxxx::cxDpy, PictOpSrc, mRID, &rcol, x, y, w, h);
	}
		
	// Draws a wide character string at position x, y in LTR direction
	void  DrawUniString(int x, int y, CColor col, const std::wstring& s, bool erase = true);
	
	// Draws a wide character string at position x, y
	void DrawString(int x, int y, CColor col, const std::wstring& s, bool erase = true)
	{
		if (cxxx::cxRTL) DrawUniString(x, y, col, cxxx::cxBidi->Convert(s), erase);
		else DrawUniString(x, y, col, s, erase);
	}		
	
	// Draws a narrow character string in the applications encoding at position x, y
	void DrawString(int x, int y, CColor col, const std::string& s, bool erase = true)
	{
		DrawString(x, y, col, cxxx::cxWiden(s), erase);
	}
	
	// Copies d to x, y of this window
	void CopyArea(Drawable d, int x, int y, int w, int h)
	{
		if (d == None) return;
		//XCopyArea(cxxx::cxDpy, d, mXID, cxxx::cxGCManager->GetGC(), 0, 0, w, h, x, y);
		Picture pic = XRenderCreatePicture(cxxx::cxDpy, d, cxxx::cxARGBFormat, 0, NULL);
		XRenderComposite(cxxx::cxDpy, PictOpSrc, pic, pic, mRID, 0, 0, 0, 0, x, y, w, h);
		XRenderFreePicture(cxxx::cxDpy, pic);
	}
	
	// Returns the text width of the first len characters of s 
	int GetTextWidth(const std::wstring& s, unsigned len) const
	{
		cxAssert(len <= s.size());
		if (len == 0) return 0;
		int width = 0;
		for (unsigned i = 0; i < len; ++i) width += GetCharWidth(s[i]);
		return width;
	}
	
	// Returns the text width of s
    int GetTextWidth(const std::wstring& s) const
	{
		return GetTextWidth(s, s.size());
	}
	
	// Returns the text height
	int GetTextHeight() const
	{
		return mFont->GetHeight();
	}
	
	// Returns the character width of c
	int GetCharWidth(wchar_t c) const;
	
	// Returns the average character width
	int GetAvgCharWidth()
	{
		return GetCharWidth('x');
	}
	
	// Changes the window attributes
	void ChangeAttributes(unsigned mask)
	{
		XChangeWindowAttributes(cxxx::cxDpy, mXID, mask, &mAttributes);
	}
	
	// Resizes the window
	void Resize(unsigned w, unsigned h)
	{
		cxAssert(w < MAX_WIDTH && h < MAX_HEIGHT);
		if ((w != GetWidth()) || (h != GetHeight())) {
			XResizeWindow(cxxx::cxDpy, mXID, w, h );
			mWidth = w;
			mHeight = h;
			if (cxxx::cxRTL) Move(mX, mY);
		}
	}
	
	// Moves the window
	void Move(int x, int y)
	{
		if (x != mX || y != mY || cxxx::cxRTL) {
			if (cxxx::cxRTL) {
				if (!IsPopup() && mParent) x = mParent->mWidth - mWidth - x;
				else x = DisplayWidth(cxxx::cxDpy, DefaultScreen(cxxx::cxDpy)) - x - mWidth;
			}
			mX = x;
			mY = y;
			XMoveWindow(cxxx::cxDpy, mXID, x, y);
		}
	}
	
	// Starts wait modus 
	void DisplayWaitCursor()
	{
#ifndef CXXX_DEBUG
		if (!IsVisible() || IsWaiting()) return ;
		if (XGrabPointer(cxxx::cxDpy, GetXID(), True, NoEventMask, GrabModeAsync, GrabModeAsync, None, mWaitCursor, CurrentTime ) == GrabSuccess ) mWindowState.set(ISWAITING);
#endif // CXXX_DEBUG
	}
	
	// Ends wait modus
	void UndisplayWaitCursor()
	{
		if (IsWaiting()) {
			XUngrabPointer(cxxx::cxDpy, CurrentTime);
			mWindowState.reset(ISWAITING);
		}
	}
	
    // Returns the mouse position relative to window rel
    void GetPointerPos(int& x, int& y, CWindow* rel)
    {
        Window wdummy;
        int idummy;
        unsigned udummy;
		if (rel == NULL) XQueryPointer(cxxx::cxDpy, mXID, &wdummy, &wdummy, &x, &y, &idummy, &idummy, &udummy);
        else XQueryPointer(cxxx::cxDpy, rel->mXID, &wdummy, &wdummy, &idummy, &idummy, &x, &y, &udummy);
		if (cxxx::cxRTL) {
			if (rel) x = rel->mWidth - x;
			else x = DisplayWidth(cxxx::cxDpy, DefaultScreen(cxxx::cxDpy)) - x;
		 }
	}

    // Translates the coordinates relative to window rel
    void TranslateCoordinates(int& x, int& y, CWindow* rel)
    {
	    int sx = x;
	    int sy = y;
	    Window dummy;
        Window dest;
        if (rel == NULL) dest = DefaultRootWindow(cxxx::cxDpy);
        else dest = rel->mXID;
		if (cxxx::cxRTL) sx = mWidth - sx;
	    int res = XTranslateCoordinates(cxxx::cxDpy, mXID, dest, sx, sy, &x, &y, &dummy);
		if (!res) cxxx::cxError << cxxx::lib::_("XTranslateCoordinates:\nFailed to translate coordinates", 1) << CError::Error;
 		if (cxxx::cxRTL) {
			if (rel) x = rel->mWidth - x;
			else x = DisplayWidth(cxxx::cxDpy, DefaultScreen(cxxx::cxDpy)) - x;
		 }
	}

	// Draws a frame
	void DrawFrame(int x, int y, int w, int h, TFrameStyle style, bool erase);

	// Draws a string containing a hot char
	void DrawHotString(int x, int y, CColor col, const std::string& s,  const CHotCharPos& hot, bool grayed = false, bool erase = true)
	{
		DrawHotString(x, y, col, cxxx::cxWiden(s), hot, grayed, erase);
	}

	void DrawHotString(int x, int y, CColor col, const std::wstring& s,  const CHotCharPos& hot, bool grayed = false, bool erase = true);
	
	// Draws strings that may occupy multiple lines
	void DrawMultiLineString(int x, int y, int w, CColor col, const std::string& s);

	// Draws an icon using the mask information and colorize with col
	void DrawIcon(int x, int y, const CIcon& p, CColor col, bool grayed = false);
	
	// Draws an icon
	void DrawIcon(int x, int y, const CIcon& p,bool grayed = false);
	
	// Scrolls the window
	void Scroll(int x, int y);
	
	// Displays a popup window
	void Popup(CMenu& mnu);
	
	//Makes this a focusable window
	void SetFocusable(bool b = true);
	
	// Returns the name of a
    static std::string GetAtomName(const Atom a)
	{
		char* name = XGetAtomName(cxxx::cxDpy, a);
		std::string ret = name;
		XFree(name);
		return ret;
	}
	
	// Tries to grab the selection
    bool GrabSelection(Atom sel, Time time)
    {
        return cxxx::cxSelManager->Grab(this, sel, time);
    }

	// Releases the selection
    void ReleaseSelection(Atom sel, Time time)
    {
        cxxx::cxSelManager->Release(sel, time);
    }

	// Returns the Selection owner
    CWindow* GetSelectionOwner(Atom sel, Time time)
    {
        return cxxx::cxSelManager->GetOwner(sel, time);
    }

	// Requests the selection
    bool RequestSelection(Atom sel, Atom typ, Atom prop, std::vector<unsigned char>& ret) {
        return cxxx::cxSelManager->Request(this, sel, typ, prop, ret);
    }

	// Converts the selection
    virtual bool ConvertSelection(Atom sel, Atom typ, std::vector<unsigned char>& ret);

	// Enables or disables this window
	virtual void Enable(bool enable)
	{
		mWindowState[ISENABLED] = enable;
	}

	// Recolors all windows in response to changes in the settings
	void RecolorWindows()
	{
		for (TWindowMap::iterator i = mWindowMap.begin(); i != mWindowMap.end(); ++i) i->second->Recolor();
	}	

	// Recolor this window
	virtual void Recolor();

protected:
	// Makes w the parent of this window
	void Reparent(CWindow* parent)
	{
		mParent = parent;
	}

	// Returns the atom with name s
    Atom GetAtomFromName(const std::string& s) const
	{
		return XInternAtom(cxxx::cxDpy, s.c_str(), False);
	}

	// Returns the xrender picture for this window
	Picture GetRID()
	{
		return mRID;
	}
		
protected:
    // flags
	std::bitset<FLAG_COUNT> mWindowState;
	// window position
	int mX;
	int mY;
	// window size
	unsigned mWidth;
	unsigned mHeight;
	XSetWindowAttributes mAttributes;
	static Cursor mWaitCursor;
    unsigned mEventMask;
    CFont* mFont;
private:
	void Init();
		
	void HLine(XRectangle& rect, int x, int y, unsigned len)
	{
		rect.x = x;
		rect.y = y;
		rect.width = len;
		rect.height = 1;
	}
		
	void VLine(XRectangle& rect, int x, int y, unsigned len)
	{
		rect.x = x;
		rect.y = y;
		rect.width = 1;
		rect.height = len;
	}
		
private:
	CWindow* mParent;
	static TWindowMap mWindowMap;
	Picture mRID;
	//
friend class CTopLevelWindow;
};

#endif // CWindow_h
