/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : CXZStreambuf.cpp
  Info   : $Id: CXZStreambuf.cpp,v 1.3 2012/06/26 08:05:52 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CXZStreambuf.h>

#ifdef CXXX_HAS_XZ

// C++ headers
#include <cstdio>
#include <algorithm>

// Namespace declarations
using std::streambuf;
using std::streamoff;
using std::streampos;
using std::streamsize;
using std::ios_base;
using std::min;
using std::copy;

CXZStreambuf::CXZStreambuf(std::streambuf& in, unsigned size)
: mIn(in),
  mBufferSize(size)
{
	mState = UNINITIALIZED;
	mInBuffer = new char[mBufferSize];
	mOutBuffer = new char[mBufferSize];
	if (!DecompressInit()) DecompressEnd();
}

/* virtual */ CXZStreambuf::~CXZStreambuf()
{
	DecompressEnd();
	delete[] mInBuffer;
	delete[] mOutBuffer;
}

/* virtual */ int CXZStreambuf::overflow(int c __attribute__ ((unused)))
{
	// write function disabled
	return EOF;
}

/* virtual */ streampos CXZStreambuf::seekoff(streamoff off, ios_base::seekdir way, ios_base::openmode which)
{
	streampos pos;
	if (way == ios_base::beg) pos = 0;
	else if (way == ios_base::cur) pos = mXZ.total_out;
	else if (way == ios_base::end) return -1; // way too expensive
	pos += off;
	return seekpos(pos, which);
}

/* virtual */ streampos CXZStreambuf::seekpos(streampos pos, ios_base::openmode which)
{
	uint64_t upos = static_cast<uint64_t>(pos);
	if ((which & ios_base::out) != 0) return -1;
	if ((which & ios_base::in) == 0) return -1;
	if (mXZ.total_out >= upos) {
		// pos has already been read
		if (static_cast<int>(mXZ.total_out - upos) <= egptr() - eback()) {
			setg(eback(), egptr() - (mXZ.total_out - upos), egptr());
			return pos;
		}
		if (!DecompressInit()) {
			DecompressEnd();
			return -1;
		}
	}
	// decompress until position
	while (mXZ.total_out < upos) {
		if (!Decompress(upos - mXZ.total_out)) break;
	}
	if (mXZ.total_out == upos) {
		setg(eback(), egptr(), egptr());
		return pos;
	} else return -1;
}

/* virtual */ streamsize CXZStreambuf::showmanyc()
{
	return egptr() - gptr();
}

/* virtual */ int CXZStreambuf::underflow()
{
	if (mState != GOOD) return EOF;
	if (gptr() < egptr()) return *gptr();
	while (egptr() == gptr()) {
		if (!Decompress(-1)) break;
	}
	if (gptr() == egptr()) return EOF;
	else return *gptr() & 0xff;
}

/* virtual */ streamsize CXZStreambuf::xsgetn(char* s, streamsize n)
{
	if (mState != GOOD || n == 0) return 0;
	unsigned ret = min(showmanyc(), n);
	if (ret != 0) {
		copy(gptr(), gptr() + ret, s);
		setg(eback(), gptr() + ret, egptr());
		if (ret != n) ret += xsgetn(s + ret, n - ret);	return ret;
	}
	while (egptr() == gptr()) {
		if (!Decompress(n)) break;
	}
	ret = showmanyc();
	cxAssert(ret <= n);
	if (ret == 0) return 0;
	copy(gptr(), egptr(), s);
	setg(eback(), egptr(), egptr());
	if (ret != n) ret += xsgetn(s + ret, n - ret);
	return ret;
}

// Initializes the compression library, returns true if successfull
bool CXZStreambuf::DecompressInit()
{
	if (mState != UNINITIALIZED) DecompressEnd();
	if (mIn.pubseekpos(0, ios_base::in) == -1) {
		// failed to rewind mIn
		return false;
	}
	// fill input buffer
	if (mIn.sgetn(mInBuffer, 6) != 6) return false;
	// set buffer pointers
	setg(mOutBuffer + 1, mOutBuffer + 1, mOutBuffer + 1);
	// initialize ZStream pointers
	mXZ.allocator = NULL;
	mXZ.next_in = reinterpret_cast<uint8_t*>(mInBuffer);
	mXZ.avail_in = 6;
	mXZ.total_out = 0;
	mXZ.next_out = reinterpret_cast<uint8_t*>(egptr());
	mXZ.avail_out = mBufferSize - 1;
	mXZ.reserved_ptr1 = NULL;
	mXZ.reserved_ptr2 = NULL;
	mXZ.reserved_int1 = 0;
	mXZ.reserved_int2 = 0;
	mXZ.reserved_enum1 = (lzma_reserved_enum)0;
	mXZ.reserved_enum2 = (lzma_reserved_enum)0;
#ifdef CXXX_LZMA_ORIGINAL
	// Lasse Collins' original declaration of lzma_strm declares these
	// members, but they lack in the debian port and maybe others
	mXZ.reserved_ptr3 = NULL;
	mXZ.reserved_ptr4 = NULL;
	mXZ.reserved_int3 = 0;
	mXZ.reserved_int4 = 0;
#endif //CXXX_LZMA_ORIGINAL			
	mXZ.internal = NULL;
	if (lzma_stream_decoder(&mXZ, UINT64_MAX, LZMA_TELL_UNSUPPORTED_CHECK | LZMA_CONCATENATED) == LZMA_OK)	{
		mState = GOOD;
		return true;
	} else return false;
}

// Decompresses a maximum of size characters, returns true if successfull
bool CXZStreambuf::Decompress(unsigned size)
{
	if (mState != GOOD) return false;
	// prepare input buffer
	if (!mXZ.avail_in) {
		mXZ.next_in = (uint8_t*)mInBuffer;
		mXZ.avail_in = mIn.sgetn(mInBuffer, mBufferSize);
		if (!mXZ.avail_in) mState = FINISH;
	}
	// prepare output buffer
	if (!mXZ.avail_out) {
		mOutBuffer[0] = *(egptr() - 1);
		mXZ.next_out = (uint8_t*)eback();
		mXZ.avail_out = mBufferSize - 1;
		setg(eback(), eback(), eback());
	}
	// honor size parameter
	if (mXZ.avail_out > size) mXZ.avail_out = size;
	// do the actual decompression
	int ret;
	
	if (mState == GOOD) {
		ret = lzma_code(&mXZ, LZMA_RUN);
		if (ret != LZMA_OK) mState = FAULT;
	} else {
		ret = lzma_code(&mXZ, LZMA_FINISH);
		if (ret == LZMA_STREAM_END) mState = END;
		else mState = FAULT;
	}

	// set streambuf pointers
	setg(eback(), gptr(), reinterpret_cast<char*>(mXZ.next_out));
	return mState != FAULT;
}

#endif // CXXX_HAS_XZ
