/*----------------------------------------------------------------------------
  Project: cxxx <http://cxplorer.sourceforge.net>
  File   : pnglite.c
  Info   : $Id: pnglite.c,v 1.7 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  This is a MODIFIED version of pnglite.c by Daniel Karling.
  CHANGES:
  CRC checking on file read has been disabled.
  Support for interlaced images has been added.
  Memory allocation is allways malloc / free; png_init has been removed.
  png_print_info has been removed; it is not needed by C++X
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>
  Copyright © 2007 Daniel Karling  <daniel.karling@gmail.com>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

/*  pnglite.c - pnglite library
    For conditions of distribution and use, see copyright notice in pnglite.h
*/

/* C headers */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* zlib header */
#include <zlib.h>

/* C++X headers */
#include <cxxx/pnglite.h>

/*
	These arrays are needed for displaying interlaced images
 	The image is loaded in seven passes.
*/

static unsigned png_starting_row[7] =  { 0, 0, 4, 0, 2, 0, 1 };
static unsigned png_starting_col[7] =  { 0, 4, 0, 2, 0, 1, 0 };
static unsigned png_row_increment[7] = { 8, 8, 8, 4, 4, 2, 2 };
static unsigned png_col_increment[7] = { 8, 8, 4, 4, 2, 2, 1 };


static size_t file_read(png_t* png, void* out, size_t size, size_t numel)
{
	size_t result;
	if(png->read_fun)
	{
		result = png->read_fun(out, size, numel, png->user_pointer);
	}
	else
	{
		if(!out)
		{
			result = fseek(png->user_pointer, (long)(size*numel), SEEK_CUR);
		}
		else
		{
			result = fread(out, size, numel, png->user_pointer);
		}
	}

	return result;
}

static size_t file_write(png_t* png, void* p, size_t size, size_t numel)
{
	size_t result;

	if(png->write_fun)
	{
		result = png->write_fun(p, size, numel, png->user_pointer);
	}
	else
	{
		result = fwrite(p, size, numel, png->user_pointer);
	}

	return result;
}

static int file_read_ul(png_t* png, unsigned *out)
{
	unsigned char buf[4];

	if(file_read(png, buf, 1, 4) != 4)
		return PNG_FILE_ERROR;

	if (out)
		*out = (buf[0]<<24) | (buf[1]<<16) | (buf[2]<<8) | buf[3];

	return PNG_NO_ERROR;
}

static int file_write_ul(png_t* png, unsigned in)
{
	unsigned char buf[4];

	buf[0] = (in>>24) & 0xff;
	buf[1] = (in>>16) & 0xff;
	buf[2] = (in>>8) & 0xff;
	buf[3] = (in) & 0xff;

	if(file_write(png, buf, 1, 4) != 4)
		return PNG_FILE_ERROR;

	return PNG_NO_ERROR;
}
	

static unsigned get_ul(unsigned char* buf)
{
	unsigned result;
	unsigned char foo[4];

	memcpy(foo, buf, 4);

	result = (foo[0]<<24) | (foo[1]<<16) | (foo[2]<<8) | foo[3];

	return result;
}

static unsigned set_ul(unsigned char* buf, unsigned in)
{
	buf[0] = (in>>24) & 0xff;
	buf[1] = (in>>16) & 0xff;
	buf[2] = (in>>8) & 0xff;
	buf[3] = (in) & 0xff;

	return PNG_NO_ERROR;
}

static int png_get_bpp(png_t* png)
{
	int bpp;

	switch(png->color_type)
	{
	case PNG_GREYSCALE:
		bpp = 1; break;
	case PNG_TRUECOLOR:
		bpp = 3; break;
	case PNG_INDEXED:
		bpp = 1; break;
	case PNG_GREYSCALE_ALPHA:
		bpp = 2; break;
	case PNG_TRUECOLOR_ALPHA:
		bpp = 4; break;
	default:
		return PNG_FILE_ERROR;
	}

	bpp *= png->depth/8;
	if (!bpp) bpp = 1;

	return bpp;
}

static int png_read_ihdr(png_t* png)
{
	unsigned length;
	unsigned char ihdr[13+4];		 /* length should be 13, make room for type (IHDR) */

	file_read_ul(png, &length);

	if(length != 13)
		return PNG_FILE_ERROR;

	if(file_read(png, ihdr, 1, 13+4) != 13+4)
		return PNG_EOF_ERROR;

	if (file_read_ul(png, NULL) != PNG_NO_ERROR)
		return PNG_EOF_ERROR;

	png->width = get_ul(ihdr+4);
	png->height = get_ul(ihdr+8);
	png->depth = ihdr[12];
	png->color_type = ihdr[13];
	png->compression_method = ihdr[14];
	png->filter_method = ihdr[15];
	png->interlace_method = ihdr[16];
	if(png->interlace_method)
	{	
		int  i;
		for(i = 0; i < 7; i++)
		{
			png->sub_image_width[i] = (png->width+png_col_increment[i] - png_starting_col[i]-1) / png_col_increment[i];
			png->sub_image_height[i] = (png->height+png_row_increment[i] - png_starting_row[i]-1) / png_row_increment[i];
		}
	}		

	return PNG_NO_ERROR;
}

static int png_write_ihdr(png_t* png)
{
	unsigned char ihdr[13+4];
	unsigned char *p = ihdr;
	unsigned crc;
	
	file_write(png, "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A", 1, 8);

	file_write_ul(png, 13);
    
	*p = 'I';			p++;
	*p = 'H';			p++;
	*p = 'D';			p++;
	*p = 'R';			p++;
	set_ul(p, png->width);	p+=4;
	set_ul(p, png->height);	p+=4;
	*p = png->depth;			p++;
	*p = png->color_type;	p++;
	*p = 0;				p++;
	*p = 0;				p++;
	*p = 0;				p++;

	file_write(png, ihdr, 1, 13+4);

	crc = crc32(0L, 0, 0);
	crc = crc32(crc, ihdr, 13+4);

	file_write_ul(png, crc);

	return PNG_NO_ERROR;
}

int png_open_read(png_t* png, png_read_callback_t read_fun, void* user_pointer)
{
	char header[8];
	int result;
	memset(png, 0, sizeof(png_t));
	png->read_fun = read_fun;
	png->write_fun = 0;
	png->user_pointer = user_pointer;

	if(!read_fun && !user_pointer)
		return PNG_WRONG_ARGUMENTS;

	if(file_read(png, header, 1, 8) != 8)
		return PNG_EOF_ERROR;

	if(memcmp(header, "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A", 8) != 0)
		return PNG_HEADER_ERROR;

	result = png_read_ihdr(png);

	png->bpp = (unsigned char)png_get_bpp(png);

	return result;
}

void png_end_read(png_t* png)
{
	if(png->png_data)
	{
		free(png->png_data);
		png->png_data = NULL;
		png->png_datalen = 0;
	}

	if(png->png_plte)
	{
		free(png->png_plte);
		png->png_plte = NULL;
		png->png_pltesize = 0;
	}
	png->width = 0;
	png->height = 0;
	
	png->color_type = 0;
	png->compression_method = 0;
	png->filter_method = 0;
	png->interlace_method = 0;
	png->bpp = 0;
	png->transparency = 0;
}

int png_open_write(png_t* png, png_write_callback_t write_fun, void* user_pointer)
{
	png->write_fun = write_fun;
	png->read_fun = 0;
	png->user_pointer = user_pointer;

	if(!write_fun && !user_pointer)
		return PNG_WRONG_ARGUMENTS;

	return PNG_NO_ERROR;
}

int png_open_file_read(png_t *png, const char* filename)
{
	FILE* fp = fopen(filename, "rb");

	if(!fp)
		return PNG_FILE_ERROR;

	return png_open_read(png, 0, fp);
}

int png_open_file_write(png_t *png, const char* filename)
{
	FILE* fp = fopen(filename, "wb");

	if(!fp)
		return PNG_FILE_ERROR;

	return png_open_write(png, 0, fp);
}

int png_open_file(png_t *png, const char* filename)
{
	return png_open_file_read(png, filename);
}

int png_close_file(png_t* png)
{
	fclose(png->user_pointer);

	return PNG_NO_ERROR;
}

static int png_init_deflate(png_t* png, unsigned char* data, int datalen)
{
	z_stream *stream;
	png->zs = malloc(sizeof(z_stream));

	stream = png->zs;

	if(!stream)
		return PNG_MEMORY_ERROR;

	memset(stream, 0, sizeof(z_stream));

	if(deflateInit(stream, Z_DEFAULT_COMPRESSION) != Z_OK)
		return PNG_ZLIB_ERROR;

	stream->next_in = data;
	stream->avail_in = datalen;

	return PNG_NO_ERROR;
}

static int png_init_inflate(png_t* png)
{
	z_stream *stream;
	png->zs = malloc(sizeof(z_stream));

	stream = png->zs;

	if(!stream)
		return PNG_MEMORY_ERROR;

		memset(stream, 0, sizeof(z_stream));
	if(inflateInit(stream) != Z_OK)
		return PNG_ZLIB_ERROR;

	stream->next_out = png->png_data;
	stream->avail_out = png->png_datalen;

	return PNG_NO_ERROR;
}

static int png_end_deflate(png_t* png)
{
	z_stream *stream = png->zs;

	if(!stream)
		return PNG_MEMORY_ERROR;

	deflateEnd(stream);

	free(png->zs);
	png->zs = 0;

	return PNG_NO_ERROR;
}

static int png_end_inflate(png_t* png)
{
	z_stream *stream = png->zs;

	if(!stream)
		return PNG_MEMORY_ERROR;

	if(inflateEnd(stream) != Z_OK)
	{
		printf("ZLIB says: %s\n", stream->msg);
		return PNG_ZLIB_ERROR;
	}

	free(png->zs);
	png->zs = NULL;

	return PNG_NO_ERROR;
}

static int png_inflate(png_t* png, char* data, int len)
{
	int result;
	z_stream *stream = png->zs;

	if(!stream)
		return PNG_MEMORY_ERROR;

	stream->next_in = (unsigned char*)data;
	stream->avail_in = len;
	
	result = inflate(stream, Z_SYNC_FLUSH);

	if(result != Z_STREAM_END && result != Z_OK)
	{
		printf("%s\n", stream->msg);
		return PNG_ZLIB_ERROR;
	}

	if(stream->avail_in != 0)
		return PNG_ZLIB_ERROR;

	return PNG_NO_ERROR;
}

static int png_deflate(png_t* png, char* outdata, int outlen, int *outwritten)
{
	int result;

	z_stream *stream = png->zs;


	if(!stream)
		return PNG_MEMORY_ERROR;

	stream->next_out = (unsigned char*)outdata;
	stream->avail_out = outlen;

	result = deflate(stream, Z_SYNC_FLUSH);

	*outwritten = outlen - stream->avail_out;

	if(result != Z_STREAM_END && result != Z_OK)
	{
		printf("%s\n", stream->msg);
		return PNG_ZLIB_ERROR;
	}

	return result;
}

static int png_write_idats(png_t* png, unsigned char* data)
{
	unsigned char *chunk;
	unsigned long written;
	unsigned long crc;
	unsigned size = png->width * png->height * png->bpp + png->height;
	
	(void)png_init_deflate;
	(void)png_end_deflate;
	(void)png_deflate;

	chunk = malloc(size);
	memcpy(chunk, "IDAT", 4);
	
	written = size;
	compress(chunk+4, &written, data, size);
	
	crc = crc32(0L, Z_NULL, 0);
	crc = crc32(crc, chunk, written+4);
	set_ul(chunk+written+4, crc);
	file_write_ul(png, written);
	file_write(png, chunk, 1, written+8);
	free(chunk);

	file_write_ul(png, 0);
	file_write(png, "IEND", 1, 4);
	crc = crc32(0L, (const unsigned char *)"IEND", 4);
	file_write_ul(png, crc);
	
	return PNG_NO_ERROR;
}

static int png_read_idat(png_t* png, unsigned firstlen) 
{
	char type[4];
	char *chunk;
	int result;
	unsigned length = firstlen;
	unsigned old_len = length;

	chunk = malloc(firstlen); 

	result = png_init_inflate(png);

	if(result != PNG_NO_ERROR)
	{
		png_end_inflate(png);
		free(chunk); 
		return result;
	}

	do
	{
		if(file_read(png, chunk, 1, length) != length)
		{
			png_end_inflate(png);
			free(chunk); 
			return PNG_FILE_ERROR;
		}

		if(file_read_ul(png, NULL) != PNG_NO_ERROR)
			return PNG_EOF_ERROR;

		result = png_inflate(png, chunk, length);

		if(result != PNG_NO_ERROR) break;
		
		file_read_ul(png, &length);

		if(length > old_len)
		{
			free(chunk); 
			chunk = malloc(length); 
			old_len = length;
		}

		if(file_read(png, type, 1, 4) != 4)
		{
			result = PNG_FILE_ERROR;
			break;
		}

	}while(memcmp(type, "IDAT", 4) == 0);

	if(memcmp(type, "IEND", 4) == 0)
		result = PNG_DONE;
	else
		file_read(png, 0, 1, length + 4);
		
	free(chunk);
	png_end_inflate(png);

	return result;
}

static int png_process_chunk(png_t* png)
{
	int result = PNG_NO_ERROR;
	unsigned i;
	char type[4];
	unsigned length;

	file_read_ul(png, &length);

	if(file_read(png, &type, 1, 4) != 4)
		return PNG_FILE_ERROR;

	if(memcmp(type, "IDAT", 4) == 0) /* if we found an idat, all other idats should be followed with no other chunks in between */
	{
		png->png_datalen = png->width * png->height * png->bpp;
		
		if(png->interlace_method)
		{
			for(i = 0; i < 7; i++)
			{
				png->png_datalen += png->sub_image_height[i];
			}
		}
		else png->png_datalen += png->height;

		png->png_data = malloc(png->png_datalen);
		
		if(!png->png_data)
			return PNG_MEMORY_ERROR;

		return png_read_idat(png, length);
	}
	else if(memcmp(type, "IEND", 4) == 0)
	{
		if(png->png_data)
			return PNG_DONE;
		else
			/* missing idat */
			return PNG_FILE_ERROR;
	}
	else if (memcmp(type, "PLTE", 4) == 0)
	{
		png->png_pltesize = length / 3;
		png->png_plte = malloc(length + png->png_pltesize);
		for (i = 0; i < png->png_pltesize * 4; i += 4)
		{
			file_read(png, &png->png_plte[i], 1, 3);
			png->png_plte[i + 3] = 0xff;
		}			
		file_read_ul(png, NULL);
	}
	else if (memcmp(type, "tRNS", 4) == 0)
	{
		if(png->color_type == PNG_INDEXED)
		{
			for (i= 0; i < length; i++)
				file_read(png, &png->png_plte[4 * i + 3], 1, 1);
			file_read_ul(png, NULL);
		}
		else
		{
			png->transparency = 1;
			if(length > 6 )
				return PNG_FILE_ERROR;
			file_read(png, png->png_trns, 1, length);
			file_read_ul(png, NULL);
			if (png->depth == 8)
			{
				png->png_trns[0] = png->png_trns[1];
				png->png_trns[1] = png->png_trns[3];
				png->png_trns[2] = png->png_trns[5];
				memset(png->png_trns + 3, 0, 3);
			}
		}
	}
	else 
	{
		file_read(png, 0, 1, length + 4);		/* unknown chunk */
	}

	return result;
}

static void png_filter_sub(int stride, unsigned char* in, unsigned char* out, int len)
{
	int i;
	unsigned char a = 0;

	for(i = 0; i < len; i++)
	{
		if(i >= stride)
			a = out[i - stride];
		
		out[i] = in[i] + a;
	}
}

static void png_filter_up(unsigned char* in, unsigned char* out, unsigned char* prev_line, int len)
{
	if(prev_line) 
	{ 
		int  i;
		for(i = 0; i < len; i++) 
		    out[i] = in[i] + prev_line[i]; 
	} 
	else 
	        memcpy(out, in, len);
}

static void png_filter_average(int stride, unsigned char* in, unsigned char* out, unsigned char* prev_line, int len)
{
	int i;
	unsigned char a = 0;
	unsigned char b = 0;
	unsigned int sum = 0;

	for(i = 0; i < len; i++)
	{
		if(prev_line)
			b = prev_line[i];

		if(i >= stride)
			a = out[i - stride];

		sum = a;
		sum += b;

		out[i] = (char)(in[i] + sum/2);
	}
}

static unsigned char png_paeth(unsigned char a, unsigned char b, unsigned char c)
{
	int p = (int)a + b - c;
	int pa = abs(p - a);
	int pb = abs(p - b);
	int pc = abs(p - c);

	int pr;

	if(pa <= pb && pa <= pc)
		pr = a;
	else if(pb <= pc)
		pr = b;
	else
		pr = c;

	return (char)pr;
}

static void png_filter_paeth(int stride, unsigned char* in, unsigned char* out, unsigned char* prev_line, int len)
{
	int i;
	unsigned char a;
	unsigned char b;
	unsigned char c;

	for(i = 0; i < len; i++)
	{
		if(prev_line && i >= stride)
		{
			a = out[i - stride];
			b = prev_line[i];
			c = prev_line[i - stride];
		}
		else
		{
			if(prev_line)
				b = prev_line[i];
			else
				b = 0;
	
			if(i >= stride)
				a = out[i - stride];
			else
				a = 0;

			c = 0;
		}

		out[i] = in[i] + png_paeth(a, b, c);
	}
}

static int png_unfilter(png_t* png, unsigned char* in, unsigned char* out, unsigned width, unsigned height)
{
	unsigned char* prev = NULL;
	
	if (png->depth > 4)
		width *= png->bpp;
	else
	{
		width *= png->depth;
		width += 7;
		width /= 8;
	}
	
	while(height--)
	{
		unsigned char filter = *in++;

		switch(filter)
		{
		case 0: /* none */
			memcpy(out, in, width);
			break;
		case 1: /* sub */
			png_filter_sub(png->bpp, in, out, width);
			break;
		case 2: /* up */
			png_filter_up(in, out, prev, width);
			break;
		case 3: /* average */
			png_filter_average(png->bpp, in, out, prev, width);
			break;
		case 4: /* paeth */
			png_filter_paeth(png->bpp, in, out, prev, width);
			break;
		default:
			return PNG_UNKNOWN_FILTER;
		}

		prev = out;
		in += width;
		out += width;
	}
	
	return PNG_NO_ERROR;
}

int png_expand(png_t* png, unsigned char* data, unsigned width, unsigned height)
{
	unsigned char* out = data + width * height;
	data += height * ((width * png->depth + 7) / 8);
	
	unsigned shift0 = (8 - width * png->depth) % 8;
	unsigned mask = (1 << png->depth) - 1;
	unsigned mult = 255 / mask;

	unsigned row;
	unsigned col;
	unsigned shift;

	unsigned char c;
	unsigned char tmp;

	c = *--data;
	
	for (row = 0; row < height; row++)
	{
		shift = shift0;
		for (col = 0; col < width; col++)
		{
			tmp = (c >> shift) & mask;
			if (png->color_type == PNG_GREYSCALE)
				tmp *= mult;
			else if (tmp >= png->png_pltesize)
			{
				png_end_read(png);
				return PNG_FILE_ERROR;
			}
			*--out = tmp;
			shift += png->depth;
			if (shift == 8)
			{
				c = *--data;
				shift = 0;
			}
		}
	}
	return PNG_NO_ERROR;
}

int png_get_data(png_t* png, unsigned char* data)
{
	int result = PNG_NO_ERROR;
	unsigned char* in;
	unsigned char* out;
	unsigned i;

	while(result == PNG_NO_ERROR)
	{
		result = png_process_chunk(png);
	}

	if(result != PNG_DONE)
	{
		png_end_read(png); 
		return result;
	}

	/* unfilter */
	if (png->interlace_method)
	{
		in = png->png_data;
		out = data;
		for (i = 0; i < 7; i++)
		{
			if (png->sub_image_width[i] && png->sub_image_height[i])
			{
				result = png_unfilter(png, in, out, png->sub_image_width[i], png->sub_image_height[i]);
				if(result != PNG_NO_ERROR)
					break;
				if (png->depth > 4)
				{
					in += png->sub_image_width[i] * png->sub_image_height[i] * png->bpp + png->sub_image_height[i];
				}
				else
				{
					in += png->sub_image_height[i] * ((png->sub_image_width[i] * png->depth + 7) / 8) + png->sub_image_height[i];
				}
				out += png->sub_image_width[i] * png->sub_image_height[i] * png->bpp;
			}
		}
	}
	else
	{
		result = png_unfilter(png, png->png_data, data, png->width, png->height);
	}
	
	/* expand to one byte per pixel */
	if (result == PNG_NO_ERROR && png->depth < 8)
	{
		if (png->interlace_method)
		{
			in = data;
			for (i = 0; i < 7; i++)
			{
				result = png_expand(png, in, png->sub_image_width[i], png->sub_image_height[i]);
				in += png->sub_image_height[i] * png->sub_image_width[i];
			}
		}
		else
		{
			result = png_expand(png, data, png->width, png->height);
		}
	}
	
	/* deinterlace the image */
	if (result == PNG_NO_ERROR && png->interlace_method)
	{
		memcpy(png->png_data, data, png->width * png->height * png->bpp);
		
		unsigned row;
		unsigned col;
		unsigned i;
		in = png->png_data;
		
		for (i = 0; i < 7; i++)
		{
			for (row = png_starting_row[i]; row < png->height; row += png_row_increment[i])
			{
				for (col = png_starting_col[i]; col < png->width; col += png_col_increment[i])
				{
					out = data + (row * png->width + col) * png->bpp;
					memcpy(out, in, png->bpp);
					in += png->bpp;
				}
			}
		}
	}
	
	if(result != PNG_NO_ERROR)
	{
		png_end_read(png); 
		return result;
	}
				
	free(png->png_data);
	png->png_data = 0;

	return result;
}

int png_set_data(png_t* png, unsigned width, unsigned height, char depth, int color, unsigned char* data)
{
	unsigned i;
	unsigned char *filtered;
	png->width = width;
	png->height = height;
	png->depth = depth;
	png->color_type = color;
	png->bpp = png_get_bpp(png);

	filtered = malloc(width * height * png->bpp + height);

	for(i = 0; i < png->height; i++)
	{
		filtered[i*png->width*png->bpp+i] = 0;
		memcpy(&filtered[i*png->width*png->bpp+i+1], data + i * png->width*png->bpp, png->width*png->bpp);
	}

	/* png_filter(png, filtered); */
	png_write_ihdr(png);
	png_write_idats(png, filtered);
	
	free(filtered);
	return PNG_NO_ERROR;
}

char* png_error_string(int error)
{
	switch(error)
	{
	case PNG_NO_ERROR:
		return "No error";
	case PNG_FILE_ERROR:
		return "Unknown file error.";
	case PNG_HEADER_ERROR:
		return "No PNG header found. Are you sure this is a PNG?";
	case PNG_IO_ERROR:
		return "Failure while reading file.";
	case PNG_EOF_ERROR:
		return "Reached end of file.";
	case PNG_CRC_ERROR:
		return "CRC or chunk length error.";
	case PNG_MEMORY_ERROR:
		return "Could not allocate memory.";
	case PNG_ZLIB_ERROR:
		return "zlib reported an error.";
	case PNG_UNKNOWN_FILTER:
		return "Unknown filter method used in scanline.";
	case PNG_DONE:
		return "PNG done";
	case PNG_NOT_SUPPORTED:
		return "The PNG is unsupported by pnglite, too bad for you!";
	case PNG_WRONG_ARGUMENTS:
		return "Wrong combination of arguments passed to png_open. You must use either a read_function or supply a file pointer to use.";
	default:
		return "Unknown error.";
	};
}

