/*----------------------------------------------------------------------------
  Project: libintlize <http://intlize.sourceforge.net>
  File   : i5e.c
  Info   : $Id: i5e.c,v 1.5 2012/03/18 20:31:35 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2010 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  software  is  provided  'as-is',  without  any  express  or  implied
  warranty.  In  no  event  will  the authors be held liable for any damages
  arising from the use of this software.

  Permission  is  granted  to  anyone  to use this software for any purpose,
  including  commercial  applications,  and  to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The  origin  of  this software must not be misrepresented; you must not
     claim  that  you  wrote the original software. If you use this software
     in  a  product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

----------------------------------------------------------------------------*/

#include "i5e.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ICONV
#include <iconv.h>
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* replaces del with ins in buf; returns 0 in case of success; 1 if del was not 
found in buf */
int replace(char *buf, const char *ins, const char *del)
{
    int diff;
	char * pos = strstr(buf, del);
	if (pos == NULL) return 1;
	diff = strlen(ins) - strlen(del);
	pos += strlen(del);
	if (diff > 0) {
		char *src = buf + strlen(buf);
		char *dst = src + diff;
		while (src >= pos) *dst-- = *src--;
	} else if (diff < 0) {
		char *src = pos;
		char *dst = src + diff;
		while (*src) *dst++ = *src++;
		*dst = *src; 
	}
	memcpy(pos - strlen(del), ins, strlen(ins));
	return 0;
}

/* appends str2 to str1 */
/* returns NULL in case of error */
char* append(char* str1, char* str2, int str2len)
{
	char* str_ret;
	int str1len;
	if (str1) str1len = strlen(str1);
	else str1len = 0;
	str_ret = realloc(str1, str1len + str2len + 1);
	if (str_ret) {
		memcpy(str_ret + str1len, str2, str2len);
		str_ret[str1len + str2len] = '\0';
	}
	return str_ret;
}			

/* replaces %L with lang in the colon-seperated path path
 returns 0 in case of success */
char* get_path(const char *lang, const char *path)
{
	int path_len = strlen(path);
	
	/* allocate sufficient space for the returned path */
	char* ret_path = malloc(path_len + strlen(lang) + 1);
	
	/* pointers to the begin and end of path */
	const char *beg = path;
	const char *end = path + path_len;

	/* sz is the size of the colon seperated path elements */
	int sz;
	
	/* temporary language variable */
	char tmp_lang[6];
	
	/* allocation failed */
	if (ret_path == NULL) return NULL;
	
	/* iterate through the path */
	do {
		const char *tmp = strchr(beg, ':');
		if (tmp) sz = tmp - beg;
		else sz = strlen(beg);
		memcpy(ret_path, beg, sz);
		ret_path[sz] = '\0';
		/* try all characters of lang */
		if (replace(ret_path, lang, "%L") == 0) {
			if (access(ret_path, R_OK) == 0) return ret_path;
			/* no match: restore the path */
			memcpy(ret_path, beg, sz);
			ret_path[sz] = '\0';
		}
		strncpy(tmp_lang, lang, 5);
		/* try with lang truncated to 5 characters */
		if (strlen(lang) > 5) {
			tmp_lang[5] = '\0';
			if (replace(ret_path, tmp_lang, "%L") == 0) {
				if (access(ret_path, R_OK) == 0) return ret_path;
				/* no match: restore the path */
				memcpy(ret_path, beg, sz);
				ret_path[sz] = '\0';
			}
		}
		/* try with lang truncated to 2 characters */
		if (strlen(lang) >= 2) {
			tmp_lang[2] = '\0';
			if (replace(ret_path, tmp_lang, "%L") == 0) {
				if (access(ret_path, R_OK) == 0) return ret_path;
			}
		}
		/* next path element */
		beg += sz + 1;
	} while (beg < end);
	
	/* give up */
	free(ret_path);
	return NULL;
}

/* the encoding table */
static const char* encoding_table[] =
{
	"",              /* 0 */
	"ISO-8859-1",    /* 1 */
	"ISO-8859-2",    /* 2 */
	"ISO-8859-3",    /* 3 */
	"ISO-8859-4",    /* 4 */
	"ISO-8859-5",    /* 5 */
	"ISO-8859-6",    /* 6 */
	"ISO-8859-7",    /* 7 */
	"ISO-8859-8",    /* 8 */
	"ISO-8859-9",    /* 9 */
	"ISO-8859-10",   /* 10 */
	"ISO-8859-11",   /* 11 */
	"ISO-8859-12",   /* 12 */
	"ISO-8859-13",   /* 13 */
	"ISO-8859-14",   /* 14 */
	"ISO-8859-15",   /* 15 */
	"ISO-8859-16",   /* 16 */
	"KOI8-R",        /* 17 */
	"KOI8-U",        /* 18 */
	"KOI8-T",        /* 19 */
	"ASCII" ,        /* 20 */
	""      ,        /* 21 */
	""      ,        /* 22 */
	""      ,        /* 23 */
	""      ,        /* 24 */
	""      ,        /* 25 */
	""      ,        /* 26 */
	""      ,        /* 27 */
	""      ,        /* 28 */
	""      ,        /* 29 */
	""      ,        /* 30 */
	""      ,        /* 31 */
	"UTF-8"          /* 32 */
};

/* returns the charset name of cs */
const char* get_charset(unsigned cs)
{
	if (cs > 32) cs = 0;
	return encoding_table[cs];
}

#ifdef HAVE_ICONV
/* converts msg from the translation file's charset to that of the application */
char* convert(iconv_t conv, char* msg)
{
	static char buf[256];
	if (conv != (iconv_t)(-1)) {
		char *ret = NULL;
		size_t inleft = strlen(msg);
		size_t outleft = 256;
#ifdef HAVE_BSD_ICONV
        const char *inbegin = msg;
#else
        char *inbegin = msg;
#endif
		char *outbegin = buf;
		int error = 0;
		while (!error) {
			error = iconv(conv, &inbegin, &inleft, &outbegin, &outleft) == (size_t)(-1);
			if (!error) break;
			else if (errno == E2BIG) {
				ret = append(ret, buf, 256 - outleft);
				outleft = 256;
				outbegin = buf;
				error = 0;
			} else if (errno == EILSEQ) {
				ret = append(ret, buf, 256 - outleft);
				outbegin = buf;
				ret = append(ret, "?", 1);
				++inbegin;
				--inleft;
				error = 0;
			}
		}
		if (error) {
			inleft = 0;
			iconv(conv, NULL, &inleft, NULL, &inleft);
		} else {
			ret = append(ret, buf, 256 - outleft);
		}
		return ret;
	} else return msg;
}
#endif

/* Creates a catalog, returns NULL in case of an error */
i5e_catalog_t i5e_create_catalog()
{
	/* allocate the struct */
	i5e_catalog_t catalog = malloc(sizeof(i5e_catalog_struct_t));
	if (catalog) {
		/* clear all catalog fields */
		catalog->msg = NULL;
		catalog->msg_cnt = 0;
		catalog->charset = 0;
		catalog->id = NULL;
		catalog->buf = NULL;
		catalog->buf_end = NULL;
		catalog->converter = NULL;
	}
	return catalog;
}

/* Releases all storage associated with the catalog */
void i5e_free_catalog(i5e_catalog_t catalog)
{
	i5e_close(catalog);
	free(catalog);
}

/* openes the catalog catalog using language lang and the path path to the message location
   if charset is not NULL, the messages will be converted to that charset */
int i5e_open(i5e_catalog_t catalog, const char *lang, const char *path, const char *charset)
{
#ifndef HAVE_ICONV
	if (charset) return 9;
#endif	
	/* Variables */
	/* The file name of the translation file */
	char *fn;
	/* The input stream */
	FILE *in;
	/* The size of the translation file */
	int sz;
	/* A temporary read buffer */
	int c;
	/* A temporary pointer to the read buffer */
	char *buf_ptr;
	/* Allocated size of the message index */
	int msg_sz = 0;
	/* The name of the charset */
	const char* cs;
	
	if (catalog == NULL) return 6;
	if (lang == NULL || *lang == '\0') return 6;
	if (path == NULL || *path == '\0') return 6;
	
	fn = get_path(lang, path);
	if (!fn) return 1;
	/* found translation file at fn */		
	in = fopen(fn, "r");
	free(fn);
	if (in == NULL) return 2;
	/* get the size of the file */
  	fseek (in , 0 , SEEK_END);
  	sz = ftell (in);
  	rewind (in);
	
	/* allocate the space */
	catalog->buf = malloc(sz);
	if (catalog->buf == NULL) {
		fclose(in);
		return 3;
	}
	
	/* pointer to the end of buf */
	catalog->buf_end = catalog->buf + sz;
	
	/* read the file magic */
	fread(catalog->buf, 1, 8, in);
	if (strcmp(catalog->buf, "intlize") != 0) {
		fclose(in);
		i5e_close(catalog);
		return 4;
	}
	
	/* read the charset */
	c = fgetc(in);
	catalog->charset = c;
	
	/* read the version id */
	buf_ptr = catalog->buf;
	while (c != 0 && c != EOF) {
		c = fgetc(in);
		*buf_ptr++ = c;
	}
	if (c == EOF) {
		fclose(in);
		i5e_close(catalog);
		return 5;
	}
	if (*catalog->buf != '\0') {
		catalog->id = strdup(catalog->buf);
		if (catalog->id == NULL) {
			fclose(in);
			i5e_close(catalog);
			return 3;
		}
	}
	
	/* read the messages */
	/* read pointer */
	buf_ptr = catalog->buf;
	do {
		if (c == 0) {
			/* allocate the storage for catalog->msg on the fly */
			if (catalog->msg_cnt == msg_sz) {
				msg_sz += 64;
				catalog->msg = realloc(catalog->msg, msg_sz * sizeof(char*));
				if (catalog->msg == NULL) {
					fclose(in);
					i5e_close(catalog);
					return 3;
				}
			}
			catalog->msg[catalog->msg_cnt++] = buf_ptr;
		}
		c = fgetc(in);
		*buf_ptr++ = (char)c;
	} while (c != EOF);
	fclose(in);
	
	/* msg_cnt is off by 1; correct this */
	catalog->msg_cnt--;
#ifdef HAVE_ICONV	
	/* initialize iconv */
	if (charset && *charset) {
		cs = get_charset(catalog->charset);
		if (*cs == '\0') return 7;
		catalog->converter = malloc(sizeof(iconv_t));
		if (catalog->converter == NULL) {
			i5e_close(catalog);
			return 3;
		}
		iconv_t *ic = catalog->converter;
		*ic = iconv_open(charset, cs);
		if (*ic == (iconv_t)(-1)) {
			i5e_close(catalog);
			return 8;
		}
	}
#endif
	return 0;
}

/* clears all fields in the catalog and frees all memory associated with it */
void i5e_close(i5e_catalog_t catalog)
{
	if (catalog->msg) {
		int i;
		/* delete the converted messages first, they aren't accessible later */
		for (i = 0; i < catalog->msg_cnt; ++i) {
			if (catalog->msg[i] < catalog->buf || catalog->msg[i] >= catalog->buf_end) free(catalog->msg[i]);
		}
		/* delete the pointers to the messages */
		free(catalog->msg);
		catalog->msg = NULL;
	}
	if (catalog->id) {
		/* delete the id field */
		free(catalog->id);
		catalog->id = NULL;
	}
	if (catalog->buf) {
		/* delete the original messages */
		free(catalog->buf);
		catalog->buf = NULL;
	}
#ifdef HAVE_ICONV
	if (catalog->converter) {
		/* close iconv and delete associated data */
		iconv_t *ic = catalog->converter;
		if (*ic != (iconv_t)(-1)) iconv_close(*ic);
		free(catalog->converter);
		catalog->converter = NULL;
	}
#endif
	catalog->msg_cnt = 0;;
	catalog->charset = 0;
	catalog->buf_end = NULL;
}

/* returns the charset-converted message for msgnum */
const char* i5e_get_converted_msg(i5e_catalog_t catalog, int msgnum)
{
#ifdef HAVE_ICONV
	if (msgnum && msgnum < catalog->msg_cnt) {
		if (catalog->msg[msgnum] >= catalog->buf && catalog->msg[msgnum] < catalog->buf_end) {
			iconv_t *ic = catalog->converter;
			if (ic && *ic != (iconv_t)(-1)) catalog->msg[msgnum] = convert(*ic, catalog->msg[msgnum]);
		}
		if (catalog->msg[msgnum]) return catalog->msg[msgnum];
	}
#endif
	return catalog->msg[0];
}

/* the error messages */
#define error_table_size 12
static const char* error_table[error_table_size] =
{
	"Intlize: Success.\n",                             /* 0 */
	"Intlize: No suitable translation file found.\n",  /* 1 */
	"Intlize: Failed to open translation file.\n",     /* 2 */
	"Intlize: Out of memory.\n",                       /* 3 */
	"Intlize: Wrong file type.\n",                     /* 4 */
	"Intlize: Unexpected end of file.\n",              /* 5 */
	"Intlize: Argument error.\n",                      /* 6 */
	"Intlize: Unknown encoding.\n",                    /* 7 */
	"Intlize: No suitable converter found.\n",         /* 8 */
	"Intlize: Charset conversion not implemented.\n",  /* 9 */
	"Intlize: Catalog not initialized.\n",             /* 10 */
	"Intlize: Index out of range.\n"		   /* 11 */
};	

/* returns the error text for the return value of i5e_open() */
const char* i5e_error_text(int errnum)
{
	if (errnum < 0 || errnum >= error_table_size) errnum = 11;
	return error_table[errnum];
}

/* returns the ID of the catalog */
const char* i5e_getid(i5e_catalog_t catalog)
{
	return catalog->id;
}

