/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CFileAssoc.cpp
  Info   : $Id: CFileAssoc.cpp,v 1.28 2012/07/17 12:16:43 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2005 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CMagic.h>
#include <cxxx/CFileOpenDialog.h>
#include <cxxx/CFileHandler.h>

// application headers
#include "CFileAssoc.h"

// C headers
#include <unistd.h>

// namespace declarations
using std::string;
using std::vector;
using cxxx::cxError;
using cxxx::cxgettext;
using cxxx::cxFileManager;

CFileAssoc::CFileAssoc(CWindow* parent, const std::string& file, unsigned type, bool remember)
: CTopLevelWindow(),
  mName(file),
  mRadio(this),
  mExec(&mRadio, IDEXEC),
  mDocument(&mRadio, IDDOCUMENT),
  mPath(this),
  mArguments(this),
  mBrowse(this, IDBROWSE, CIcon::BROWSE, "Browse"),
  mTerm(this, IDTERM),
  mRemember(this, IDREMEMBER),
  mOK(this, IDOK),
  mCancel(this, IDCANCEL)
{
	TransientFor(parent->GetTopLevelWindow());
	SetCaption(_("File Association", 129));
	Resize(300, 250);
	FixSize();
	if (mName.empty()) {
		cxError << _("CFileAssoc: No filename given.", 126) << CError::Error;
		return;
	}
	if (mName[mName.size() - 1] == '/') mName.resize(mName.size() - 1);
	string::size_type pos = file.rfind('/');
	if (++pos == 0) mName = file;
	else mName = file.substr(pos);
	mType = CMagic::GetMimeType(type);
	mIcon.Create(CMagic::GetIcon(type), 32);
	mExec.SetName(_("Is executable", 2));
	mExec.Move(10, 50);
	mDocument.SetName(_("Open with", 5));
	mDocument.Move(10, 70);
	mDocument.SetCheck(true);
	mPath.SetText("");
	mPath.Resize(250, mPath.GetHeight());
	mPath.Move(10, 95);
	mArguments.SetText("");
	mArguments.Resize(280, mArguments.GetHeight());
	mArguments.Move(10, 140);
	mBrowse.Move(270, 95);
	mTerm.SetName(_("Use terminal", 127));
	mTerm.Move(10, 165);
	mRemember.SetName(_("Remember association with this type of file", 128));
	mRemember.Move(10, 185);
	if (!remember) mRemember.Enable(false);
	mOK.SetName(_("&OK", 3));
	mOK.Move(10, 215);
	mCancel.SetName(_("&Cancel", 4));
	mCancel.Move(210, 215);
	mExec.Show();
	mDocument.Show();
	mPath.Show();
	mArguments.Show();
	mBrowse.Show();
	mTerm.Show();
	mRemember.Show();
	mOK.Show();
	mCancel.Show();
	SetFocusWindow(&mPath);
}

/* virtual */ void CFileAssoc::OnExpose(XExposeEvent& e __attribute__ ((unused)))
{
	CColor col(TC_FRAME_TEXT);
	DrawIcon(10, 10, mIcon);
	DrawString(52, 10, col, mName);
	DrawString(52, 30, col, mType);
	DrawString(10, 120, col, _("Program arguments", 1));
}

/* virtual */ void CFileAssoc::OnMessage(unsigned msgid, TMessage type, unsigned value __attribute__ ((unused)))
{
	switch (msgid) {
	case IDOK:
		mReturn = OK;
		mShutDown = true;
		break;
	case IDCANCEL:
		OnQuit();
		break;
	case IDEXEC:
		if (type == MT_CHECK) mPath.Enable(false);
		else if (type == MT_UNCHECK) mPath.Enable(true);
		break;
	case IDDOCUMENT:
		break;
	case IDBROWSE:
		OnBrowse();
		break;
	case IDTERM:
		break;
	case IDREMEMBER:
		break;
	default:
		break;
	}
}

/* virtual */ bool CFileAssoc::OnUpdate(unsigned msgid) const
{
	bool ret = false;
	switch (msgid) {
	case IDCANCEL:
	case IDEXEC:
	case IDDOCUMENT:
	case IDBROWSE:
	case IDTERM:
	case IDREMEMBER:
		ret = true;
	default:
		break;
	}
	bool ena = mExec.IsChecked();
	if (!ena) {
		// first check if this is an absolute filename
		struct stat st;
		if (stat(mPath.GetText().c_str(), &st) == 0) {
			if (S_ISREG(st.st_mode)) {
				ena = access(mPath.GetText().c_str(), X_OK) == 0;
			}
		}
		if (!ena) {
			// no - use the search path
			const vector<string>& tmp = cxFileManager->GetSearchPath();
			string path;
			string file = mPath.GetText();
			for (vector<string>::const_iterator it = tmp.begin(); it < tmp.end(); ++it) {
				path = cxFileManager->GetAbsolutePath(file, *it);
				if (stat(path.c_str(), &st) != 0) continue;
				if (!S_ISREG(st.st_mode)) continue;
				if (access(path.c_str(), X_OK) == 0) {
					ena = true;
					break;
				}
			}
		}
	}
	const_cast<CPushButton&>(mOK).Enable(ena);
	return ret;
}

/* virtual */ void CFileAssoc::OnKey(wchar_t key, unsigned mask)
{
	CTopLevelWindow::OnKey(key, mask);
	if (mFocusWindow == &mPath || mFocusWindow == &mArguments) {
		if (key == L'\n' && mOK.IsEnabled()) {
			mReturn = OK;
			mShutDown = true;
		}
	}
}

/* virtual */ void CFileAssoc::OnQuit()
{
	mReturn = CANCEL;
	mShutDown = true;
}

void CFileAssoc::OnBrowse()
{
	if (!mPath.GetText().empty()) {
		mExecdir = mPath.GetText();
		struct stat st;
		if (stat(mExecdir.c_str(), &st) != 0 || !S_ISDIR(st.st_mode)) {
			string::size_type pos = mExecdir.rfind('/');
			if (pos != string::npos) mExecdir.resize(pos);
			else mExecdir = "/";
		}
	}
	mPath.SetText(mExecdir);
	vector<unsigned> filter;
	filter.push_back(TMagic::BASHSCT);
	filter.push_back(TMagic::DIRECTORY);
	filter.push_back(TMagic::ELFEXE);
	filter.push_back(TMagic::PERLSCT);
	filter.push_back(TMagic::SHSCT);
	CFileOpenDialog dlg(this, mExecdir, filter);
	int x = 10;
	int y = 10;
	TranslateCoordinates(x, y, NULL);
	dlg.Move(x, y);
	dlg.Show();
	if (dlg.DoModal() == CFileOpenDialog::OK) mPath.SetText(dlg.GetPath());
}
