/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CGeneralSheet.cpp
  Info   : $Id: CGeneralSheet.cpp,v 1.46 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "CGeneralSheet.h"
#include "CXplorerList.h"
#include "CXplorerFrame.h"
#include "CXplorerMenu.h"

// C++ headers
#include <cassert>
#include <cerrno>
#include <cstdio>
#include <sstream>

// C headers
#include <unistd.h>
#include <dirent.h>

// namespace declarations
using std::string;
using std::wstring;
using std::vector;
using std::ostringstream;
using std::back_insert_iterator;
using std::time_put;
using std::numpunct;
using std::use_facet;
using std::uint64_t;
using cxxx::cxError;
using cxxx::cxgettext;
using cxxx::cxLocale;
using cxxx::cxFileManager;
using cxxx::cxWiden;

CGeneralSheet::CGeneralSheet(CWindow* parent, CTapeArchive* arch, const string& path)
: CWindow(parent),
  mEdit(this),
  mPath(path),
  mStatus(NULL),
  mCalculate(false)
{
	Resize(390, 225);
	if (arch) {
		if (path == arch->GetName()) mPath = "";
		else if (path.size() > arch->GetName().size()) mPath = mPath.substr(arch->GetName().size() + 1);
		mStatus = new struct stat;
		if (!arch->Stat(mPath, *mStatus)) {
			delete mStatus;
			mStatus = NULL;
		}
	} else {
		struct stat st;
		if (lstat(path.c_str(), &st) == 0) {
			mStatus = new struct stat;
			memcpy(mStatus, &st, sizeof(struct stat));
		}
	}
	mEdit.Resize(300, 20);
	mEdit.Move(85, 20 - GetTextHeight());
	string::size_type begin = 0;
	string::size_type size = mPath.rfind('/');
	mLocation = cxWiden(mPath.substr(0, size));
	unsigned w;
	if (size == 0) {
		w = 0;
		mLocation = L"/";
	} else {
		w = GetTextWidth(mLocation);
		if (w > 295) {
			w += GetTextWidth(L"...");
			while (w > 290) w -= GetCharWidth(mLocation[begin++]);
			mLocation = L"..." + mLocation;
		}
	}
	string p = mPath.substr(0, size);
	++size;
	mName = mPath.substr(size);
	if (p == cxFileManager->GetTrashDir()) {
		mName = cxFileManager->GetFileName(cxFileManager->GetOrigin(mPath.substr(size)));
		mEdit.Enable(false);
		mLocation = cxWiden(_("Trash", 163));
	} else {
		mName = mPath.substr(size);
		if (arch || access(mPath.substr(0, size - 1).c_str(), W_OK)) mEdit.Enable(false);
	}
	mEdit.SetText(mName);
	if (mStatus == NULL) {
		cxError << _("Failed to read file status.\n{1}", 6) << strerror(errno) << CError::Warn;
		mType = "?";
		mSize = "?";
		mATime = "?";
		mMTime = "?";
	} else {
 		if (S_ISLNK(mStatus->st_mode)) {
			char buf[PATH_MAX];
			mType = _("Symbolic Link", 7);
			mMime = mMagic.GetMimeType(TMagic::SYMLINK);
			if (arch) mTo = arch->ReadLink(mPath);
			else {
				assert(mStatus->st_size < PATH_MAX);
				unsigned len = readlink(path.c_str(), buf, PATH_MAX);
				buf[len] = 0;
				mTo = buf;
			}
		} else {
			unsigned typ;
			if (arch) typ = arch->GetFileType(mPath);
			else typ = mMagic.GetFileType(mPath);
			mType = mMagic.GetDescription(typ);
			mMime = mMagic.GetMimeType(typ);
			if (mMime.empty()) mMime = _("not available", 8);
			if (arch) {
				uint64_t size;
				size = arch->GetFileSize(mPath);
				UpdateSize(size);
				mCalculate = false;
			}
			else {
				mSize = _("Calculating...", 9);
				mCalculate = true;
			}
		}
		ostringstream os;
		os.imbue(cxLocale);
		struct tm ft;
		const char* tpl("%x   %X");
		localtime_r(&mStatus->st_atime, &ft);
		const time_put<char>& tp(use_facet<time_put<char> >(cxLocale));
		use_facet<time_put<char> >(cxLocale).put(os, os, ' ', &ft, tpl, tpl + 7 /* strlen(tpl) */);
		mATime = os.str();
		os.str("");
		localtime_r(&mStatus->st_mtime, &ft);
		tp.put(os, os, ' ', &ft, tpl, tpl + 7 /* strlen(tpl) */);
		mMTime = os.str();
	}
	mEdit.Show();
}

void CGeneralSheet::UpdateSize(uint64_t u)
{
	ostringstream os;
	if (u < (uint64_t)999) {
		os << u << " ";
	} else {
		char sz = 'K';
		u *= 100;
		u += 512;
		u >>= 10;
		if (u > (uint64_t)99999) {
			u += 512;
			u >>= 10;
			sz = 'M';
		}
		if (u > (uint64_t)99999) {
			u += 512;
			u >>= 10;
			sz = 'G';
		}
		os << u / 100;
		os << use_facet<numpunct<char> >(cxLocale).decimal_point();
		u %= 100;
		os << u / 10;
		os << u % 10 << " " << sz;
	}
	os << "Bytes";
	mSize = os.str();
	ClearArea(95, 100, 225, 20);
	DrawString(95, 100, CColor(TC_FRAME_TEXT), mSize);
}

/* virtual */ void CGeneralSheet::OnMap(XMapEvent& e)
{
	CWindow::OnMap(e);
	GetTopLevelWindow()->SetFocusWindow(&mEdit);
	if (!mCalculate) return ;
	mCalculate = false;
	if (e.type == MapNotify) UpdateSize(GetSize(mPath, *mStatus));
}

/* virtual */ void CGeneralSheet::OnExpose(XExposeEvent& e __attribute__ ((unused)))
{
	int y = 0;
	int w = GetWidth();
	int h = GetHeight();
	DrawFrame(0, y, w, h, FS_SHEET, false);
	DrawLine(1, 136, w - 3, 136, CColor(TC_BRIGHT));
	DrawLine(2, 135, w - 3, 135, CColor(TC_DARK));
	CColor col(TC_FRAME_TEXT);
	y += 10;
	DrawString(10, y, col, _("Name:", 10));
	y += 30;
	DrawString(10, y, col, _("Type:", 11));
	DrawString(95, y, col, mType);
	y += 20;
	DrawString(10, y, col, _("MIME:", 12));
	DrawString(95, y, col, mMime);
	y += 20;
	DrawString(10, y, col, _("Location:", 13));
	DrawString(95, y, col, mLocation);
	y += 20;
	if (!mTo.empty()) {
		DrawString(10, y, col, _("Target:", 14));
		DrawString(95, y, col, mTo);
	} else {
		DrawString(10, y, col, _("Size:", 15));
		DrawString(95, y, col, mSize);
	}
	y += 60;
	DrawString(10, y, col, _("Modified:", 16));
	DrawString(95, y, col, mMTime);
	y += 20;
	DrawString(10, y, col, _("Accessed:", 17));
	DrawString(95, y, col, mATime);
}

uint64_t CGeneralSheet::GetDirSize(const string& dir)
{
	uint64_t size(0);
	if (!GetTopLevelWindow()->IsRunning()) return size;
	if (access(dir.c_str(), R_OK) == 0) {
		vector<string> files = cxFileManager->ReadDirectory(dir);
		string fn;
		struct stat st;
		for (vector<string>::iterator i = files.begin(); i != files.end(); ++i) {
			fn = dir + "/" + *i;
			if (lstat(fn.c_str(), &st)) continue;
			size += GetSize(fn, st);
		}
	}
	return size;
}

// return value: true if dialog can close
bool CGeneralSheet::DoOK()
{
	if (mEdit.IsEnabled() && mEdit.GetText() != mName) {
		string::size_type pos = mPath.rfind('/');
		string to = mPath.substr(0, pos);
		to += '/';
		to += mEdit.GetText();
		if (rename(mPath.c_str(), to.c_str())) {
			if (errno == ENOTEMPTY) errno = EEXIST;
			cxError << _("Failed to rename {1} to {2}\n{3}", 18) << mPath << to << strerror(errno) << CError::Error;
			return false;
		}
	}
	return true;
}

/* virtual */ void CGeneralSheet::OnFocus(bool in)
{
	if (in) mEdit.SetFocus();
}
