/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CPermissionSheet.cpp
  Info   : $Id: CPermissionSheet.cpp,v 1.30 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "CPermissionSheet.h"
#include "CXplorerList.h"
#include "CXplorerMenu.h"

// C++ headers
#include <cerrno>

// C headers
#include <pwd.h>
#include <grp.h>
#include <unistd.h>

// Namespace declarations
using std::string;
using std::vector;
using cxxx::cxError;
using cxxx::cxgettext;

CPermissionSheet::CPermissionSheet(CWindow* parent, CTapeArchive* arch, const string& path)
: CWindow(parent),
  mUID(0),
  mGID(0),
  mPath(path),
  mAccess(NOACCESS),
  mGroupList(this, 0),
  mUserList(this, 0)
{
	// Read the file status
	if (arch) {
		if (path == arch->GetName()) mPath = "";
		else if (path.size() > arch->GetName().size()) mPath = path.substr(arch->GetName().size() + 1);
		if (!arch->Stat(mPath, mStatus)) mAccess = NOSTATUS;
	} else if (lstat(path.c_str(), &mStatus) != 0) mAccess = NOSTATUS;
	// Get Ownership of file and calculate permissions
	if (mAccess != NOSTATUS) {
		mUID = mStatus.st_uid;
		mGID = mStatus.st_gid;
		uid_t id = geteuid();
		if (id == 0) mAccess = ROOTACCESS;
		else if (id == mUID) mAccess = USERACCESS;
		else mAccess = NOACCESS;
		struct passwd* pwd = getpwuid(mUID);
		if (pwd != NULL) mUserName = pwd->pw_name;
		struct group* grp = getgrgid(mGID);
		if (grp != NULL) mGroupName = grp->gr_name;
	}
	if (mAccess == USERACCESS) {
		// fill in groups
		unsigned count = getgroups(0, NULL);
		gid_t* gids = new gid_t[count];
        if (getgroups(count, gids) < 0) count = 0;
		gid_t sel = 0;
		for (unsigned i = 0; i < count; i++) {
			mGroupList.AppendItem(getgrgid(gids[i])->gr_name);
			if (gids[i] == mGID) sel = i;
		}
		mGroupList.Select(sel);
        delete[] gids;
	} else if (mAccess == ROOTACCESS) {
		unsigned sel = 0;
		bool selected = false;
		setgrent();
		struct group* grp = getgrent();
		while (grp) {
			mGroupList.AppendItem(grp->gr_name);
			if (!selected) {
				if (grp->gr_gid == mGID) selected = true;
				else ++sel;
			}
			grp = getgrent();
		}
		if (!selected) sel = 0;
		mGroupList.Select(sel);
		endgrent();
		sel = 0;
		selected = false;
		setpwent();
		struct passwd* usr = getpwent();
		while (usr) {
			mUserList.AppendItem(usr->pw_name);
			if (!selected) {
				if (usr->pw_uid == mUID) selected = true;
				else ++sel;
			}
			usr = getpwent();
		}
		if (!selected) sel = 0;
		mUserList.Select(sel);
		endpwent();
	}		
	// build check button array
	if (mAccess != NOSTATUS) {
		for (unsigned i = 0; i <= STICK; ++i) {
			mCheckButton.push_back(new CCheckButton(this, i));
			mCheckButton[i]->Move(65 + 60 * (i % 4), 30 + 25 * (i / 4));
		}
		mCheckButton[RUSR]->SetCheck(mStatus.st_mode & S_IRUSR);
		mCheckButton[WUSR]->SetCheck(mStatus.st_mode & S_IWUSR);
		mCheckButton[XUSR]->SetCheck(mStatus.st_mode & S_IXUSR);
		mCheckButton[SUID]->SetCheck(mStatus.st_mode & S_ISUID);
		mCheckButton[RGRP]->SetCheck(mStatus.st_mode & S_IRGRP);
		mCheckButton[WGRP]->SetCheck(mStatus.st_mode & S_IWGRP);
		mCheckButton[XGRP]->SetCheck(mStatus.st_mode & S_IXGRP);
		mCheckButton[SGID]->SetCheck(mStatus.st_mode & S_ISGID);
		mCheckButton[ROTH]->SetCheck(mStatus.st_mode & S_IROTH);
		mCheckButton[WOTH]->SetCheck(mStatus.st_mode & S_IWOTH);
		mCheckButton[XOTH]->SetCheck(mStatus.st_mode & S_IXOTH);
		mCheckButton[STICK]->SetCheck(mStatus.st_mode & S_ISVTX);
	}
	if (mAccess <= NOACCESS) {
		for (vector<CCheckButton*>::iterator i = mCheckButton.begin(); i < mCheckButton.end(); ++i) {
			(*i)->SetFocusable(false);
			(*i)->Enable(false);
		}
	}
	if (mAccess != NOSTATUS) {
		if (S_ISDIR(mStatus.st_mode)) mXAction = _("Enter", 27);
		else mXAction = _("Execute", 28);
		for (vector<CCheckButton*>::iterator i = mCheckButton.begin(); i < mCheckButton.end(); ++i) (*i)->Show();
	}
	mGroupList.Move(70, 142);
	mUserList.Move(70, 122);
	if (mAccess >= USERACCESS) mGroupList.Show();
	if (mAccess == ROOTACCESS) mUserList.Show();
}

/* virtual */ CPermissionSheet::~CPermissionSheet()
{
	for (vector<CCheckButton*>::iterator i = mCheckButton.begin(); i < mCheckButton.end(); ++i) delete *i;
}

/* virtual */ void CPermissionSheet::OnExpose(XExposeEvent& e)
{
	ClearArea(e);
	CColor col(TC_FRAME_TEXT);
	if (mAccess != NOSTATUS) {
		int x = 0;
		int y = 0;
		int w = GetWidth();
		int h = GetHeight();
		DrawFrame(x, y, w, h, FS_SHEET, false);
		DrawLine(1, 111, w - 3, 111, CColor(TC_BRIGHT));
		DrawLine(2, 110, w - 3, 110, CColor(TC_DARK));
		DrawString(10, 10, col, _("Class", 29));
		DrawString(65, 10, col, _("Read", 30));
		DrawString(125, 10, col, _("Write", 31));
		DrawString(185, 10, col, mXAction);
		DrawString(245, 10, col, _("Special", 32));
		y = 30 + (mCheckButton[0]->GetHeight() - GetTextHeight()) / 2;
		DrawString(10, y, col, _("User", 33));
		DrawString(275, y, col, _("Set UID", 34));
		y += 25;
		DrawString(10, y, col, _("Group", 35));
		DrawString(275, y, col, _("Set GID", 36));
		y += 25;
		DrawString(10, y, col, _("Other", 37));
		DrawString(275, y, col, _("Sticky", 38));
		DrawString(10, 125, col, _("User:", 39));
		if (mAccess < ROOTACCESS) DrawString(75, 125, col, mUserName);
		DrawString(10, 145, col, _("Group:", 40));
		if (mAccess < USERACCESS) DrawString(75, 145, col, mGroupName);
	} else {
		DrawString(10, 30, col, _("Could not read file information.", 41));
	}
}

// returns true if dialog can close
bool CPermissionSheet::DoOK()
{
	if (mAccess < USERACCESS) return true;
	mode_t oldst = mStatus.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX);
	mode_t newst = 0;
	if (mCheckButton[RUSR]->IsChecked()) newst |= S_IRUSR;
	if (mCheckButton[WUSR]->IsChecked()) newst |= S_IWUSR;
	if (mCheckButton[XUSR]->IsChecked()) newst |= S_IXUSR;
	if (mCheckButton[RGRP]->IsChecked()) newst |= S_IRGRP;
	if (mCheckButton[WGRP]->IsChecked()) newst |= S_IWGRP;
	if (mCheckButton[XGRP]->IsChecked()) newst |= S_IXGRP;
	if (mCheckButton[ROTH]->IsChecked()) newst |= S_IROTH;
	if (mCheckButton[WOTH]->IsChecked()) newst |= S_IWOTH;
	if (mCheckButton[XOTH]->IsChecked()) newst |= S_IXOTH;
	if (mCheckButton[SUID]->IsChecked()) newst |= S_ISUID;
	if (mCheckButton[SGID]->IsChecked()) newst |= S_ISGID;
	if (mCheckButton[STICK]->IsChecked()) newst |= S_ISVTX;
	if (oldst != newst) {
		if (chmod(mPath.c_str(), newst)) {
			bool e = true;
#ifdef EFTYPE
			if (errno == EFTYPE) {
				newst &= ~S_ISVTX;
				if (!chmod(mPath.c_str(), newst)) e = false;
			}
#endif // EFTYPE
			if (e) {
				cxError << _("Failed to change Mode bits.\n{1}", 42) << strerror(errno) << CError::Warn;
				return false;
			}
		}
	}
	assert(!mGroupList.GetText().empty());
	gid_t newgrp = getgrnam(mGroupList.GetText().c_str())->gr_gid;
	uid_t newusr = mUID;
	if (mAccess == ROOTACCESS) {
		assert (!mUserList.GetText().empty());
		newusr = getpwnam(mUserList.GetText().c_str())->pw_uid;
	}
	if (newgrp != mGID || newusr != mUID) {
		if (chown(mPath.c_str(), newusr, newgrp)) {
			cxError << _("Failed to set new file owner group.\n{1}", 43) << strerror(errno) << CError::Warn;
			return false;
		}
	}
	return true;
}
