/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CTapeArchive.cpp
  Info   : $Id: CTapeArchive.cpp,v 1.36 2012/09/16 17:49:44 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2007 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "CTapeArchive.h"

// C++X headers
#include <cxxx/CError.h>
#include <cxxx/CFileHandler.h>

// C++ headers
#include <cassert>
#include <cerrno>

// C headers
#include <unistd.h>

// Namespace declarations
using std::string;
using std::wstring;
using std::ios;
using std::vector;
using std::ofstream;
using cxxx::cxError;
using cxxx::cxFileManager;
using cxxx::cxgettext;

void CTapeArchive::ReadHeader()
{
	ClearFileStatus();
	mIn.seekg(mOffset);
	mName = ReadString(100);
	if (mName == "./") {
		mOffset += 512;
		ReadHeader();
		return;
	}
	if (mName.size() >= 2 && mName.compare(0, 2, "./") == 0) mName = mName.substr(2);
	if (!mIn.seekg(mOffset + (ios::pos_type)100)) return;
	mFileStatus.st_mode = ReadOctal(8);
	// not too portable, but works at least for gcc on linux
	mFileStatus.st_uid = ReadOctal(8);
	mFileStatus.st_gid = ReadOctal(8);
	mFileStatus.st_size = ReadOctal(12);
	mFileStatus.st_mtime = ReadOctal(12);
	mChecksum = ReadOctal(8);
	mIn.seekg(mOffset + (ios::pos_type)156);
	char ftyp;
	mIn.get(ftyp);
	switch (ftyp) {
		case '\0':
		case '0':
		case '7':
			mFileStatus.st_mode |= S_IFREG;
			break;
		case '1':
			// hard link! we don't care
		case '2':
			mFileStatus.st_mode |= S_IFLNK;
			break;
		case '3':
			mFileStatus.st_mode |= S_IFCHR;
			break;
		case '4':
			mFileStatus.st_mode |= S_IFBLK;
			break;
		case '5':
			mFileStatus.st_mode |= S_IFDIR;
			break;
		case '6':
			mFileStatus.st_mode |= S_IFIFO;
			break;
		default:
			break;
	} 
	mLinkName = ReadString(100);
	if (mArchiveType == UNINITIALIZED) {
		mIn.seekg(mOffset + (ios::pos_type)257);
		char magic[8];
		mIn.read(magic, 8);
		if (strcmp(magic, "ustar  ") == 0) mArchiveType = GNUTAR;
		else if (strcmp(magic, "ustar") == 0) mArchiveType = POSIXORSTAR;
		else mArchiveType = UNKTAR;
	}
	mIn.seekg(mOffset + (ios::pos_type)265);
	mUserName = ReadString(32);
	mIn.seekg(mOffset + (ios::pos_type)297);
	mGroupName = ReadString(32);
	mIn.seekg(mOffset + (ios::pos_type)329);
	mDevMajor = ReadOctal(8);
	mDevMinor = ReadOctal(8);
	if (mArchiveType == POSIXORSTAR) {
		char buf[5];
		mIn.seekg(mOffset + (ios::pos_type)475);
		mIn.read(buf, 2);
		mIn.seekg(mOffset + (ios::pos_type)487);
		mIn.read(buf + 2, 2);
		mIn.seekg(mOffset + (ios::pos_type)499);
		mIn.get(buf[4]);
		if (buf[0] == 0 && buf[1] >= '0' && buf[1] < '8' && buf[2] == ' '
			&& buf[3] >= '0' && buf[3] < '8' && buf [4] == ' ') mArchiveType = STAR;
		else mArchiveType = POSIXTAR;
	}
	mIn.seekg(mOffset + (ios::pos_type)345);
	string prefix;
	if (mArchiveType == POSIXTAR) prefix = ReadString(155);
	else if (mArchiveType == STAR) {
		prefix = ReadString(131);
		mIn.seekg(mOffset + (ios::pos_type)476);
		mFileStatus.st_atime = ReadOctal(12);
		mFileStatus.st_ctime = ReadOctal(12);
	} else if (mArchiveType == GNUTAR) {
		mFileStatus.st_atime = ReadOctal(12);
		mFileStatus.st_ctime = ReadOctal(12);
		//mOffset = ReadOctal();
		/*
			handle sparse files
		*/
	}
	if (!prefix.empty()) mName = prefix + mName;
	switch (ftyp) {
	case 'L':
		// long file name
		ReadSpecialHeaderL();
		break;
	case 'x':
		// posix extended header
		ReadSpecialHeaderx();
		break;
	default:
		break;
	}
}
void CTapeArchive::ReadSpecialHeaderL()
{
	mOffset += 512;
	mIn.seekg(mOffset);
	string name;
	getline(mIn, name, '\0');
	assert(mFileStatus.st_size == (int)name.size() + 1);
	mOffset += (mFileStatus.st_size + 511) & -512u;
	ReadHeader();
	mName = name;
	if (mName.size() >= 2 && mName.compare(0, 2, "./") == 0) mName = mName.substr(2);
}

void CTapeArchive::ReadSpecialHeaderx()
{
	mOffset += 512;
	mIn.seekg(mOffset);
	string uname, gname, path, linkpath, buf;
	timespec atime, ctime, mtime;
	atime.tv_sec = 0;
	ctime.tv_sec = 0;
	mtime.tv_sec = 0;
	atime.tv_nsec = 0;
	ctime.tv_nsec = 0;
	mtime.tv_nsec = 0;
	ios::pos_type end = mOffset + (ios::pos_type)mFileStatus.st_size;
	while (mIn && mIn.tellg() < end) {
		mIn.ignore(10, ' ');
		getline(mIn, buf, '=');
		if (buf == "atime") atime = ReadTime();
		else if (buf == "ctime") ctime = ReadTime();
		else if (buf == "mtime") mtime = ReadTime();
		else if (buf == "uname") getline(mIn, uname);
		else if (buf == "gname") getline(mIn, gname);
		else if (buf == "path") getline(mIn, path);
		else if (buf == "linkpath") getline(mIn, linkpath);
		else mIn.ignore(10000, '\n');
	}
	mOffset = (end + (ios::pos_type)511) & -512u;
	ReadHeader();
	if (atime.tv_sec != 0) {
		mFileStatus.st_atime = atime.tv_sec;
#ifdef _STATBUF_ST_NSEC
 #ifdef __USE_MISC
		mFileStatus.st_atim.tv_nsec = atime.tv_nsec;
 #else // __USE_MISC
		mFileStatus.st_atimensec = atime.tv_nsec;
 #endif // __USE_MISC
#endif // _STATBUF_ST_NSEC
	}
	if (ctime.tv_sec != 0) {
		mFileStatus.st_ctime = ctime.tv_sec;
#ifdef _STATBUF_ST_NSEC
 #ifdef __USE_MISC
		mFileStatus.st_ctim.tv_nsec = ctime.tv_nsec;
 #else // __USE_MISC
		mFileStatus.st_ctimensec = ctime.tv_nsec;
 #endif // __USE_MISC
#endif // _STATBUF_ST_NSEC
	}
	if (mtime.tv_sec != 0) {
		mFileStatus.st_mtime = mtime.tv_sec;
#ifdef _STATBUF_ST_NSEC
 #ifdef __USE_MISC
		mFileStatus.st_mtim.tv_nsec = mtime.tv_nsec;
 #else // __USE_MISC
		mFileStatus.st_mtimensec = mtime.tv_nsec;
 #endif // __USE_MISC
#endif // _STATBUF_ST_NSEC
	}
	if (!uname.empty()) mUserName = uname;
	if (!gname.empty()) mGroupName = gname;
	if (!path.empty()) mName = path;
	if (!linkpath.empty()) mLinkName = linkpath;	
	if (mName.size() >= 2 && mName.compare(0, 2, "./") == 0) mName = mName.substr(2);
}

vector<CFileListItem> CTapeArchive::ListDirectory(const std::string& dir)
{
	string cmp = dir;
	if (!cmp.empty()) cmp += '/';
	vector<CFileListItem> ret;
	vector<string> dirs;
	CIcon icon;
	mOffset = 0;
	ReadHeader();
	while (mIn && !mName.empty()) {
		mOffset += 512;
		if (cmp.empty() || mName.compare(0, cmp.size(), cmp) == 0) {
			string name;
			if (mName.size() > cmp.size()) {
				name = mName.substr(cmp.size());
				bool inter = true;
				string::size_type pos = name.find('/');
				if (pos >= name.size() - 1) inter = false;
				name = name.substr(0, pos);
				if (!inter) {
					unsigned type;
					mIn.seekg(mOffset);
					if (S_ISLNK(mFileStatus.st_mode)) type = TMagic::UNKNOWN;
					else type = mMagic.GetFileType(*mIn.rdbuf(), mFileStatus, name);
					CFileListItem item(icon, name, mFileStatus, type, 0, 0);
					if (S_ISLNK(mFileStatus.st_mode)) item.IsLink(true);
					ret.push_back(item);
				} else {
					bool doit = true;
					for (vector<string>::iterator i = dirs.begin(); i != dirs.end(); ++i) {
						if (name == *i) {
							doit = false;
							break;
						}
					}
					if (doit) dirs.push_back(name);
				}
			}
		}
		mOffset += (mFileStatus.st_size + 511) & -512u;
		ReadHeader();
	}
	// fill in implicit directories
	// we do not have any information about them at all, so let's give them the same
	// status as the archive has.
	struct stat st = mArchiveStatus;
	st.st_mode &= ~S_IFMT;
	st.st_mode |= S_IFDIR;
	for (vector<string>::iterator i = dirs.begin(); i != dirs.end(); ++i) {
		bool doit = true;
		for (vector<CFileListItem>::iterator j = ret.begin(); j != ret.end(); ++j) {
			if (*i == j->GetText()) {
				doit = false;
				break;
			}
		}
		if (doit) {
			CFileListItem item(icon, *i, mArchiveStatus, TMagic::DIRECTORY, 0, 0);
			ret.push_back(item);
		}
	}
	return ret;
}

bool CTapeArchive::FindFile(const std::string& file)
{
	mOffset = 0;
	ReadHeader();
	if (!IsValid()) return false;
	while (mIn && !mName.empty()) {
		mOffset += 512;
		if (mName[mName.size() - 1] == '/') {
			if (mName.compare(0, mName.size() - 1, file) == 0)  return true;
		} else {
			if (mName == file) return true;
		}		
		mOffset += (mFileStatus.st_size + 511) & -512u;
		ReadHeader();
	}
	if (!mIn) mIn.clear();
	return false;
}


// extracts the file file from the archive to a temporary file.
// returns the name of the temporary file in case of success, else an empty string.
string CTapeArchive::Extract(const string& file)
{
	string ret;
	if (file != mName && !FindFile(file)) return ret;
	// found it; create a temporary file for extraction
	char tmpnam[] = "/tmp/cxxx_tmp_XXXXXX";
	int fd = mkstemp(tmpnam);
	if (fd == -1) {
		cxError << _("Failed to open temporary file for extraction.", 143) << CError::Error;
		return ret;
	}
	char buf[512];
	unsigned written = 0;
	unsigned left = mFileStatus.st_size;
	unsigned chunk;
	mIn.seekg(mOffset);
	while (left) {
		if (left > 512) chunk = 512;
		else chunk = left;
		mIn.read(buf, chunk);
		if (!mIn) {
			written = -1u;
			break;
		}
		left -= chunk;
		while (chunk != 0 || errno == EINTR) {
			written = write(fd, buf, chunk);
			if (written != -1u) chunk -= written;
			else {
				cxError << _("Failed to write temporary file", 141) << CError::Error;
				left = 0;
				break;
			}
		}
	}
	close(fd);
	if (written == -1u) unlink(tmpnam);
	else {
		ret = tmpnam;
		// try to change owner and group; don't care if looses
		chown(ret.c_str(), mFileStatus.st_uid, mFileStatus.st_gid);
		// try to set permissions; don't care if looses
		chmod(ret.c_str(), mFileStatus.st_mode);
	}		
	return ret;
}

unsigned CTapeArchive::GetDirSize(const string& dir)
{
	unsigned size = 0;
	string dirname(dir);
	if (!dirname.empty()) dirname += '/';
	mOffset = 0;
	ReadHeader();
	while (mIn && !mName.empty()) {
		mOffset += 512;
		if (dirname.empty() || mName.compare(0, dirname.size(), dirname) == 0) {
			size += mFileStatus.st_size;
		}
		mOffset += (mFileStatus.st_size + 511) & -512u;
		ReadHeader();
	}
	return size;
}

void CTapeArchive::Extract(const string& file, const string& todir)
{
	ofstream out;
	mOffset = 0;
	ReadHeader();
	string srcpath;
	if (!file.empty() && file[0] == '/') srcpath = file.substr(1);
	else srcpath = file;
	string::size_type namepos = srcpath.rfind('/') + 1;
	while (mIn && !mName.empty()) {
		mOffset += 512;
		if (srcpath.empty() || mName.compare(0, srcpath.size(), srcpath) == 0) {
			if (mName.size() == srcpath.size() || mName[srcpath.size()] == '/') {
				string name = todir;
				name += "/" + mName.substr(namepos);
				string::size_type pos = todir.size();
				while ((pos = name.find('/', pos)) != string::npos) {
					string dir = name.substr(0, ++pos);
					if (access(dir.c_str(), F_OK) != 0) mkdir(dir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);  
				}	
				if (S_ISDIR(mFileStatus.st_mode)) {
					if (access(name.c_str(), F_OK) != 0) mkdir(name.c_str(), mFileStatus.st_mode);
					// try to change owner, group and permissions
					chown(name.c_str(), mFileStatus.st_uid, mFileStatus.st_gid);
					chmod(name.c_str(), mFileStatus.st_mode);
				} else if (S_ISLNK(mFileStatus.st_mode)) {
					if (!symlink(mLinkName.c_str(), name.c_str())) {
						// try to change owner, group and permissions
						chown(name.c_str(), mFileStatus.st_uid, mFileStatus.st_gid);
						chmod(name.c_str(), mFileStatus.st_mode);
					}
				} else {
					out.open(name.c_str());
					if (out.good() && mIn.good()) {
						char buf[1024];
						mIn.seekg(mOffset);
						unsigned left = mFileStatus.st_size;
						unsigned chunk;
						while (left) {
							if (left > 1024) chunk = 1024;
							else chunk = left;
							left -= chunk;
							mIn.read(buf, chunk);
							out.write(buf, chunk);
						}
						out.close();
						// try to change owner, group and permissions
						chown(name.c_str(), mFileStatus.st_uid, mFileStatus.st_gid);
						chmod(name.c_str(), mFileStatus.st_mode);
					}
					if (mName.size() == srcpath.size()) break;
				}
			}
		}
		mOffset += (mFileStatus.st_size + 511) & -512u;
		ReadHeader();
	}
}
