/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CTapeArchive.h
  Info   : $Id: CTapeArchive.h,v 1.21 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2007 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

#ifndef CTapeArchive_h
#define CTapeArchive_h

// C++X headers
#include <cxxx/CListItem.h>
#include <cxxx/CMagic.h>

// C++ headers
#include <string>
#include <fstream>
#include <vector>

class CTapeArchive 
{
private:
	enum
	{
		// Archive types
		UNINITIALIZED,
		INVALID,
		GNUTAR,
		POSIXTAR,
		STAR,
		POSIXORSTAR,
		UNKTAR
	};

public:
	// Constructor
	CTapeArchive()
	{
		Close();
	}

	// Opens fn and uses fake name name
	void Open(const std::string& name, const std::string& fn)
	{
		Close();
		mArchiveName = name;
		mArchivePath = fn;
		mIn.open(fn.c_str());
		if (!mIn) {
			cxxx::cxError << cxxx::_("Failed to open Archive {1}.", 146) << name << CError::Error;
			mArchiveType = INVALID;
		}
		if (stat(fn.c_str(), &mArchiveStatus)) {
			cxxx::cxError << cxxx::_("Failed to retrieve file status of Archive {1}.", 142) << name << CError::Error;
			mArchiveType = INVALID;
		}
		mOffset = 0;
		ReadHeader();
		mOffset = 512;
	}

	// Close archive
	void Close()
	{
		mArchiveName.clear();
		mArchiveType = UNINITIALIZED;
		mOffset = 0;
		mChecksum = 0;
		mDevMajor = 0;
		mDevMinor = 0;
		mIn.close();
		mIn.clear();
	}	

	// Lists directory dir of the archive
	std::vector<CFileListItem> ListDirectory(const std::string& dir);

	// Extracts file and returns the file name of the extracted file
	std::string Extract(const std::string& file);

	// Extracts file to todir
	void Extract(const std::string& file, const std::string& todir);

	// Makes this archive invalid
	void Invalidate()
	{
		mArchiveType = INVALID;
	}

	// Returns whether this archive is valid
	bool IsValid() const
	{
		return mArchiveType > INVALID;
	}
	
	// Returns the (fake) name of the archive
	const std::string& GetName() const
	{
		return mArchiveName;
	}
	
	// Returns the file name of the archive
	const std::string& GetPath() const
	{
		return mArchivePath;
	}
	
	// Like stat for normal files
	bool Stat(const std::string& file, struct stat& statret)
	{
		if (mName != file && !FindFile(file)) return false;
		memcpy(&statret, &mFileStatus, sizeof(struct stat));
		return true;
	}	
		
	// Returns the type of file 
	unsigned GetFileType(const std::string& file)
	{
  
		if (mName != file && !FindFile(file)) return TMagic::LOCKED;
 	  	mIn.seekg(mOffset);
  	 	return mMagic.GetFileType(*mIn.rdbuf(), mFileStatus, file);
	}
   	
	// Returns the size of file
	unsigned GetFileSize(const std::string& file)
	{
		if (file != mName && !FindFile(file)) return 0;
		if (S_ISDIR(mFileStatus.st_mode)) return GetDirSize(file);
		return mFileStatus.st_size;
	}
	
	// Returns the link target of file
	const std::string& ReadLink(const std::string& file)
	{
		if (file != mName && !FindFile(file)) return cxxx::cxEmptyString;
		return mLinkName;
	}
	
private:
	// Reads a file header
	void ReadHeader();

	// Helper functions
	unsigned ReadOctal(unsigned size)
	{
		char buf[12];
		mIn.read(buf, size);
		unsigned u = 0;
		unsigned i = 0;
		while (buf[i] == ' ') ++i;
		while (i < size && buf[i] >= '0' && buf[i] < '9') {
			u <<= 3;
			u += (buf[i] & 0x0f);
			++i;
		}
		return u;
	}

	std::string ReadString(unsigned size)
	{
		std::string s;
		char c;
		while (mIn.get(c) && c && s.size() < size) s += c;
		return s;
	}

	timespec ReadTime()
	{
		char c;
		timespec ts;
		ts.tv_sec = 0;
		ts.tv_nsec = 0;
		while (mIn.get(c) && c >= '0' && c <= '9') {
			ts.tv_sec *= 10;
			ts.tv_sec += c -'0';
		}
		if (c != '.' && c != '\n') {
			ts.tv_sec = 0;
			return ts;
		}
		unsigned mul = 1000000000;
		while (mIn.get(c) && c >= '0' && c <= '9') {
			ts.tv_nsec += (unsigned)(c - '0') * mul;
			mul /= 10;
		}
		if (c != '\n') {
			ts.tv_sec = 0;
			ts.tv_nsec = 0;
		}
		return ts;
	}

	void ReadSpecialHeaderL();
	void ReadSpecialHeaderx();
	void ClearFileStatus()
	{
		memcpy(&mFileStatus, &mArchiveStatus, sizeof(struct stat));
	}

	bool FindFile(const std::string& file);
	unsigned GetDirSize(const std::string& dir);
	
private:
	std::string mName;
	std::string mLinkName;
	std::string mArchiveName;
	std::string mArchivePath;
	struct stat mFileStatus;
	struct stat mArchiveStatus;
	std::string mUserName;
	std::string mGroupName;
	unsigned mChecksum;
	unsigned mDevMajor;
	unsigned mDevMinor;
	unsigned mArchiveType;
	std::ios::pos_type mOffset;
	std::ifstream mIn;
	CMagic mMagic;
};

#endif // CTapeArchive_h
