/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CTermConfig.cpp
  Info   : $Id: CTermConfig.cpp,v 1.7 2012/07/17 12:16:43 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// C++X headers
#include <cxxx/CFileOpenDialog.h>
#include <cxxx/CFileHandler.h>

// application headers
#include "CTermConfig.h"

// C headers
#include <unistd.h>

// namespace declarations
using std::string;
using std::vector;
using cxxx::cxError;
using cxxx::cxgettext;
using cxxx::cxFileManager;

CTermConfig::CTermConfig(CWindow* parent, const string& cmd, const string& opt)
: CTopLevelWindow(),
  mPath(this),
  mArguments(this),
  mBrowse(this, IDBROWSE, CIcon::BROWSE, _("Browse", 167)),
  mOK(this, IDOK),
  mCancel(this, IDCANCEL)
{
	TransientFor(parent->GetTopLevelWindow());
	SetCaption(_("Set Terminal Command", 174));
	Resize(300, 145);
	FixSize();
	const char* term;
	if (!cmd.empty() && IsExecutable(cmd)) term = cmd.c_str();
	else {
		term = getenv("TERM");
		if (!term || !IsExecutable(term)) {
			if (IsExecutable("xterm")) term = "xterm";
			else if (IsExecutable("rxvt")) term = "rxvt";
			else if (IsExecutable("urxvt")) term = "urxvt";
			else term = "";
		}
	}
	mPath.SetText(term);
	mPath.Resize(250, mPath.GetHeight());
	mPath.Move(10, 30);
	if (opt.empty()) mArguments.SetText("-hold -e");
	else mArguments.SetText(opt);
	mArguments.Resize(280, mArguments.GetHeight());
	mArguments.Move(10, 75);
	mBrowse.Move(270, 30);
	mOK.SetName(_("&OK", 3));
	mOK.Move(10, 110);
	mCancel.SetName(_("&Cancel", 4));
	mCancel.Move(210, 110);
	mPath.Show();
	mArguments.Show();
	mBrowse.Show();
	mOK.Show();
	mCancel.Show();
	SetFocusWindow(&mPath);
}

/* virtual */ void CTermConfig::OnExpose(XExposeEvent& e __attribute__ ((unused)))
{
	CColor col(TC_FRAME_TEXT);
	DrawString(10, 10, col, _("Terminal command", 173));
	DrawString(10, 55, col, _("Program arguments to run commands", 177));
}

/* virtual */ void CTermConfig::OnMessage(unsigned msgid, TMessage type __attribute__ ((unused)), unsigned value __attribute__ ((unused)))
{
	switch (msgid) {
	case IDOK:
		mReturn = OK;
		mShutDown = true;
		break;
	case IDCANCEL:
		OnQuit();
		break;
	case IDBROWSE:
		OnBrowse();
		break;
	default:
		break;
	}
}

/* virtual */ bool CTermConfig::OnUpdate(unsigned msgid) const
{
	bool ret = false;
	switch (msgid) {
	case IDCANCEL:
	case IDBROWSE:
		ret = true;
	default:
		break;
	}
	const_cast<CPushButton&>(mOK).Enable(IsExecutable(mPath.GetText()));
	return ret;
}

/* virtual */ void CTermConfig::OnKey(wchar_t key, unsigned mask)
{
	CTopLevelWindow::OnKey(key, mask);
	if (mFocusWindow == &mPath || mFocusWindow == &mArguments) {
		if (key == L'\n' && mOK.IsEnabled()) {
			mReturn = OK;
			mShutDown = true;
		}
	}
}

/* virtual */ void CTermConfig::OnQuit()
{
	mReturn = CANCEL;
	mShutDown = true;
}

void CTermConfig::OnBrowse()
{
	if (!mPath.GetText().empty()) {
		mExecdir = mPath.GetText();
		struct stat st;
		if (stat(mExecdir.c_str(), &st) != 0 || !S_ISDIR(st.st_mode)) {
			string::size_type pos = mExecdir.rfind('/');
			if (pos != string::npos) mExecdir.resize(pos);
			else mExecdir = "/";
		}
	}
	mPath.SetText(mExecdir);
	vector<unsigned> filter;
	filter.push_back(TMagic::BASHSCT);
	filter.push_back(TMagic::DIRECTORY);
	filter.push_back(TMagic::ELFEXE);
	filter.push_back(TMagic::PERLSCT);
	filter.push_back(TMagic::SHSCT);
	CFileOpenDialog dlg(this, mExecdir, filter);
	int x = 10;
	int y = 10;
	TranslateCoordinates(x, y, NULL);
	dlg.Move(x, y);
	dlg.Show();
	if (dlg.DoModal() == CFileOpenDialog::OK) mPath.SetText(dlg.GetPath());
}

bool CTermConfig::IsExecutable(const string& file) const
{
	if (file.empty()) return false;
	if (file[0] == '/') return access(mPath.GetText().c_str(), X_OK) == 0;
	// use the search path
	const vector<string>& tmp = cxFileManager->GetSearchPath();
	string path;
	for (vector<string>::const_iterator it = tmp.begin(); it < tmp.end(); ++it) {
		path = cxFileManager->GetAbsolutePath(file, *it);
		if (access(path.c_str(), X_OK) == 0) return true;
	}
	return false;
}
	
