/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CTrashListControl.cpp
  Info   : $Id: CTrashListControl.cpp,v 1.20 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2007 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "CTrashListControl.h"
#include "CUntrashDialog.h"
#include "CPropertyDialog.h"

// C++X headers
#include <cxxx/CFileHandler.h>

// C++ headers
#include <cassert>

// Namespace declarations
using cxxx::cxgettext;
using cxxx::cxFileManager;
using cxxx::cxWiden;
using cxxx::cxNarrow;
using std::string;
using std::wstring;
using std::max;
using std::ifstream;

/* virtual */ void CTrashListControl::Append(const std::string& path, const std::string& name)
{
	unsigned flags = 0;
	string alias = cxFileManager->GetOrigin(name);
	if (alias.empty()) {
		// the file has been moved here, but is not trashed. Try to trash it.
		string tmppath = cxFileManager->GetAbsolutePath(name, cxFileManager->GetTrashDir());
		string tmpname = cxFileManager->Trash(tmppath);
		tmppath = cxFileManager->GetAbsolutePath(tmpname, cxFileManager->GetTrashDir());
		// failed to trash: ignore
		if (tmpname.empty()) return;
		// succeeded to trash: append to list
		Append(tmppath, tmpname);
		return;
	}
	string text = cxFileManager->GetFileName(alias);
    unsigned typ;
	struct stat st;
	if (lstat(path.c_str(), &st)) {
		typ = TMagic::LOCKED;
		flags |= CFileListItem::LOCKED;
	} else {
		if (S_ISLNK(st.st_mode)) {
			flags |= CFileListItem::LINKED;
			if (stat(path.c_str(), &st)) {
				typ = TMagic::LOCKED;
				flags |= CFileListItem::LOCKED;
			}
		}
		ifstream in;
		in.open(path.c_str());
		if (!in) {
			typ = TMagic::LOCKED;
			flags |= CFileListItem::LOCKED;
		} else typ = mMagic.GetFileType(*in.rdbuf(), st, alias);
	}
	if (flags & CFileListItem::LOCKED) {
		st.st_size = 0;
		st.st_mtime = 0;
	}
	CIcon icon(CMagic::GetIcon(typ), mIconSize);
	unsigned h = max(mIconHeight, (unsigned)GetTextHeight() + 2 * TEXT_BORDER);
	unsigned w = mIconWidth + GetTextWidth(cxWiden(text)) + 2 * TEXT_BORDER;
	text += '\0' + name + '\0' + cxFileManager->GetFilePath(alias);
	CFileListItem it(icon, text, st, typ, w, h);
	if (flags & CFileListItem::LOCKED) it.IsLocked(true);
	if (flags & CFileListItem::LINKED) it.IsLink(true);
	if (flags & CFileListItem::DIRECTORY) it.IsDirectory(true);
    mList.push_back(it);
}

/* virtual */ void CTrashListControl::DrawItem(unsigned idx)
{
	assert(idx < mList.size());
	CFileListItem item = GetItem(idx);
	unsigned row;
	unsigned col;
	// translate index to row and column
	col = idx / mRow.size();
	row = idx % mRow.size();
	int x;
	int y;
	if (col) x = mColumn[col - 1];
	else x = 0;
	if (row) y = mRow[row - 1];
	else y = 0;
	unsigned w = mColumn[col] - x;
	unsigned h = mRow[row] - y;
	WindowToScreen(x, y);
	ClearArea(x, y, w, h);
	int xi, yi;
	unsigned wi, hi;
	xi = x + ICON_BORDER + mIconWidth / 2;
	yi = y + h / 2;
	CColor color(TC_BACKGND_TEXT);
	if (item.GetIcon().IsValid()) DrawIcon(xi - item.GetIcon().GetWidth() / 2, yi - item.GetIcon().GetHeight() / 2, item.GetIcon());
	if (item.IsDirectory()) DrawIcon(xi - mLinkOverlay.GetWidth() / 2, yi - mLinkOverlay.GetHeight() / 2, mLinkOverlay);
	if ((item.GetType() & TMagic::COMPRESSED) != 0) DrawIcon(xi - mZipOverlay.GetWidth() / 2, yi - mZipOverlay.GetHeight() / 2, mZipOverlay);
	xi = x + mIconWidth + TEXT_BORDER;
	yi = y + (h - GetTextHeight()) / 2 -TEXT_BORDER;
	wi = item.GetWidth() - mIconWidth;
	hi = GetTextHeight() + 2 * TEXT_BORDER;
	if (item.IsSelected()) {
		if (IsFocused()) {
			color = CColor(TC_TITLE_TEXT);
			FillRectangle(xi, yi, wi, hi, CColor(TC_TITLE));
		} else {
			DrawDottedRectangle(xi, yi, wi, hi, color);
		}
	} //else SetTextGC(CGC::BACKGND);
	xi += TEXT_BORDER;
	yi += TEXT_BORDER;
	if (!item.GetText().empty()) DrawString(xi, yi, color, item.GetText().c_str(), !item.IsSelected());
	// restore the border
	if ((x < BORDER_WIDTH) || (x + w + BORDER_WIDTH >= GetWidth()) || (y + h + BORDER_WIDTH >= GetHeight())) {
		DrawFrame();
	}		
}

void CTrashListControl::OnProperties()
{
	if (GetSelCount() != 1) return ;
	int x(10);
	int y(10);
	GetParent()->TranslateCoordinates(x, y, NULL);
	const char* fn = GetItem(GetSelection().front()).GetText().c_str();
	fn += strlen(fn) + 1;
	string filename = cxFileManager->GetAbsolutePath(fn, cxFileManager->GetTrashDir());
	CPropertyDialog dlg(this, NULL, filename);
	dlg.Move(x, y);
	dlg.Show();
	dlg.DoModal();
}

void CTrashListControl::CreateContextMenu()
{
	// --- CONTEXT ---
	mContext.Entry(CUntrashDialog::IDEDIT_DEL, _("Delete", 73), CIcon::EDITDELETE);
	mContext.Seperator();
	mContext.Entry(CUntrashDialog::IDOPEN_WITH, _("Open with ...", 74));
	mContext.Seperator();
	mContext.Entry(CUntrashDialog::IDLISTPROP, _("Properties ...", 75));
}
