/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CXplorerFrame.cpp
  Info   : $Id: CXplorerFrame.cpp,v 1.118 2012/08/06 17:36:40 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "Xplorer.h"
#include "CXplorerFrame.h"
#include "CFileAssoc.h"
#include "CTapeArchive.h"
#include "CUntrashDialog.h"
#include "CTermConfig.h"

// C++X headers
#include <cxxx/CHSplitterWindow.h>
#include <cxxx/CProfile.h>
#include <cxxx/CMessageBox.h>
#include <cxxx/CApplication.h>
#include <cxxx/CFileHandler.h>
#include <cxxx/CDeskTopEntry.h>

// C++ headers
#include <cerrno>
#include <cstring>
#include <fstream>
#include <cctype>
#include <sstream>

// X headers
#include <X11/cursorfont.h>

// C headers
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>

// Namespace declarations
using std::string;
using std::vector;
using std::ofstream;
using std::ostringstream;
using cxxx::cxApp;
using cxxx::cxProfile;
using cxxx::cxError;
using cxxx::cxFileManager;
using cxxx::cxHomeDir;
using cxxx::cxDpy;
using cxxx::filesort::ASCENDING;
using cxxx::filesort::DIRFIRST;
using cxxx::filesort::HIDDEN;
using cxxx::filesort::NAME;
using cxxx::cxgettext;
using cxxx::cxSelManager;
using cxxx::cxNarrow;

Atom CXplorerFrame::TEXT_URI_LIST = None;
Atom CXplorerFrame::DND_USERNAME = None;
Atom CXplorerFrame::USERNAME = None;
Atom CXplorerFrame::HOSTNAME = None;

CXplorerFrame::CXplorerFrame()
: CTopLevelWindow(),
  mTree(this, IDTREE),
  mList(this, IDLIST),
  mMenu(this),
  mTool(this),
  mSplitter(this, IDSPLITTER),
  mStatusBar(this),
  mFlags(VIEW10K),
  mTreeWidth(150),
  mMagicFile(NULL),
  mIsArchive(false)
{
	SetMinSize(250, 200);
	TEXT_URI_LIST = GetAtomFromName("text/uri-list");
	DND_USERNAME = GetAtomFromName("text/x-xdnd-username");
	USERNAME = GetAtomFromName("USER");
	HOSTNAME = GetAtomFromName("HOST");
	string magfile = cxProfile->GetFileName();
	string::size_type pos = magfile.rfind('/');
	magfile.resize(pos);
	assert(!magfile.empty());
	magfile += "/action.conf";
	magfile = cxProfile->Read("Magic File", magfile);
	if (access(magfile.c_str(), R_OK | W_OK) != 0) {
		if (errno != ENOENT) cxError << _("Could not access magic configuration file {1}\n{2}", 148) << magfile << strerror(errno) << CError::Error;
		else {
			ofstream ost(magfile.c_str());
			if (!ost) cxError << _("Could not create magic configuration file {1}\n{2}", 151) << magfile << strerror(errno) << CError::Error;
		}
	}
	mMagicFile = new CProfile(magfile);
	mExecPath = cxProfile->Read("Executables Path", "/");
	mTerminalCmd = cxProfile->Read("Terminal Command", "");
	mTerminalArgs = cxProfile->Read("Terminal Arguments", "");
	SetFlags(cxProfile->Read("Flags", mFlags));
	SetSortKey(cxProfile->Read("Sort Flags", ASCENDING | NAME | DIRFIRST));
	mTreeWidth = cxProfile->Read("Tree Width", mTreeWidth);
	mMenu.SetDirList(cxProfile->Read("Recently Visited", ""));
	int w = cxProfile->Read("Width", 400);
	int h = cxProfile->Read("Height", 200);
	Resize(w, h);
	mStatusBar.AddField('L', 5000);
	if (mFlags & VIEW_MULTICOLUMN) {
		mList.SetStyle(CListControl::MULTI_COLUMN_VIEW, 16, 16);
		mMenu.Check(IDVIEW_MULI_COLUMN);
	} else {
		mList.SetStyle(CListControl::ICON_VIEW, 32, 32);
		mMenu.Check(IDVIEW_ICON);
	}		
	SetDirectory("/");
	mTree.Show();
	mList.Show();
	mMenu.Show();
	mSplitter.Show();
	if (mFlags & TOOLBAR) {
		mTool.Show();
		mMenu.Check(IDTOOL);
	}
	if (mFlags & STATUSBAR) {
		mStatusBar.Show();
		mMenu.Check(IDSTATUS);
	}
	if (mSortFlags & HIDDEN) mMenu.Check(IDSHOW_HIDDEN);
	SetFocusWindow(&mList);
	Show();
}

/* virtual */ CXplorerFrame::~CXplorerFrame()
{
	cxProfile->Write("Flags", mFlags);
	cxProfile->Write("Sort Flags", mSortFlags);
	cxProfile->Write("Tree Width", mTreeWidth);
	cxProfile->Write("Width", GetWidth());
	cxProfile->Write("Height", GetHeight());
	cxProfile->Write("Executables Path", mExecPath);
	cxProfile->Write("Terminal Command", mTerminalCmd);
	cxProfile->Write("Terminal Arguments", mTerminalArgs);
	cxProfile->Write("Recently Visited", mMenu.GetDirList());
	delete mMagicFile;
	cxProfile->Flush();
}

void CXplorerFrame::SetDirectory(std::string dir)
{
	// strip trailing slash
	if (dir.size() > 1 && dir[dir.size() - 1] == '/') dir.resize(dir.size() - 1);
	// is this an archive?
	mIsArchive = false;
	struct stat st;
	if (access(dir.c_str(), F_OK) == 0) {
		if (stat(dir.c_str(), &st)) {
			cxError << _("Failed to read file status of {1}", 145) << dir << CError::Error;
			return;
		}
		if (!S_ISDIR(st.st_mode)) mIsArchive = true;
	} else mIsArchive = true;
	if (mIsArchive) {
		// maybe a tape archive
		// if we're lucky, the archive still exists
		if (mArchive.IsValid() && dir.size() > mArchive.GetName().size()) {
			if (dir.compare(0, mArchive.GetName().size(), mArchive.GetName()) != 0) mIsArchive = false;
		} else mIsArchive = false;
		if (!mIsArchive) {
			// archive not available, try to create one
			string tmp = dir;
			while (access(tmp.c_str(), F_OK) != 0) tmp = cxFileManager->GetFilePath(tmp);
			string name = tmp;
			// tmpdir holds an existing file, see if it's an archive
			CMagic mag;
			unsigned typ = mag.GetFileType(tmp);
			// decompress if necessary
			if (typ & TMagic::COMPRESSED) {
				tmp = cxFileManager->Decompress(tmp);
				if (tmp.empty()) return;
				typ &= TMagic::UNCOMPRESSED;
			}
			if (typ == TMagic::POSIXTAR || typ == TMagic::GNUTAR || typ == TMagic::UNKTAR)	{
				// gotcha!
				mArchive.Open(name, tmp);
				mIsArchive = mArchive.IsValid();
			}
		}
		if (!mIsArchive) {
			cxError << _("CXplorerFrame::SetDirectory({1}): Not a directory", 140) << dir << CError::Error;
			return;
		}
	}
	if (mFlags & STATUSBAR) mStatusBar.SetText(0, _("Reading directory...", 121));
	if (mIsArchive) {
		mCaption = cxFileManager->GetFileName(mArchive.GetName());
		mCaption += ":/";
		mCaption += dir.substr(mArchive.GetName().size());
	} else mCaption = dir;
	mCaption += " - Xplorer";
	SetCaption(mCaption);
	mCurrentDir = dir;
	if (mIsArchive) {
		mList.SetDirectory(dir, &mArchive);
		mTree.SetDirectory(dir, &mArchive);
	} else {
		mList.SetDirectory(dir, NULL);
		mTree.SetDirectory(dir, NULL);
		mMenu.SetDirectory(dir);
		mHistory.SetCurrent(dir);
	}
	UpdateStatusBar();
}

void CXplorerFrame::Configure()
{
	mSplitter.SetRange(MIN_CONTROL_WIDTH, GetWidth() - MIN_CONTROL_WIDTH);
	int x, y, w, h, s;
	x = 0;
	y = 0;
	w = GetWidth();
	h = GetHeight();
	s = mTreeWidth;
	mMenu.Resize(w, 20);
	mMenu.Move(x, y);
	y += 20;
	h -= 20;
	if (mFlags & TOOLBAR) {
		mTool.Resize(w, 30);
		mTool.Move(x, y);
		y += 30;
		h -= 30;
	}
	if (mFlags & STATUSBAR) {
		h -= mStatusBar.GetHeight();
		mStatusBar.Move(x, y + h);
		mStatusBar.Resize(w, mStatusBar.GetHeight());
	}
	mTree.Resize(s, h);
	mTree.Move(x, y);
	s += x;
	mSplitter.Resize(5, h);
	mSplitter.Move(s, y);
	s += 5;
	mList.Resize(w - s, h);
	mList.Move(s, y);
}

/* virtual */ void CXplorerFrame::OnShortcut(wchar_t key, unsigned state __attribute__ ((unused)))
{
	key = toupper(key);
	switch (key) {
	case 'A':  //select all
		OnMessage(IDEDIT_SELA, MT_ACCELERATOR, 0);
		break;
	case 'C':  //copy
		OnMessage(IDEDIT_COPY, MT_ACCELERATOR, 0);
		break;
	case 'H':  // show hidden files
		OnMessage(IDSHOW_HIDDEN, MT_ACCELERATOR, 0);
		break;
	case 'Q':  //quit
		OnMessage(IDFILE_EXIT, MT_ACCELERATOR, 0);
		break;
	case 'T':  //open terminal
		OnMessage(IDOPEN_TERM, MT_ACCELERATOR, 0);
		break;
	case 'V':  //paste
		OnMessage(IDEDIT_PASTE, MT_ACCELERATOR, 0);
		break;
	case 'X':  //cut
		OnMessage(IDEDIT_CUT, MT_ACCELERATOR, 0);
		break;
	default:
		break;
	}
}

/* virtual */ void CXplorerFrame::OnKey(wchar_t key, unsigned state)
{
	switch (key) {
	case L'\x7f':
		if (state & ShiftMask) OnMessage(IDEDIT_DEL, MT_ACCELERATOR, 0);
		else OnMessage(IDEDIT_TRASH, MT_ACCELERATOR, 0);
		return ;
	case CK_Left:
		if (state & Mod1Mask) OnMessage(IDNAV_BACK, MT_ACCELERATOR, 0);
		break;
	case CK_Up:
		if (state & Mod1Mask) OnMessage(IDNAV_UP, MT_ACCELERATOR, 0);
		break;
	case CK_Right:
		if (state & Mod1Mask) OnMessage(IDNAV_FORWARD, MT_ACCELERATOR, 0);
		break;
	default:
		CTopLevelWindow::OnKey(key, state);
	}
}

/* virtual */ void CXplorerFrame::OnMessage(unsigned msgid, TMessage type, unsigned value)
{
	switch (msgid) {
	case IDTREE:
		if (mTree.GetSelCount() == 1) {
			if (type == MT_SELECTION_CHANGED) {
				string dir(mTree.GetPath(mTree.GetSelection().front()));
				if (access(dir.c_str(), R_OK | X_OK) == 0) {
					SetDirectory(dir);
					//mHistory.SetCurrent(mCurrentDir);
				}
			} else if (type == MT_CONFIRM_SELECTION) {
				unsigned idx = mTree.GetSelection().front();
				if (access(mTree.GetPath(idx).c_str(), R_OK | X_OK) == 0) {
					if (mTree.GetItem(idx).IsOpened()) mTree.Close(idx);
					else mTree.Open(idx);
				}
			}
		}
		break;
	case IDLIST:
		if (type == MT_CONFIRM_SELECTION) OnListSelConfirmed();
		else if (type == MT_SELECTION_CHANGED && (mFlags & STATUSBAR) != 0) UpdateStatusBar();
		break ;
	case IDTOOL:
		if (type == MT_CHECK) {
			mFlags |= TOOLBAR;
			mTool.Show();
			Configure();
		} else if (type == MT_UNCHECK) {
			mFlags &= ~TOOLBAR;
			mTool.Hide();
			Configure();
		}
		break;
	case IDSTATUS:
		if (type == MT_CHECK) {
			mFlags |= STATUSBAR;
			mStatusBar.Show();
			Configure();
		} else if (type == MT_UNCHECK) {
			mFlags &= ~STATUSBAR;
			mStatusBar.Hide();
			Configure();
		}
		break;
	case IDSPLITTER:
		if (type == MT_MOTION) {
			mTreeWidth = value;
			mTree.Resize(value, mTree.GetHeight());
			mTree.Move(0, mTree.GetYPos());
			mSplitter.Move(value, mSplitter.GetYPos());
			value += mSplitter.GetWidth();
			mList.Resize(GetWidth() - value, mList.GetHeight());
			mList.Move(value, mList.GetYPos());
		}
		break;
	case IDSHOW_HIDDEN:
		if (type == MT_CHECK && (mSortFlags & HIDDEN) == 0) {
			SetSortKey(mSortFlags | HIDDEN);
			Reload();
		} else if (type == MT_UNCHECK && (mSortFlags & HIDDEN) != 0) {
			SetSortKey(mSortFlags & ~HIDDEN);
			Reload();
		} else if (type == MT_ACCELERATOR) {
			SetSortKey(mSortFlags ^ HIDDEN);
			Reload();
		}
		break;
	case IDVIEW_REFRESH:
		Reload();
		break;
	case IDVIEW_SORT:
		OnSortDialog();
		break;
	case IDOPEN_TERM:
		OnOpenTerminal();
		break;
	case IDCONFIGURE_TERM:
		OnConfigureTerminal();
		break;
	case IDNEW_DIR:
		OnNewDialog(CNewDialog::DIRECTORY);
		break;
	case IDNEW_TEXT:
		OnNewDialog(CNewDialog::TEXT_FILE);
		break;
	case IDNAV_FORWARD:
		if (mHistory.CanForward() == 0) break;
		SetDirectory(mHistory.GetNext());
		break;
	case IDNAV_BACK:
		if (mHistory.CanBack() == 0) break;
		SetDirectory(mHistory.GetPrev());
		break;
	case IDNAV_UP:
		{
			string tmp(cxFileManager->GetFilePath(mCurrentDir));
			if (tmp.empty()) tmp += '/';
			SetDirectory(tmp);
		}
		break;
	case IDNAV_HOME:
		SetDirectory(cxHomeDir);
		break;
	case IDFILE_EXIT:
		cxApp->Quit();
		break;
	case IDEDIT_CUT:
        OnCut();
        break;
    case IDEDIT_COPY:
        OnCopy();
        break;
    case IDEDIT_PASTE:
        OnPaste();
        break;
    case IDEDIT_DEL:
        OnDelete();
        break;
    case IDEDIT_TRASH:
        OnTrash();
        break;
	case IDEDIT_SELA:
		mList.SelectAll();
		UpdateStatusBar();
		break;
	case IDEDIT_SELR:
		mList.InvertSelection();
		UpdateStatusBar();
		break;
	case IDLISTPROP:
		mList.OnProperties();
		break;
	case IDTREEPROP:
		mTree.OnProperties();
		break;
	case IDFILE_OPEN:
		OnOpenFile();
		break;
	case IDABOUT_CXXX:
		OnAboutCxxx();
		break;
	case IDABOUT_XPLORER:
		OnAboutXplorer();
		break;
	case IDOPEN_WITH:
		OnOpenWith();
		break;
	case IDVIEW_PREVIEW:
		OnPreviewDialog();
		break;
	case IDSETTINGS:
		OnSettings();
		break;
	case IDVIEW_MULI_COLUMN:
		if ((mFlags & VIEW_MULTICOLUMN) == 0) {
			mList.SetStyle(CListControl::MULTI_COLUMN_VIEW, 16, 16);
			mFlags |= VIEW_MULTICOLUMN;
			Reload();
		}
		break;
	case IDVIEW_ICON:
		if (mFlags & VIEW_MULTICOLUMN) {
			mList.SetStyle(CListControl::ICON_VIEW, 32, 32);
			mFlags &= ~VIEW_MULTICOLUMN;
			Reload();
		}
		break;
	case IDTRASH:
		{	
			int x = 10;
			int y = 10;
			mTree.TranslateCoordinates(x, y, NULL);
			CUntrashDialog ud(this);
			ud.Move(x, y);
			ud.Show();
			ud.DoModal();
		}
		break;	
    default:
		if (msgid >= IDNAV_RECENT && msgid < IDNAV_RECENT + CRecentMenu::RECENTSIZE) {
			string tmp = mMenu.GetDirectory(msgid);
			if (!tmp.empty()) SetDirectory(tmp);
		}
		break;
	}
}

void CXplorerFrame::OnConfigureTerminal()
{
	int x = 10;
	int y = 10;
	mTree.TranslateCoordinates(x, y, NULL);
	CTermConfig dlg(this, mTerminalCmd, mTerminalArgs);
	dlg.Move(x, y);
	dlg.Show();
	dlg.SetBinPath(mExecPath);
	if (dlg.DoModal() == CTermConfig::OK) {
		mTerminalCmd = dlg.GetTerminalCommand();
		mTerminalArgs = dlg.GetTerminalArgs();
	}
}

const string& CXplorerFrame::GetTerminalCommand()
{
	if (mTerminalCmd.empty()) OnConfigureTerminal();
	if (!mTerminalCmd.empty() && mTerminalCmd[0] != '/') {
		string cmd;
		const vector<string>& sp = cxFileManager->GetSearchPath();
		for (vector<string>::const_iterator i = sp.begin(); i != sp.end(); ++i) {
			cmd = *i + '/' + mTerminalCmd;
			if (access(cmd.c_str(), X_OK) == 0) break;
		}
		mTerminalCmd = cmd;
	}
	if (access(mTerminalCmd.c_str(), X_OK)) cxError<<_("Cannot open terminal", 176)<<CError::Error;
	return mTerminalCmd;
}

/* virtual */ bool CXplorerFrame::OnUpdate(unsigned msgid) const
{
	bool ret;
	switch (msgid) {
	case IDEDIT_CUT:
	case IDEDIT_TRASH:
	case IDEDIT_DEL:
		ret = mList.CanPaste() && mList.GetSelCount() > 0;
		break;
	case IDEDIT_COPY:
		ret = mList.GetSelCount() > 0;
		break;
	case IDOPEN_WITH:
		ret = mList.GetSelCount() == 1;
		break;
	case IDEDIT_PASTE:
	case IDNEW_DIR:
	case IDNEW_TEXT:
		ret = mList.CanPaste();
		break;
	case IDNAV_HOME:
	case IDTOOL:
	case IDSTATUS:
	case IDSHOW_HIDDEN:
	case IDVIEW_REFRESH:
	case IDVIEW_SORT:
	case IDVIEW_PREVIEW:
	case IDOPEN_TERM:
	case IDCONFIGURE_TERM:
	case IDFILE_EXIT:
	case IDEDIT_SELA:
	case IDEDIT_SELR:
	case IDABOUT_CXXX:
	case IDABOUT_XPLORER:
	case IDTREEPROP:
	case IDFILE_OPEN:
	case IDSETTINGS:
	case IDVIEW_ICON:
	case IDVIEW_MULI_COLUMN:
	case IDTRASH:
		ret = true;
		break;
	case IDLISTPROP:
		ret = mList.GetSelCount() <= 1;
		break;
	case IDNAV_FORWARD:
		ret = mHistory.CanForward();
		break;
	case IDNAV_BACK:
		ret = mHistory.CanBack();
		break;
	case IDNAV_UP:
		ret = mCurrentDir != "/";
		break;
	default:
		if (msgid >= IDNAV_RECENT && msgid < IDNAV_RECENT + CRecentMenu::RECENTSIZE) ret = true;
		else ret = false;
		break;
	}
	return ret;
}

void CXplorerFrame::OpenFile(string path, unsigned type)
{
	// decompression changes path, save it to display the correct path 
	// when (eventually) calling GetFileAssoc
	const string savepath(path);
	if (type == TMagic::DIRECTORY || type == TMagic::TARDIR) {
		SetDirectory(path);
		return ;
	}
	if (mIsArchive) {
		assert(path.size() >= mArchive.GetName().size());
		assert(path.compare(0, mArchive.GetName().size(), mArchive.GetName()) == 0);
		if (path.size() == mArchive.GetName().size()) path = "";
		else path = path.substr(mArchive.GetName().size() + 1);
		string tmp = mArchive.Extract(path);
		if (tmp.empty()) {
			cxError << _("Failed to extract file {1} from archive.", 147) << path << CError::Error;
			return;
		} else path = tmp;
	}
	if (type == TMagic::DESKTOP || type == TMagic::KDEDESK) {
		CDeskTopEntry de(path);
		string cmd = de.GetExec();
		if (!cmd.empty()) {
			// TODO: prepare string
			cxFileManager->Execute(cmd);
		}
		return;
	} 
	if (type & TMagic::COMPRESSED) {
		type &= TMagic::UNCOMPRESSED;
		path = cxFileManager->Decompress(path);
		if (path.empty()) return;
	}
	if (type == TMagic::ELFEXE) {
		string cmd = path;
		string::size_type pos = 0;
		while ((pos = cmd.find(' ', pos)) < cmd.size()) {
			cmd.replace(pos, 1, "\\ ");
			pos += 2;
		}
		cxFileManager->Execute(cmd);
		return;
	}
	if ((type == TMagic::POSIXTAR || type == TMagic::GNUTAR || type == TMagic::UNKTAR) && !mIsArchive) {
		mArchive.Open(savepath, path);
		SetDirectory(savepath);
		return;
	}
	if (type == TMagic::EMPTY || type == TMagic::LOCKED || type == TMagic::LOCKDIR) return;
	string cmd(mMagicFile->Read(CMagic::GetMimeType(type), "?"));
	if (cmd == "?") {
		cmd = GetFileAssoc(savepath, type);
		if (cmd.empty()) return;
	} else if (cmd == "*") return;
	string::size_type pos = 0;
	while ((pos = path.find(' ', pos)) < path.size()) {
		path.replace(pos, 1, "\\ ");
		pos += 2;
	}
	while ((pos = cmd.find("%s")) != string::npos) cmd.replace(pos, 2, path);
	cxFileManager->Execute(cmd, mCurrentDir);
}

void CXplorerFrame::OnCut()
{
	if (mList.GetSelCount() == 0 || !GrabSelection(CSelection::CLIPBOARD, CurrentTime)) return ;
	vector<string> alias = cxFileManager->Trash(GetListSelection());
	mSelection.clear();
	string path = cxFileManager->GetTrashDir();
	if (!path.empty() && path[path.size() - 1] != '/') path += '/';
	unsigned pathlen = path.size();
	for (vector<string>::iterator i = alias.begin(); i != alias.end(); ++i) {
		path.resize(pathlen);
		path += *i;
		mSelection.push_back(path);
	}
	mList.DeselectAll();
	Reload();
}

void CXplorerFrame::OnDelete()
{
	string text, caption;
	if (mList.GetSelCount() == 1) {
		text = _("Really delete file \'%s\'?\n\nThere is no way to restore the file!", 59);
		string::size_type pos = text.find("%s");
		if (pos != string::npos) text.replace(pos, 2, mList.GetItem(mList.GetSelection().front()).GetText());
		caption = _("Delete File", 60);
	} else {
		text = _("Really delete the selected files?\n\nThere is no way to restore the files!", 61);
		caption = _("Delete Files", 62);
	}
	CMessageBox mb (text, caption, CMessageBox::OK | CMessageBox::CANCEL, CIcon::WARNING);
	mb.FocusButton(CMessageBox::CANCEL);
	if (mb.DoModal() ==	CMessageBox::OK) {
		cxFileManager->Delete(GetListSelection());
		Reload();
	}
}

void CXplorerFrame::OnCopy()
{
	if (mList.GetSelCount() == 0 || !GrabSelection(CSelection::CLIPBOARD, CurrentTime)) return;
	if (mIsArchive) {
		string clip = cxSelManager->GetClipSelDirectory();
		if (access(clip.c_str(), F_OK) != 0) {
			if (mkdir(clip.c_str(), S_IRWXU) != 0) return;
		}
		clip += '/';
		if (!cxFileManager->ClearDirectory(clip)) return;
		string fn;
		string archdir = mCurrentDir.substr(mArchive.GetName().size()) + "/";
		mSelection.clear();
		for (vector<unsigned>::const_iterator i = mList.GetSelection().begin(); i < mList.GetSelection().end(); ++i) {
			fn = mList.GetItem(*i).GetText();
			mArchive.Extract(archdir + fn, clip);
			mSelection.push_back(clip + fn);			
		}			
	} else mSelection = GetListSelection();
}

vector<string> CXplorerFrame::GetListSelection()
{
    vector<string> v;
    for (vector<unsigned>::const_iterator i = mList.GetSelection().begin(); i != mList.GetSelection().end(); ++i) v.push_back(mList.GetPath(*i));
    return v;
}

bool CXplorerFrame::GetSelection(vector<string>& ret)
{
	string name;
	vector<unsigned char> v;
	if (RequestSelection(CSelection::CLIPBOARD, USERNAME, USERNAME, v) || RequestSelection(CSelection::CLIPBOARD, DND_USERNAME, USERNAME, v)) {
		name.assign((const char*)&v[0], v.size());
	}
	if (!name.empty()) {
		while (iscntrl(name[name.size() - 1])) name.resize(name.size() - 1);
		if (name != cxApp->GetUsername()) return false;
	}
	if (!RequestSelection(CSelection::CLIPBOARD, TEXT_URI_LIST, TEXT_URI_LIST, v)) return false;
	vector<unsigned char>::iterator i1, i2;
	i1 = v.begin();
	while (i1 < v.end()) {
		i2 = std::find(i1, v.end(), '\n');
		string s((char*)&v[i1 - v.begin()], i2 - i1);
		while (!s.empty() && iscntrl(s[s.size() - 1])) s.resize(s.size() - 1);
		if (!s.empty() && cxFileManager->ToLocalFS(s)) ret.push_back(s);
		i1 = ++i2;
	}
	return true;
}

/* virtual */ bool CXplorerFrame::ConvertSelection(Atom sel, Atom typ, std :: vector<unsigned char>& ret)
{
	if (sel != CSelection::CLIPBOARD) return false;
	ret.clear();
	const unsigned char* sbeg(NULL);
	const unsigned char* send(NULL);
	if (typ == CSelection::TARGETS) {
		Atom a;
		sbeg = (unsigned char*)&a;
		send = sbeg + sizeof(Atom);
		a = DND_USERNAME;
		copy(sbeg, send, back_inserter(ret));
		a = TEXT_URI_LIST;
		copy(sbeg, send, back_inserter(ret));
		a = USERNAME;
		copy(sbeg, send, back_inserter(ret));
		a = HOSTNAME;
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	if (typ == TEXT_URI_LIST) {
		string s;
		for (vector<string>::iterator i = mSelection.begin(); i != mSelection.end(); ++i) {
			s += "file://";
			s += cxApp->GetHostname();
			s += *i;
			s += "\n";
		}
		sbeg = (const unsigned char*)&s[0];
		send = sbeg + s.size();
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	if (typ == DND_USERNAME || typ == USERNAME) {
		sbeg = (const unsigned char*)(cxApp->GetUsername().data());
		send = sbeg + cxApp->GetUsername().size();
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	if (typ == HOSTNAME) {
		sbeg = (const unsigned char*)(cxApp->GetHostname().data());
		send = sbeg + cxApp->GetHostname().size();
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	return false;
}

void CXplorerFrame::UpdateStatusBar()
{
	if (!(mFlags & STATUSBAR)) return;
	ostringstream os;
	unsigned dirs(0);
	unsigned files(0);
	unsigned size(0);
	if (mList.GetSelCount() == 0) {
		os << _("Total:", 118) << " ";
		for (unsigned i = 0; i < mList.GetItemCount(); ++i) {
			CFileListItem item(mList.GetItem(i));
			if (item.GetType() == TMagic::DIRECTORY || item.GetType() == TMagic::LOCKDIR || item.GetType() == TMagic::TARDIR) ++dirs;
			else {
				size += item.GetSize();
				++files;
			}
		}
	} else {
		os << _("Selection:", 119) << " ";
		for (unsigned i = 0; i < mList.GetSelCount(); ++i) {
			CFileListItem item(mList.GetItem(mList.GetSelection()[i]));
			if (item.GetType() == TMagic::DIRECTORY || item.GetType() == TMagic::LOCKDIR || item.GetType() == TMagic::TARDIR) ++dirs;
			else {
				size += item.GetSize();
				++files;
			}
		}
	}
	if (dirs != 0) {
		os << dirs << " ";
		if (dirs > 1) os << _("directories", 120);
		else os << _("directory", 21);
		if (files != 0) os << ", ";
	}
	if (files != 0)	{
		os << files << " ";
		if (files > 1) os << _("files", 116);
		else os << _("file", 117);
		os << " (" << size << " B)";
	}
	mStatusBar.SetText(0, os.str());
}

string CXplorerFrame::GetFileAssoc(const string& name, unsigned type)
{
	string ret;
	bool rem = (type != TMagic::BACKUP && type != TMagic::DESKTOP && type != TMagic::KDEDESK && type != TMagic::NAUTLINK && type != TMagic::UNKARCH && type != TMagic::UNKELF && type != TMagic::UNKIFF && type != TMagic::UNKNOWN && type != TMagic::UNKRIFF && type != TMagic::UNKTAR);
	int x = 10;
	int y = 10;
	mTree.TranslateCoordinates(x, y, NULL);
	CFileAssoc dlg(this, name, type, rem);
	dlg.SetBinPath(mExecPath);
	dlg.Move(x, y);
	dlg.Show();
	if (dlg.DoModal() == CFileAssoc::OK) {
		if (dlg.NeedTerminal()) ret = mTerminalCmd + " " + mTerminalArgs + " ";
		if (dlg.IsExecutable()) {
			ret += "%s";
			if (!dlg.GetArguments().empty()) ret += " " + dlg.GetArguments();
		} else {
			string execpath = dlg.GetPath();
			string::size_type pos = 0;
			while ((pos = execpath.find(' ', pos)) < execpath.size()) {
				execpath.replace(pos, 1, "\\ ");
				pos += 2;
			}
			ret += execpath;
			if (!dlg.GetArguments().empty()) ret += " " + dlg.GetArguments();
			ret	+= " %s";
			execpath = dlg.GetPath();
			pos = execpath.rfind('/');
			if (pos != string::npos) {
				execpath.resize(pos);
				if (execpath.empty()) execpath = "/";
				if (execpath != mExecPath) mExecPath = execpath;
			}
		}
		if (dlg.RememberAssoc()) {
			mMagicFile->Write(CMagic::GetMimeType(type), ret);
			mMagicFile->Flush();
		}
	} else ret = "";
	return ret;
}

void CXplorerFrame::OnOpenWith()
{
	if (mList.GetSelCount() == 1) {
		unsigned idx = mList.GetSelection().front();
		CFileListItem sel = mList.GetItem(idx);
		string cmd = GetFileAssoc(mList.GetPath(idx), sel.GetType());
		if (cmd.empty()) return;
		string::size_type pos = 0;
		string file = cxFileManager->GetAbsolutePath(sel.GetText(), mCurrentDir);
		if (mIsArchive) {
			assert(file.size() >= mArchive.GetName().size());
			assert(file.compare(0, mArchive.GetName().size(), mArchive.GetName()) == 0);
			if (file.size() == mArchive.GetName().size()) file = "";
			else file = file.substr(mArchive.GetName().size() + 1);
			file = mArchive.Extract(file);
			if (file.empty()) {
				cxError << _("Failed to extract file {1} from archive.", 147);
				return;
			}
		}
		while ((pos = file.find(' ', pos)) < file.size()) {
			file.replace(pos, 1, "\\ ");
			pos += 2;
		}
		while ((pos = cmd.find("%s")) != string::npos) cmd.replace(pos, 2, file);
		cxFileManager->Execute(cmd, mCurrentDir);
	}
}

void CXplorerFrame::Select(const string& filename)
{
	for (unsigned i = 0; i < mList.GetItemCount(); ++i) {
		if (mList.GetItem(i).GetText() == filename) {
			mList.Select(i);
			break;
		}
	}
}
