/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CXplorerList.cpp
  Info   : $Id: CXplorerList.cpp,v 1.80 2012/09/12 16:01:43 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2001 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "Xplorer.h"
#include "CXplorerList.h"
#include "CXplorerFrame.h"

// C++X headers
#include <cxxx/CMenuItem.h>
#include <cxxx/CApplication.h>
#include <cxxx/CFileHandler.h>
#include <cxxx/CDeskTopEntry.h>
#include <cxxx/CImage.h>
#include <cxxx/CError.h>

// C++ headers
#include <cerrno>
#include <cstdio> // rename
#include <fstream>
#include <vector>

// X headers
#include <X11/Xatom.h>

// C headers
#include <unistd.h>
#include <stdio.h> // access

// Namespace declarations
using std::ios;
using std::ifstream;
using std::ofstream;
using std::string;
using std::wstring;
using std::streampos;
using std::vector;
using std::max;
using cxxx::cxApp;
using cxxx::cxFileManager;
using cxxx::cxgettext;
using cxxx::cxError;
using cxxx::cxSelManager;
using cxxx::cxNarrow;
using cxxx::cxWiden;

CXplorerList::CXplorerList(CWindow* parent, unsigned msgid)
: CFileListControl(parent, msgid),
  mContext(GetTopLevelWindow(), ""),
  mContextCreate(GetTopLevelWindow(), ""),
  mDnDSelection(-1u),
  mPostLoadTimer(this, IDPOSTLOAD),
  mXplorerFlags(0),
  mMaxPreviewSize(0),
  mArchive(NULL)
{
	mOpenDir.Create(CIcon::FOLDER_OPEN, mIconSize);
	mActiveExe.Create(CIcon::ACTIVE_EXEC, mIconSize);
	mMultipleSelection = true;
	CreateContextMenu();
	SetupDnD();
}

void CXplorerList::CreateContextMenu()
{
	// --- CONTEXT ---
	mContext.Entry(mContextCreate, _("Create New", 63), CIcon::FILENEW);
	mContext.Seperator();
	mContext.Entry(CXplorerFrame::IDNAV_UP, _("Up", 64), CIcon::UP);
	mContext.Entry(CXplorerFrame::IDNAV_BACK, _("Back", 65), CIcon::BACK);
	mContext.Entry(CXplorerFrame::IDNAV_FORWARD, _("Forward", 66), CIcon::FORWARD);
	mContext.Entry(CXplorerFrame::IDVIEW_REFRESH, _("Reload", 67), CIcon::RELOAD);
	mContext.Seperator();
	mContext.Entry(CXplorerFrame::IDEDIT_CUT, _("Cut", 69), CIcon::EDITCUT);
	mContext.Entry(CXplorerFrame::IDEDIT_COPY, _("Copy", 70), CIcon::EDITCOPY);
	mContext.Entry(CXplorerFrame::IDEDIT_PASTE, _("Paste", 71), CIcon::EDITPASTE);
	mContext.Seperator();
	mContext.Entry(CXplorerFrame::IDEDIT_TRASH, _("Move to trash", 72), CIcon::TRASHCAN_EMPTY);
	mContext.Entry(CXplorerFrame::IDEDIT_DEL, _("Delete", 73), CIcon::EDITDELETE);
	mContext.Seperator();
	mContext.Entry(CXplorerFrame::IDOPEN_WITH, _("Open with ...", 74));
	mContext.Seperator();
	mContext.Entry(CXplorerFrame::IDLISTPROP, _("Properties ...", 75));

	// --- CONTEXT --- CREATE NEW ---
	mContextCreate.Entry(CXplorerFrame::IDNEW_DIR, _("Directory", 76), CIcon::FOLDERNEW);
	mContextCreate.Entry(CXplorerFrame::IDNEW_TEXT, _("Text File", 77), CIcon::FILENEW);
}

/* virtual */ void CXplorerList::OnMessage(unsigned msgid, TMessage type, unsigned value)
{
	if (msgid == IDPOSTLOAD) {
		OnPostLoad();
		if (!mPostLoadList.empty()) mPostLoadTimer.SetTime(100);
	} else CFileListControl::OnMessage(msgid, type, value);
}

/* virtual */ void CXplorerList::ReadDirectory()
{
	mPostLoadTimer.Stop();
	mPostLoadList.clear();
	if (mArchive == NULL) CFileListControl::ReadDirectory();
	else {
		ReadArchive();
		Sort();
	}
}

void CXplorerList::ReadArchive()
{
	if (mCurrentDir.compare(0, mArchive->GetName().size(), mArchive->GetName()) != 0) {
		cxError << _("CXplorerList::ReadArchive(): Wrong archive.", 144) << CError::Error;
		return;
	}
	string dir;
	if (mCurrentDir.size() > mArchive->GetName().size()) dir = mCurrentDir.substr(mArchive->GetName().size() + 1); 
	vector<CFileListItem> v = mArchive->ListDirectory(dir);
	if (!mArchive->IsValid()) return;
	mFileWatch.Stop();
	mWriteAccess = false;
	for (vector<CFileListItem>::iterator i = v.begin(); i < v.end(); ++i) {
		// Append will not work here
		// The CFileListItems are o.k., except icon, width and height
		if (i->GetType() == TMagic::DIRECTORY) i->SetType(TMagic::TARDIR);
		CIcon icon(CMagic::GetIcon(i->GetType()), mIconSize);
		struct stat st;
		st.st_size = i->GetSize();
		st.st_mtime = i->GetTime();
		unsigned h;
		unsigned w;
		vector<CTextSizeHint> hint;
		wstring tmp(cxWiden(i->GetText()));
		if (mStyle == MULTI_COLUMN_VIEW) {
			h = max(mIconHeight, (unsigned)GetTextHeight() + 2 * TEXT_BORDER);
			w = mIconWidth + GetTextWidth(tmp) + 2 * TEXT_BORDER;
		} else {
			// mStyle == ICON_VIEW
			w = ICON_ITEM_WIDTH;
			// Text may be multiline, do the calculation
			//string tmp(i->GetText());
			unsigned d;
			unsigned pos;
			while (!tmp.empty()) {
				d = tmp.size() + 1;
				pos = 0;
				do {
					d = (d + 1) /2;
					if (pos + d <= tmp.size() && (unsigned)GetTextWidth(tmp, pos + d) <= w) pos += d;
				} while (d > 1);
				hint.push_back(CTextSizeHint(pos, GetTextWidth(tmp, pos) + 2 * TEXT_BORDER));
				tmp = tmp.substr(pos);
			}
			h = mIconHeight + (GetTextHeight() + 2 * TEXT_BORDER) * hint.size();
		}
		CFileListItem item(icon, i->GetText(), st, i->GetType(), w, h);
		item.IsLocked(i->IsLocked());
		item.IsLink(i->IsLink());
		item.IsDirectory(i->IsDirectory());
		for (vector<CTextSizeHint>::iterator vi = hint.begin(); vi != hint.end(); ++vi) item.SetTextHint(vi->mLength, vi->mWidth);
		mList.push_back(item);
	}
	Update();
}

/* virtual */ void CXplorerList::Append(const string& path, const string& name)
{
	CFileListControl::Append(path, name);
	CFileListItem it = (CFileListItem&)mList.back();
	if (it.GetSize() < mMaxPreviewSize) {
		bool tst;
		switch (it.GetType()) {
		case TMagic::DESKTOP:
		case TMagic::KDEDESK:
			tst = (mXplorerFlags & CXplorerFrame::VIEWDESKTOP);
			break;
		case TMagic::XPM:
		case TMagic::XCUR:
			tst = (mXplorerFlags & CXplorerFrame::VIEWXPM);
			break;
		case TMagic::PAM:
		case TMagic::PPM:
		case TMagic::PPMT:
		case TMagic::PGM:
		case TMagic::PGMT:
		case TMagic::PBM:
		case TMagic::PBMT:
			tst = mXplorerFlags & CXplorerFrame::VIEWPPM;
			break;
		case TMagic::MSBMP:
			tst = mXplorerFlags & CXplorerFrame::VIEWBMP;
			break;
		case TMagic::PNG:
			tst = mXplorerFlags & CXplorerFrame::VIEWPNG;
			break;
#ifdef CXXX_HAS_GIF
		case TMagic::GIF:
		case TMagic::GIF1:
			tst = mXplorerFlags & CXplorerFrame::VIEWGIF;
			break;
#endif // CXXX_HAS_GIF
#ifdef CXXX_HAS_JPEG
		case TMagic::JPEG:
			tst = mXplorerFlags & CXplorerFrame::VIEWJPG;
			break;
#endif // CXXX_HAS_JPEG
#ifdef CXXX_HAS_TIFF	
		case TMagic::TIFF:
			tst = mXplorerFlags & CXplorerFrame::VIEWTIFF;
			break;
#endif // CXXX_HAS_TIFF
		default:
			tst = false;
			break;
		}
		if (tst) {
			if (mPostLoadList.empty()) mPostLoadTimer.SetTime(100);
			mPostLoadList.push_back(it);
		}
	}
}

void CXplorerList::OnPostLoad()
{
	unsigned size = 0;
	while (!mPostLoadList.empty() && size < PREVIEWCHUNK) {
		CFileListItem item = mPostLoadList.back();
		mPostLoadList.pop_back();
		string path = cxFileManager->GetAbsolutePath(item.GetText(), mCurrentDir);
		if (item.GetType() == TMagic::DESKTOP || item.GetType() == TMagic::KDEDESK) {
			CDeskTopEntry ent;
			ent.Read(path);
			path = ent.GetIcon();
			if (!path.empty() && path[0] != '/') path = CIcon::FindFileName(path, mIconSize);
		}
		if (!path.empty()) {
			CIcon icon;
			icon.Create(path, mIconSize, mIconSize, CImage::SCALE_ASPECT | CImage::SCALE_DOWN);
			if (icon.IsValid()) {
				item.SetIcon(icon);
				unsigned idx = GetIndex(item);
				if (idx < mList.size()) DrawItem(idx);
			}
		}
		size += item.GetSize();
	}
}

/* virtual */ bool CXplorerList::OnDragInit(unsigned button, Atom& action)
{
	if (mArchive && (button == 1 || button == 3)) {
		string clip = cxSelManager->GetDnDSelDirectory();
		if (access(clip.c_str(), F_OK) != 0) {
			if (mkdir(clip.c_str(), S_IRWXU) != 0) return false;
		}
		clip +=  '/';
		if (!cxFileManager->ClearDirectory(clip)) return false;
		string fn;
		string archdir = mCurrentDir.substr(mArchive->GetName().size()) + "/";
		for (vector<unsigned>::const_iterator i = GetSelection().begin(); i != GetSelection().end(); ++i) {
			fn = GetItem(*i).GetText();
			mArchive->Extract(archdir + fn, clip);
		}
	}
	if (button == 1) {
        if (mWriteAccess && !mArchive) action = CDnDManager::DND_MOVE;
        else action = CDnDManager::DND_COPY;
		return true;
	}
    if (button == 3) {
		action = CDnDManager::DND_ASK;
		return true;
	}
    return false;
}

/* virtual */ bool CXplorerList::ConvertSelection(Atom sel, Atom typ, vector<unsigned char>& ret)
{
    if (sel != DND_SELECTION) return false;
    ret.clear();
    const unsigned char* sbeg;
    const unsigned char* send;
    if (typ == CSelection::TARGETS) {
        Atom a;
        sbeg = (unsigned char*)&a;
        send = sbeg + sizeof(Atom);
        a = CXplorerFrame::DND_USERNAME;
        copy(sbeg, send, back_inserter(ret));
        a = CXplorerFrame::TEXT_URI_LIST;
        copy(sbeg, send, back_inserter(ret));
        return true;
    }
    if (typ == CXplorerFrame::TEXT_URI_LIST) {
        string s;
	string clip = "file://" + cxApp->GetHostname();
		if (mArchive) {
			clip += cxSelManager->GetDnDSelDirectory() + "/";
			for (vector<unsigned>::const_iterator i = GetSelection().begin(); i != GetSelection().end(); ++i) {
				s += clip + GetItem(*i).GetText() + "\r\n";
			}
		} else {
			for (vector<unsigned>::const_iterator i = GetSelection().begin(); i != GetSelection().end(); ++i) {
				s += clip + GetPath(*i) + "\r\n";
			}
		}
        sbeg = (const unsigned char*)&s[0];
        send = sbeg + s.size();
        copy(sbeg, send, back_inserter(ret));
        return true;
    }
    if (typ == CXplorerFrame::DND_USERNAME) {
        sbeg = (const unsigned char*)(cxApp->GetUsername().data());
        send = sbeg + cxApp->GetUsername().size();
        copy(sbeg, send, back_inserter(ret));
        return true;
    }
    return false;
}

/* virtual */ Atom CXplorerList::GetDnDAction(int x, int y, Atom action)
{
	if (x < BORDER_WIDTH || y < BORDER_WIDTH || x > (int)mViewWidth + BORDER_WIDTH || y > (int)mViewHeight + BORDER_WIDTH) action = None;
	if (action != DND_COPY && action != DND_MOVE && action != DND_LINK && action != DND_ASK) action = None;
	if ((mDnDSelection < GetItemCount()) && mDnDIcon.IsValid()) {
		GetItem(mDnDSelection).SetIcon(mDnDIcon);
		mDnDIcon.Clear();
		DrawItem(mDnDSelection);
	}
	if (action == None) return None;
	mDnDSelection = ItemFromPos(x, y);
	bool wacc = mWriteAccess;
	unsigned w, h;
	GetBounds(x, y, w, h);
	SetIgnoreRect(x, y, w, h);
	if (mDnDSelection < GetItemCount()) {
		CFileListItem& it = GetItem(mDnDSelection);
		if (it.GetType() == TMagic::DIRECTORY) {
			wacc = (access(GetPath(mDnDSelection).c_str(), W_OK | X_OK) == 0);
			if (wacc) {
				mDnDIcon = it.GetIcon();
				it.SetIcon(mOpenDir);
			}
		} else if (it.GetType() == TMagic::ELFEXE) {
			wacc = (access(GetPath(mDnDSelection).c_str(), X_OK) == 0);
			if (wacc) {
				mDnDIcon = it.GetIcon();
				it.SetIcon(mActiveExe);
				action = DND_COPY;
			}
		}
		if ((mDnDSelection < GetItemCount()) && mDnDIcon.IsValid()) {
			bool sel = GetItem(mDnDSelection).IsSelected();
			bool foc = mWindowState.test(ISFOCUSED);
			if (!sel) GetItem(mDnDSelection).Select(true);
			if (foc) mWindowState.reset(ISFOCUSED);
			DrawItem(mDnDSelection);
			if (!sel) GetItem(mDnDSelection).Select(false);
			if (foc) mWindowState.set(ISFOCUSED);
		}
	}
	if (wacc) return action;
	else return None;
}

/* virtual */ bool CXplorerList::OnDropInit(vector<Atom> mime)
{
    vector<Atom>::iterator it = std::find(mime.begin(), mime.end(), CXplorerFrame::DND_USERNAME);
    string user;
    if (it != mime.end()) {
		user = GetDnDUser();
		if (!user.empty() && user != cxApp->GetUsername()) return false;
	}
    it = std::find(mime.begin(), mime.end(), CXplorerFrame::TEXT_URI_LIST);
    if (it == mime.end()) return false;
    return GetDnDFiles(mDnDFiles);
}

/* virtual */ bool CXplorerList::OnDrop(Atom action)
{
	bool isdir = false;
	if (mDnDSelection < GetItemCount()) {
		if (mDnDIcon.IsValid()) {
			GetItem(mDnDSelection).SetIcon(mDnDIcon);
			DrawItem(mDnDSelection);
			mDnDIcon.Clear();
		}
		if (GetItem(mDnDSelection).GetType() ==TMagic::DIRECTORY) isdir = true;
		if (GetItem(mDnDSelection).GetType() == TMagic::ELFEXE && action == DND_COPY) {
			string cmd = GetPath(mDnDSelection);
			string::size_type pos = 0;
			while ((pos = cmd.find(' ', pos)) < cmd.size()) {
				cmd.replace(pos, 1, "\\ ");
				pos += 2;
			}
			cmd += ' ';
           	string::size_type cmdlen = cmd.size();
			for (vector<string>::iterator i = mDnDFiles.begin(); i != mDnDFiles.end(); ++i) {
				cmd.resize(cmdlen);
				cmd += *i;
				pos = cmdlen;
				while ((pos = cmd.find(' ', pos)) < cmd.size()) {
					cmd.replace(pos, 1, "\\ ");
					pos += 2;
				}
				cxFileManager->Execute(cmd, mCurrentDir);
			}
			return true;
		}
   	}
	if (action == None) return true;
	string to;
	if (isdir) to = GetPath(mDnDSelection);
	else to = mCurrentDir;
	if (action == DND_MOVE) {
		cxFileManager->Move(mDnDFiles, to);
		return true;
	} else if (action == DND_COPY) {
		cxFileManager->Copy(mDnDFiles, to);
		return true;
	} else if (action == DND_LINK) {
		cxFileManager->Link(mDnDFiles, to);
		return true;
	}
	return false;
}

bool CXplorerList::GetDnDFiles(vector<string>& ret)
{
	ret.clear();
	vector<unsigned char> v;
	if (!RequestSelection(DND_SELECTION, CXplorerFrame::TEXT_URI_LIST, DND_SELECTION, v)) return false;
	vector<unsigned char>::iterator i1, i2;
	i1 = v.begin();
	while (i1 < v.end()) {
		i2 = std::find(i1, v.end(), '\n');
		string s((char*)&v[i1 - v.begin()], i2 - i1);
		while (!s.empty() && iscntrl(s[s.size() - 1])) s.resize(s.size() - 1);
		if (!s.empty() && cxFileManager->ToLocalFS(s)) ret.push_back(s);
		i1 = ++i2;
	}
	return !ret.empty();
}

/* virtual */ bool CXplorerList::CanAcceptAction(Atom action)
{
	return (action == DND_MOVE || action == DND_COPY || action == DND_LINK || action == DND_ASK);
}

/* virtual */ void CXplorerList::OnDropFinished()
{
	static_cast<CXplorerFrame*>(GetParent())->Reload();
}

bool CXplorerList::GetActionList(vector<Atom>& actions, vector<string>& desc)
{
	if (mDnDSelection < GetItemCount() && GetItem(mDnDSelection).GetType() == TMagic::ELFEXE) {
		actions.push_back(DND_COPY);
		desc.push_back(_("Execute here", 78));
	} else {
		actions.push_back(DND_COPY);
		desc.push_back(_("Copy here", 79));
		if (mWriteAccess) {
			actions.push_back(DND_MOVE);
			desc.push_back(_("Move here", 80));
		}
		if (!mArchive) {
			actions.push_back(DND_LINK);
			desc.push_back(_("Link here", 81));
		}
	}
	return true;
}

void CXplorerList::SetFlags(unsigned flags)
{
	mXplorerFlags = flags & CXplorerFrame::VIEWFLAGS;
	switch (flags & CXplorerFrame::VIEWSIZE) {
	case CXplorerFrame::VIEW10K:
		mMaxPreviewSize = 10000;
		break;
	case CXplorerFrame::VIEW30K:
		mMaxPreviewSize = 30000;
		break;
	case CXplorerFrame::VIEW100K:
		mMaxPreviewSize = 100000;
		break;
	case CXplorerFrame::VIEW300K:
		mMaxPreviewSize = 300000;
		break;
	case CXplorerFrame::VIEW1M:
		mMaxPreviewSize = 1000000;
		break;
	case CXplorerFrame::VIEWALL:
		mMaxPreviewSize = -1u;
		break;
	default:
		mMaxPreviewSize = 0;
	}
	if ((flags & ~CXplorerFrame::VIEWSIZE) == 0) mMaxPreviewSize = 0;
	Reload();
}
	
/* virtual */ void CXplorerList::SetStyle(unsigned style, unsigned wico, unsigned hico)
{
	CFileListControl::SetStyle(style, wico, hico);
	mOpenDir.Create(CIcon::FOLDER_OPEN, mIconSize);
	mActiveExe.Create(CIcon::ACTIVE_EXEC, mIconSize);
}

string CXplorerList::GetDnDUser()
{
	string s;
	vector<unsigned char> v;
	if (RequestSelection(DND_SELECTION, CXplorerFrame::DND_USERNAME, DND_SELECTION, v)) {
		while (!v.empty() && iscntrl(v.back())) v.pop_back();
		s.assign((char*)&v[0], v.size());
	}
	return s;
}
