/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : CXplorerTree.cpp
  Info   : $Id: CXplorerTree.cpp,v 1.34 2012/09/12 16:01:43 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2002 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// Application headers
#include "Xplorer.h"
#include "CXplorerTree.h"
#include "CXplorerFrame.h"

// C++X headers
#include <cxxx/CFileHandler.h>
#include <cxxx/CApplication.h>

// C++ headers
#include <cstdlib> // getenv
#include <cerrno>

// C headers
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>

// Namespace declarations
using std::string;
using std::vector;
using cxxx::cxApp;
using cxxx::cxError;
using cxxx::cxFileManager;
using cxxx::cxgettext;
using cxxx::cxNarrow;

void CXplorerTree::CreateContextMenu()
{
	// --- CONTEXT ---
	mContextMenu.Entry(CXplorerFrame::IDNAV_UP, _("Up", 64), CIcon::UP);
	mContextMenu.Entry(CXplorerFrame::IDNAV_BACK, _("Back", 65), CIcon::BACK);
	mContextMenu.Entry(CXplorerFrame::IDNAV_FORWARD, _("Forward", 66), CIcon::FORWARD);
	mContextMenu.Entry(CXplorerFrame::IDVIEW_REFRESH, _("Reload", 67), CIcon::RELOAD);
	mContextMenu.Seperator();
	mContextMenu.Entry(CXplorerFrame::IDTREEPROP, _("Properties ...", 75));
}

/* virtual */ void CXplorerTree::OnContext()
{
	Popup(mContextMenu);
}

/* virtual */ bool CXplorerTree::OnDragInit(unsigned button, Atom& action)
{
	if (button == 1) {
		if (access(GetItem(GetSelection().front()).GetText().c_str(), W_OK | X_OK) == 0) action = CDnDManager::DND_MOVE;
		else action = CDnDManager::DND_COPY;
		return true;
	} else if (button == 3) {
		action = CDnDManager::DND_ASK;
		return true;
	}
	return false;
}

/* virtual */ bool CXplorerTree::OnDrop(Atom action)
{
	if (mDnDSelection < GetItemCount()) {
		if (mDnDIcon.IsValid()) {
			GetItem(mDnDSelection).SetIcon(mDnDIcon);
			DrawItem(mDnDSelection);
			mDnDIcon.Clear();
			DrawFrame();
		}
	}
	if (action == None) return true;
	string to(GetPath(mDnDSelection));
	if (action == DND_MOVE) cxFileManager->Move(mDnDFiles, to);
	else if (action == DND_COPY) cxFileManager->Copy(mDnDFiles, to);
	else if (action == DND_LINK) cxFileManager->Link(mDnDFiles, to);
	else if (action == DND_PRIVATE) return false;
	static_cast<CXplorerFrame*>(GetParent())->Reload();
	return true;
}

/* virtual */ bool CXplorerTree::OnDropInit(std::vector<Atom> mime)
{
	vector<Atom>::iterator it = std::find(mime.begin(), mime.end(), CXplorerFrame::DND_USERNAME);
	string user;
	if (it != mime.end()) {
		user = GetDnDUser();
		if (!user.empty() && user != cxApp->GetUsername()) return false;
	}
	it = std::find(mime.begin(), mime.end(), CXplorerFrame::TEXT_URI_LIST);
	if (it == mime.end()) return false;
	return GetDnDFiles(mDnDFiles);
}

/* virtual */ bool CXplorerTree::CanAcceptAction(Atom action)
{
	return (action == DND_MOVE || action == DND_COPY || action == DND_LINK || action == DND_ASK);
}

/* virtual */ bool CXplorerTree::CanPerformAction(Atom action)
{
	return (action == DND_MOVE || action == DND_COPY || action == DND_LINK);
}

/* virtual */ Atom CXplorerTree::GetDnDAction(int x, int y, Atom action)
{
	if (x < BORDER_WIDTH || y < BORDER_WIDTH || x >= (int)mViewWidth + BORDER_WIDTH || y >= (int)mViewHeight + BORDER_WIDTH) action = None;
	if (action != DND_COPY && action != DND_MOVE && action != DND_LINK && action != DND_ASK && action != DND_PRIVATE) action = None;
	bool needframe(false);
	if ((mDnDSelection < GetItemCount()) && mDnDIcon.IsValid()) {
        GetItem(mDnDSelection).SetIcon(mDnDIcon);
        mDnDIcon.Clear();
        DrawItem(mDnDSelection);
		needframe = true;
	}
	if (action != None) {
		mDnDSelection = ItemFromPos(x, y);
		unsigned w, h;
		GetBounds(x, y, w, h);
		SetIgnoreRect(x, y, w, h);
		if (mDnDSelection < GetItemCount()) {
			if (access(GetPath(mDnDSelection).c_str(), W_OK | X_OK) != 0) action = None;
			else {
				mDnDIcon = GetItem(mDnDSelection).GetIcon();
				GetItem(mDnDSelection).SetIcon(mOpenDir);
				DrawItem(mDnDSelection);
				DrawDottedRectangle (x + mIconWidth, y - TEXT_BORDER + (mIconHeight - GetTextHeight()) / 2, GetItem(mDnDSelection).GetWidth() - mIconWidth - 1, GetTextHeight() + 2 * TEXT_BORDER - 1, CColor(TC_BACKGND_TEXT));
				needframe = true;
			}
		} else action = None;
	}
	if (needframe) DrawFrame();
	return action;
}

/* virtual */ void CXplorerTree::OnDropFinished()
{
	static_cast<CXplorerFrame*>(GetParent())->Reload();
}

/* virtual */ bool CXplorerTree::GetActionList(std::vector<Atom>& actions, std::vector<std::string>& desc)
{
	actions.push_back(DND_COPY);
	desc.push_back(_("Copy here", 79));
	actions.push_back(DND_MOVE);
	desc.push_back(_("Move here", 80));
	actions.push_back(DND_LINK);
	desc.push_back(_("Link here", 81));
	return true;
}

bool CXplorerTree::GetDnDFiles(vector<string>& ret)
{
	ret.clear();
	vector<unsigned char> v;
	if (!RequestSelection(DND_SELECTION, CXplorerFrame::TEXT_URI_LIST, DND_SELECTION, v)) return false;
	vector<unsigned char>::iterator i1, i2;
	i1 = v.begin();
	while (i1 < v.end()) {
		i2 = std::find(i1, v.end(), '\n');
		string s((char*)&v[i1 - v.begin()], i2 - i1);
		while (!s.empty() && iscntrl(s[s.size() - 1])) s.resize(s.size() - 1);
		if (!s.empty() && cxFileManager->ToLocalFS(s)) ret.push_back(s);
		i1 = ++i2;
    }
    return !ret.empty();
}

string CXplorerTree::GetDnDUser()
{
	string s;
	vector<unsigned char> v;
	if (RequestSelection(DND_SELECTION, CXplorerFrame::DND_USERNAME, DND_SELECTION, v)) {
		while (!v.empty() && iscntrl(v.back())) v.pop_back();
		s.assign((char*)&v[0], v.size());
	}
	return s;
}

/* virtual */ bool CXplorerTree::ConvertSelection(Atom sel, Atom typ, vector<unsigned char>& ret)
{
	if (sel != DND_SELECTION) return false;
	ret.clear();
	const unsigned char* sbeg(NULL);
	const unsigned char* send(NULL);
	if (typ == CSelection::TARGETS) {
		Atom a;
		sbeg = (unsigned char*)&a;
		send = sbeg + sizeof(Atom);
		a = CXplorerFrame::DND_USERNAME;
		copy(sbeg, send, back_inserter(ret));
		a = CXplorerFrame::TEXT_URI_LIST;
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	if (typ == CXplorerFrame::TEXT_URI_LIST) {
		string s("file://");
		s += cxApp->GetHostname();
		s += GetPath(GetSelection().front());
		s += "\r\n";
		sbeg = (const unsigned char*)&s[0];
		send = sbeg + s.size();
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	if (typ == CXplorerFrame::DND_USERNAME) {
		sbeg = (const unsigned char*)(cxApp->GetUsername().data());
		send = sbeg + cxApp->GetUsername().size();
		copy(sbeg, send, back_inserter(ret));
		return true;
	}
	return false;
}
