/*----------------------------------------------------------------------------
  Project: Xplorer <http://cxplorer.sourceforge.net>
  File   : Xplorer.cpp
  Info   : $Id: Xplorer.cpp,v 1.23 2012/07/17 12:16:43 gkraeuter Exp $
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
  Copyright © 2003 - 2012 Gerhard Kräuter <gkraeuter@users.sourceforge.net>

  This  program  is  free software: you can redistribute it and/or modify it
  under  the  terms  of  the  GNU General Public License as published by the
  Free Software Foundation,  either  version  3  of the License, or (at your
  option) any later version.

  This  program  is  distributed  in  the  hope  that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public License
  for more details.

  You  should  have  received a copy of the GNU General Public License along
  with this program. If not, see  <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------*/

// application headers
#include "Xplorer.h"
#include "CXplorerFrame.h"

// C++X headers
#include <cxxx/Cxxx.h>
#include <cxxx/CApplication.h>

// C++ headers
#include <iostream>
#include <string>
#include <stdexcept>

// C headers
#include <langinfo.h>
#include <unistd.h>

// namespace declarations
using std::string;
using std::vector;
using std::cerr;
using std::cout;
using std::endl;
using std::strcmp;
using std::locale;
using std::runtime_error;
using cxxx::cxgettext;
using cxxx::cxAppIntl;
using cxxx::cxCxxxIntl;
using cxxx::cxLocale;
using cxxx::cxHomeDir;
using cxxx::cxCxxxLocalePath;
using cxxx::cxAppLocalePath;
using cxxx::cxSysIconDir;
using cxxx::cxFileManager;

int main(int argc, char **argv)
{
	// test the presence of cxxx locale directory
	if (access(LOCALEDIR "/C/LC_MESSAGES/cxxx.msg", R_OK) || access(DATADIR "/cxxx/icons", X_OK)) {
		// no! libcxxx is not installed!
		// read the current working directory
		char buf[PATH_MAX + 1];
		const char* cwd = getcwd(buf, PATH_MAX);
		if (!cwd) {		
			cerr << "Failed to read the current working directory" << endl;
			cwd = "/";
		}
		// find the absolute path to this application
		CFileHandler filehandler;
		string apppath = filehandler.GetAbsolutePath(argv[0], cwd);
		// this may be a link; resolve it
		ssize_t len = readlink(apppath.c_str(), buf, PATH_MAX);
		while (len != -1) {
			buf[len] = 0;
			apppath = filehandler.GetAbsolutePath(buf, filehandler.GetFilePath(apppath));
			len = readlink(apppath.c_str(), buf, PATH_MAX);
		}
		string bindir = filehandler.GetFilePath(apppath);
		string prefix = filehandler.GetFilePath(bindir);
		string podir;
		do {
			podir = prefix + "/po";
			if (access(podir.c_str(), F_OK) == 0) break;
			prefix = filehandler.GetFilePath(prefix);
		} while (prefix != "/");
		cxCxxxLocalePath = prefix + "/po/%L/cxxx.msg";
		cxSysIconDir = prefix + "/icons";
		cxAppLocalePath = prefix + "/po/%L/xplorer.msg";
	}
	// Parsing the arguments	
	string filename;
	for (int i = 1; i < argc; ++i) {
		// Print versions to stdout
		if (strcmp(argv[i], "--version") == 0 || strcmp(argv[i], "-v") == 0) {
			cout << "Xplorer version " << xplorer_VERSION << ", compiled against C++X version " << CXXX_VERSION << "." << endl;
			exit (0);
		// Determine the file type and print to stdout
		} else if (strcmp(argv[i], "--type") == 0 || strcmp(argv[i], "-t") == 0) {
			IntlInit(xplorer::BOTH);
			if (++i < argc) {
				filename = argv[i];
				CMagic mag;
				CFileHandler filehandler;
				unsigned typ = mag.GetFileType(filename);
				filename = filehandler.GetFileName(filename);
				if (filename.empty()) filename = "/";
				cout << filename << ": " << mag.GetMimeType(typ);
				switch (typ & TMagic::COMPRESSED) {
				case TMagic::GZCOMPRESSED:
					cout << ", GZIP compressed";
					break;
				case TMagic::BZCOMPRESSED:
					cout << ", BZIP2 compressed";
					break;
				case TMagic::XZCOMPRESSED:
					cout << ", XZ compressed";
					break;
				default:
					break;
				}
				cout << "." << endl;
			} else cerr << cxxx::_("Xplorer: error parsing command line", 170) << endl;
			exit(EXIT_SUCCESS);
		} else if (*argv[i] == '-') {
			// Print usage to stdout
			IntlInit(xplorer::APP);
			cerr << cxxx::_("Usage:\n", 168);
			cerr << cxxx::_("xplorer [OPTION | FILE]\n", 169);
			cerr << cxxx::_("where OPTION is one of the following:\n", 172);
			cerr << cxxx::_("-v, --version\nprints the versions of xplorer and libcxxx and exits.\n", 171);
			cerr << cxxx::_("-?, --help\nprint this message and exits.\n", 139);
			cerr << cxxx::_("FILE can be an absolute or relative file name or one of the following:\n", 138);
			cerr << cxxx::_("ROOT for the root directory\n", 164);
			cerr << cxxx::_("HOME for the users home directory\n", 165);
			cerr << cxxx::_("CWD for the current working directory\n\n", 166) << endl;
			exit(0);
		} else {
			filename = argv[i];
		}
	}
	// Fire up xplorer
	CApplication app("xplorer");
	string id = "Xplorer ";
	id += xplorer_VERSION;
	if (id != cxAppIntl.get_id()) cerr << "Xplorer: Version mismatch in localization file." << endl;
	if (filename == "ROOT" || filename.empty()) filename = "/";
	else if (filename == "HOME") filename = cxHomeDir;
	else if (filename[0] != '/') {
		char buf[PATH_MAX + 1];
		const char* cwd = getcwd(buf, PATH_MAX);
		if (!cwd) {
			cerr << "Failed to read the current working directory" << endl;
			cwd = "/";
		}
		if (filename == "CWD") filename = cwd;
		else filename = cxFileManager->GetAbsolutePath(filename, cwd);
	}
	string dirname = filename;
	filename.clear();
	CMagic mag;
	while (mag.GetFileType(dirname) != TMagic::DIRECTORY) {
		filename = cxFileManager->GetFileName(dirname);
		dirname = cxFileManager->GetFilePath(dirname);
	}
	CXplorerFrame frame;
	frame.SetDirectory(dirname);
	if (!filename.empty()) frame.Select(filename);
	app.Run();
	return 0;
}

void xplorer::IntlInit(unsigned which)
{
	// Initialize Internationalization
	// cxCxxxIntl is normally initialized in CApplication::CApplication, but that would open X
	if (cxCxxxLocalePath.empty()) cxCxxxLocalePath = LOCALEDIR "/%L/LC_MESSAGES/cxxx.msg";
	if (cxAppLocalePath.empty()) cxAppLocalePath = LOCALEDIR "/%L/LC_MESSAGES/xplorer.msg";
	try {
		locale loc("");
		cxLocale = loc;
	}
	catch (const runtime_error& e) {
		cerr << "Xplorer:\nFailed to set locale:\n" << e.what() << "\nDefaulting to \'C\' locale." << endl;
	}
	locale::global(cxLocale);
	if (which != APP) {
		int err = cxCxxxIntl.open(cxLocale.name().c_str(), cxCxxxLocalePath.c_str(), nl_langinfo(CODESET));
		if (err) {
			cerr << "Xplorer: Failed to open C++X localization file. Trying C locale." << endl;
			err = cxCxxxIntl.open("C", cxCxxxLocalePath.c_str(), nl_langinfo(CODESET));
			if (err) {
				cerr << cxCxxxIntl.error_text(err) << endl;
				exit (EXIT_FAILURE);
			}
		}
		string id = "C++X ";
		id += CXXX_VERSION;
		if (id != cxCxxxIntl.get_id()) cerr << "Xplorer: Version mismatch in C++X localization file." << endl;
	}
	if (which != LIB) {
		int err = cxAppIntl.open(cxLocale.name().c_str(), cxAppLocalePath.c_str(), nl_langinfo(CODESET));
		if (err) {
			cerr << "Xplorer: Failed to open localization file. Trying C locale." << endl;
			err = cxAppIntl.open("C", cxAppLocalePath.c_str(), nl_langinfo(CODESET));
			if (err) {
				cerr << cxAppIntl.error_text(err) << endl;
				exit (EXIT_FAILURE);
			}
		}
		string id = "Xplorer ";
		id += xplorer_VERSION;
		if (id != cxAppIntl.get_id()) cerr << "Xplorer: Version mismatch in localization file." << endl;
	}
}
